/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  ActiveInterval.java
//
//=============================================================================

package xmto2xmta;

import java.util.Vector;
import xmtutil.*;
import xmto2xmta.timegraph.*;

/*********************************************************************** 
 * Class containing a specific life cycle information for an object.
 *
 * <p><b>History:</b><ul>
 * <li>27 Mar 2001 - Created file.
 * </ul><p>
 *********************************************************************/
public class ActiveInterval {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	// These are the only meaningful fill type values in ActiveInterval.
	public static final int FILL_FREEZE = 2;
	public static final int FILL_TRANSITION = 4;

	/***********************************************************************
	 * Input structures
	 **********************************************************************/
	private TimingAttribute computedBeginTime_ = null;
	private TimingAttribute observedBeginTime_ = null;
	private TimingAttribute activeDuration_ = null;
	private Vector repeatTimes_ = null;	// vector of repeat begin times
	private TimingAttribute fillEndTime_ = null;
	private int fillType_ = 0;
	private TimingAttribute intrinsicDuration_ = null;
	private LifeCycle lifeCycle = null;
	private Vector elasids = null;
	private Vector simpleDurations = null;
	private TimingAttribute activeEnd_ = null;
	private TimingAttribute computedEndTime = null;

/***********************************************************************
 **********************************************************************/
	public ActiveInterval(LifeCycle lifeCycle) {
		this.lifeCycle = lifeCycle;
	}

/***********************************************************************
 * from observed begin to end of last simple duration
 **********************************************************************/
public TimingAttribute getActiveDuration() {
	return activeDuration_;
}

/***********************************************************************
 * End of the active interval before freeze
 **********************************************************************/
public TimingAttribute getActiveEndTime() {
	return activeEnd_;
}

public TimingAttribute getActiveEnd() {
	return getActiveEndTime();
}

/***********************************************************************
 **********************************************************************/
public TimingAttribute getComputedBeginTime() {
	return computedBeginTime_;
}

/***********************************************************************
 **********************************************************************/
public TimingAttribute getFillEndTime() {
	return fillEndTime_;
}

/***********************************************************************
 **********************************************************************/
public int getFillType() {
	return fillType_;
}

/***********************************************************************
 * observed begin time
 **********************************************************************/
public TimingAttribute getObservedBeginTime() {
	return observedBeginTime_;
}

/***********************************************************************
 * deprecated
 **********************************************************************/
public TimingAttribute getSceneBeginTime() {
	return getObservedBeginTime();
}

/***********************************************************************
 * end of fill time
 **********************************************************************/
public TimingAttribute getSceneEndTime() {
	return getFillEndTime();
}

/***********************************************************************
 * except the first iteration.
 **********************************************************************/
public Vector getRepeatTimes() {  // vector of begin times
	return repeatTimes_;
}

public Vector getELASids() {  // vector ELASid
	return elasids;
}

/***********************************************************************
 * return a vector of simple durations
 **********************************************************************/
public Vector getSimpleDurations() {
	return simpleDurations;
}

/***********************************************************************
 * If asIs == true, don't try to compute the intrinsic dur.
 **********************************************************************/
public TimingAttribute getIntrinsicDuration() {
	return getIntrinsicDuration(false);
}

public TimingAttribute getIntrinsicDuration(boolean asIs) {
	if (asIs) {
		if (intrinsicDuration_ == null) {
			return new TimingAttribute(TimingAttribute.INVALID);
		} else {
			return intrinsicDuration_;
		}
	} else if (intrinsicDuration_ != null) {
		return intrinsicDuration_;
	} else {
		if (lifeCycle == null) {
			intrinsicDuration_ = new TimingAttribute(TimingAttribute.INVALID);
		} else {
			Container timeContainer = lifeCycle.getTimeContainer();
			if ((timeContainer != null) && (timeContainer instanceof MediaTimeContainer)) {
				Time intrinsicDur = ((TimeContainer) timeContainer).computeIntrinsicDur();
				if (intrinsicDur.isResolved()) {
					intrinsicDuration_ = LifeCycle.convertFrom(intrinsicDur);
				} else {
					intrinsicDuration_ = new TimingAttribute(TimingAttribute.INVALID);
				}
			} else {
				intrinsicDuration_ = new TimingAttribute(TimingAttribute.INVALID);
			}
		}
		return intrinsicDuration_;
	}
}

public TimingAttribute getComputedEndTime() {
	return computedEndTime;
}

/***********************************************************************
 **********************************************************************/
public void setActiveDuration(TimingAttribute activeDuration) {
	activeDuration_ = activeDuration;
}

/***********************************************************************
 **********************************************************************/
public void setActiveEnd(TimingAttribute activeEnd) {
	activeEnd_ = activeEnd;
}

/***********************************************************************
 **********************************************************************/
public void setComputedBeginTime(TimingAttribute computedBeginTime) {
	computedBeginTime_ = computedBeginTime;
}

/***********************************************************************
 **********************************************************************/
public void setFillEndTime(TimingAttribute fillEndTime) {
	fillEndTime_ = fillEndTime;
}

/***********************************************************************
 **********************************************************************/
public void setFillType(int fillType) {
	fillType_ = fillType;
}

public void setObservedBeginTime(TimingAttribute observedBeginTime) {
	observedBeginTime_ = observedBeginTime;
}

/***********************************************************************
 **********************************************************************/
public void setRepeatTimes(Vector repeatTimes) {  // vector of TimingAttributes
	repeatTimes_ = repeatTimes;
}

/***********************************************************************
 **********************************************************************/
public void setIntrinsicDuration(TimingAttribute intrinsicDuration) {
	intrinsicDuration_ = intrinsicDuration;
}

public void setLifeCycle(LifeCycle lifeCycle) {
	this.lifeCycle = lifeCycle;
}

public LifeCycle getLifeCycle() {
	return lifeCycle;
}

public void setELASids(Vector elasids) { // vector of ELASid
	this.elasids = elasids;
}

public void setSimpleDurations(Vector simpleDurations) { // vector of TimingAttributes
	this.simpleDurations = simpleDurations;
}

public void setComputedEndTime(TimingAttribute computedEndTime) {
	this.computedEndTime = computedEndTime;
}

/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}

}
