/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  EventModel.java
//   
//=============================================================================
package xmto2xmta;

import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import xmtutil.MediaObjectInformation;
import xmto2xmta.resources.Xmto2XmtaRes;
import xmto2xmta.tree.audio;
import xmto2xmta.tree.ConvertElement;
import xmto2xmta.tree.ExternalElement;
import xmto2xmta.tree.TimingElement;
import xmto2xmta.tree.TreeElement;
import xmto2xmta.tree.video;
import xmtutil.DOMXMTO;
import xmtutil.TimingAttribute;

/*********************************************************************** 
 * Class to group together the various type of timing elements.
 *
 * <p><b>History:</b><ul>
 * <li>23 Feb 2001 - Created file.
 * <li>27 Mar 2001 - Add check if no end time then end event could be offset 
 *                   from begin attrtibute that is an event.
 * <li>20 Apr 2000 - Remove resolve model, replaced by TimeScheduler.
 *                 - add simple support for begin event (for &lt;set&gt;)
 * </ul><p>
 *********************************************************************/
public class EventModel {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	/***********************************************************************
	 * Input structures
	 **********************************************************************/
	private Xmto2Xmta o2aConverter_ = null;
	private Hashtable xmtoIdTable_ = null;
	private Hashtable convertIndexTable_ = null;
	private TimeScheduler timeScheduler_ = null;
	private ConvertElement root_ = null;

	/***********************************************************************
	 * Hash table Data structure for event times.
	 * 
	 * The prime number 89 was chosen as the hash table initialize size because it was
	 * mentioned as a better starting hash size for optimization than the default 101.
	 * The element2ClassMapTable_ table is composed of (key, data)
	 *   The key is the convert tree element indexes (unique id).
	 *   The data is the Data structure for resolved times.
	 **********************************************************************/
	private Hashtable eventTable_ = new Hashtable(89);

	/*********************************************************************** 
	 **********************************************************************/
	private ResourceBundle xmto2XmtaRes_ = ResourceBundle.getBundle("xmto2xmta.resources.Xmto2XmtaRes", Locale.getDefault());
/*********************************************************************** 
 * Create the Event Model. 
 * @param converter      converter with pointers to the convert tree, 
 *                       the xmto id hash table and the convert tree index hash table.
 * @param xmtoIds        table of the convert tree Elements.  The hash index is the xmto id 
 *                       attribute.  Not all convert tree elements are contained in this table,
 *                       for example the elements which don't have xmto ids.  Also this table
 *                       may have elemnents removed if an element isn't supported or has an error
 *                       which can't be converted.
 * @param convertIndexes table of the convert tree Elements.  The hash index is the unique
 *                       index for each convert element (like an id, but not to be confused with the 
 *                       xmto id.  This table may have elemnents removed if an element isn't 
 *                       supported or has an error which can't be converted.
 * @param resolveModel   Model holding all the resolved times for convert tree timing elements.
 *********************************************************************/
public EventModel(Xmto2Xmta converter, Hashtable xmtoIds, Hashtable convertIndexes, ConvertElement root, TimeScheduler timeScheduler) {
	o2aConverter_ = converter; // DOn't know if this is necessary or not but pass it along
	xmtoIdTable_ = xmtoIds;
	convertIndexTable_ = convertIndexes;
	timeScheduler_ = timeScheduler;
	root_ = root;
}
/***********************************************************************
 * Return the active duration of the specified element.
 * JSB: don't know if this method makes any sense in the EventModel.  I don't know
 * if the active duration can't be resoved in the ResolveModel and yet can
 * be resolved in the EventModel?  For now keep the method in.
 **********************************************************************/
public TimingAttribute getActiveDuration(Integer convertTreeIndex) {
	Object element = convertIndexTable_.get(convertTreeIndex);
	if (null != element && element instanceof TimingElement) {
		return ((TimingElement) element).getXmtoAttrDur();
	} else
		return new TimingAttribute(TimingAttribute.INDEFINITE_VALUE);
}
/*********************************************************************** 
 * Get the begining scene time for the begin event.
 * The begin event can occur across some range of time.  This range is dependent on range
 * of this element and the range of the element which is clicked (sensor element).
 * <p>
 * Since this element is a click event, its range can't be determined via the lifecycle of the 
 * element, but try anyway.  If we can't find the range from the life cycle use the range of 
 * the parent/ancestor timing element.  The parent/ancestor range can be used to limit this
 * element's range.
 * <p>
 * The the range of this element also depends on the duration of the element which is 
 * clicked (sensor element).  Thus the range of time is limited to the active duration of 
 * the sensor element.  
 * <p>
 * The range of the begin event is the intersection of sensor duration and the 
 * element/parent/ancestor element duration.  
 * <p>
 * This routine is returns the earliest time in which the begin event can occur, 
 * the begining of the range.  
 **********************************************************************/
public TimingAttribute getBeginSceneBeginTime(Integer convertTreeIndex) {
	if (!isBeginEvent(convertTreeIndex))
		return null;

	// Get the convert tree object
	Object object = convertIndexTable_.get(convertTreeIndex);
	if (null == object || !(object instanceof TimingElement))
		return null;
	TimingElement timingElement = (TimingElement) object;

	// Get the sensor object (convert tree object) which causes the begin event
	TimingAttribute begin = ((TimingElement) object).getXmtoAttrBegin();
	String id = begin.getIdValue();
	object = xmtoIdTable_.get(id);
	TimingElement sensorTimingElement = null;
	TimingAttribute sensorBeginTime = null;
	if (null != object && object instanceof TimingElement) {
		sensorTimingElement = (TimingElement) object;


		// Get the starting of it as the earliest time this object can begin.
		Integer sensorIndex = sensorTimingElement.getIndex();
		Vector lifeCycles = timeScheduler_.getLifeCycles(sensorIndex);
		if (lifeCycles.size() > 0) {
			LifeCycle lifeCycle = (LifeCycle) lifeCycles.firstElement();
			Vector activeIntervals = lifeCycle.getActiveIntervals();
			if (activeIntervals.size() > 0) {
				ActiveInterval activeInterval = (ActiveInterval) activeIntervals.firstElement();
				sensorBeginTime = activeInterval.getSceneBeginTime();
			}
		}
	}

	// Get the begin time of the element/parent/ancestor timing element.
	// Check the element if there is a begin time
	TimingAttribute elementBeginTime = null;
	Vector lifeCycles = timeScheduler_.getLifeCycles(convertTreeIndex);
	if (lifeCycles.size() > 0) {
		LifeCycle lifeCycle = (LifeCycle) lifeCycles.firstElement();
		Vector activeIntervals = lifeCycle.getActiveIntervals();
		if (activeIntervals.size() > 0) {
			ActiveInterval activeInterval = (ActiveInterval) activeIntervals.firstElement();
			elementBeginTime = activeInterval.getSceneBeginTime();
		}
	}
	// If we didn't get a begin time from the element, which is expected.
	// Then get it from its parent/ancestor timing element.
	// Find the parent/ancestor timing element
	TreeElement treeElement = timingElement.getParent();
	while (null != treeElement && !(treeElement instanceof TimingElement)) {
		treeElement = treeElement.getParent();
	}
	if (null != treeElement) {
		TimingElement parent = (TimingElement) treeElement;

		// Now get the begin time for the parent
		lifeCycles = timeScheduler_.getLifeCycles(parent.getIndex());
		if (lifeCycles.size() > 0) {
			LifeCycle lifeCycle = (LifeCycle) lifeCycles.firstElement();
			Vector activeIntervals = lifeCycle.getActiveIntervals();
			if (activeIntervals.size() > 0) {
				ActiveInterval activeInterval = (ActiveInterval) activeIntervals.firstElement();
				elementBeginTime = activeInterval.getSceneBeginTime();
			}
		}
	}

	// return the first time
	if (null == sensorBeginTime)
		return elementBeginTime;
	if (null == elementBeginTime)
		return sensorBeginTime;
	if (sensorBeginTime.lessThan(elementBeginTime))
		return sensorBeginTime;
	else
		return elementBeginTime;
}
/***********************************************************************
 * Get the begining scene time for the end event
 * The end event can occur across some range of times, usually this is the duration of the
 * element which is clicked on.  Thus the range of time is often the active duration of that
 * element.  This routine is returns the earliest time of the end event can occur.
 * 20 Apr 2001 - Comment out this method since it end event isn't supported.
 *               Logic kept in as reference.
 **********************************************************************/
public TimingAttribute getBeginSceneEndTime(Integer convertTreeIndex) {
	//       if (!isEndEvent(convertTreeIndex))
	//       	return null;
	//
	//       // Get the convert tree object
	//       Object object = convertIndexTable_.get(convertTreeIndex);
	//       if (null == object || !(object instanceof TimingElement))
	//       	return null;
	//       TimingElement timingElement = (TimingElement) object;
	//
	//       // Get the convert tree object which causes the begin event
	//       TimingAttribute time = ((TimingElement) object).getXmtoAttrEnd();
	//       if (null == time)
	//       	time = ((TimingElement) object).getXmtoAttrBegin();
	//       switch (time.getTimeValueType()) {
	//       	case TimingAttribute.EVENT_VALUE :
	//       		{
	//       			String id = time.getIdValue();
	//       			object = xmtoIdTable_.get(id);
	//       			if (null == object || !(object instanceof TimingElement))
	//       				return null;
	//       			TimingElement sensorTimingElement = (TimingElement) object;
	//       			// Get the starting of it as the earliest time this object can begin.
	//       			Integer sensorIndex = sensorTimingElement.getIndex();
	//       			if (resolveModel_.isBeginResolved(sensorIndex)) {
	//       				TimingAttribute sensorTime = resolveModel_.getSceneBeginTime(sensorIndex);
	//       				return sensorTime;
	//       			}
	//       			break;
	//       		}
	//       	case TimingAttribute.MULTI_VALUE :
	//       		{
	//       			for (Enumeration enum = time.getMultiTimingAttributes().elements(); enum.hasMoreElements();) {
	//       				TimingAttribute endTime = (TimingAttribute) enum.nextElement();
	//       				switch (endTime.getTimeValueType()) {
	//       					case TimingAttribute.EVENT_VALUE :
	//       						String id = endTime.getIdValue();
	//       						object = xmtoIdTable_.get(id);
	//       						if (null == object || !(object instanceof TimingElement))
	//       							return null;
	//       						TimingElement sensorTimingElement = (TimingElement) object;
	//       						// Get the starting of it as the earliest time this object can begin.
	//       						Integer sensorIndex = sensorTimingElement.getIndex();
	//       						if (resolveModel_.isBeginResolved(sensorIndex)) {
	//       							TimingAttribute sensorTime = resolveModel_.getSceneBeginTime(sensorIndex);
	//       							return sensorTime;
	//       						}
	//       						break;
	//       					default :
	//       						break;
	//       				}
	//       			}
	//       			break;
	//       		}
	//       	default :
	//       		break;
	//       }
	return null;
}
/*********************************************************************** 
 * Get the ending scene time for the begin event
 * The begin event can occur across some range of times, usually this is the duration of the
 * element which is clicked on.  Thus the range of time is often the active duration of that
 * element.  This routine is returns the latest time in which the begin event can occur.  
 * It might be indefinite, thus it might never start if the user never clicks.
 
 * Get the ending scene time for the begin event.
 * The begin event can occur across some range of time.  This range is dependent on range
 * of this element and the range of the element which is clicked (sensor element).
 * <p>
 * Since this element is a click event, its range can't be determined via the lifecycle of the 
 * element, but try anyway.  If we can't find the range from the life cycle use the range of 
 * the parent/ancestor timing element.  The parent/ancestor range can be used to limit this
 * element's range.
 * <p>
 * The the range of this element also depends on the duration of the element which is 
 * clicked (sensor element).  Thus the range of time is limited to the active duration of 
 * the sensor element.  
 * <p>
 * The range of the begin event is the intersection of sensor duration and the 
 * element/parent/ancestor element duration.  
 * <p>
 * This routine is returns the latest time in which the begin event can occur, 
 * the end of the range.  It might be indefinite, thus it might never start if 
 * the user never clicks.
 **********************************************************************/
public TimingAttribute getEndSceneBeginTime(Integer convertTreeIndex) {
	// Get the convert tree object
	Object object = convertIndexTable_.get(convertTreeIndex);
	if (null == object || !(object instanceof TimingElement))
		return null;
	TimingElement timingElement = (TimingElement) object;

	// Get the sensor object (convert tree object) which causes the begin event
	TimingAttribute begin = ((TimingElement) object).getXmtoAttrBegin();
	String id = begin.getIdValue();
	object = xmtoIdTable_.get(id);
	TimingElement sensorTimingElement = null;
	TimingAttribute sensorEndTime = null;
	if (null != object && object instanceof TimingElement) {
		sensorTimingElement = (TimingElement) object;


		// Get the starting of it as the earliest time this object can begin.
		Integer sensorIndex = sensorTimingElement.getIndex();
		Vector lifeCycles = timeScheduler_.getLifeCycles(sensorIndex);
		if (lifeCycles.size() > 0) {
			LifeCycle lifeCycle = (LifeCycle) lifeCycles.firstElement();
			Vector activeIntervals = lifeCycle.getActiveIntervals();
			if (activeIntervals.size() > 0) {
				ActiveInterval activeInterval = (ActiveInterval) activeIntervals.firstElement();
				sensorEndTime = activeInterval.getSceneEndTime();
			}
		}
	}

	// Get the end time of the element/parent/ancestor timing element.
	// Check the element if there is a begin time
	TimingAttribute elementEndTime = null;
	Vector lifeCycles = timeScheduler_.getLifeCycles(convertTreeIndex);
	if (lifeCycles.size() > 0) {
		LifeCycle lifeCycle = (LifeCycle) lifeCycles.firstElement();
		Vector activeIntervals = lifeCycle.getActiveIntervals();
		if (activeIntervals.size() > 0) {
			ActiveInterval activeInterval = (ActiveInterval) activeIntervals.firstElement();
			elementEndTime = activeInterval.getSceneEndTime();
		}
	}
	// If we didn't get a begin time from the element, which is expected.
	// Then get it from its parent/ancestor timing element.
	// Find the parent/ancestor timing element
	TreeElement treeElement = timingElement.getParent();
	while (null != treeElement && !(treeElement instanceof TimingElement)) {
		treeElement = treeElement.getParent();
	}
	if (null != treeElement) {
		TimingElement parent = (TimingElement) treeElement;

		// Now get the begin time for the parent
		lifeCycles = timeScheduler_.getLifeCycles(parent.getIndex());
		if (lifeCycles.size() > 0) {
			LifeCycle lifeCycle = (LifeCycle) lifeCycles.firstElement();
			Vector activeIntervals = lifeCycle.getActiveIntervals();
			if (activeIntervals.size() > 0) {
				ActiveInterval activeInterval = (ActiveInterval) activeIntervals.firstElement();
				elementEndTime = activeInterval.getSceneEndTime();
			}
		}
	}

	// return the first time
	if (null == sensorEndTime)
		return elementEndTime;
	if (null == elementEndTime)
		return sensorEndTime;
	if (sensorEndTime.greaterThan(elementEndTime))
		return sensorEndTime;
	else
		return elementEndTime;
}
/*********************************************************************** 
 * Get the ending scene time for the end event
 * The end event can occur across some range of times, usually this is the duration of the
 * element which is clicked on.  Thus the range of time is often the active duration of that
 * element.  This routine is returns the latest time in which the end event can occur.  
 * It might be indefinite, thus it might never end.
 * 20 Apr 2001 - Comment out this method since it end event isn't supported.
 *               Logic kept in as reference.
 **********************************************************************/
public TimingAttribute getEndSceneEndTime(Integer convertTreeIndex) {
	//       if (!isEndEvent(convertTreeIndex))
	//       	return null;
	//
	//       // Get the convert tree object
	//       Object object = convertIndexTable_.get(convertTreeIndex);
	//       if (null == object || !(object instanceof TimingElement))
	//       	return null;
	//       TimingElement timingElement = (TimingElement) object;
	//
	//       // Get the convert tree object which causes the begin event
	//       TimingAttribute time = ((TimingElement) object).getXmtoAttrEnd();
	//       if (null == time)
	//       	time = ((TimingElement) object).getXmtoAttrBegin();
	//       switch (time.getTimeValueType()) {
	//       	case TimingAttribute.EVENT_VALUE :
	//       		{
	//       			String id = time.getIdValue();
	//       			object = xmtoIdTable_.get(id);
	//       			if (null == object || !(object instanceof TimingElement))
	//       				return null;
	//       			TimingElement sensorTimingElement = (TimingElement) object;
	//       			// Get the starting of it as the earliest time this object can begin.
	//       			Integer sensorIndex = sensorTimingElement.getIndex();
	//       			if (resolveModel_.isEndResolved(sensorIndex)) {
	//       				TimingAttribute sensorTime = resolveModel_.getSceneEndTime(sensorIndex);
	//       				return sensorTime;
	//       			}
	//       			break;
	//       		}
	//       	case TimingAttribute.MULTI_VALUE :
	//       		{
	//       			for (Enumeration enum = time.getMultiTimingAttributes().elements(); enum.hasMoreElements();) {
	//       				TimingAttribute endTime = (TimingAttribute) enum.nextElement();
	//       				switch (endTime.getTimeValueType()) {
	//       					case TimingAttribute.EVENT_VALUE :
	//       						String id = endTime.getIdValue();
	//       						object = xmtoIdTable_.get(id);
	//       						if (null == object || !(object instanceof TimingElement))
	//       							return null;
	//       						TimingElement sensorTimingElement = (TimingElement) object;
	//       						// Get the starting of it as the earliest time this object can begin.
	//       						Integer sensorIndex = sensorTimingElement.getIndex();
	//       						if (resolveModel_.isEndResolved(sensorIndex)) {
	//       							TimingAttribute sensorTime = resolveModel_.getSceneEndTime(sensorIndex);
	//       							return sensorTime;
	//       						}
	//       						break;
	//       					default :
	//       						break;
	//       				}
	//       			}
	//       			break;
	//       		}
	//       	default :
	//       		break;
	//       }
	return null;
}
/***********************************************************************
 * This is the determines the event which kicks off the begining of the element.
 * This is the begin event, often a element.click time attribute.  This
 * often will contain a time offset from the event (element.click+40s).
 **********************************************************************/
public TimingAttribute getSceneBeginTime(Integer convertTreeIndex) {
	if (isBeginEvent(convertTreeIndex)) {
		//
		Object object = convertIndexTable_.get(convertTreeIndex);
		if (null != object && object instanceof TimingElement) {
			// If this is an event then return the event.
			return ((TimingElement) object).getXmtoAttrBegin();
		}
	}

	// Not an event so get the timing from the lifecycle timeinterval
	Vector lifeCycles = timeScheduler_.getLifeCycles(convertTreeIndex);
	if (lifeCycles.size() > 0) {
		LifeCycle lifeCycle = (LifeCycle) lifeCycles.firstElement();
		Vector activeIntervals = lifeCycle.getActiveIntervals();
		if (activeIntervals.size() > 0) {
			ActiveInterval activeInterval = (ActiveInterval) activeIntervals.firstElement();
			return activeInterval.getSceneBeginTime();
		}
	}

	//
	return null;
}
/*********************************************************************** 
 * This is the determines the event which kicks off the ending of the element.
 * This is the end event, often a element.click time attribute.  This
 * often will contain a time offset from the event (element.click+40s).
 * This can also be a multi value time attribute, often with a event value and
 * a clock value attributes.  The event is the standard click plus offset.  The
 * clock value would be often also be included because of the ending of a parent
 * element.  Thus it would end when a click is recieved or the &lt;par&gt; element ends.
 * 20 Apr 2001 - Comment out this method since it end event isn't supported.
 *               Logic kept in as reference.
 **********************************************************************/
public TimingAttribute getSceneEndTime(Integer convertTreeIndex) {
	//       if (!isEndEvent(convertTreeIndex))
	//       	return null;
	//
	//       //
	//       Object object = convertIndexTable_.get(convertTreeIndex);
	//       if (null == object || !(object instanceof TimingElement))
	//       	return null;
	//       TimingElement timingElement = (TimingElement) object;
	//
	//       // If there is an end attribute then use it.
	//       TimingAttribute time = timingElement.getXmtoAttrEnd();
	//       if (null != time)
	//       	return time;
	//
	//       // There isn't an event attribute then the begin should be an event and the end is an offset from it.
	//       time = timingElement.getXmtoAttrBegin();
	//       if (null == time)
	//       	return null;
	//
	//       // Get the duration.  Then end should be the begin + the duration
	//       TimingAttribute dur = timingElement.getXmtoAttrDur();
	//       if (null != dur)
	//       	return time.add(dur);
	//
	//       // There isn't a duration so try and figure one out.  If this isn't an element with an 
	//       // intrinsic time (video or sound) then the duration is 0.  If duration is 0 then the end
	//       // is the same as the begin.
	//       String src = null;
	//       if (timingElement instanceof video || timingElement instanceof audio) {
	//       	src = ((ExternalElement) timingElement).getXmtoAttrSrc();
	//       }
	//       if (null == src)
	//       	return time;
	//
	//       // we have a source file so see if we can deterime intrinsic duration from source file.
	//       MediaObjectInformation moi = new MediaObjectInformation(src);
	//       switch (moi.getMediaType()) {
	//       	case MediaObjectInformation.VIDEO :
	//       	case MediaObjectInformation.AUDIO :
	//       		long duration = moi.getDuration();
	//       		dur = new TimingAttribute(TimingAttribute.CLOCK_VALUE, duration);
	//       		return time.add(dur);
	//       }
	//
	//       //
	//       return time;
	return null;
}
/***********************************************************************
 **********************************************************************/
public boolean isBeginEvent(Integer convertTreeIndex) {
	Object object = convertIndexTable_.get(convertTreeIndex);
	if (null != object && object instanceof TimingElement) {
		TimingAttribute time = ((TimingElement) object).getXmtoAttrBegin();
		if (null == time)
			return false;
		switch (time.getTimeValueType()) {
			case TimingAttribute.EVENT_VALUE :
				return true;
			default :
				return false;
		}
	} else
		return false;
}
/***********************************************************************
 * is the end timing attribute an event or not.  If there isn't an
 * end attribute check to see if the begin is an event.
 * 20 Apr 2001 - Comment out this method since it end event isn't supported.
 *               Logic kept in as reference.
 **********************************************************************/
public boolean isEndEvent(Integer convertTreeIndex) {
	//	Object object = convertIndexTable_.get(convertTreeIndex);
	//	if (null != object && object instanceof TimingElement) {
	//		TimingAttribute time = ((TimingElement) object).getXmtoAttrEnd();
	//		if (null == time) {
	//			time = ((TimingElement) object).getXmtoAttrBegin();
	//			if (null == time)
	//				return false;
	//		}
	//		switch (time.getTimeValueType()) {
	//			case TimingAttribute.EVENT_VALUE :
	//			case TimingAttribute.MULTI_VALUE :
	//				return true;
	//			default :
	//				return false;
	//		}
	//	} else
	return false;
}
 /*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
 static public void setDebug(boolean debug) {
		debug_ = debug;
 } 
}
