/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  LifeCycle.java
//
//=============================================================================
package xmto2xmta;

import java.util.*;
import xmtutil.*;
import xmto2xmta.timegraph.*;

/*********************************************************************** 
 * Element life cycle.
 *
 * <p><b>History:</b><ul>
 * <li>15 Mar 2001  - Created file.
 * </ul><p>
 *********************************************************************/

public class LifeCycle {

	private Container timeContainer = null;
	private Vector activeIntervals = new Vector();
	private ELASid parentELASid = null;

	public LifeCycle(ContainerLife containerLife) {
		timeContainer = containerLife.getContainer();
		ELASid tmpELASid = containerLife.getParentELASid();
		if (tmpELASid.getElementIndex().intValue() > 0) {
			parentELASid = tmpELASid;
		}
		ActiveInterval activeInterval = null;
		Vector repeatTimes = new Vector();
		Vector simpleDurations = new Vector();
		Time fillEndTime = new Time(0);
		for (Enumeration i = containerLife.enumerateIntervals(); i.hasMoreElements();) {
			LifeInterval lifeInterval = (LifeInterval) i.nextElement();
			if (lifeInterval.getType() == LifeInterval.ACTIVE) {
				if (activeInterval != null) {
					if (repeatTimes.size() > 0) {
						activeInterval.setRepeatTimes(repeatTimes);
					}
					activeInterval.setSimpleDurations(simpleDurations);
//					if (!fillEndTime.isEqual(new Time(0))) {
						activeInterval.setFillEndTime(convertFrom(fillEndTime));
//					}
					activeIntervals.add(activeInterval);
					repeatTimes = new Vector();
					simpleDurations = new Vector();
					fillEndTime = new Time(0);
				}
				activeInterval = new ActiveInterval(this);
				activeInterval.setComputedBeginTime(convertFrom(lifeInterval.getBeginTime()));				
				activeInterval.setObservedBeginTime(convertFrom(lifeInterval.getObservedBegin()));
				Time offset = Time.subtract(lifeInterval.getObservedBegin(), lifeInterval.getBeginTime());
				if (lifeInterval.getIntrinsicDur().isResolved()) {
					activeInterval.setIntrinsicDuration(convertFrom(lifeInterval.getIntrinsicDur()));
				} else {
					activeInterval.setIntrinsicDuration(null);
				}
				activeInterval.setFillType(0);
				activeInterval.setELASids(lifeInterval.getELASids());
				Time repeatBegin = new Time(lifeInterval.getBeginTime());
				Time activeDur = new Time(0);
				Time truncatedTime = new Time(0);
				int iteration = 0;
				for (Enumeration s = lifeInterval.getSimpleDurs().elements(); s.hasMoreElements();) {
					if (iteration > 0) {
						repeatTimes.add(convertFrom(repeatBegin));
					}
					Time simpleDur = (Time) s.nextElement();
					activeDur = Time.add(activeDur, simpleDur);
					repeatBegin = Time.add(repeatBegin, simpleDur);
					Time effectiveDur = simpleDur;
					if (iteration == 0) {
						effectiveDur = Time.subtract(simpleDur, offset);
					}
					simpleDurations.add(convertFrom(effectiveDur));
					iteration++;
					if (iteration == lifeInterval.getSimpleDurs().size()) {
						truncatedTime = Time.subtract(lifeInterval.getLastComputedSimpleDur(),
																simpleDur);
					}
				}
				fillEndTime = Time.add(lifeInterval.getBeginTime(), activeDur);
				activeDur = Time.subtract(activeDur, offset);
				activeInterval.setActiveDuration(convertFrom(activeDur));
				Time activeEnd = Time.add(lifeInterval.getObservedBegin(), activeDur);
				activeInterval.setActiveEnd(convertFrom(activeEnd));
				activeInterval.setComputedEndTime(convertFrom(Time.add(activeEnd, truncatedTime)));
				activeInterval.setRepeatTimes(repeatTimes);
			} else {	// it's a fill interval
				activeInterval.setFillType(lifeInterval.getType());
				fillEndTime = lifeInterval.getEndTime();
			}
		}
		if (activeInterval != null) {
			if (repeatTimes.size() > 0) {
				activeInterval.setRepeatTimes(repeatTimes);
			}
			activeInterval.setSimpleDurations(simpleDurations);
//			if (!fillEndTime.isEqual(new Time(0))) {
				activeInterval.setFillEndTime(convertFrom(fillEndTime));
//			}
			activeIntervals.add(activeInterval);
		}			
	}

	public Vector getActiveIntervals() {
		return activeIntervals;
	}

	public void setActiveIntervals(Vector activeIntervals) {
		this.activeIntervals = activeIntervals;
	}

	public Container getTimeContainer() {
		return timeContainer;
	}

	public ELASid getParentELASid() {
		return parentELASid;
	}

	/***********************************************************************
	 * Convert a Time object to a TimingAttribute object.
	 ***********************************************************************/
	public static TimingAttribute convertFrom(Time time) {
		TimingAttribute timingAttribute = new TimingAttribute(TimingAttribute.INDEFINITE_VALUE, 0);
		try {
			if (time.isDefinite()) {
				timingAttribute = new TimingAttribute(TimingAttribute.CLOCK_VALUE, time.getValue());
			}
		} catch (Exception e) {}
		return timingAttribute;
	}

}

