/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  TimeScheduler.java
// 
//=============================================================================
package xmto2xmta; 

import java.text.MessageFormat;
import java.util.*;
import java.io.PrintWriter;
import xmto2xmta.tree.*;
import xmtutil.*;
import xmto2xmta.resources.Xmto2XmtaRes;
import xmto2xmta.timegraph.*;
import java.lang.reflect.Method;

/*********************************************************************** 
 * Class to determine resolve the timing for the XMT-O tree.
 *
 * <p><b>History:</b><ul>
 * <li>27 Mar 2001 - Created file.
 * <li>04 Jun 2001 - Updated to latest XMTO spec.  background changed to backdrop
 * </ul><p>
 *********************************************************************/
public class TimeScheduler {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static public boolean debug_ = false;
	static public long PLAYTIME = 600000;	// def simulation time in ms
	static public final Integer ROOT_ID = new Integer(0);
	static private final String backdropClassName = "xmto2xmta.tree.backdrop";
	static private final String getXmtoAttrSrcMethodName = "getXmtoAttrSrcCanonical";

	/***********************************************************************
	 * Input structures
	 **********************************************************************/
	private Xmto2Xmta o2aConverter_ = null;
	private Hashtable xmtoIdTable_ = null;
	private Hashtable convertIndexTable_ = null;
	private ConvertElement treeRoot_ = null;
	private long playTime = TimeScheduler.PLAYTIME;

	private TimeGraph timeGraph = null;
	private PrintWriter writer = null;
	private PrintWriter errWriter = null;
	private Class backdropClass = null;
	private Method getXmtoAttrSrcMethod = null;
	private SimpleTimeGraph simpleTimeGraph = null;
	private boolean useSimpleTimeGraph = false;

	/***********************************************************************
	 * Resource file for the messages.
	 **********************************************************************/
	private ResourceBundle xmto2XmtaRes_ = ResourceBundle.getBundle("xmto2xmta.resources.Xmto2XmtaRes", Locale.getDefault());
	/*********************************************************************** 
	 * Create the Resolv Model. Resolve as many timings as possible.
	 * @param converter      converter with pointers to the convert tree, 
	 *                       the xmto id hash table and the convert tree index hash table.
	 * @param xmtoIds        table of the convert tree Elements.  The hash index is the xmto id 
	 *                       attribute.  Not all convert tree elements are contained in this table,
	 *                       for example the elements which don't have xmto ids.  Also this table
	 *                       may have elemnents removed if an element isn't supported or has an error
	 *                       which can't be converted.
	 * @param convertIndexes table of the convert tree Elements.  The hash index is the unique
	 *                       index for each convert element (like an id, but not to be confused with the 
	 *                       xmto id.  This table may have elemnents removed if an element isn't 
	 *                       supported or has an error which can't be converted.
	 * @param treeRoot       root of the convert tree.
	 *********************************************************************/
	public TimeScheduler(Xmto2Xmta converter, Hashtable xmtoIds, Hashtable convertIndexes, ConvertElement treeRoot) {
		this(converter, xmtoIds, convertIndexes, treeRoot, TimeScheduler.PLAYTIME);
	}

	public TimeScheduler(Xmto2Xmta converter, Hashtable xmtoIds, Hashtable convertIndexes, ConvertElement treeRoot, long playTime) {
		o2aConverter_ = converter;
		xmtoIdTable_ = xmtoIds;
		convertIndexTable_ = convertIndexes;
		treeRoot_ = treeRoot;
		this.playTime = playTime;
		writer = new PrintWriter(new Tracer(this, ErrorEvent.TRACE));
		errWriter = new PrintWriter(new Tracer(this, ErrorEvent.ERROR), true);
		try {
			backdropClass = Class.forName(backdropClassName);
			getXmtoAttrSrcMethod = backdropClass.getMethod(getXmtoAttrSrcMethodName, null);
		} catch (Exception e) {}	

		createTimeGraph();
		printLifeCycles();
		
		createSimpleTimeGraph();
		if (useSimpleTimeGraph) {
			printLifeCycles();
			printEventActiveIntervals();
		}

		writer.flush();
		errWriter.flush();
	}

	/**********************************************************************
	 * Set the default simulation play time in milliseconds
	*********************************************************************/
	public static void setPlayTime(long playTime) {
		PLAYTIME = playTime;
	}

	/***********************************************************************
	 * Get the life cycles for the specified element.
	 * @param convertTreeIndex the index of the element to be look up
	 * @return a Vector containing the life cycles of the element.
	 *         If this element is not resolved then an empty list is returned.
	 *         If this element doesn't have any life cycles then an empty
	 *         list is returned.
	 **********************************************************************/
	public Vector getLifeCycles(Integer convertTreeIndex) {
		Vector lifeCycles = new Vector();
		if (convertTreeIndex == null) return lifeCycles;

		// get the time container for the index number
		Object container = null;
		if (useSimpleTimeGraph) {
			container = simpleTimeGraph.getContainerByID(convertTreeIndex);
		} else {
			container = timeGraph.getContainerByID(convertTreeIndex);
		}
		if (container == null) {
			return lifeCycles;
		}

		// get the container's life history and create a vector of LifeCycle objects.
		Vector lifeHistory = null;
		if (useSimpleTimeGraph) {
			lifeHistory = ((SimpleTimeContainer) container).getLifeHistory();
		} else {
			lifeHistory = ((TimeContainer) container).getLifeHistory();
		}
		if (lifeHistory == null) {
			return lifeCycles;
		}
		for (Enumeration e = lifeHistory.elements(); e.hasMoreElements();) {
			ContainerLife containerLife = (ContainerLife) e.nextElement();
			LifeCycle lifeCycle = new LifeCycle(containerLife);
			lifeCycles.add(lifeCycle);
		}
		return lifeCycles;
	}

	/***********************************************************************
	 * Get the event interval for an element.
	 * @param convertTreeIndex the index of the element to be look up
	 * @return EventActiveInterval.
	 **********************************************************************/
	public EventActiveInterval getEventActiveInterval(Integer convertTreeIndex) {
		if (!useSimpleTimeGraph || convertTreeIndex == null || simpleTimeGraph == null) return null;
		// get the time container for the index number
		SimpleTimeContainer container = (SimpleTimeContainer) simpleTimeGraph.getContainerByID(convertTreeIndex);
		if (container == null) {
			return null;
		}
		EventActiveInterval eventInterval = null;

		TimingAttribute beginFixedTime = null;
		TimingAttribute beginEventInfo = null;
		TimingAttribute endFixedTime = null;
		TimingAttribute endEventInfo = null;

		Interval beginTime = container.getBeginTime();
		Interval endTime = container.getEndTime();
		if (beginTime.isResolved() && endTime.isResolved()) {
			if (beginTime.point()) {
				beginFixedTime = convert(beginTime.getBeginTime());
			} else {
				beginFixedTime = convert(container.getBeginConstrainTime());
				Condition condition = container.getInheritedBeginCondition();
				beginEventInfo = createEventInfo(beginTime, condition);
			}
			if (endTime.point()) {
				endFixedTime = convert(endTime.getBeginTime());
			} else {
				endFixedTime = convert(container.getEndConstrainTime());
				Condition condition = container.getInheritedEndCondition();
				endEventInfo = createEventInfo(endTime, condition);
			}
			eventInterval = new EventActiveInterval();
			eventInterval.setObservedBeginTime(beginFixedTime);
			eventInterval.setFillEndTime(endFixedTime);
			eventInterval.setBeginEvent(beginEventInfo);
			eventInterval.setEndEvent(endEventInfo);
			eventInterval.setELASid(container.getELASid());
			eventInterval.setSimpleDuration(convert(container.getSimpleDur()));
			SimpleTimeContainer parent = (SimpleTimeContainer) container.getParent();
			if (parent != null) {
				eventInterval.setParentELASid(parent.getELASid());
			}
		}
		return eventInterval;
	}

	private TimingAttribute createEventInfo(Interval interval, Condition condition) {
		if (interval == null || condition == null || !interval.isResolved() ||
				interval.point()) return null;
		Time beginTime = interval.getBeginTime();
		Time endTime = interval.getEndTime();
		Vector timingAttributes = new Vector();
		timingAttributes.add(convert(beginTime));
		timingAttributes.add(convert(endTime));
		TimingAttribute eventAttribute = convertEventCondition(condition);
		if (eventAttribute == null) return null;
		timingAttributes.add(eventAttribute);
		return new TimingAttribute(timingAttributes);
	}

	public void printEventActiveIntervals() {
		writer.println();
		writer.println("printing event intervals");
		for (Enumeration e = convertIndexTable_.keys(); e.hasMoreElements();) {
			Integer i = (Integer) e.nextElement();
			writer.println();
			writer.println("getting event active interval for element "+i);
			EventActiveInterval eventInterval = getEventActiveInterval(i);
			if (eventInterval == null) {
				writer.println("no event active interval for element "+i);
				continue;
			}
			writer.println("simple dur="+eventInterval.getSimpleDuration());
			writer.println("observed begin="+eventInterval.getObservedBeginTime());
			writer.println("fill end time="+eventInterval.getFillEndTime());
			writer.println("begin event="+eventInterval.getBeginEvent());
			writer.println("end event="+eventInterval.getEndEvent());
			ELASid elasid = eventInterval.getELASid();
			writer.print("elasid="+elasid);
			ELASid parentELASid = eventInterval.getParentELASid();
			writer.print(",  parent elasid="+parentELASid);
			writer.println();
			writer.println("fill type="+eventInterval.getFillType());
		}
	}			

	/*********************************************************************** 
	 * Enable debugging information to be printed to the console. 
	 *
	 * @param debug     sets debug information is printed
	 **********************************************************************/
	static public void setDebug(boolean debug) {
		debug_ = debug;
	}

	static public boolean getDebug() {
		return debug_;
	}

	private void createTimeGraph() {
		// The root of the convert tree is a XMT-O element.
		// Find the body of the document, which should be one
		// of the immediate children.

		writer.println("creating timegraph");
		timeGraph = new TimeGraph(writer, errWriter);
		TimeContainer timeGraphRoot = new ParTimeContainer(timeGraph, ROOT_ID, null);
		timeGraph.setRoot(timeGraphRoot);
		timeGraph.setContainerID(ROOT_ID, timeGraphRoot);
		writer.println("created root, loading the rest of time graph");
		loadTimeGraph(timeGraphRoot, treeRoot_);
		writer.println("initializing time graph");
		timeGraph.initialize();
		writer.println("playing time graph, max simulation time(ms)="+playTime);
		timeGraph.play(new Time(playTime));
		
	} // createTimeGraph()

	private void loadTimeGraph(TimeContainer timeGraphParent, ConvertElement treeParent) {
        writer.println("tree element index="+treeParent.getIndex()+", class="+treeParent.getClass().getName());
		for (Enumeration e = treeParent.getChildren().elements(); e.hasMoreElements(); ) {
			ConvertElement treeChild = (ConvertElement) e.nextElement();
			if (treeChild instanceof TimingElement) {
				Integer treeChildIndex = treeChild.getIndex();
				TimeContainer timeGraphChild = null;
				if (treeChild instanceof ParContainable) {
					if (treeChild instanceof MediaElement) {
						timeGraphChild = new MediaTimeContainer(timeGraph, treeChildIndex, timeGraphParent);
					} else {
						timeGraphChild = new ParTimeContainer(timeGraph, treeChildIndex, timeGraphParent);
					}
				} else if (treeChild instanceof SeqContainable) {
					timeGraphChild = new SeqTimeContainer(timeGraph, treeChildIndex, timeGraphParent);
				} else if (treeChild instanceof AnimationElement) {
					timeGraphChild = new AnimationTimeContainer(timeGraph, treeChildIndex, timeGraphParent);
				} else {  // assume media time container (or simple timed media)
					timeGraphChild = new MediaTimeContainer(timeGraph, treeChildIndex, timeGraphParent);
				}
				timeGraphParent.addChild(timeGraphChild);
				timeGraph.setContainerID(treeChildIndex, timeGraphChild);
				loadTimeContainerAttributes((TimingElement) treeChild, timeGraphChild, timeGraphParent);
				loadTimeGraph(timeGraphChild, treeChild);
			} else {  // non-timing element??
				writer.println("non-timing element found");
				loadTimeGraph(timeGraphParent, treeChild);
			}
		}
	}  
    
	private void loadTimeContainerAttributes(TimingElement convertElement,
															TimeContainer graphElement,
															TimeContainer graphParent) {

		TimingAttribute beginAttr = convertElement.getXmtoAttrBegin();
		Vector conditions = createConditions(beginAttr, true,
												graphElement, graphParent);
		graphElement.setBeginConditions(conditions);
		TimingAttribute endAttr = convertElement.getXmtoAttrEnd();
		conditions = createConditions(endAttr, false,
												graphElement, graphParent);
		graphElement.setEndConditions(conditions);

		TimingAttribute durAttr = convertElement.getXmtoAttrDur();
		if (durAttr != null) {
			int durType = TimeContainer.DURTYPE_UNSPECIFIED;
			Time dur = null;
			switch (durAttr.getTimeValueType()) {
				case TimingAttribute.INDEFINITE_VALUE:
					durType = TimeContainer.DURTYPE_TIMEVALUE;
					dur = new Time(Time.INDEFINITE);
					break;
				case TimingAttribute.CLOCK_VALUE:
					durType = TimeContainer.DURTYPE_TIMEVALUE;
					dur = new Time(durAttr.getClockValue());
					break;
				case TimingAttribute.MEDIA_VALUE:
					durType = TimeContainer.DURTYPE_MEDIA;
					dur = new Time();
					break;
			}
			if (durType != TimeContainer.DURTYPE_UNSPECIFIED) {
				graphElement.setDur(durType, dur);
			}
		}

		try {
			String repeatCountString = convertElement.getXmtoAttrRepeatCount();
			if (repeatCountString.equals("indefinite")) {
				graphElement.setRepeatCountIndefinite();
			} else {
				float repeatCount = new Float(convertElement.getXmtoAttrRepeatCount()).floatValue();
				if (repeatCount > 0.0) {
					graphElement.setRepeatCount(repeatCount);
				} else {
					writer.println("invalid repeat count ignored="+repeatCount);
				}
			}
		} catch (Exception e) {}
	
		TimingAttribute repeatDurAttr = convertElement.getXmtoAttrRepeatDur();
		if (repeatDurAttr != null) {
			Time repeatDur = null;
			switch (repeatDurAttr.getTimeValueType()) {
				case TimingAttribute.INDEFINITE_VALUE:
					repeatDur = new Time(Time.INDEFINITE);
					break;
				case TimingAttribute.CLOCK_VALUE:
					repeatDur = new Time(repeatDurAttr.getClockValue());
					break;
			}
			graphElement.setRepeatDur(repeatDur);
		}

		if (!(graphElement instanceof SeqTimeContainer)) {
			String endsyncString = convertElement.getXmtoAttrEndsync();
			if (endsyncString != null) {
				int endsync = -1;
				Integer endsyncID = new Integer(-1);
				if (endsyncString.equals("first")) {
					endsync = TimeContainer.ENDSYNC_FIRST;
				} else if (endsyncString.equals("last")) {
					endsync = TimeContainer.ENDSYNC_LAST;
				} else if (endsyncString.equals("all")) {
					endsync = TimeContainer.ENDSYNC_ALL;
				} else if (endsyncString.equals("media")) {
					endsync = TimeContainer.ENDSYNC_MEDIA;
				} else {
					try {
						endsyncID = new Integer(endsyncString);
						endsync = TimeContainer.ENDSYNC_ID;
					} catch (Exception e) {}
				}
				if (endsync != -1) {
					graphElement.setEndsyncRule(endsync, endsyncID);
				}
			}
	   }

		TimingAttribute minAttr = convertElement.getXmtoAttrMin();
		if (minAttr != null) {
			int minType = TimeContainer.MINMAX_TIMEVALUE;
			Time min = null;
			switch (minAttr.getTimeValueType()) {
				case TimingAttribute.CLOCK_VALUE:
					minType = TimeContainer.MINMAX_TIMEVALUE;
					min = new Time(minAttr.getClockValue());
					break;
				case TimingAttribute.MEDIA_VALUE:
					if (!(graphElement instanceof MediaTimeContainer)) break;
					minType = TimeContainer.MINMAX_MEDIA;
					min = new Time();
					break;
			}
			if (min != null) {
				graphElement.setMin(minType, min);
			}
		}

		TimingAttribute maxAttr = convertElement.getXmtoAttrMax();
		if (maxAttr != null) {
			int maxType = TimeContainer.MINMAX_TIMEVALUE;
			Time max = null;
			switch (maxAttr.getTimeValueType()) {
				case TimingAttribute.INDEFINITE_VALUE:
					maxType = TimeContainer.MINMAX_TIMEVALUE;
					max = new Time(Time.INDEFINITE);
					break;
				case TimingAttribute.CLOCK_VALUE:
					maxType = TimeContainer.MINMAX_TIMEVALUE;
					max = new Time(maxAttr.getClockValue());
					break;
				case TimingAttribute.MEDIA_VALUE:
					if (!(graphElement instanceof MediaTimeContainer)) break;
					maxType = TimeContainer.MINMAX_MEDIA;
					max = new Time();
					break;
			}
			if (max != null) {
				graphElement.setMax(maxType, max);
			}
		}

		String fillDefString = convertElement.getXmtoAttrFillDefault();
		if (fillDefString != null) {
			writer.println("got filldefault="+fillDefString+" for id="+graphElement.getID());
			int fillDef = TimeContainer.FILL_INHERIT;
			if (fillDefString.equals("remove")) {
				fillDef = TimeContainer.FILL_REMOVE;
			} else if (fillDefString.equals("freeze")) {
				fillDef = TimeContainer.FILL_FREEZE;
			} else if (fillDefString.equals("hold")) {
				fillDef = TimeContainer.FILL_HOLD;
			} else if (fillDefString.equals("auto")) {
				fillDef = TimeContainer.FILL_AUTO;
			} else if (fillDefString.equals("transition")) {
				fillDef = TimeContainer.FILL_TRANSITION;
			}
			if (fillDef != TimeContainer.FILL_INHERIT) {
				graphElement.setFillDefault(fillDef);
			}
		}

		String fillString = convertElement.getXmtoAttrFill();
		if (fillString != null) {
			writer.println("got fill="+fillString+" for id="+graphElement.getID());
			int fill = TimeContainer.FILL_DEFAULT;
			if (fillString.equals("remove")) {
				fill = TimeContainer.FILL_REMOVE;
			} else if (fillString.equals("freeze")) {
				fill = TimeContainer.FILL_FREEZE;
			} else if (fillString.equals("hold")) {
				fill = TimeContainer.FILL_HOLD;
			} else if (fillString.equals("auto")) {
				fill = TimeContainer.FILL_AUTO;
			} else if (fillString.equals("transition")) {
				fill = TimeContainer.FILL_TRANSITION;
			}
			if (fill != TimeContainer.FILL_DEFAULT) {
				graphElement.setFill(fill);
			}
		}

		String restartDefString = convertElement.getXmtoAttrRepeatRestartDefault();
		if (restartDefString != null) {
			int restartDef = TimeContainer.RESTART_INHERIT;
			if (restartDefString.equals("always")) {
				restartDef = TimeContainer.RESTART_ALWAYS;
			} else if (restartDefString.equals("whenNotActive")) {
				restartDef = TimeContainer.RESTART_WHENNOTACTIVE;
			} else if (restartDefString.equals("never")) {
				restartDef = TimeContainer.RESTART_NEVER;
			}
			if (restartDef != TimeContainer.RESTART_INHERIT) {
				graphElement.setRestartDefault(restartDef);
			}
		}
		String restartString = convertElement.getXmtoAttrRepeatRestart();
		if (restartString != null) {
			int restart = TimeContainer.RESTART_DEFAULT;
			if (restartString.equals("always")) {
				restart = TimeContainer.RESTART_ALWAYS;
			} else if (restartString.equals("whenNotActive")) {
				restart = TimeContainer.RESTART_WHENNOTACTIVE;
			} else if (restartString.equals("never")) {
				restart = TimeContainer.RESTART_NEVER;
			}
			if (restart != TimeContainer.RESTART_INHERIT) {
				graphElement.setRestart(restart);
			}
		}
		if (graphElement instanceof MediaTimeContainer) {   
		    String mediaSrc = null;
		    if (convertElement instanceof ExternalElement) {
			    mediaSrc = ((ExternalElement)convertElement).getXmtoAttrSrcCanonical();
			} else if ((backdropClass != null) && backdropClass.isInstance(convertElement)) {
				try {
			   	mediaSrc = (String) getXmtoAttrSrcMethod.invoke(convertElement, null);
				} catch (Exception e) {}
			}
			if (mediaSrc != null) {
			  	((MediaTimeContainer)graphElement).setMediaSrc(mediaSrc);
			}
		}
	}	

	private Vector createConditions(TimingAttribute attr,
							boolean begin,
							TimeContainer graphElement, TimeContainer graphParent) {
		Vector conditions = new Vector();
		if (attr != null) {
			Vector attrList = null;
			int attrType = attr.getTimeValueType();
			if (attrType == TimingAttribute.MULTI_VALUE) {
				attrList = attr.getMultiTimingAttributes();
			} else {
				attrList = new Vector();
				attrList.add(attr);
			}
			Integer myID = graphElement.getID();
			for (Enumeration e = attrList.elements(); e.hasMoreElements(); ) {
				attr = (TimingAttribute) e.nextElement();
				Condition condition = new Condition();
				int conditionType = -1;
				Integer syncbaseID = null;
				long clockValue = 0;
				int eventSymbol = 0;
				String idValue = null;
				boolean offsetCondition = false;
				switch (attr.getTimeValueType()) {
					case TimingAttribute.CLOCK_VALUE:
						clockValue = attr.getClockValue();
						if ((graphParent instanceof ParTimeContainer) ||
								(graphParent instanceof MediaTimeContainer)) {
							conditionType = Condition.SYNCBASE_BEGIN;
							syncbaseID = graphParent.getID();
							offsetCondition = true;
						} else if (graphParent instanceof SeqTimeContainer) {
							Container prevSibling = graphParent.getPrevChild(graphElement);
							if (prevSibling != null) {
								conditionType = Condition.SYNCBASE_END;
								syncbaseID = prevSibling.getID();
							} else {
								conditionType = Condition.SYNCBASE_BEGIN;
								syncbaseID = graphParent.getID();
								offsetCondition = true;
							} 
						}
						break;
					case TimingAttribute.SYNCBASE_VALUE:
						clockValue = attr.getClockValue();
						idValue = attr.getIdValue();
						if (idValue != null) {
							ConvertElement syncbaseElement = (ConvertElement) xmtoIdTable_.get(idValue);
							if (syncbaseElement != null) {
								syncbaseID = syncbaseElement.getIndex();
								if (begin && myID.equals(syncbaseID)) {
									writer.println(myID+" begin syncbase cannot be itself");
								} else {
									if (attr.getTimeSymbol() == TimingAttribute.TIME_SYMBOL_BEGIN) {
										conditionType = Condition.SYNCBASE_BEGIN;
									} else {
										if (!begin && myID.equals(syncbaseID)) {
											writer.println(myID+" end syncbase cannot be end of itself");
										} else {
											conditionType = Condition.SYNCBASE_END;
										}
									}
								}
							}
						}
						break;
					case TimingAttribute.SYNC_TO_PREV_VALUE:
						clockValue = attr.getClockValue();
						Container prevSibling = graphParent.getPrevChild(graphElement);
						if (prevSibling != null) {
							syncbaseID = prevSibling.getID();
							if (attr.getTimeSymbol() == TimingAttribute.TIME_SYMBOL_BEGIN) {
								conditionType = Condition.SYNCBASE_BEGIN;
							} else {
								conditionType = Condition.SYNCBASE_END;
							}
						} else {
							conditionType = Condition.SYNCBASE_BEGIN;
							syncbaseID = graphParent.getID();
						} 
						break;
				    case TimingAttribute.EVENT_VALUE:
						 clockValue = attr.getClockValue();
						 idValue = attr.getIdValue();
						 if (idValue != null) {
							 ConvertElement syncbaseElement = (ConvertElement) xmtoIdTable_.get(idValue);
						 	 if (syncbaseElement != null) {
						 		 syncbaseID = syncbaseElement.getIndex();
						 	 } else {
						 		 writer.println(myID+" undefined element ID reference in condition");
						 		 break;
							 }
						  } else {
								syncbaseID = myID;
						  }
				        eventSymbol = attr.getEventSymbol();
				        switch (eventSymbol) {
				            case TimingAttribute.SYMBOL_EVENT_BEGIN_EVENT:
									 if (begin && myID.equals(syncbaseID)) {
										 writer.println(myID+" begin event base cannot be itself");
									 } else {
				               	conditionType = Condition.BEGIN_EVENT;
									 }
				                break;
				            case TimingAttribute.SYMBOL_EVENT_END_EVENT:
									 if (myID.equals(syncbaseID)) {
										 writer.println(myID+" begin or end event base cannot be end of itself");
									 } else {
				               	conditionType = Condition.END_EVENT;
				               	clockValue = attr.getClockValue();
									 }
				                break;
				            case TimingAttribute.SYMBOL_EVENT_REPEAT_EVENT:
									 if (myID.equals(syncbaseID)) {
										 writer.println(myID+" begin or end event base cannot be based on a repeat of itself");
									 } else {
				               	conditionType = Condition.REPEAT;
				               	 clockValue = attr.getClockValue();
									 }
				                break;
				            case TimingAttribute.SYMBOL_EVENT_CLICK:
				            case TimingAttribute.SYMBOL_EVENT_ACTIVATE_EVENT:
									 if (begin && myID.equals(syncbaseID)) {
										 writer.println(myID+" begin event base cannot be on activation of itself");
									 } else {
				               	 conditionType = Condition.ACTIVATE_EVENT;
				               	clockValue = attr.getClockValue();
									 }
				                break;
				        }
				        clockValue = attr.getClockValue();
				        break;
				    case TimingAttribute.MEDIA_MARKER_VALUE:
				        conditionType = Condition.MEDIAMARKER;
				        break;
					case TimingAttribute.INDEFINITE_VALUE:
						if (begin) break;
						conditionType = Condition.INDEFINITE;
						break;
				}
				if (conditionType != -1) {
					condition.setType(conditionType);
					condition.setSyncbaseID(syncbaseID);
					condition.setOffsetValue(clockValue);
					condition.setOffsetCondition(offsetCondition);
					conditions.add(condition);
				}
			}
		}
		if (begin && (conditions.size() == 0)) {
			int conditionType = -1;
			Integer syncbaseID = null;
			boolean offsetCondition = false;
			if ((graphParent instanceof ParTimeContainer) ||
					(graphParent instanceof MediaTimeContainer)) {
				conditionType = Condition.SYNCBASE_BEGIN;
				syncbaseID = graphParent.getID();
				offsetCondition = true;
			} else if (graphParent instanceof SeqTimeContainer) {
				Container prevSibling = graphParent.getPrevChild(graphElement);
				if (prevSibling != null) {
					conditionType = Condition.SYNCBASE_END;
					syncbaseID = prevSibling.getID();
				} else {
					conditionType = Condition.SYNCBASE_BEGIN;
					syncbaseID = graphParent.getID();
					offsetCondition = true;
				} 
			}
			Condition condition = new Condition();
			condition.setType(conditionType);
			condition.setSyncbaseID(syncbaseID);
			condition.setOffsetCondition(offsetCondition);
			conditions.add(condition);
		}

		return conditions;
	}								

	private void createSimpleTimeGraph() {
		writer.println("creating simple timegraph");
		simpleTimeGraph = new SimpleTimeGraph(writer, errWriter);
		simpleTimeGraph.initialize(timeGraph);
		if (simpleTimeGraph.isSupported()) {
			useSimpleTimeGraph = true;
			simpleTimeGraph.play();
		}
	}

	private TimingAttribute convert(Time time) {
		TimingAttribute timingAttribute = new TimingAttribute(TimingAttribute.INVALID, 0);
		if (time == null || !time.isResolved()) return timingAttribute;
		try {
			if (time.isDefinite()) {
				timingAttribute = new TimingAttribute(TimingAttribute.CLOCK_VALUE, time.getValue());
			} else {
				timingAttribute = new TimingAttribute(TimingAttribute.INDEFINITE_VALUE, 0);
			}
		} catch (Exception e) {
				timingAttribute = new TimingAttribute(TimingAttribute.INDEFINITE_VALUE, 0);
		}
		return timingAttribute;
	}

	private TimingAttribute convertEventCondition(Condition condition) {
		if (condition == null || condition.getType() != Condition.ACTIVATE_EVENT) return null;
		TimingAttribute timingAttribute = null;
		int symbol = TimingAttribute.SYMBOL_EVENT_CLICK;
		int valueType = TimingAttribute.EVENT_VALUE;
		long clockValue = condition.getOffsetValue();
		if (convertIndexTable_ == null) return null;
		ConvertElement convertElement = (ConvertElement) convertIndexTable_.get(condition.getSyncbaseID());
		if (convertElement == null) return null;
		String xmtoID = convertElement.getXmtoAttrId();
		if (xmtoID == null) return null;		
		return new TimingAttribute(null, null, valueType, clockValue, xmtoID, symbol);
	}

	/**********************************************************************
	*********************************************************************/
	public void fireErrorEvent(int errorLevel, int errorId, String message) {
		o2aConverter_.fireErrorEvent(new ErrorEvent(o2aConverter_, errorLevel, errorId, message));
	}

	private void printLifeCycles() {
		for (Enumeration e = convertIndexTable_.keys(); e.hasMoreElements();) {
			Integer i = (Integer) e.nextElement();
			writer.println();
			writer.println("getting element "+i);
			Vector lifeCycles = getLifeCycles(i);
			writer.println("number of life cycles="+lifeCycles.size());
			int timeCount = 0;
			for (Enumeration l = lifeCycles.elements(); l.hasMoreElements();) {
				LifeCycle lifeCycle = (LifeCycle) l.nextElement();
				writer.println("life cycle #"+timeCount);
				writer.println("parent ELASid="+lifeCycle.getParentELASid());
				Vector activeIntervals = lifeCycle.getActiveIntervals();
				writer.println("number of active intervals="+activeIntervals.size());
				int lifeCount = 0;
				for (Enumeration c = activeIntervals.elements(); c.hasMoreElements();) {
					ActiveInterval activeInterval = (ActiveInterval) c.nextElement();
					writer.println("active interval #"+lifeCount);
					writer.println("active dur="+activeInterval.getActiveDuration());
					writer.println("active end time="+activeInterval.getActiveEnd());
					writer.println("computed end time="+activeInterval.getComputedEndTime());
					writer.println("computed begin="+activeInterval.getComputedBeginTime());
					writer.println("observed begin="+activeInterval.getObservedBeginTime());
					writer.println("scene begin="+activeInterval.getSceneBeginTime());
					writer.println("scene end="+activeInterval.getSceneEndTime());
					writer.println("intrinsic dur(as is)="+activeInterval.getIntrinsicDuration(true));
					writer.println("intrinsic dur(computed)="+activeInterval.getIntrinsicDuration());
					Vector repeatTimes = activeInterval.getRepeatTimes();
					writer.println("number of repeats="+repeatTimes.size());
					writer.println("time container ID="+activeInterval.getLifeCycle().getTimeContainer().getID());
					int repeatCount = 1;
					for (Enumeration r = repeatTimes.elements(); r.hasMoreElements();) {
						TimingAttribute repeatTime = (TimingAttribute) r.nextElement();
						writer.println("repeat #"+repeatCount+" time="+repeatTime);
						repeatCount++;
					}
					Vector simpleDurations = activeInterval.getSimpleDurations();
					writer.println("number of simpleDurations="+simpleDurations.size());
					writer.print("simple durations=");
					for (Enumeration sd = simpleDurations.elements(); sd.hasMoreElements();) {
						TimingAttribute simpleDuration = (TimingAttribute) sd.nextElement();
						writer.print(simpleDuration+", ");
					}
					writer.println();
					writer.print("elasids=");
					Vector elasids = activeInterval.getELASids();
					for (Enumeration id = elasids.elements(); id.hasMoreElements();) {
						ELASid elasid = (ELASid) id.nextElement();
						writer.print(elasid+", ");
					}
					writer.println();
					writer.println("fill type="+activeInterval.getFillType());
					writer.println("fill end time="+activeInterval.getFillEndTime());
					lifeCount++;
				}
				timeCount++;
			}
		}
	}

}
