/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  GroupMapper.java
//
//=============================================================================
package xmto2xmta.mapgroup;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import xmto2xmta.ErrorEvent;
import xmto2xmta.MapInterface;
import xmto2xmta.Xmto2Xmta;
import xmto2xmta.tree.ConvertElement;

/*********************************************************************** 
 * Class to determine resolve the timing for the XMT-O tree.
 *
 * <p><b>History:</b><ul>
 * <li>27 Mar 2001 - Created file.
 * </ul><p>
 *
 *********************************************************************/
public class GroupMapper {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static public boolean debug_ = false;

	/***********************************************************************
	 * Input structures
	 **********************************************************************/
	private Xmto2Xmta xmto2xmta_ = null;
	private Hashtable xmtoIdTable_ = null;
	private Hashtable convertIndexTable_ = null;
	private ConvertElement treeRoot_ = null;

	/***********************************************************************
	 * Hash table containing the layout mapping classes which correspond to 
	 * the XMT-O tree elements elements.
	 * The prime number 89 was chosen as the hash table initialize size because it was
	 * mentioned as a better starting hash size for optimization than the default 101.
	 * The layoutTable_ table is composed of (key, data)
	 *   The key is the XMT-O tree element index.
	 *   The data is the layout mapping class.
	 **********************************************************************/
	private Hashtable mpTable_ = new Hashtable(89);

	/***********************************************************************
	 * Used to create Lists for class mappings
	 **********************************************************************/
	private String treeClassesPackage_ = "xmto2xmta.tree.";
	private String mpClassesPackage_ = "xmto2xmta.mapgroup.";
	private Vector treeClasses_ = new Vector();
	private Vector mpClasses_ = new Vector();
	private String[] treeClassNames_ = {
	/* */ "body",
	/* */ "group",
	/* */ "region",
	/* */};
	private String[] mpClassNames_ = {
	/* */ "GrpBody",
	/* */ "GrpGroup",
	/* */ "GrpRegion",
	/* */};
/*********************************************************************** 
 * Create the Resolv Model. Resolve as many timings as possible.
 * @param converter      converter with pointers to the convert tree, 
 *                       the xmto id hash table and the convert tree index hash table.
 * @param xmtoIds        table of the convert tree Elements.  The hash index is the xmto id 
 *                       attribute.  Not all convert tree elements are contained in this table,
 *                       for example the elements which don't have xmto ids.  Also this table
 *                       may have elemnents removed if an element isn't supported or has an error
 *                       which can't be converted.
 * @param convertIndexes table of the convert tree Elements.  The hash index is the unique
 *                       index for each convert element (like an id, but not to be confused with the 
 *                       xmto id.  This table may have elemnents removed if an element isn't 
 *                       supported or has an error which can't be converted.
 * @param treeRoot       root of the convert tree.
 *********************************************************************/
public GroupMapper(Xmto2Xmta xmto2xmta, Hashtable xmtoIds, Hashtable convertIndexes, ConvertElement treeRoot) {
	xmto2xmta_ = xmto2xmta;
	xmtoIdTable_ = xmtoIds;
	convertIndexTable_ = convertIndexes;
	treeRoot_ = treeRoot;

	//
	initializeClassLists();
	
	//
	doMapping();
}
/**********************************************************************
*********************************************************************/
private void doMapping() {
	visitElements(treeRoot_);
}
/**********************************************************************
 *********************************************************************/
public void fireErrorEvent(int errorLevel, int errorId, String message) {
	xmto2xmta_.fireErrorEvent(new ErrorEvent(xmto2xmta_, errorLevel, errorId, message));
}
/*******************************************************************
 *******************************************************************/
private void initializeClassLists() {
	if (treeClassNames_.length != mpClassNames_.length)
		fireErrorEvent(ErrorEvent.FATAL_ERROR, 0, "GroupMapper.initializeClassLists() - Internal error mapping lists different sizes");

	//
	for (int i = 0; i < treeClassNames_.length; i++) {
		Class treeClass = null;
		Class mpClass = null;

		// Try and load the classes	
		try {
			treeClass = Class.forName(treeClassesPackage_ + treeClassNames_[i]);
		} catch (Exception e) {
			if (debug_)
				fireErrorEvent(ErrorEvent.TRACE, 0, "GroupMapper.initializeClassLists() - Unable to create class=" + (treeClassesPackage_ + treeClassNames_[i]) + " exception=" + e);
		}
		try {
			mpClass = Class.forName(mpClassesPackage_ + mpClassNames_[i]);
		} catch (Exception e) {
			if (debug_)
				fireErrorEvent(ErrorEvent.TRACE, 0, "GroupMapper.initializeClassLists() - Unable to create class=" + (mpClassesPackage_ + mpClassNames_[i]) + " exception=" + e);
		}

		// If there were no problems then add the files to the list
		if (null != treeClass && null != mpClass) {
			treeClasses_.add(treeClass);
			mpClasses_.add(mpClass);
		}
	}
	if (debug_) {
		for (int i = 0; i < treeClasses_.size(); i++) {
			Class treeClass = (Class) treeClasses_.get(i);
			Class mpClass = (Class) mpClasses_.get(i);
			fireErrorEvent(ErrorEvent.TRACE, 0, "GroupMapper.initializeClassLists() - Created treeClass[" + i + "]=" + treeClass + "" + " mpClass[" + i + "]=" + mpClass);
		}
	}
}
/*******************************************************************
 *******************************************************************/
private void mapClass(ConvertElement element) {
	for (int i = 0; i < treeClasses_.size(); i++) {
		Class treeClass = (Class) treeClasses_.get(i);
		if (treeClass.isInstance(element)) {
			Class mpClass = (Class) mpClasses_.get(i);
			try {
				// Create the class 
				MapInterface mpObject = (MapInterface) mpClass.newInstance();
				mpObject.init(xmto2xmta_, element);

				// Add the class to the list of layout classes.
				mpTable_.put(element.getIndex(), mpObject);

				// Do the layout on the class
				mpObject.doMapping();
				return;
			} catch (Exception e) {
				if (debug_) {
					fireErrorEvent(ErrorEvent.TRACE, 0, "GroupMapper.mapClass() - unable to create class=" + mpClass.getName() + " exception=" + e);
					e.printStackTrace();
				}
			}
		}
	}
	if (debug_)
		fireErrorEvent(ErrorEvent.TRACE, 0, "GroupMapper.mapClass() - unable to find class for node=" + element.getClass().getName());
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
/***********************************************************************
 * Call Validate on the convert element and traverse through the children.
 **********************************************************************/
private void visitElements(ConvertElement element) {
	mapClass(element);

	// Now go through the validation on the children
	Vector clone = element.getChildren();
	for (Enumeration children = clone.elements(); children.hasMoreElements();) {
		ConvertElement child = (ConvertElement) children.nextElement();
		visitElements(child);
	}
}
}
