/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  GrpRegion.java
//
//=============================================================================
package xmto2xmta.mapgroup;

import xmto2xmta.tree.TreeElement;
import xmto2xmta.tree.RegionContainerElement;
import xmto2xmta.tree.region;
import xmto2xmta.tree.ConvertElement;
import xmto2xmta.timegraph.ELASid;
import java.text.MessageFormat;
import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import xmtutil.DOMXMTA;
import xmtutil.DOMXMTO;
import xmtutil.DOMXMTOParameter;
import xmto2xmta.ErrorEvent;
import xmto2xmta.IdManager;
import xmto2xmta.MapRoot;
import xmto2xmta.Xmto2Xmta;
import xmto2xmta.resources.Xmto2XmtaRes;
import xmtutil.DOMXMTAParameter;

/*********************************************************************** 
 * Class to handle the conversion of the XMT-O element.  The class
 * name is XMTO instead of XMT-O since java doesn't allow a dash character 
 * in the class name.
 *
 * <p><b>History:</b><ul>
 * <li>13 Dec 2000 - Created file.
 * <li>30 Apr 2001 - Add switch to generate new format for XMT-A files.
 * </ul><p>
 *********************************************************************/
public class GrpRegion extends MapRoot  {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	/*********************************************************************** 
	 **********************************************************************/
	private region treeRegion_ = null;

	/*********************************************************************** 
	 **********************************************************************/
	private ResourceBundle xmto2XmtaRes_ = ResourceBundle.getBundle("xmto2xmta.resources.Xmto2XmtaRes", Locale.getDefault());
/***********************************************************************
 * Create the group infrastructure. 
 * The infrastructure contains the par node for the insertion.
 * The par node for the insertion contains the information for the element.  This also 
 * takes care of the position in the ordered group field (for z-order).
 * <pre>
 *   &lt;par begin="0"&gt;
 *     &lt;Insert atNode="???"&gt;
 *       &lt;Transform2D DEF="???"&gt;
 *         &lt;children&gt;
 *           &lt;Layer2D DEF="???" size="???"&gt;
 *             &lt;children&gt;
 *               &lt;OrderedGroup DEF="???"&gt;
 *                 &lt;children/&gt;
 *               &lt;/OrderedGroup&gt;
 *               &lt;OrderedGroup DEF="???"&gt;
 *                 &lt;children/&gt;
 *               &lt;/OrderedGroup&gt;
 *             &lt;/children&gt;
 *           &lt;/Layer2D&gt;
 *         &lt;/children&gt;
 *       &lt;/Transform2D&gt;
 *     &lt;/Insert&gt;
 *     &lt;Insert atField="order" atNode="???" position="END" value="???"/&gt;
 *   &lt;/par&gt;
 * </pre> 
 **********************************************************************/
protected boolean createInsertMapping() {
	//
	Document xmtaDoc = treeRegion_.getXmtaDoc();
	Hashtable xmtaIdTable = treeRegion_.getXmtaIds();

	// Find the Xmta parent group element which this node should be inserted into.
	RegionContainerElement xmtaParentElement = null;
	TreeElement parent = treeRegion_.getParent();
	while (null != parent) {
		if (parent instanceof RegionContainerElement) {
			xmtaParentElement = (RegionContainerElement) parent;
			// break out of the loop
			break;
		}
		parent = parent.getParent();
	}
	if (null == xmtaParentElement) {
		// There always should be at least a body so we should never get here.  
		// If we have then the user has previously ignored error messages.
		// Thus send another error message and continue on.
		fireErrorEvent(ErrorEvent.ERROR, 0, xmto2XmtaRes_.getString(Xmto2XmtaRes.MSG_ERROR_INVALID_DOCUMENT_STRUCTURE), treeRegion_.getXmtoElement());
		treeRegion_.remove();
		return false;
	}

	// Get the ordered group of the parent we are going to be inserted into.
	String parentLayoutOrderedGroupId = xmtaParentElement.getXmtaLayoutOrderedGroupId();

	// The par node for the insertion contains the top level group which will contain the
	// information for the element.  This also takes care of the position in the ordered group
	// field (for z-order).
	// <par begin="0">
	//   <Insert atNode="???">
	//     <Transform2D DEF="???">
	//       <children>
	//         <Layer2d DEF="???">
	//           <children>
	//             <OrderedGroup DEF="???">
	//               <children/>
	//             </OrderedGroup>
	//             <OrderedGroup DEF="???">
	//               <children/>
	//             </OrderedGroup>
	//           </children>
	//         </Layer2d>
	//       <children/>
	//     </Transform2D>
	//   </Insert>
	//   <Insert atField="order" atNode="???" position="END" value="???"/>
	// </par>

	// Create the par for the insert command. Don't place it in the XMT-A tree yet.
	Element parInsertNode = xmtaDoc.createElement(DOMXMTA.par);
	treeRegion_.addToParNode(parInsertNode);
	// Set the insert time.
	parInsertNode.setAttribute(DOMXMTA.begin, DOMXMTAParameter.time2Param(0));

	// Create insert Node 
	Element insertAtNode = xmtaDoc.createElement(DOMXMTA.Insert);
	insertAtNode.setAttribute(DOMXMTA.atNode, parentLayoutOrderedGroupId);
	parInsertNode.appendChild(insertAtNode);

	// Create transform2d to allow the object to be moved
	Element transform2D = xmtaDoc.createElement(DOMXMTA.Transform2D);
	String transform2dId = treeRegion_.getXmtaTransform2dId();
	transform2D.setAttribute(DOMXMTA.DEF, transform2dId);
	xmtaIdTable.put(transform2dId, transform2D);
	// Set the translation offset
	String translationAttribute = treeRegion_.getXmtoAttrTranslation();
	if (null != translationAttribute || !translationAttribute.equals(""))
		transform2D.setAttribute(DOMXMTA.translation, translationAttribute);
	insertAtNode.appendChild(transform2D);

	//
	Node transform2dChildren = transform2D.appendChild(xmtaDoc.createElement(DOMXMTA.children));

	// Create Layer2d to allow the children of the object to be clipped to the size of the object.
	Element layer2D = xmtaDoc.createElement(DOMXMTA.Layer2D);
	String layer2dId = treeRegion_.getXmtaLayer2dId();
	layer2D.setAttribute(DOMXMTA.DEF, layer2dId);
	xmtaIdTable.put(layer2dId, layer2D);
	// Set the translation offset
	String sizeAttribute = treeRegion_.getXmtoAttrSize();
	if (null != sizeAttribute || !sizeAttribute.equals(""))
		layer2D.setAttribute(DOMXMTA.size, sizeAttribute);
	transform2dChildren.appendChild(layer2D);

	//
	Node layer2dChildren = layer2D.appendChild(xmtaDoc.createElement(DOMXMTA.children));

	// Create the Ordered Group for the media in the scene
	Element mediaOrderedGroup = xmtaDoc.createElement(DOMXMTA.OrderedGroup);
	String mediaOrderedGroupId = treeRegion_.getXmtaOrderedGroupId();
	mediaOrderedGroup.setAttribute(DOMXMTA.DEF, mediaOrderedGroupId);
	xmtaIdTable.put(mediaOrderedGroupId, mediaOrderedGroup);
	layer2dChildren.appendChild(mediaOrderedGroup);

	//
	mediaOrderedGroup.appendChild(xmtaDoc.createElement(DOMXMTA.children));

	// Create the Ordered Group for sub regions
	Element layoutOrderedGroup = xmtaDoc.createElement(DOMXMTA.OrderedGroup);
	String layoutOrderedGroupId = treeRegion_.getXmtaLayoutOrderedGroupId();
	layoutOrderedGroup.setAttribute(DOMXMTA.DEF, layoutOrderedGroupId);
	xmtaIdTable.put(layoutOrderedGroupId, layoutOrderedGroup);
	layer2dChildren.appendChild(layoutOrderedGroup);

	//
	layoutOrderedGroup.appendChild(xmtaDoc.createElement(DOMXMTA.children));

	// Insert to modify the OrderedGroup's atField.
	// Create insert Field.  
	// Insert group isn't known yet so don't fill it in.
	// Also don't know the zOrder so don't fill it in.
	Element insertAtField = xmtaDoc.createElement(DOMXMTA.Insert);
	insertAtField.setAttribute(DOMXMTA.atNode, parentLayoutOrderedGroupId);
	insertAtField.setAttribute(DOMXMTA.atField, DOMXMTA.order);
	insertAtField.setAttribute(DOMXMTA.position, DOMXMTA.END);
	// Set the z-order (order field on insert atField tag)
	String orderAttribute = getOrderAttribute();
	insertAtField.setAttribute(DOMXMTA.value, orderAttribute);
	parInsertNode.appendChild(insertAtField);

	//
	return true;
}
/*******************************************************************
 *******************************************************************/
public void doMapping() {
	if (debug_)
		fireErrorEvent(ErrorEvent.TRACE, 0, "GrpRegion.doMapping() - entered convertElement=" + getConvertElement());
	boolean success = true;

	//
	success = createInsertMapping();

	//
	if (debug_)
		fireErrorEvent(ErrorEvent.TRACE, 0, "GrpRegion.doMapping() - exiting success=" + success);
}
/*******************************************************************
 *******************************************************************/
public String getOrderAttribute() {
	// Set the z-order (order field on insert atField tag)
	String orderAttribute = treeRegion_.getXmtoAttrZindex();
	if (null != orderAttribute && !orderAttribute.equals(""))
		return orderAttribute;

	// We don't have an order.  Thus in order to make sure it is on top of the stack give 
	// it a z-order index the same as it's previous sibling.	
	// First get the list of siblings. NOTE: the list includes this object.
	Vector siblings = treeRegion_.getParent().getChildren();

	// Find the index of this object
	int orderIndex = siblings.indexOf(treeRegion_);

	// go backward through the previous siblings looking for one the last one with a z-index
	for (int i = orderIndex - 1; i >= 0; i--) {
		Object sibling = siblings.elementAt(i);
		if (sibling instanceof region) {
			region siblingRegion = (region) sibling;
			String order = siblingRegion.getXmtoAttrZindex();
			if (null != order && !order.equals(""))
				return order;
		}
	}

	// Couldn't find a previous sibling with a z-order.  Check the subsequent siblings
	// if they have one.  If they do then set this object to the lowest z-index.
	// Go foward through the subsequent siblings looking for the one with the lowest z-index.
	int siblingsCount = siblings.size();
	Integer lowestZindex = null;
	for (int i = orderIndex + 1; i < siblingsCount; i++) {
		Object sibling = siblings.elementAt(i);
		if (sibling instanceof region) {
			region siblingRegion = (region) sibling;
			String order = siblingRegion.getXmtoAttrZindex();
			if (null != order && !order.equals("")) {
				Integer zindex = DOMXMTOParameter.param2Integer(order);
				if (null != zindex) {
					if (null == lowestZindex || 0 > zindex.compareTo(lowestZindex))
						lowestZindex = zindex;
				}
			}
		}
	}

	// if found a zindex for a sibling then return it.
	if (null != lowestZindex)
		return lowestZindex.toString();

	// aren't any z indexes so use a default value of 1
	return "1";
}
/*******************************************************************
 *******************************************************************/
public void init(Xmto2Xmta xmto2xmta, ConvertElement element) throws Exception {
	super.init(xmto2xmta, element);

	//
	if (!(element instanceof region)) {
		if (debug_)
			fireErrorEvent(ErrorEvent.TRACE, 0, "GrpRegion.init() - element=" + element + " is not an instance of region");
		throw new Exception("");
	}

	//
	treeRegion_ = (region) element;

	//
	if (debug_)
		fireErrorEvent(ErrorEvent.TRACE, 0, "GrpRegion.init() - treeRegion_=" + treeRegion_);
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
/***********************************************************************
 **********************************************************************/
public String toString() {
	return "<region>" + super.toString() + "</region>";
}
}
