/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  MedExternalElement.java
//
//=============================================================================
package xmto2xmta.mapmedia;

import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import xmtutil.MediaObjectInformation;
import xmto2xmta.ActiveInterval;
import xmto2xmta.ErrorEvent;
import xmto2xmta.EventModel;
import xmto2xmta.IdManager;
import xmto2xmta.LifeCycle;
import xmto2xmta.TimeScheduler;
import xmto2xmta.Xmto2Xmta;
import xmto2xmta.resources.ExternalElementRes;
import xmto2xmta.resources.Xmto2XmtaRes;
import xmto2xmta.timegraph.ELASid;
import xmto2xmta.EventActiveInterval;
import xmtutil.DOMXMTA;
import xmtutil.DOMXMTAParameter;
import xmtutil.DOMXMTO;
import xmtutil.TimingAttribute;
/*********************************************************************** 
 * Class to group together the various type media elements whose data is external to the XMT file.
 *
 * <p><b>History:</b><ul>
 * <li>13 Dec 2000 - Created file.
 * <li>09 Feb 2001 - Fix bug should support indefininte duration (end time)
 * <li>20 Jun 2001 - Remove support for old XMT-A format.
 * <li>28 Jun 2001 - Changed SLConfigDescriptor to use predefined option.
 * <li>28 May 2002 - odid url prefix is od:// in new spec.
 * </ul><p>
 *********************************************************************/
public abstract class MedExternalElement extends MedMediaElement {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	/*********************************************************************** 
	 **********************************************************************/
	static protected final String ODID_URL_PREFIX = "od://";
	static protected final String FILE_URL_PREFIX = "";
	// static protected final String FILE_URL_PREFIX = "file:///";

	/*********************************************************************** 
	 **********************************************************************/
	private xmto2xmta.tree.ExternalElement treeExternalElement_ = null;

	/*********************************************************************** 
	 **********************************************************************/
	private ResourceBundle xmto2XmtaRes_ = ResourceBundle.getBundle("xmto2xmta.resources.Xmto2XmtaRes", Locale.getDefault());
	private ResourceBundle externalElementRes_ = ResourceBundle.getBundle("xmto2xmta.resources.ExternalElementRes", Locale.getDefault());
/***********************************************************************
 * Create the media Element infrastructure.
 * The infrastructure contains the par node for the deletion.
 * The par for the deletion, deletes the top level group for the element.  It also
 * takes care of the position in the ordered group field (for z-order)
 * <pre>
 *   &lt;par begin="???"&gt;
 *     &lt;Delete atNode="???"/&gt;
 *     &lt;Delete atField="order" atNode="???"/&gt;
 *   &lt;/par&gt;
 * </pre>
 **********************************************************************/
protected boolean createDeleteMapping(ELASid elasId) {
	boolean success = super.createDeleteMapping(elasId);
	if (!success)
		return success;

	// JSB: TBD
	// for now ignore events
	//
	//	if (endResolved) {
	//	} else {
	//		// Not so easy way.  In this case we have an event so end with a conditional and
	//		// set route from touch sensor.
	//		success = timeMapEndEvent();
	//		if (!success)
	//			return success;
	//	}

	TimeScheduler timeScheduler = treeExternalElement_.getXmto2Xmta().getTimeScheduler();
	Vector lifeCycles = timeScheduler.getLifeCycles(elasId.getElementIndex());
	LifeCycle lifeCycle = (LifeCycle) lifeCycles.elementAt(elasId.getLifeCycleIndex().intValue());
	Vector activeIntervals = lifeCycle.getActiveIntervals();
	ActiveInterval activeInterval = (ActiveInterval) activeIntervals.elementAt(elasId.getIntervalIndex().intValue());
	TimingAttribute deleteTime = null;
	int sdIndex = elasId.getSimpleDurIndex().intValue();
	if (activeInterval.getELASids().size() - 1 == sdIndex)
		deleteTime = activeInterval.getFillEndTime();
	else
		deleteTime = (TimingAttribute) activeInterval.getRepeatTimes().elementAt(sdIndex);
	// If this simple duration doesn't end then we don't need to worry about deleting it.
	if (deleteTime.getTimeValueType() == TimingAttribute.INDEFINITE_VALUE)
		return true;
	else
		return generateDeleteMapping(elasId, deleteTime);
}

/***********************************************************************
 * Create the media Element infrastructure.
 * The infrastructure contains the par node for the deletion.
 * The par for the deletion, deletes the top level group for the element.  It also
 * takes care of the position in the ordered group field (for z-order)
 * <pre>
 *   &lt;par begin="???"&gt;
 *     &lt;Delete atNode="???"/&gt;
 *     &lt;Delete atField="order" atNode="???"/&gt;
 *   &lt;/par&gt;
 * </pre>
 **********************************************************************/
protected boolean createDeleteMapping(ELASid elasId, ELASid parentELASid, TimingAttribute deleteTime) {
	boolean success = super.createDeleteMapping(elasId, parentELASid, deleteTime);
	if (!success)
		return success;
	else
		return generateDeleteMapping(elasId, deleteTime);
}

private boolean generateDeleteMapping(ELASid elasId, TimingAttribute deleteTime) {
	Document xmtaDoc = getConvertElement().getXmtaDoc();

	// Nest the remove par under the par that creates the object. 
	// <par begin="???">
	//   <ObjectDescriptorRemove objectDescriptorId="???"/>
	// </par>
	Element parRemove = xmtaDoc.createElement(DOMXMTA.par);
	parRemove.setAttribute(DOMXMTA.begin, DOMXMTAParameter.time2Param(deleteTime.getClockValue()));
	treeExternalElement_.addToParOd(parRemove);

	//
	Element objectDescriptorRemove = xmtaDoc.createElement(DOMXMTA.ObjectDescriptorRemove);
	String objectDescriptorId = treeExternalElement_.getXmtaObjectDescriptorId(elasId);
	objectDescriptorRemove.setAttribute(DOMXMTA.objectDescriptorId, objectDescriptorId);
	parRemove.appendChild(objectDescriptorRemove);

	//
	return true;
}
/***********************************************************************
 * Create the External Element infrastructure.
 * The infrastructure contains the par node for the insertion of the OD.
 * <pre>
 *   &lt;par begin="???"&gt;
 *     &lt;ObjectDescriptorUpdate&gt;
 *       &lt;OD&gt;
 *         &lt;ObjectDescriptor ObjectDescriptorID="???"&gt;
 *           &lt;esDescr&gt;
 *             &lt;ES_Descriptor ES_ID="???"&gt;
 *               &lt;StreamSource url="???"/&gt;
 *               &lt;decConfigDescr&gt;
 *                 &lt;DecoderConfigDescriptor bufferSizeDB="auto" objectTypeIndication="???" streamType="???"&gt;
 *                   &lt;decSpecificInfo&gt;
 *                     &lt;DecoderSpecificInfo type="auto"/&gt;
 *                   &lt;/decSpecificInfo&gt;
 *                 &lt;/DecoderConfigDescriptor&gt;
 *               &lt;/decConfigDescr&gt;
 *               &lt;slConfigDescr&gt;
 *                 &lt;SLConfigDescriptor&gt;
 *                   &lt;predefined value="2"/&gt;
 *                 &lt;/SLConfigDescriptor&gt;
 *               &lt;/slConfigDescr&gt;
 *             &lt;/ES_Descriptor&gt;
 *           &lt;/esDescr&gt;
 *         &lt;/ObjectDescriptor&gt;
 *       &lt;OD/&gt;
 *     &lt;/ObjectDescriptorUpdate&gt;
 *   &lt;/par&gt;
 * </pre>
 **********************************************************************/
protected boolean createInsertMapping(ELASid elasId) {
	boolean success = super.createInsertMapping(elasId);
	if (!success)
		return success;

	// JSB: TBD
	// for now ignore events
	//	// Attach the switch into the group
	//	if (beginResolved) {
	//		// Easy way timing is resolved so attach switch directly
	//		groupChildren_.appendChild(switch_);
	//	} else {
	//		// Not so easy way.  In this case we have an event so attach switch inside conditional and
	//		// set route from touch sensor.
	//		success = timeMapBeginEvent();
	//		if (!success)
	//			return success;
	//	}

	// Get the insert time.
	TimeScheduler timeScheduler = getConvertElement().getXmto2Xmta().getTimeScheduler();
	Vector lifeCycles = timeScheduler.getLifeCycles(elasId.getElementIndex());
	LifeCycle lifeCycle = (LifeCycle) lifeCycles.elementAt(elasId.getLifeCycleIndex().intValue());
	Vector activeIntervals = lifeCycle.getActiveIntervals();
	ActiveInterval activeInterval = (ActiveInterval) activeIntervals.elementAt(elasId.getIntervalIndex().intValue());
	TimingAttribute insertTime = null;
	int sdIndex = elasId.getSimpleDurIndex().intValue();
	if (0 == sdIndex) {
		// Insert time is the observed time not the computed time.  This is the time
		// the node is placed in the Bifs tree for the z-ordering.
		insertTime = activeInterval.getComputedBeginTime();
	} else
		insertTime = (TimingAttribute) activeInterval.getRepeatTimes().elementAt(sdIndex - 1);

	return generateInsertMapping(elasId, insertTime);
}

/***********************************************************************
 * Create the External Element infrastructure.
 * The infrastructure contains the par node for the insertion of the OD.
 * <pre>
 *   &lt;par begin="???"&gt;
 *     &lt;ObjectDescriptorUpdate&gt;
 *       &lt;OD&gt;
 *         &lt;ObjectDescriptor ObjectDescriptorID="???"&gt;
 *           &lt;esDescr&gt;
 *             &lt;ES_Descriptor ES_ID="???"&gt;
 *               &lt;StreamSource url="???"/&gt;
 *               &lt;decConfigDescr&gt;
 *                 &lt;DecoderConfigDescriptor bufferSizeDB="auto" objectTypeIndication="???" streamType="???"&gt;
 *                   &lt;decSpecificInfo&gt;
 *                     &lt;DecoderSpecificInfo type="auto"/&gt;
 *                   &lt;/decSpecificInfo&gt;
 *                 &lt;/DecoderConfigDescriptor&gt;
 *               &lt;/decConfigDescr&gt;
 *               &lt;slConfigDescr&gt;
 *                 &lt;SLConfigDescriptor&gt;
 *                   &lt;predefined value="2"/&gt;
 *                 &lt;/SLConfigDescriptor&gt;
 *               &lt;/slConfigDescr&gt;
 *             &lt;/ES_Descriptor&gt;
 *           &lt;/esDescr&gt;
 *         &lt;/ObjectDescriptor&gt;
 *       &lt;OD/&gt;
 *     &lt;/ObjectDescriptorUpdate&gt;
 *   &lt;/par&gt;
 * </pre>
 **********************************************************************/
protected boolean createInsertMapping(ELASid elasId, ELASid parentELASid, TimingAttribute insertTime, TimingAttribute simpleDuration) {
	boolean success = super.createInsertMapping(elasId, parentELASid, insertTime, simpleDuration);
	if (!success)
		return success;
	else
		return generateInsertMapping(elasId, insertTime);
}

private boolean generateInsertMapping(ELASid elasId, TimingAttribute insertTime) {
	Document xmtaDoc = treeExternalElement_.getXmtaDoc();
	Hashtable xmtaIdTable = treeExternalElement_.getXmtaIds();

	// Create the OD Update for the new object 
	// <par begin="???">
	//   <ObjectDescriptorUpdate>
	//     <OD>
	//       <ObjectDescriptor ObjectDescriptorID="???">
	//         <esDescr>
	//           <ES_Descriptor ES_ID="???">
	//             <StreamSource url="???"/>
	//             <decConfigDescr>
	//               <DecoderConfigDescriptor bufferSizeDB="auto" objectTypeIndication="???" streamType="???">
	//                 <decSpecificInfo>
	//                   <DecoderSpecificInfo type="auto"/>
	//                 </decSpecificInfo>
	//               </DecoderConfigDescriptor> 
	//             </decConfigDescr>
	//             <slConfigDescr>
	//               <SLConfigDescriptor>
	//                 <predefined value="2"/>
	//               </SLConfigDescriptor>
	//             </slConfigDescr>
	//           </ES_Descriptor>
	//         </esDescr>
	//       </ObjectDescriptor>
	//     <OD/>
	//   </ObjectDescriptorUpdate>
	// </par>


	Element parUpdate = xmtaDoc.createElement(DOMXMTA.par);
	// It is still possible that the start is before the begining of the presentation, ie. t<0s
	// If this is the case then set the start time to t=0 and issue a warning to the user.
	if (insertTime.lessThan(0)) {
		parUpdate.setAttribute(DOMXMTA.begin, DOMXMTAParameter.time2Param(0));
		// report back the warning
		MessageFormat msgFormat = new MessageFormat(externalElementRes_.getString(ExternalElementRes.MSG_WARNING_NEGATIVE_SCENE_START));
		StringBuffer buffer = new StringBuffer();
		buffer.append("<").append(treeExternalElement_.getXmtoElement().getNodeName());
		String id = treeExternalElement_.getXmtoAttrId();
		if (null != id && !id.equals(""))
			buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
		buffer.append(">");
		Object[] msgArgs = {buffer};
		String message = msgFormat.format(msgArgs);
		fireErrorEvent(ErrorEvent.WARNING, 0, message, treeExternalElement_.getXmtoElement());
	} else {
		parUpdate.setAttribute(DOMXMTA.begin, DOMXMTAParameter.time2Param(insertTime.getClockValue()));
	}
	treeExternalElement_.addToParOd(parUpdate);

	//
	Node objectDescriptorUpdate = parUpdate.appendChild(xmtaDoc.createElement(DOMXMTA.ObjectDescriptorUpdate));

	//
	Node oD = objectDescriptorUpdate.appendChild(xmtaDoc.createElement(DOMXMTA.OD));

	// write ObjectDescriptor
	Element objectDescriptor = xmtaDoc.createElement(DOMXMTA.ObjectDescriptor);
	// make the switch and transform ids based on object id or make them unique.
	String objectDescriptorId = treeExternalElement_.getXmtaObjectDescriptorId(elasId);
	objectDescriptor.setAttribute(DOMXMTA.objectDescriptorID, objectDescriptorId);
	xmtaIdTable.put(objectDescriptorId, objectDescriptor);
	oD.appendChild(objectDescriptor);

	//
	Node esDescr = null;
	Node descr = objectDescriptor.appendChild(xmtaDoc.createElement(DOMXMTA.Descr));
	esDescr = descr.appendChild(xmtaDoc.createElement(DOMXMTA.esDescr));

	// write ESDescriptor
	Element es_Descriptor = xmtaDoc.createElement(DOMXMTA.ES_Descriptor);
	String esDescriptorId = treeExternalElement_.getXmtaEsDescriptorId(elasId);
	es_Descriptor.setAttribute(DOMXMTA.ES_ID, esDescriptorId);
	xmtaIdTable.put(esDescriptorId, es_Descriptor);
	esDescr.appendChild(es_Descriptor);

	//
	Node decConfigDescr = es_Descriptor.appendChild(xmtaDoc.createElement(DOMXMTA.decConfigDescr));

	// write DecoderConfigDescriptor
	Element decoderConfigDescriptor = xmtaDoc.createElement(DOMXMTA.DecoderConfigDescriptor);
	decoderConfigDescriptor.setAttribute(DOMXMTA.bufferSizeDB, DOMXMTA.auto);
	// Determine the type of the file.
	MediaObjectInformation moi = new MediaObjectInformation(treeExternalElement_.getXmtoAttrSrcCanonical());
	decoderConfigDescriptor.setAttribute(DOMXMTA.objectTypeIndication, Integer.toString(moi.getMediaObjectType()));
	decoderConfigDescriptor.setAttribute(DOMXMTA.streamType, Integer.toString(moi.getMediaStreamType()));
	decConfigDescr.appendChild(decoderConfigDescriptor);

	//
	Node decSpecificInfo = decoderConfigDescriptor.appendChild(xmtaDoc.createElement(DOMXMTA.decSpecificInfo));
	// write DecoderSpecificInfo
	Element decoderSpecificInfo = xmtaDoc.createElement(DOMXMTA.DecoderSpecificInfo);
	decoderSpecificInfo.setAttribute(DOMXMTA.type, DOMXMTA.auto);
	decSpecificInfo.appendChild(decoderSpecificInfo);

	// Create the Sl Config Descriptor
	Node slConfigDescr = es_Descriptor.appendChild(xmtaDoc.createElement(DOMXMTA.slConfigDescr));
	//
	Element slConfigDescriptor = xmtaDoc.createElement(DOMXMTA.SLConfigDescriptor);
	slConfigDescr.appendChild(slConfigDescriptor);
	Element predefined = xmtaDoc.createElement(DOMXMTA.predefined);
	predefined.setAttribute(DOMXMTA.value, "2");
	slConfigDescriptor.appendChild(predefined);
	//  write the StreamSource element
	Element streamSource = xmtaDoc.createElement(DOMXMTA.StreamSource);
	String uri = treeExternalElement_.getXmtoAttrSrc().replace('\\', '/');
	streamSource.setAttribute(DOMXMTA.url, FILE_URL_PREFIX + uri);
	es_Descriptor.appendChild(streamSource);

	//
	return true;
}
/*******************************************************************
 *******************************************************************/
public void init(Xmto2Xmta xmto2xmta, xmto2xmta.tree.ConvertElement element) throws Exception {
	super.init(xmto2xmta, element);

	//
	if (!(element instanceof xmto2xmta.tree.ExternalElement)) {
		if (debug_)
			fireErrorEvent(ErrorEvent.TRACE, 0, "MedExternalElement.init() - element=" + element + " is not an instance of xmto2xmta.tree.ExternalElement");
		throw new Exception("");
	}

	//
	treeExternalElement_ = (xmto2xmta.tree.ExternalElement) element;

	//
	if (debug_)
		fireErrorEvent(ErrorEvent.TRACE, 0, "MedExternalElement.init() - treeExternalElement_=" + treeExternalElement_);
}

/*********************************************************************** 
 * Attach the switch for the element based on the begin event.  The begining 
 * of the object is based on an event thus attach switch 
 * The begining of the object is based on an event thus attach switch 
 * inside conditional and set the route from a touch sensor.  If there
 * is an offset time then also need a timer conditional.
 **********************************************************************/
protected boolean createEventInsertMapping(ELASid elasId, ELASid parentELASid) {
	boolean success = super.createEventInsertMapping(elasId, parentELASid);
	if (!success)
		return success;

	// Get the insert time.
	TimeScheduler timeScheduler = getConvertElement().getXmto2Xmta().getTimeScheduler();
	EventActiveInterval eventInterval = timeScheduler.getEventActiveInterval(elasId.getElementIndex());
	TimingAttribute eventAttribute = eventInterval.getBeginEvent();
	TimingAttribute observedBegin = eventInterval.getObservedBeginTime();
	if (eventAttribute == null || eventAttribute.getTimeValueType() != TimingAttribute.MULTI_VALUE) {
		return false;
	}
	Vector timingAttributes = eventAttribute.getMultiTimingAttributes();
	if (timingAttributes.size() < 3) {
		return false;
	}
	TimingAttribute sceneInsertTime = (TimingAttribute) timingAttributes.elementAt(0);
	// For right now we only support time values
	boolean error = false;
	if (null == sceneInsertTime)
	  	error = true;
	else
	  	switch (sceneInsertTime.getTimeValueType()) {
	     	case TimingAttribute.CLOCK_VALUE :
	     		break;
	     	default :
	     		error = true;
	   }
	if (null == observedBegin)
	  	error = true;
	else
	  	switch (observedBegin.getTimeValueType()) {
	     	case TimingAttribute.CLOCK_VALUE :
	     	case TimingAttribute.INDEFINITE_VALUE :
	     		break;
	     	default :
	     		error = true;
	   }
	if (!error) {
		if (observedBegin.getTimeValueType() == TimingAttribute.CLOCK_VALUE && observedBegin.greaterThan(sceneInsertTime)) {
			sceneInsertTime = observedBegin;
		}
		return generateInsertMapping(elasId, sceneInsertTime);

	} else {
  	// report back the problem
   	MessageFormat msgFormat = new MessageFormat(xmto2XmtaRes_.getString(Xmto2XmtaRes.MSG_ERROR_TIMING_PROBLEM));
     	StringBuffer buffer = new StringBuffer();
     	buffer.append(getConvertElement().getXmtoElement().getNodeName());
     	if (null != getConvertElement().getXmtoAttrId() && !getConvertElement().getXmtoAttrId().equals(""))
     		buffer.append(" ").append(DOMXMTO.id).append("=\"").append(getConvertElement().getXmtoAttrId()).append("\"");
     	Object[] msgArgs = {buffer};
     	String message = msgFormat.format(msgArgs);
     	fireErrorEvent(ErrorEvent.ERROR, 0, message, getConvertElement().getXmtoElement());
     	// Since we don't support the timings for this element remove it from the tree
     	getConvertElement().remove();
     	return false;
	}
}

/*********************************************************************** 
 * Map the timing of the end of the object based on the end event(s).
 * The end could be a single event or more likely a list of end events
 * or clock timing.  The first one which occurs will remove the event and
 * the rest will be ignored (or fall through without effect).
 **********************************************************************/
protected boolean createEventDeleteMapping(ELASid elasId, ELASid parentELASid) {
	boolean success = super.createEventDeleteMapping(elasId, parentELASid);
	if (!success)
		return success;

	// Get the delete time.
	TimeScheduler timeScheduler = getConvertElement().getXmto2Xmta().getTimeScheduler();
	EventActiveInterval eventInterval = timeScheduler.getEventActiveInterval(elasId.getElementIndex());
	TimingAttribute eventAttribute = eventInterval.getEndEvent();
	TimingAttribute sceneDeleteTime = eventInterval.getFillEndTime();
	if (eventAttribute == null || eventAttribute.getTimeValueType() != TimingAttribute.MULTI_VALUE) {
		return false;
	}
	Vector timingAttributes = eventAttribute.getMultiTimingAttributes();
	if (timingAttributes.size() < 3) {
		return false;
	}
	TimingAttribute sceneInsertTime = (TimingAttribute) timingAttributes.elementAt(0);
	// For right now we only support time values
	boolean error = false;
	if (null == sceneInsertTime)
	  	error = true;
	else
	  	switch (sceneInsertTime.getTimeValueType()) {
	     	case TimingAttribute.CLOCK_VALUE :
	     		break;
	     	default :
	     		error = true;
	   }
	if (sceneDeleteTime != null) {
	  	switch (sceneDeleteTime.getTimeValueType()) {
	     	case TimingAttribute.CLOCK_VALUE :
	     	case TimingAttribute.INDEFINITE_VALUE :
	     		break;
	     	default :
	     		error = true;
	   }
	}

   if (error) {
  	// report back the problem
   	MessageFormat msgFormat = new MessageFormat(xmto2XmtaRes_.getString(Xmto2XmtaRes.MSG_ERROR_TIMING_PROBLEM));
     	StringBuffer buffer = new StringBuffer();
     	buffer.append(getConvertElement().getXmtoElement().getNodeName());
     	if (null != getConvertElement().getXmtoAttrId() && !getConvertElement().getXmtoAttrId().equals(""))
     		buffer.append(" ").append(DOMXMTO.id).append("=\"").append(getConvertElement().getXmtoAttrId()).append("\"");
     	Object[] msgArgs = {buffer};
     	String message = msgFormat.format(msgArgs);
     	fireErrorEvent(ErrorEvent.ERROR, 0, message, getConvertElement().getXmtoElement());
     	// Since we don't support the timings for this element remove it from the tree
     	getConvertElement().remove();
     	return false;
    }
	
	if (!generateDeleteMapping(elasId, sceneInsertTime))
		return false;

	if (sceneDeleteTime != null && sceneDeleteTime.getTimeValueType() == TimingAttribute.CLOCK_VALUE) {
		return generateDeleteMapping(elasId, sceneDeleteTime);
	}
	return true;
}

/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
}
