/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  Condition.java
//
//=============================================================================
package xmto2xmta.timegraph;

import java.util.*;

/*********************************************************************** 
 * Syncbase or event-based condition.
 *
 * <p><b>History:</b><ul>
 * <li>15 Mar 2001 - Created file.
 * </ul><p>
 *********************************************************************/
public class Condition {
	
	public static final int UNKNOWN = 0;
	public static final int SYNCBASE_BEGIN = 1;
	public static final int SYNCBASE_END = 2;
	public static final int BEGIN_EVENT = 3;
	public static final int END_EVENT = 4;
	public static final int ACTIVATE_EVENT = 5;
	public static final int REPEAT = 6;
	public static final int BASEANDOFFSET = 100;
	public static final int ACCESSKEY = 101;
	public static final int MEDIAMARKER = 102;
	public static final int WALLCLOCK = 103;
	public static final int INDEFINITE = 104;

	public static final String[] TYPE_NAME = {
		"unknown", "syncbase_begin", "syncbase_end", "begin_event",
		"end_event", "activate_event", "repeat",
		"accessKey", "mediaMarker", "wallclock", "indefinite", };

	private int type = UNKNOWN;
	private long offsetValue = 0;
	private Integer syncbaseID = new Integer(0);
	private Container syncbase = null;
	private long repeatCount = 1;
	private String marker = null;
	private String accessKeyValue = null;
	private String wallClockValue = null;
	private boolean offsetCondition = false;

	private HashSet instanceTimes = new HashSet();	// of Time
	private Time currentInstanceTime = null;

	public void setType(int type) {
		this.type = type;
	}

	public int getType() {
		return type;
	}

	public void setSyncbaseID(Integer syncbaseID) {
		this.syncbaseID = syncbaseID;
	}

	public Integer getSyncbaseID() {
		return syncbaseID;
	}

	public void setSyncbase(Container syncbase) {
		this.syncbase = syncbase;
	}

	public Container getSyncbase() {
		return syncbase;
	}

	/********************************************************
	 * The offset here means the offset value from the syncbase
	 * in the parent simple time.  E.g. begin=foo.begin+offset.
	 *********************************************************/
	public void setOffsetValue(long offsetValue) {
		this.offsetValue = offsetValue;
	}

	public long getOffsetValue() {
		return offsetValue;
	}

	public void setRepeatCount(long repeatCount) {
		this.repeatCount = repeatCount;
	}
	
	public long getRepeatCount() {
		return repeatCount;
	}
	
	public void setMarker(String marker) {
	    this.marker = marker;
	}
	
	public String getMarker() {
	    return marker;
	}
	
	public void setAccessKeyValue(String accessKeyValue) {
	    this.accessKeyValue = accessKeyValue;
	}
	
	public String getAccessKeyValue() {
	    return accessKeyValue;
	}
	
	public void setWallClockValue(String wallClockValue) {
	    this.wallClockValue = wallClockValue;
	}
	
	public String getWallClockValue() {
	    return wallClockValue;
	}

	/*****************************************************
	 * The instanceTimes is a set of Time objects.
	 *****************************************************/
	public void setInstanceTimes(HashSet instanceTimes) {
		if (instanceTimes == null) {
			this.instanceTimes = new HashSet();
		} else {
			this.instanceTimes = instanceTimes;
		}
	}

	public HashSet getInstanceTimes() {
		return instanceTimes;
	}

	public void setCurrentInstanceTime(Time instanceTime) {
		currentInstanceTime = instanceTime;
	}
	public Time getCurrentInstanceTime() {
		return currentInstanceTime;
	}

	public boolean isEventType() {
		switch (type) {
			case Condition.BEGIN_EVENT:
			case Condition.END_EVENT:
			case Condition.ACTIVATE_EVENT:
			case Condition.REPEAT:
			case Condition.ACCESSKEY:
				return true;
			default:
				return false;
		}
	}

	public boolean isSyncbaseType() {
		switch (type) {
			case Condition.SYNCBASE_BEGIN:
			case Condition.SYNCBASE_END:
			case Condition.MEDIAMARKER:
			case Condition.WALLCLOCK:
				return true;
			default:
				return false;
		}
	}

	public boolean hasBaseAndOffset() {
		if (type > Condition.UNKNOWN && type <= Condition.BASEANDOFFSET) {
			return true;
		} else {
			return false;
		}
	}

	public boolean isIndefinite() {
		return type == Condition.INDEFINITE;
	}

	public boolean isDeterministic() {
		switch (type) {
			case Condition.ACTIVATE_EVENT:
			case Condition.ACCESSKEY:
			case Condition.MEDIAMARKER:
			case Condition.WALLCLOCK:
				return false;
			default:
				return true;
		}
	}

	public boolean getOffsetCondition() {
		return offsetCondition;
	}

	public void setOffsetCondition(boolean offsetCondition) {
		this.offsetCondition = offsetCondition;
	}

	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("type="+Condition.TYPE_NAME[type]);
		if (hasBaseAndOffset()) {
			buf.append(" syncbase="+syncbaseID.toString());
			buf.append(" offset="+String.valueOf(offsetValue));
		}
		if (instanceTimes.size() > 0) {
			buf.append(" instance times=");
			for (Iterator i = instanceTimes.iterator(); i.hasNext(); ) {
				Time instanceTime = (Time) i.next();
				buf.append(instanceTime.toString());
				if (currentInstanceTime != null && instanceTime.isEqual(currentInstanceTime)) {
					buf.append(" current, ");
				} else {
					buf.append(", ");
				}
			}
		}
		return buf.toString();
	}

}   // class Condition
