/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  Container.java
//
//=============================================================================
package xmto2xmta.timegraph;

import java.util.*;
import java.io.*;

/*********************************************************************** 
 * Generic container class (a node) in timing graph.
 *
 * <p><b>History:</b><ul>
 * <li>15 Mar 2001 - Created file.
 * </ul><p>                                                                   
 *********************************************************************/

public class Container {
   private Container parent = null;
   private Vector children = null;  // of Containers
   private Integer id = null;
	private TimeGraph timeGraph = null; 
	protected PrintWriter writer = null; 
	protected PrintWriter errWriter = null; 

	public Container() {
		this(null, new Integer(-1), null);
	}

   public Container(TimeGraph timeGraph, Integer id, Container parent) {
		this.timeGraph = timeGraph;
      this.parent = parent;
		children = new Vector();
		this.id = id;
		if (timeGraph != null) {
			writer = timeGraph.getPrintWriter();
			errWriter = timeGraph.getErrPrintWriter();
		}
	}

	public void setTimeGraph(TimeGraph timeGraph) {
		this.timeGraph = timeGraph;
		if (timeGraph != null) {
			writer = timeGraph.getPrintWriter();
			errWriter = timeGraph.getErrPrintWriter();
		}
	}

	public TimeGraph getTimeGraph() {
		return timeGraph;
	}

	public void setParent(Container parent) {
		this.parent = parent;
	}

	public Container getParent() {
		return parent;
	}

	public void setID(Integer id) {
		this.id = id;
	}

	public Integer getID() {
		return id;
	}

	public Container getRoot() {
		return timeGraph.getRoot();
	}

	public void addChild(Container child) {
		if (!children.contains(child)) {
			children.addElement(child);
			child.setParent(this);
		}
	}

	public Vector getChildren() {
		return children;
	}

	public Enumeration children() {
		return children.elements();
	}

	/*******************************************************
	 * Return the previous sibling of the given child.
	 * If the given child is not found, or it's the first
	 * child, return null.
	 *******************************************************/
	public Container getPrevChild(Container child) {
		if (child == null) return null;
		int childIndex = children.indexOf(child);
		if (childIndex < 1) {
			return null;
		}
		return (Container) children.elementAt(childIndex - 1);
	}

	/*******************************************************
	 * Return the next sibling of the given child.
	 * If the given child is not found, or it's the last
	 * child, return null.
	 *******************************************************/
	public Container getNextChild(Container child) {
		if (child == null) return null;
		int childIndex = children.indexOf(child);
		if ((childIndex == -1) || (childIndex == children.size() -1)) {
			return null;
		}
		return (Container) children.elementAt(childIndex + 1);
	}

	public boolean isCommonAscendantOf(Container container1,
													Container container2) {
		if (container1 == null || container2 == null) return false;
		Container commonAscendant = container1.getCommonAscendant(container2);		
		return isAscendantOf(commonAscendant);
	}

	public boolean isAscendantOf(Container otherContainer) {
		boolean result = false;
		Container curContainer = otherContainer;
		while (curContainer != null) {
			if (curContainer == this) {
				result = true;
				break;
			}
			curContainer = curContainer.getParent();
		}
		return result;
	}

	/*******************************************************
	 * Find the nearest common ascendant of this container
	 * and another container in the time graph.
	 * Note that if one of the two containers is a parent
	 * or an ascendant of the other, it is considered a
	 * common ascendant.
	 *******************************************************/
	public Container getCommonAscendant(Container otherContainer) {
		Hashtable visitedAscendants = new Hashtable();
		Container myAscendant = this;
		Container herAscendant = otherContainer;
		Container commonAscendant = null;
		Container root = null;
		while (true) {
			if (myAscendant != null) {
				if (visitedAscendants.containsKey(myAscendant)) {
					commonAscendant = myAscendant;
					break;
				} else {
					visitedAscendants.put(myAscendant, "");
					Container tmp = myAscendant.getParent();
					if (tmp == null) {
						root = myAscendant;
					}
					myAscendant = tmp;
				}
			} else {
				if (herAscendant == null) { // bad tree.
					break;
				}
			}
			if (herAscendant != null) {
				if (visitedAscendants.containsKey(herAscendant)) {
					commonAscendant = herAscendant;
					break;
				} else {
					visitedAscendants.put(herAscendant, "");
					Container tmp = herAscendant.getParent();
					if (tmp == null) {
						root = herAscendant;
					}
					herAscendant = tmp;
				}
			}
		}
		if (commonAscendant == null) {
			// bad tree!
			commonAscendant = root;
		}
		return commonAscendant;
	}

	public void hasEvent() {
	}
				
}	// class Container
