/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  SimpleTimeContainer.java
//
//=============================================================================
package xmto2xmta.timegraph;

import java.util.*;

/*********************************************************************** 
 * Time container in timing graph. Parent of all actual timed elements.
 *
 * <p><b>History:</b><ul>
 * <li>15 Mar 2001 - Created file.
 * </ul><p>
 *********************************************************************/

public class SimpleTimeContainer extends SimpleContainer {

	/**************************************************************
	 * The following section contains timing attributes from the
	 * XMT-O doc element.  Once set,these values cannot change throughout
	 * the life of the document.
	 **************************************************************/
	private TimeContainer timeContainer = null;

	private Time dur = new Time();
	private int durType = TimeContainer.DURTYPE_UNSPECIFIED;
	private Time simpleDur = new Time();
	private Time activeDur = new Time();
	private Interval beginTime = new Interval();
	private Interval endTime = new Interval();
	private Interval endConditionTime = new Interval();
	private Time intrinsicDur = new Time();

	private Vector lifeHistory = new Vector();	// history of life cycles
	private ContainerLife containerLife = null;	// current life cycle
	private LifeInterval lifeInterval = null;		// current active interval in the current life cycle
	private ELASid elasid = null;
	private Condition inheritedBeginCondition = null;		// inherited nondeterministic condition
	private Condition inheritedEndCondition = null;		// inherited nondeterministic condition
	private Time beginConstrainTime = new Time(Time.INDEFINITE);
	private Time endConstrainTime = new Time(Time.INDEFINITE);

	// constructors
	public SimpleTimeContainer() {
		this(null, new Integer(-1), null);
	}

	public SimpleTimeContainer(Integer id, Container parent) {
		this(null, id, parent);
	}

	public SimpleTimeContainer(SimpleTimeGraph timeGraph, Integer id, Container parent) {
		super(timeGraph, id, parent);
		createLifeContainer();
	}

	public SimpleTimeContainer(SimpleTimeGraph timeGraph, Container parent, TimeContainer timeContainer)
				throws UnsupportedElementException {
		super(timeGraph, timeContainer.getID(), parent);
		this.timeContainer = timeContainer;
		createLifeContainer();
		load(timeContainer);
	}

	private void load(TimeContainer timeContainer) throws UnsupportedElementException {
		if (timeContainer == null) throw new UnsupportedElementException();
		if (timeContainer.getBeginConditions().size() > 1) {
			throw new UnsupportedElementException();
		}
		if (timeContainer.getEndConditions().size() > 1) {
			throw new UnsupportedElementException();
		}
		durType = timeContainer.getDurType();
		if (durType != TimeContainer.DURTYPE_TIMEVALUE && durType != TimeContainer.DURTYPE_UNSPECIFIED) {
			throw new UnsupportedElementException();
		}
		dur = timeContainer.getDur();
		if (durType == TimeContainer.DURTYPE_UNSPECIFIED && !(timeContainer instanceof MediaTimeContainer)) {
			durType = TimeContainer.DURTYPE_TIMEVALUE;
			dur = new Time(Time.INDEFINITE);
		}
		if (timeContainer.repeatCountSpecified() || timeContainer.repeatDurSpecified()) {
			throw new UnsupportedElementException();
		}
		this.timeContainer = timeContainer;
		inheritedBeginCondition = (Condition) timeContainer.getBeginConditions().elementAt(0);
		switch (inheritedBeginCondition.getType()) {
			case Condition.SYNCBASE_BEGIN:
			case Condition.SYNCBASE_END:
			case Condition.ACTIVATE_EVENT:
				break;
			default:
				throw new UnsupportedElementException();
		}
		if (timeContainer.getEndConditions().size() > 0) {
			inheritedEndCondition = (Condition) timeContainer.getEndConditions().elementAt(0);
			switch (inheritedEndCondition.getType()) {
				case Condition.SYNCBASE_BEGIN:
				case Condition.SYNCBASE_END:
				case Condition.ACTIVATE_EVENT:
					break;
				default:
					throw new UnsupportedElementException();
			}
		}
		if (!(timeContainer instanceof MediaTimeContainer) && hasEventConditions()) {
			throw new UnsupportedElementException();
		}

		simpleDur = computeSimpleDur();
	}

	private void createLifeContainer() {
		containerLife = new ContainerLife(new Time(0), new Time(Time.INDEFINITE));
		containerLife.setContainer(timeContainer);
		elasid = new ELASid(getID(), new Integer(0), new Integer(0), new Integer(0));
		// lifeInterval.getELASids().add(elasid);
		if (getParent() != null) {
			containerLife.setParentELASid(((SimpleTimeContainer) getParent()).getELASid());
		}

		// containerLife.setParentELASid(parentELASid);
		lifeHistory.add(containerLife);
	}

	public Vector getBeginConditions() {
		return timeContainer.getBeginConditions();
	}

	public Vector getEndConditions() {
		return timeContainer.getEndConditions();
	}

	public int getDurType() {
		return durType;
	}

	public Time getDur() {
		return dur;
	}

	public int getMinType() {
		return timeContainer.getMinType();
	}

	public Time getMin() {
		return timeContainer.getMin();
	}

	public int getMaxType() {
		return timeContainer.getMaxType();
	}

	public Time getMax() {
		return timeContainer.getMax();
	}

	public boolean endSpecified() {
		return (inheritedEndCondition != null);
	}

	public boolean durSpecified() {
		return (durType != TimeContainer.DURTYPE_UNSPECIFIED);
	}

	public Vector getSyncbaseDependents() {
		return timeContainer.getSyncbaseDependents();
	}

	public Vector getEventDependents() {
		return timeContainer.getEventDependents();
	}

	public SimpleTimeContainer getSimpleTimeContainer(TimeContainer timeContainer) {
		return (SimpleTimeContainer) (getSimpleTimeGraph().getContainerByID( timeContainer.getID() ));
	}

	public Time getSimpleDur() {
		return simpleDur;
	}

	public Time getActiveDur() {
		return activeDur;
	}

	public Interval getBeginTime() {
		return beginTime;
	}

	public Interval getEndTime() {
		return endTime;
	}

	public Interval getEndConditionTime() {
		return endConditionTime;
	}

	public void setSimpleDur(Time time) {
		simpleDur = time;
	}

	public void setActiveDur(Time time) {
		activeDur = time;
	}

	public void setBeginTime(Interval time) {
		beginTime = time;
	}

	public void setEndTime(Interval time) {
		endTime = time;
	}

	public Time getIntrinsicDur() {
		return timeContainer.getIntrinsicDur();
	}

	public Condition getInheritedBeginCondition() {
		return inheritedBeginCondition;
	}

	public Condition getInheritedEndCondition() {
		return inheritedEndCondition;
	}

	public Vector getLifeHistory() {
		return lifeHistory;
	}

	public ELASid getELASid() {
		return elasid;
	}

	public TimeContainer getTimeContainer() {
		return timeContainer;
	}

	public Time getBeginConstrainTime() {
		return beginConstrainTime;
	}

	public Time getEndConstrainTime() {
		return endConstrainTime;
	}

	/****************************************************************
	 * Called by a syncbase element when it changed its begin or end time.
	 ****************************************************************/
	public void update(SimpleTimeContainer syncbase) {
		boolean changed = false;
		if (syncbase == null) {
			writer.println(getID()+" start");
			changed = true;
		} else {
			writer.println(getID()+" update from " + syncbase.getID());
			Interval newBeginTime = computeBeginTime(syncbase);
			if (!newBeginTime.equals(beginTime)) {
				beginTime = newBeginTime;
				changed = true;
				writer.println(getID()+" new begin time="+beginTime);
			}
			Interval newEndConditionTime = computeEndConditionTime(syncbase);
			if (!newEndConditionTime.equals(endConditionTime)) {
				endConditionTime = newEndConditionTime;
				changed = true;
				writer.println(getID()+" new end condition time="+endConditionTime);
			}
			Interval newEndTime = computeEndTime();
			if (!newEndTime.equals(endTime)) {
				endTime = newEndTime;
				changed = true;
			}
			if (adjust()) {
				changed = true;
			}
		}
		if (changed) {
			constrainChildren();
			constrainSiblings();
			if (beginTime.point() && endTime.point()) {
				LifeInterval lifeInterval = new LifeInterval(beginTime.getBeginTime(), endTime.getBeginTime());
				lifeInterval.getELASids().add(elasid);
				lifeInterval.setObservedBegin(beginTime.getBeginTime());
				lifeInterval.setIntrinsicDur(timeContainer.getIntrinsicDur());
				lifeInterval.getSimpleDurs().add(Time.subtract(endTime.getBeginTime(), beginTime.getBeginTime()));
				containerLife.setLifeIntervals(new Vector());
				containerLife.addInterval(lifeInterval);
			}
				
			writer.println(getID()+" calling children update() ");
			for (Enumeration e = children(); e.hasMoreElements();) {
				SimpleTimeContainer child = (SimpleTimeContainer) e.nextElement();
				if (!child.getID().equals(getID())) {
					child.update(this);
				}
			}
			writer.println(getID()+" calling dependents update() ");
			for (Enumeration e = getSyncbaseDependents().elements(); e.hasMoreElements();) {
				SimpleTimeContainer dependent = getSimpleTimeContainer( (TimeContainer) e.nextElement() );
				if (!dependent.getID().equals(getID())) {
					dependent.update(this);
				}
			}
			for (Enumeration e = getEventDependents().elements(); e.hasMoreElements();) {
				SimpleTimeContainer dependent = getSimpleTimeContainer( (TimeContainer) e.nextElement() );
				if (!dependent.getID().equals(getID())) {
					dependent.update(this);
				}
			}
		}			
	}
		
	/****************************************************************
	 ****************************************************************/
	private Time computeSimpleDur() {
		if (simpleDur.isResolved()) return simpleDur;
		writer.println(getID()+" computing simple dur");
		if (getDurType() == TimeContainer.DURTYPE_UNSPECIFIED) {
			if (endSpecified()) {
				simpleDur = new Time(Time.INDEFINITE);
				writer.println(getID()+" simple dur = indefinite");
			} else {
				simpleDur = computeIntrinsicDur();
				intrinsicDur = simpleDur;
				writer.println(getID()+" simple dur = implicit dur =" + simpleDur);
			}
		} else if (getDurType() == TimeContainer.DURTYPE_TIMEVALUE) {
			simpleDur = getDur();
			writer.println(getID()+" simple dur = time value = " + simpleDur);
		} else if (getDurType() == TimeContainer.DURTYPE_MEDIA) {
			simpleDur = computeIntrinsicDur();
			intrinsicDur = simpleDur;
			writer.println(getID()+" simple dur = media (implicit) = " + simpleDur);
		} else {
			writer.println(getID()+" invalid durtype="+getDurType());
		}
		return simpleDur;
	}

	private Time computeIntrinsicDur() {
		writer.println(getID()+" computing intrinsic dur");
		if (timeContainer instanceof MediaTimeContainer) {
			return ((MediaTimeContainer) timeContainer).computeIntrinsicDur();
		} else {
			return new Time();
		}
	}


	/****************************************************************
	 ****************************************************************/
	private Interval evaluateBeginCondition(SimpleTimeContainer syncbase) {
		Interval newBeginTime = beginTime;
		Condition condition = (Condition) getBeginConditions().elementAt(0);
		SimpleTimeContainer conditionSyncbase = getSimpleTimeContainer((TimeContainer) condition.getSyncbase());
		if (syncbase.getID().equals(conditionSyncbase.getID())) {
			newBeginTime = evaluateCondition(condition, true);
		}
		return newBeginTime;
	}

	/****************************************************************
	 ****************************************************************/
	private Interval evaluateEndCondition(SimpleTimeContainer syncbase) {
		Interval newEndConditionTime = endConditionTime;
		if (getEndConditions().size() > 0) {
			Condition condition = (Condition) getEndConditions().elementAt(0);
			SimpleTimeContainer conditionSyncbase = getSimpleTimeContainer((TimeContainer) condition.getSyncbase());
			if (syncbase.getID().equals(conditionSyncbase.getID())) {
				newEndConditionTime = evaluateCondition(condition, false);
			}
		}
		return newEndConditionTime;
	}

	/****************************************************************
	 ****************************************************************/
	private Interval evaluateCondition(Condition condition, boolean begin) {
		Interval timeValue = new Interval();
		SimpleTimeContainer syncbase = getSimpleTimeContainer((TimeContainer) condition.getSyncbase());
		if (condition.isDeterministic()) {
			if (condition.getType() == Condition.SYNCBASE_BEGIN) {
				timeValue = syncbase.getBeginTime();
			} else if (condition.getType() == Condition.SYNCBASE_END) {
				timeValue = syncbase.getEndTime();
			}
			if (timeValue.point()) {
				timeValue = timeValue.add(timeValue, new Time(condition.getOffsetValue()));
			} else {
				if (begin) {
					inheritedBeginCondition = syncbase.getInheritedBeginCondition();
					if (inheritedBeginCondition != null) {
						inheritedBeginCondition.setOffsetValue(
							((Condition) getBeginConditions().elementAt(0)).getOffsetValue() +
							inheritedBeginCondition.getOffsetValue());
					}
				} else {
					inheritedEndCondition = syncbase.getInheritedEndCondition();
					if (inheritedEndCondition != null && getEndConditions().size() > 0) {
						inheritedEndCondition.setOffsetValue(
							((Condition) getEndConditions().elementAt(0)).getOffsetValue() +
							inheritedEndCondition.getOffsetValue());
					}
				}
			}
		} else {
			if (condition.getType() == Condition.BEGIN_EVENT ||
					condition.getType() == Condition.END_EVENT ||
					condition.getType() == Condition.ACTIVATE_EVENT) {
				Interval beginTimeValue = syncbase.getBeginTime();
				Interval endTimeValue = syncbase.getEndTime();
				if (!beginTimeValue.point()) {
					beginTimeValue = Interval.add(beginTimeValue,
												new Time(((Condition) syncbase.getBeginConditions().elementAt(0)).getOffsetValue()));
				}
				if (!endTimeValue.point()) {
					endTimeValue = Interval.add(endTimeValue,
												new Time(((Condition) syncbase.getEndConditions().elementAt(0)).getOffsetValue()));
				}
				timeValue = Interval.union(beginTimeValue, endTimeValue);
			}
		}
		return timeValue;
	}

	/****************************************************************
	 ****************************************************************/
	private Interval computeEndTime() {
		Interval newEndTime = new Interval();
		if (endSpecified()) {
			newEndTime = endConditionTime;
		} else {
			newEndTime = Interval.add(beginTime, computeSimpleDur());
		}
		writer.println(getID()+" end time="+newEndTime);
		return newEndTime;
	}

	/****************************************************************
	 ****************************************************************/
	private Interval computeBeginTime(SimpleTimeContainer syncbase) {
		return evaluateBeginCondition(syncbase);
	}

	/****************************************************************
	 ****************************************************************/
	private Interval computeEndConditionTime(SimpleTimeContainer syncbase) {
		return evaluateEndCondition(syncbase);
	}

	/****************************************************************
	 ****************************************************************/
	private boolean adjust() {
		boolean changed = false;
		if (beginTime.getEndTime().isGreater(endConstrainTime)) {
			beginTime.setEndTime(endConstrainTime);
			if (beginTime.negative()) {
				beginTime = new Interval();
			}
			changed = true;
		}
		if (endTime.getEndTime().isGreater(endConstrainTime)) {
			endTime.setEndTime(endConstrainTime);
			if (endTime.negative()) {
				endTime = new Interval(endConstrainTime, endConstrainTime);
			}
			changed = true;
		}
		if (endTime.getEndTime().isLess(beginTime.getBeginTime())) {
			beginTime = new Interval();
			endTime = new Interval();
			changed = true;
		}
		if (changed) {
			writer.println(getID()+" adjusted begin="+beginTime);
			writer.println(getID()+" adjusted end="+endTime);
		}
		return changed;
	}

	private void constrainChildren() {
		writer.println(getID()+" constraining children times");
		for (Enumeration e = children(); e.hasMoreElements();) {
			SimpleTimeContainer child = (SimpleTimeContainer) e.nextElement();
			if (!child.getID().equals(getID())) {
				child.constrain(beginTime.getBeginTime(), endTime.getEndTime());
			}
		}
	}

	private void constrainSiblings() {
		if (!beginTime.point() || getParent() == null) return;
		SimpleTimeContainer parent = (SimpleTimeContainer) getParent();
		if (parent.getTimeContainer() instanceof SeqTimeContainer) {
			SimpleTimeContainer prev = (SimpleTimeContainer) parent.getPrevChild(this);
			if (prev != null) {
				prev.constrain(null, beginTime.getBeginTime());
			}
		}
	}
	
	public void constrain(Time beginConstrainTime, Time endConstrainTime) {
		if (beginConstrainTime != null) {
			this.beginConstrainTime = beginConstrainTime;
		}
		if (endConstrainTime != null) {
			this.endConstrainTime = endConstrainTime;
		}
	}

	public boolean hasEventConditions() {
		for (Enumeration e = getBeginConditions().elements(); e.hasMoreElements();) {
			Condition condition = (Condition) e.nextElement();
			if (condition.isEventType()) {
				return true;
			}
		}
		for (Enumeration e = getEndConditions().elements(); e.hasMoreElements();) {
			Condition condition = (Condition) e.nextElement();
			if (condition.isEventType()) {
				return true;
			}
		}
		return false;
	}


}
