/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  SimpleTimeGraph.java
//
//=============================================================================
package xmto2xmta.timegraph;

import java.util.*;
import java.io.*;
import xmto2xmta.ErrorEvent;

/*********************************************************************** 
 * Simple Time graph class.  Time graph models the timing relationships.
 *
 * <p><b>History:</b><ul>
 * <li>15 Mar 2001 - Created file.
 * </ul><p>
 *********************************************************************/

public class SimpleTimeGraph {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	private Container root = null;
	private Hashtable idTab = new Hashtable();
	private PrintWriter writer = null;
	private PrintWriter errWriter = null;
	private boolean hasEventConditions = false;
	private boolean supported = false;

	public SimpleTimeGraph(PrintWriter writer, PrintWriter errWriter) {
		this.writer = writer;
		this.errWriter = errWriter;
   }
    
	public void setPrintWriter(PrintWriter writer) {
		this.writer = writer;
	}

	public void setErrPrintWriter(PrintWriter errWriter) {
		this.errWriter = errWriter;
	}

	public void setRoot(Container root) {
		this.root = root;
		idTab.put(root.getID(), root);
	}

	public Container getRoot() {
		return root;
	}

	public void setContainerID(Integer ID, Container container) {
		idTab.put(ID, container);
	}

	public Container getContainerByID(Integer ID) {
		return (Container) idTab.get(ID);
	}

	public Hashtable getIDTab() {
		return idTab;
	}

	public PrintWriter getPrintWriter() {
		return writer;
	}

	public PrintWriter getErrPrintWriter() {
		return errWriter;
	}

	public SimpleTimeContainer getSimpleTimeContainer(TimeContainer timeContainer) {
		return (SimpleTimeContainer) getContainerByID(timeContainer.getID());
	}

	public boolean isSupported() {
		return hasEventConditions & supported;
	}

	public void initialize(TimeGraph timeGraph) {
	    
		TimeContainer root = (TimeContainer) timeGraph.getRoot();
		try {
			SimpleTimeContainer simpleRoot = new SimpleTimeContainer(this, (Container) null, root);
			setRoot(simpleRoot);
			setContainerID(root.getID(), simpleRoot);
			loadChildren(simpleRoot, root);
		} catch (UnsupportedElementException e) {
			return;
		}
		supported = true;
	}
	
	private void loadChildren(SimpleTimeContainer simpleParent, TimeContainer parent) throws UnsupportedElementException {
		if (simpleParent.hasEventConditions()) {
			hasEventConditions = true;
		}
		for (Enumeration e = parent.children(); e.hasMoreElements();) {
			TimeContainer child = (TimeContainer) e.nextElement();
			SimpleTimeContainer simpleChild = new SimpleTimeContainer(this, simpleParent, child);
			simpleParent.addChild(simpleChild);
			setContainerID(child.getID(), simpleChild);
			loadChildren(simpleChild, child);
		}
	}
			

	public void play() {

		printTimeGraph(true);

		SimpleTimeContainer rootTimeContainer = (SimpleTimeContainer) root;
		rootTimeContainer.setBeginTime(new Interval(new Time(0), new Time(0)));
		rootTimeContainer.update(null);
		printTimeGraph(true);
	}

	private void printTimeGraph(boolean all) {
		writer.println();
		writer.println("supported = "+isSupported());
		if (isSupported()) {
			printContainer(root, 0, writer, all);
			printContainerHistory(root, 1, writer);
		}
		writer.flush();
	}

	private void printContainer(Container container, int depth, PrintWriter out, boolean all) {
		SimpleTimeContainer timeContainer = (SimpleTimeContainer) container;
		if (timeContainer != null) {
			printIndent(depth, out);
			out.println("CONTAINER ID="+timeContainer.getID());

			Vector conditions = timeContainer.getBeginConditions();
			for (Enumeration e = conditions.elements(); e.hasMoreElements(); ) {
				Condition condition = (Condition) e.nextElement();
				printIndent(depth+1, out);
				out.println("begin " + condition);
			}
			conditions = timeContainer.getEndConditions();
			for (Enumeration e = conditions.elements(); e.hasMoreElements(); ) {
				Condition condition = (Condition) e.nextElement();
				printIndent(depth+1, out);
				out.println("end " + condition);
			}
			if (all) {
				Vector dependents = timeContainer.getSyncbaseDependents();
				printIndent(depth+1, out);
				out.print("syncbase dependents ");
				for (Enumeration e = dependents.elements(); e.hasMoreElements(); ) {
					SimpleTimeContainer dependent = getSimpleTimeContainer( (TimeContainer) e.nextElement() );
					out.print(dependent.getID()+", ");
				}
				out.println();
				dependents = timeContainer.getEventDependents();
				printIndent(depth+1, out);
				out.print("event dependents ");
				for (Enumeration e = dependents.elements(); e.hasMoreElements(); ) {
					SimpleTimeContainer dependent = getSimpleTimeContainer( (TimeContainer) e.nextElement() );
					out.print(dependent.getID()+", ");
				}
				out.println();

				printIndent(depth+1, out);
				out.print("dur type="+ TimeContainer.DURTYPE_NAME[timeContainer.getDurType()]);
				if (timeContainer.getDurType() == TimeContainer.DURTYPE_TIMEVALUE) {
					out.println(", value="+timeContainer.getDur());
				} else {
					out.println();
				}
				printIndent(depth+1, out);
				out.print("min type="+ TimeContainer.MINMAX_NAME[timeContainer.getMinType()]);
				if (timeContainer.getMinType() == TimeContainer.MINMAX_TIMEVALUE) {
					out.print(", value="+timeContainer.getMin()+", ");
				} else {
					out.print(", ");
				}
				out.print("max type="+ TimeContainer.MINMAX_NAME[timeContainer.getMaxType()]);
				if (timeContainer.getMaxType() == TimeContainer.MINMAX_TIMEVALUE) {
					out.println(", value="+timeContainer.getMax());
				} else {
					out.println();
				}
			}			
			printIndent(depth+1, out);
			out.println("simple dur="+timeContainer.getSimpleDur()
							+", intrinsic="+timeContainer.getIntrinsicDur());
			printIndent(depth+1, out);
			out.println("end condition time="+timeContainer.getEndConditionTime());
			printIndent(depth+1, out);
			out.println("begin time="+timeContainer.getBeginTime());
			printIndent(depth+1, out);
			out.println("end time="+timeContainer.getEndTime());
			printIndent(depth+1, out);
			out.println("begin constrain time="+timeContainer.getBeginConstrainTime());
			printIndent(depth+1, out);
			out.println("end constrain time="+timeContainer.getEndConstrainTime());
			printIndent(depth+1, out);
			out.println("inherited begin condition="+timeContainer.getInheritedBeginCondition());
			printIndent(depth+1, out);
			out.println("inherited end condition="+timeContainer.getInheritedEndCondition());
			for (Enumeration e = container.children(); e.hasMoreElements(); ) {
				printContainer((Container) e.nextElement(), depth+1, out, all);
			}
		}
	}

	private void printContainerHistory(Container container, int depth, PrintWriter out) {
		SimpleTimeContainer timeContainer = (SimpleTimeContainer) container;
		if (timeContainer != null) {
			printLifeHistory(timeContainer.getID(), timeContainer.getLifeHistory(), depth, out);
			for (Enumeration e = container.children(); e.hasMoreElements(); ) {
				printContainerHistory((Container) e.nextElement(), depth, out);
			}
		}
	}

	private void printLifeHistory(Integer id, Vector history, int depth, PrintWriter out) {
		printIndent(depth, out);
		out.println("life history for container id="+id);
		for (Enumeration e = history.elements(); e.hasMoreElements();) {
			ContainerLife containerLife = (ContainerLife) e.nextElement();
			printIndent(depth+1, out);
			out.println("life cycle "+containerLife);
			printIndent(depth+1, out);
			out.println("parent ID="+containerLife.getParentELASid());
			for (Enumeration i = containerLife.getLifeIntervals().elements(); i.hasMoreElements();) {
				LifeInterval lifeInterval = (LifeInterval) i.nextElement();
				printIndent(depth+2, out);
				if (lifeInterval.getType() == LifeInterval.ACTIVE) {
					out.print("active interval "+lifeInterval+" observed begin="+lifeInterval.getObservedBegin()+
					" intrinsic dur="+lifeInterval.getIntrinsicDur()+" simple durs=");
					for (Enumeration j = lifeInterval.getSimpleDurs().elements(); j.hasMoreElements();) {
						Time simpleDur = (Time) j.nextElement();
						out.print(simpleDur+", ");
					}
					out.println();
					printIndent(depth+2, out);
					out.print("elasids=");
					for (Enumeration k = lifeInterval.getELASids().elements(); k.hasMoreElements();) {
						ELASid elasid = (ELASid) k.nextElement();
						out.print(elasid+", ");
					}
				} else {
					out.print("fill interval "+lifeInterval+", type="+TimeContainer.FILL_NAME[lifeInterval.getType()]);
				}
				out.println();
			}
		}
	}
			
	private void printIndent(int depth, PrintWriter out) {
		for (int i = 0; i < depth*3; i++) out.print(' ');
	}

}   // class SimpleTimeGraph
