/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  Time.java
//
//=============================================================================
package xmto2xmta.timegraph;

import java.util.*;

/*********************************************************************** 
 * Represents a point in time, and arithmetic operations on time.
 *
 * <p><b>History:</b><ul>
 * <li>15 Mar 2001 - Created file.
 * </ul><p>
 *********************************************************************/

public class Time implements Comparable {

	public static String UNRESOLVED = "unresolved";
	public static String INDEFINITE = "indefinite";
	public static long INFINITY = Long.MAX_VALUE;

	private static String INFINITY_STRING = "infinity";
	private static final int ADD = 0;
	private static final int SUBTRACT = 1;
	private static final int MULTIPLY = 2;

	private boolean resolved = false;
	private boolean indefinite = false;

	private Long value = new Long(0);		// in milliseconds
	
	// ctors
	public Time(Time timeInstance) {
		setTime(timeInstance);
	}	// ctor Time(Time)

	public Time(long value) {
		resolved = true;
		indefinite = false;
		this.value = new Long(value);
	}

	public Time() {					// unresolved
		setUnResolved();
	}

	public Time(String special) {
		setTime(special);
	}

	public boolean isResolved() {
		return resolved;
	}

	public boolean isIndefinite() throws UnresolvedTimeException {
		if (resolved) {
			return indefinite;
		} else {
			throw new UnresolvedTimeException();
		}
	}

	public boolean isDefinite() {
		return (resolved && !indefinite);
	}

	public long getValue() throws UnresolvedTimeException,
								IndefiniteTimeException {
		if (resolved) {
			if (indefinite) {
				throw new IndefiniteTimeException();
			} else {
				return value.longValue();
			}
		} else {
			throw new UnresolvedTimeException();
		}
	}

	public boolean isEqual(Time wrt) {
		try {
			if (!resolved) {
				if (!wrt.isResolved()) {
					return true;
				} else {
					return false;
				}
			} else if (indefinite) {
				if (wrt.isResolved() && wrt.isIndefinite()) {
					return true;
				} else {
					return false;
				}
			} else {
				if (!wrt.isResolved() || wrt.isIndefinite()) {
					return false;
				} else {
					return value.longValue() == wrt.getValue();
				}
			}
		} catch (Exception e) {	// shouldn't happen!
			System.out.println("Time pos=2");
			e.printStackTrace();
			return false;
		}
	}	// isEqual()

	public boolean isGreater(Time wrt) {
		try {
			if (!resolved) {
				if (wrt.isResolved()) {
					return true;
				} else {
					return false;
				}
			} else if (indefinite) {
				if (!wrt.isResolved() || wrt.isIndefinite()) {
					return false;
				} else {
					return true;
				}
			} else {
				if (!wrt.isResolved() || wrt.isIndefinite()) {
					return false;
				} else {
					return value.longValue() > wrt.getValue();
				}
			}
		} catch (Exception e) {	// shouldn't happen!
			System.out.println("Time pos=1");
			e.printStackTrace();
			return false;
		}
	}

	public boolean isGreaterEqual(Time wrt) {
		return isGreater(wrt) | isEqual(wrt);
	}

	public boolean isLess(Time wrt) {
		return !isGreaterEqual(wrt);
	}

	public boolean isLessEqual(Time wrt) {
		return !isGreater(wrt);
	}

	protected void setUnResolved() {
		resolved = false;
		indefinite = false;
		value = new Long(0);
	}

	protected void setIndefinite() {
		resolved = true;
		indefinite = true;
		value = new Long(0);
	}

	public void setTime(String special) {
		if (special != null) {
			if (special.equals(Time.UNRESOLVED)) {
				setUnResolved();
			} else if (special.equals(Time.INDEFINITE)) {
				setIndefinite();
			} else {
				setUnResolved();
			}
		} else {
			setUnResolved();
		}
	}

	public void setTime(long value) {
		this.value = new Long(value);
		resolved = true;
		indefinite = false;
	}

	public void setTime(Time timeInstance) {
		if (timeInstance != null) {
			try {
				resolved = timeInstance.isResolved();
				if (resolved) {
					indefinite = timeInstance.isIndefinite();
					if (!indefinite) {
						value = new Long(timeInstance.getValue());
					}
				}
			} catch (Exception e) {	// shouldn't happen!
				System.out.println("Time pos=3");
			e.printStackTrace();
				setUnResolved();
			}
		} else {
			setUnResolved();
		}
	}

	private Time add(Time time) {
		setTime(add(this, time));
		return this;
	}

	private Time subtract(Time time) {
		setTime(subtract(this, time));
		return this;
	}

	private Time multiply(Time time) {
		setTime(multiply(this, time));
		return this;
	}

	private static Time arithmetic(Time time1, Time time2, int op) {
		Time result = new Time();
		if (time1.isResolved() && time2.isResolved()) {
			try {
				switch (op) {
					case Time.ADD:
						if (time1.isIndefinite() || time2.isIndefinite()) {
							result.setIndefinite();
						} else {
							result.setTime(time1.getValue() + time2.getValue());
						}
						break;
					case Time.SUBTRACT:
						if (time1.isIndefinite() || time2.isIndefinite()) {
							result.setIndefinite();
						} else {
							result.setTime(time1.getValue() - time2.getValue());
						}
						break;
					case Time.MULTIPLY:
						if ((time1.isDefinite() && time1.getValue() == 0) ||
								(time2.isDefinite() && time2.getValue() == 0)) {
							result.setTime(0);
						} else if (time1.isDefinite() && time2.isDefinite()) {
							result.setTime(time1.getValue() * time2.getValue());
						} else {
							result.setIndefinite();
						}
						break;
				}
			} catch (Exception e) {	// shouldn't happen!
				System.out.println("Time pos=4");
			e.printStackTrace();
				result.setUnResolved();
			}
		} else {
			result.setUnResolved();
		}
		return result;
	}

	public static Time add(Time time1, Time time2) {
		return arithmetic(time1, time2, Time.ADD);
	}
	
	public static Time subtract(Time time1, Time time2) {
		return arithmetic(time1, time2, Time.SUBTRACT);
	}

	public static Time multiply(Time time1, Time time2) {
		return arithmetic(time1, time2, Time.MULTIPLY);
	}

	public static Time add(Time time1, long offset) {
		return arithmetic(time1, new Time(offset), Time.ADD);
	}
	
	public static Time subtract(Time time1, long offset) {
		return arithmetic(time1, new Time(offset), Time.SUBTRACT);
	}

	/*******************************************************
	 * Multiply the given time by the repeat count.
	 * The repeat count is a floating point number >= 0.
	 * The Time value is a long integer that represents
	 * milliseconds.  This is why we need this special method.
	 * Repeat count == -1 means it's an indefinite value.
	 *******************************************************/
	public static Time multiply(Time time1, float repeatCount) {
		try {
			if (time1.isResolved()) {
				if (repeatCount == (float) -1.0) {
					return Time.multiply(time1, new Time(Time.INDEFINITE));
				} else if (repeatCount == (float) 0.0) {
					return Time.multiply(time1, new Time(0));
				} else {
					if (time1.isIndefinite()) {
						return new Time(Time.INDEFINITE);
					} else {  // finally definite values all around
						long newValue = new Float(time1.getValue() * repeatCount).longValue();
						return new Time(newValue);
					}
				}
			} else {
				return new Time();
			}
		} catch (Exception e) {
			return new Time();
		}
	}
			
	public static Time min(Vector times) {	// min of Time
		if (times == null) {
			return null;
		}
		Time minTime = null;
		for (Enumeration enum = times.elements(); enum.hasMoreElements();) {
			Time curTime = (Time) enum.nextElement();
			if ((minTime == null) || curTime.isLess(minTime)) {
				minTime = curTime;
			}
		}
		return minTime;
	}	

	public static Time max(Vector times) { // max of Time
		if (times == null) {
			return null;
		}
		Time maxTime = null;
		for (Enumeration enum = times.elements(); enum.hasMoreElements();) {
			Time curTime = (Time) enum.nextElement();
			if ((maxTime == null) || curTime.isGreater(maxTime)) {
				maxTime = curTime;
			}
		}
		return maxTime;
	}

	public static Time remainder(Time time1, Time time2) {
		try {
			if (time1 == null || time2 == null ||
				!time1.isResolved() || !time2.isResolved()) {
				return new Time();
		    }
			if (time1.isIndefinite() || time2.isIndefinite()) {
				return new Time(Time.INDEFINITE);
			}
			long modulo = time2.getValue();
			if (modulo == 0) {
				return new Time();
			}
			return new Time(time1.getValue() % modulo);
		} catch (Exception e) {
			return new Time();
		}
	}

	public String toString() {
		if (!resolved) {
			return Time.UNRESOLVED;
		} else if (indefinite) {
			return Time.INDEFINITE;
		} else if (value.longValue() == Time.INFINITY) {
			return Time.INFINITY_STRING;
		} else if (value.longValue() == -Time.INFINITY) {
			return "-" + Time.INFINITY_STRING;
		} else {
			return value.toString();
		}
	}

	public int compareTo(Object object) {
		int result = 1;
		if (object instanceof Time) {
			Time time = (Time) object;
			if (this.isEqual(time)) result = 0;
			if (this.isLess(time)) result = -1;
			if (this.isGreater(time)) result = 1;
		}
		return result;
	}

	public int hashCode() {
		if (!resolved) return UNRESOLVED.hashCode();
		if (indefinite) return INDEFINITE.hashCode();
		return value.hashCode();
	}

	public boolean equals(Object object) {
		if (object instanceof Time) {
			return this.isEqual((Time) object);
		} else {
			return false;
		}
	}

}	// class Time

