/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  TimeContainer.java
//
//=============================================================================
package xmto2xmta.timegraph;

import java.util.*;

/*********************************************************************** 
 * Time container in timing graph. Parent of all actual timed elements.
 *
 * <p><b>History:</b><ul>
 * <li>15 Mar 2001 - Created file.
 * </ul><p>
 *********************************************************************/

public class TimeContainer extends Container {

	public static final int STATE_DEAD = 0;
	public static final int STATE_ACTIVE = 1;
	public static final int STATE_WAITING = 2;

	public static final String[] STATE_NAME = {
		"dead", "active", "waiting", };

	public static final int FILL_DEFAULT = 0;
	public static final int FILL_REMOVE = 1;
	public static final int FILL_FREEZE = 2;
	public static final int FILL_HOLD = 3;
	public static final int FILL_TRANSITION = 4;
	public static final int FILL_AUTO = 5;
	public static final int FILL_INHERIT = 6;

	public static final String[] FILL_NAME = {
		"default", "remove", "freeze", "hold", "transition", "auto", "inherit", };

	public static final int RESTART_DEFAULT = 0;
	public static final int RESTART_ALWAYS = 1;
	public static final int RESTART_WHENNOTACTIVE = 2;
	public static final int RESTART_NEVER = 3;
	public static final int RESTART_INHERIT = 4;

	public static final String[] RESTART_NAME = {
		"default", "always", "whenNotActive", "never", "inherit", };

	public static final int ENDSYNC_MEDIA = 0;
	public static final int ENDSYNC_LAST = 1;
	public static final int ENDSYNC_ALL = 2;
	public static final int ENDSYNC_FIRST = 3;
	public static final int ENDSYNC_ID = 4;	

	public static final String[] ENDSYNC_NAME = {
		"media", "last", "all", "first", "id", };

	public static final int DURTYPE_UNSPECIFIED = 0;
	public static final int DURTYPE_TIMEVALUE = 1;
	public static final int DURTYPE_MEDIA = 2;

	public static final String[] DURTYPE_NAME = {
		"unspecified", "timeValue", "media", };

	public static final int MINMAX_TIMEVALUE = 0;
	public static final int MINMAX_MEDIA = 1;

	public static final String[] MINMAX_NAME = {
		"timeValue", "media", };

	public static final int BEGIN = 0;  // begin sync array index
	public static final int END = 1;		// end sync array index

	/**************************************************************
	 * The following section contains timing attributes from the
	 * XMT-O doc element.  Once set,these values cannot change throughout
	 * the life of the document.
	 **************************************************************/
	private Vector[] syncConditions = { new Vector(), new Vector() };
	private int durType = DURTYPE_UNSPECIFIED;
	private Time dur = new Time();
	private boolean repeatCountSpecified = false;
	private float repeatCount = (float) 1.0;  // -1.0 means indefinite
	private boolean repeatCountIndefinite = false;
	private boolean repeatDurSpecified = false;
	private Time repeatDur = new Time();
	private int restart = TimeContainer.RESTART_ALWAYS;
	private int restartDefault = TimeContainer.RESTART_ALWAYS;
	private int fill = TimeContainer.FILL_AUTO;
	private int fillDefault = TimeContainer.FILL_AUTO;
	private int minType = TimeContainer.MINMAX_TIMEVALUE;
	private Time min = new Time(0);
	private int maxType = TimeContainer.MINMAX_TIMEVALUE;
	private Time max = new Time(Time.INDEFINITE);
	private int endsyncRule = TimeContainer.ENDSYNC_MEDIA;
	private Integer endsyncID = new Integer(-1);
	private Vector syncbaseDependents = new Vector();	// of TimeContainer
	private Vector eventDependents = new Vector();	// of TimeContainer

	private InstanceTimeList[] instanceTimeList = { new InstanceTimeList(), new InstanceTimeList() };
	private HashMap globalToLocalTimeCache = new HashMap();
	private Vector lifeHistory = new Vector();	// history of life cycles
	private ContainerLife containerLife = null;	// current life cycle
	private LifeInterval lifeInterval = null;		// current active interval in the current life cycle
	private LifeInterval fillInterval = null;		// current fill interval, if any
	private boolean fillTime = false;				// true if currently doing a fill

	private int state = TimeContainer.STATE_DEAD;	// element current state

	private Time intrinsicDur = new Time();
	private Time implicitDur = new Time();
	private Time simpleDur = new Time();
	private Time naturalEnd = new Time();
	private Interval currentInterval = null;
	private int accIterationCount = 0;	// number of iterations so far
	private Time accIterationDur = new Time(0);	// time of the most recent iteration
	private boolean hasStarted = false;		// true if element has ever begun
	private Time lifeBegin = new Time();	// beginning time of this life cycle
	private Time lastActiveEnd = new Time();
	private Time curSamplingTime = new Time(0);
	private boolean siblingSaysNoFill = false;
	private boolean endSpecifiedDetermined = false;
	private boolean endSpecifiedResult = false;

	// constructors
	public TimeContainer() {
		this(null, new Integer(-1), null);
	}

	public TimeContainer(Integer id, Container parent) {
		this(null, id, parent);
	}

	public TimeContainer(TimeGraph timeGraph, Integer id, Container parent) {
		super(timeGraph, id, parent);
	}

	public void setBeginConditions(Vector conditions) {
		syncConditions[BEGIN] = conditions;
	}

	public Vector getBeginConditions() {
		return syncConditions[BEGIN];
	}

	public void setEndConditions(Vector conditions) {
		syncConditions[END] = conditions;
	}

	public Vector getEndConditions() {
		return syncConditions[END];
	}

	public void setSyncbaseDependents(Vector dependents) {
		syncbaseDependents = dependents;
	}

	public Vector getSyncbaseDependents() {
		return syncbaseDependents;
	}

	public void setEventDependents(Vector dependents) {
		eventDependents = dependents;
	}

	public Vector getEventDependents() {
		return eventDependents;
	}

	public void setDur(int durType, Time dur) {
		this.durType = durType;
		this.dur = dur;
	}

	public int getDurType() {
		return durType;
	}

	public Time getDur() {
		return dur;
	}

	public void setRepeatCount(float repeatCount) {
		if (repeatCount > 0.0) {
			this.repeatCount = repeatCount;
			repeatCountSpecified = true;
			repeatCountIndefinite = false;
		}
	}

	public void setRepeatCountIndefinite() {
		repeatCountIndefinite = true;
		repeatCountSpecified = true;
	}

	public boolean isRepeatCountSpecified() {
		return repeatCountSpecified;
	}

	public boolean isRepeatCountIndefinite() {
		return repeatCountIndefinite;
	}

	public float getRepeatCount() {
		return repeatCount;
	}

	public void setRepeatDur(Time repeatDur) {
		this.repeatDur = repeatDur;
		repeatDurSpecified = true;
	}

	public boolean isRepeatDurSpecified() {
		return repeatDurSpecified;
	}

	public Time getRepeatDur() {
		return repeatDur;
	}

	public void setMin(int minType, Time min) {
		this.minType = minType;
		this.min = min;
	}

	public int getMinType() {
		return minType;
	}

	public Time getMin() {
		return min;
	}

	public void setMax(int maxType, Time max) {
		this.maxType = maxType;
		this.max = max;
	}

	public int getMaxType() {
		return maxType;
	}

	public Time getMax() {
		return max;
	}

	public void setRestart(int restart) {
		if (restart == TimeContainer.RESTART_DEFAULT) {
			restart = restartDefault;
		}
	}

	public int getRestart() {
		return restart;
	}

	public void setRestartDefault(int restartDefault) {
		if (restartDefault == TimeContainer.RESTART_INHERIT) {
			if (getParent() != null) {
				this.restartDefault = ((TimeContainer)getParent()).getRestartDefault();
			} else {
				this.restartDefault = TimeContainer.RESTART_ALWAYS;
			}
		}
		if (restart == TimeContainer.RESTART_DEFAULT) {
			restart = this.restartDefault;
		}
	}

	public int getRestartDefault() {
		return restartDefault;
	}

	public void setFill(int fill) {
		if (fill == TimeContainer.FILL_DEFAULT) {
			fill = fillDefault;
		} else {
			this.fill = fill;
		}
	}

	public int getFill() {
		return fill;
	}

	public void setFillDefault(int fillDefault) {
		if (fillDefault == TimeContainer.FILL_INHERIT) {
			if (getParent() != null) {
				this.fillDefault = ((TimeContainer)getParent()).getFillDefault();
			} else {
				this.fillDefault = TimeContainer.FILL_AUTO;
			}
		}
		if (fill == TimeContainer.FILL_DEFAULT) {
			fill = this.fillDefault;
		}
	}

	public int getFillDefault() {
		return fillDefault;
	}

	public void setEndsyncRule(int endsyncRule, Integer endsyncID) {
		if (endsyncRule != this.endsyncRule) {
			this.endsyncRule = endsyncRule;
			if (endsyncRule == TimeContainer.ENDSYNC_ID) {
				this.endsyncID = endsyncID;
			}
		}
	}

	public void setEndsyncRule(int endsyncRule) {
		this.setEndsyncRule(endsyncRule, new Integer(-1));
	}

	public int getEndsyncRule() {
		return endsyncRule;
	}

	public Integer getEndsyncID() {
		return endsyncID;
	}

	public void setIntrinsicDur(Time intrinsicDur) {
		this.intrinsicDur = intrinsicDur;
	}

	public Time getIntrinsicDur() {
		return intrinsicDur;
	}

	public void setImplicitDur(Time implicitDur) {
		this.implicitDur = implicitDur;
	}

	public Time getImplicitDur() {
		return implicitDur;
	}

	public void setSimpleDur(Time simpleDur) {
		this.simpleDur = simpleDur;
	}

	public Time getSimpleDur() {
		return simpleDur;
	}

	public Time getAccIterationDur() {
		return accIterationDur;
	}

	public int getAccIterationCount() {
		return accIterationCount;
	}

	public boolean isActiveOrDead() {
		return ((state == TimeContainer.STATE_ACTIVE) || (state == TimeContainer.STATE_DEAD));
	}

	public boolean hasStarted() {
		return hasStarted;
	}	

	public Interval getCurrentInterval() {
		return currentInterval;
	}

	public boolean hasResolvedBegin(Time from) {
		// return (instanceTimeList[BEGIN].hasResolvedTimeAfter(from));
		return (currentInterval != null);
	}

	public boolean endSpecified() {
		// return (syncConditions[END].size() > 0);
		if (!endSpecifiedDetermined) {
			boolean deterministic = false;  // assume false and only need to find one condition
			for (Enumeration e = syncConditions[END].elements(); e.hasMoreElements();) {
				Condition condition = (Condition) e.nextElement();
				if (condition.isDeterministic()) {
					deterministic = true;
					break;
				}
			}
			if (deterministic) {
				endSpecifiedResult = true;
			} else {
				endSpecifiedResult = false;
			}
			endSpecifiedDetermined = true;
		}
		return endSpecifiedResult;
	}

	public boolean durSpecified() {
		return (durType != TimeContainer.DURTYPE_UNSPECIFIED);
	}

	public boolean repeatDurSpecified() {
		return repeatDurSpecified;
	}

	public boolean repeatCountSpecified() {
		return repeatCountSpecified;
	}

	public int getState() {
		return state;
	}

	public InstanceTimeList getBeginInstanceTimeList() {
		return instanceTimeList[BEGIN];
	}

	public InstanceTimeList getEndInstanceTimeList() {
		return instanceTimeList[END];
	}

	public Vector getLifeHistory() {
		return lifeHistory;
	}

	/**************************************************************
	 * The main entry point for the time container.  The time graph
	 * calls it at the next scheduled event time for this container.
	 **************************************************************/
	public void sample() {
		switch (state) {
			case TimeContainer.STATE_ACTIVE:
				Time[] newEndTime = calcActiveEnd(currentInterval.getBeginTime(), currentInterval.getEndTime());
				if (!newEndTime[0].isEqual(currentInterval.getEndTime())) {
					currentInterval.setEndTime(newEndTime[0]);
					sendChangeIntervalNotice();
				}
				naturalEnd = newEndTime[1];
				writer.println(getID()+" natural end="+naturalEnd);
				if (currentInterval.getEndTime().isDefinite()) {
					if (getCurrentTime().isGreaterEqual(currentInterval.getEndTime())) {
						// end the current interval
						endCurInterval();
					} else if (getCurrentTime().isEqual(naturalEnd)) {
						endNaturalInterval();
					} else {
						Time nextMileStone = Time.add(currentInterval.getBeginTime(), Time.add(accIterationDur, simpleDur));
						writer.println(getID()+" temp next mile stone="+nextMileStone);
						if (nextMileStone.isGreater(currentInterval.getEndTime())) {
							nextMileStone = currentInterval.getEndTime();
						}
						writer.println(getID()+" final next mile stone="+nextMileStone);
						if (nextMileStone.isDefinite()) {
							if (nextMileStone.isEqual(getCurrentTime())) {
								if (nextMileStone.isEqual(currentInterval.getEndTime())) {
									endCurInterval();
								} else if (nextMileStone.isEqual(naturalEnd)) {
									endNaturalInterval();
								} else {
									repeatSimpleDuration();
								}
							} else if (nextMileStone.isGreater(getCurrentTime())) {
								nextSamplingTime(nextMileStone, "next simple end");
							} else {
								writer.println(getID()+" fullfilling min time");
								nextSamplingTime(currentInterval.getEndTime(), "end of current interval");
							}
						}
					}
				} else {	// indefinite or unresolved active end
					writer.println(getID()+" active end is not definite.");
					Time nextMileStone = Time.add(currentInterval.getBeginTime(), Time.add(accIterationDur, simpleDur));
					if (nextMileStone.isDefinite()) {
						if (nextMileStone.isEqual(getCurrentTime())) {
							repeatSimpleDuration();
						} else if (nextMileStone.isGreater(getCurrentTime())) {
							nextSamplingTime(nextMileStone, "next simple end");
						} else {
							writer.println(getID()+" next mile stone cannot be less than now");
						}
					} else {
						try {
							if (nextMileStone.isIndefinite()) {
								nextSamplingTime(new Time(Time.INFINITY), "indefinite end");
							}
						} catch (Exception e) {}
					}
				}
				break;
			case TimeContainer.STATE_WAITING:
				if (currentInterval != null) {
					if (currentInterval.getBeginTime().isLessEqual(getCurrentTime())) {
						startNewInterval();
					} else {
						nextSamplingTime(currentInterval.getBeginTime(), "start next interval");
					}
				} else {
					writer.println(getID()+" current interval is null. nothing to do");
				}
				break;
		}
	}

	public void nextSamplingTime(Time globalTime, String msg) {
			writer.println(getID()+" next sampling at "+globalTime+" "+msg);
			curSamplingTime = globalTime;
			getTimeGraph().scheduleSamplingTime(this, globalTime);
	}

	public void removeSamplingTime(Time globalTime) {
		getTimeGraph().removeSamplingTime(this, globalTime);
	}

	/***************************************************************
	 * Called by the parent when the parent's simple duration begins
	 * at the start of a new active duration or at each repeat.
	 * The element resets itself to a pristine state ready to start a
	 * new life cycle.  'whoStarted' is the highest ancestor that started
	 * the chain of reset events.
	 ***************************************************************/
	public void reset(TimeContainer whoRestarted, ELASid parentELASid) {
		writer.println(getID()+" resetting");

		containerLife = new ContainerLife(getCurrentTime(), new Time(Time.INDEFINITE));
		containerLife.setContainer(this);
		containerLife.setParentELASid(parentELASid);
		lifeHistory.add(containerLife);

		globalToLocalTimeCache.clear();

		if (whoRestarted == null) { // it was from the time graph
			whoRestarted = this;
		}

		for (Enumeration e = children(); e.hasMoreElements();) {
			TimeContainer child = (TimeContainer) e.nextElement();
			child.removeSyncbaseInstanceTimes(whoRestarted);
		}
		resyncSyncbaseInstanceTimes(whoRestarted, syncConditions[BEGIN], instanceTimeList[BEGIN]);
		resyncSyncbaseInstanceTimes(whoRestarted, syncConditions[END], instanceTimeList[END]);

		currentInterval = null;
		hasStarted = false;
		lastActiveEnd = new Time(-Time.INFINITY);
		lifeBegin = getCurrentTime();
		fillTime = false;
		siblingSaysNoFill = false;
		state = TimeContainer.STATE_WAITING;

		// compute the first interval and notify dependents
		currentInterval = getNextActiveInterval(lastActiveEnd, getCurrentTime(), false);
		if (currentInterval != null) {
			sendNewIntervalNotice();
			if (currentInterval.getBeginTime().isGreaterEqual(getCurrentTime())) {
				state = TimeContainer.STATE_WAITING;
				nextSamplingTime(currentInterval.getBeginTime(), "start first interval "+currentInterval);
			} else {
				startNewInterval();
			}
		}
	}

	/******************************************************************
	 * When an element is reset, must resync the syncbase instance times.
	 * The syncbase instance times are in the global time.  Convert the
	 * global time to the local time.
	 *******************************************************************/
	private void resyncSyncbaseInstanceTimes(TimeContainer whoStarted, Vector conditions,
															InstanceTimeList instanceTimeList) {
		for (Enumeration e = conditions.elements(); e.hasMoreElements();) {
			Condition condition = (Condition) e.nextElement();
			if (condition.isSyncbaseType()) {
				TimeContainer syncbase = (TimeContainer) condition.getSyncbase();
				if (syncbase != null) {
					if (!condition.getOffsetCondition()) {
						for (Iterator iter = condition.getInstanceTimes().iterator(); iter.hasNext();) {
							Time instanceTime = (Time) iter.next();
							convertFromGlobalToLocalTime(instanceTime);
						}
					} else {	// it's offset based begin. recompute the instance time
						for (Iterator i = condition.getInstanceTimes().iterator(); i.hasNext();) {
							Time instanceTime = (Time) i.next();
							instanceTimeList.removeTime(instanceTime, condition);
						}
						condition.setInstanceTimes(new HashSet());
						Time newInstanceTime = Time.add(
										getCurrentTime(), new Time(condition.getOffsetValue()));
						condition.getInstanceTimes().add(newInstanceTime);
						instanceTimeList.putTime(newInstanceTime, condition);
						condition.setCurrentInstanceTime(newInstanceTime);
					}
				}	// else it's a special provision for the root container
			} else if (condition.isIndefinite()) {
				condition.getInstanceTimes().add(new Time(Time.INDEFINITE));
				instanceTimeList.putTime(new Time(Time.INDEFINITE), condition);
			}
		}
	}

	/******************************************************************
	 * Called by the parent when an ancestor is reset.
	 * When an element is reset,  all descendants' syncbase instance times 
	 * are removed when the resetting element is a common ascendant of
	 * both the syncbase and the dependent.
	 *******************************************************************/
	public void removeSyncbaseInstanceTimes(TimeContainer whoRestarted) {
		if (whoRestarted == null) return;
		writer.println(getID()+" removing syncbase instance times, restart by " + whoRestarted.getID());
		removeSyncbaseInstanceTimes(whoRestarted, syncConditions[BEGIN], instanceTimeList[BEGIN]);
		removeSyncbaseInstanceTimes(whoRestarted, syncConditions[END], instanceTimeList[END]);
		for (Enumeration e = children(); e.hasMoreElements();) {
			TimeContainer child = (TimeContainer) e.nextElement();
			child.removeSyncbaseInstanceTimes(whoRestarted);
		}
	}

	private void removeSyncbaseInstanceTimes(TimeContainer whoRestarted,
															Vector conditions,
															InstanceTimeList instanceTimeList) {
		for (Enumeration e = conditions.elements(); e.hasMoreElements();) {
			Condition condition = (Condition) e.nextElement();
			if (condition.isSyncbaseType()) {
				TimeContainer syncbase = (TimeContainer) condition.getSyncbase();
				if (syncbase != null) {
					if (whoRestarted.isCommonAscendantOf(this, syncbase)) {
						for (Iterator iter = condition.getInstanceTimes().iterator(); iter.hasNext();) {
							Time instanceTime = (Time) iter.next();
							instanceTimeList.removeTime(instanceTime, condition);
							iter.remove();
						}
						condition.setCurrentInstanceTime(null);
					}
				}	// else it's a special provision for the root container
			}
		}
	}

	/**************************************************************
	 * Called by the parent when the parent's simple duration ends
	 * at the end of the parent's active interval or at each repeat.
	 * We must stop the current active interval and stop all
	 * the descendants of this element as well.
	 ***************************************************************/
	public void stop(ELASid elasid) {
		writer.println(getID()+" stopping");
		switch (state) {
			case TimeContainer.STATE_ACTIVE:
				// send change interval notice to dependents
				currentInterval.setEndTime(getCurrentTime());
				sendChangeIntervalNotice();
				stopCurInterval();
				break;
		   case TimeContainer.STATE_WAITING:
				// send delete interval notice to dependents
				if (currentInterval != null) {
					sendDeleteIntervalNotice();
				}
				break;
		}
		state = TimeContainer.STATE_DEAD;
		containerLife.setEndTime(getCurrentTime());
		containerLife.setParentELASid(elasid);
		stopFill();
		// containerLife = null;
	}
	
	public void stop() {
		stop(null);
	}

	/********************************************************
	 * Element is scheduled to start a new active interval now.
	 ********************************************************/
	public void startNewInterval() {
		writer.println(getID()+" starting a new interval");

		stopPrevSiblingFill();
		stopFill();
		lifeInterval = new LifeInterval(currentInterval.getBeginTime(), new Time(Time.INDEFINITE));
		if (currentInterval.getBeginTime().isLess(lifeBegin)) {
			lifeInterval.setObservedBegin(lifeBegin);
		} else {
			lifeInterval.setObservedBegin(currentInterval.getBeginTime());
		}
		lifeInterval.setIntrinsicDur(intrinsicDur);
		containerLife.addInterval(lifeInterval);

		clearAllEventInstanceTimes(syncConditions[BEGIN], instanceTimeList[BEGIN]);
		clearAllEventInstanceTimes(syncConditions[END], instanceTimeList[END]);
		implicitDur = new Time();
		simpleDur = new Time();
		accIterationDur = new Time(0);
		accIterationCount = 0;

		// send beginEvent notice to dependents here
		state = TimeContainer.STATE_ACTIVE;
		hasStarted = true;	// element now has at least one active interval
		// start the first simple duration
		if (currentInterval.getBeginTime().isEqual(getCurrentTime())) {
			beginSimpleDuration();
		} else { // pretend it started in the past if a media element
			fastForwardMedia();
		}
	}

	/****************************************************************
	 * When an element restarts, all event-based instance times must be
	 * cleared.
	 ****************************************************************/
	private void clearAllEventInstanceTimes(Vector conditions, InstanceTimeList instanceTimeList) {
		for (Enumeration e = conditions.elements(); e.hasMoreElements();) {
			Condition condition = (Condition) e.nextElement();
			if (condition.isEventType()) {
				for (Iterator i = condition.getInstanceTimes().iterator(); i.hasNext();) {
					Time instanceTime = (Time) i.next();
					instanceTimeList.removeTime(instanceTime, condition);
				}
				condition.setInstanceTimes(new HashSet());
				condition.setCurrentInstanceTime(null);
			}
		}
	}

	/********************************************************
	 * Element is scheduled to end the current active interval now.
	 ********************************************************/
	public void endCurInterval() {
		writer.println(getID()+" ending cur interval");
		stopCurInterval();
		if (restart == TimeContainer.RESTART_NEVER) {
			currentInterval = null;
			state = TimeContainer.STATE_WAITING;
		} else {
			// compute the next interval and notify dependents
			boolean skipZeroSizeInterval = false;
			if (currentInterval.getDuration().isEqual(new Time(0)) && currentInterval.getEndTime().isEqual(getCurrentTime())) {
				skipZeroSizeInterval = true;
			}
			writer.println(getID()+" in endCurInterval, skipZeroSizeInterval="+skipZeroSizeInterval);
			currentInterval = getNextActiveInterval(getCurrentTime(), getCurrentTime(), skipZeroSizeInterval);
			if (currentInterval != null) {
				sendNewIntervalNotice();
				if (currentInterval.getBeginTime().isGreaterEqual(getCurrentTime())) {
					state = TimeContainer.STATE_WAITING;
					nextSamplingTime(currentInterval.getBeginTime(), "start next interval");
				} else {
					startNewInterval();
				}
			} else {
				state = TimeContainer.STATE_WAITING;
			}
		}
		// notify parent if parent needs endsync calculation.
		TimeContainer parent = (TimeContainer) getParent();
		if (parent != null) {
			parent.childHasEnded();
		}
	}

	/********************************************************
	 * Stop the currently running active interval now.
	 ********************************************************/
	public void stopCurInterval() {
		writer.println(getID()+" stopping cur interval");
		// send endEvent notice to dependents here	
		writer.println(getID()+" begin="+currentInterval.getBeginTime()+", "+
			"natural end="+naturalEnd+", current time="+getCurrentTime());
		if (getCurrentTime().isLessEqual(naturalEnd)) {
			endSimpleDuration();
			lifeInterval.setEndTime(currentInterval.getEndTime());
			if (currentInterval.getBeginTime().isLess(lifeBegin)) {
				lifeInterval.setObservedBegin(lifeBegin);
			} else {
				lifeInterval.setObservedBegin(currentInterval.getBeginTime());
			}
			lifeInterval = null;
		}
		lastActiveEnd = getCurrentTime();

		removeSamplingTime(curSamplingTime);
		removeSamplingTime(new Time(Time.INFINITY));
		startFill();

	}

	/********************************************************
	 * End the current natural (no min) interval
	 ********************************************************/
	public void endNaturalInterval() {
		writer.println(getID()+" ending natural interval");
		endSimpleDuration();
		lifeInterval.setEndTime(naturalEnd);
		if (currentInterval.getBeginTime().isLess(lifeBegin)) {
			lifeInterval.setObservedBegin(lifeBegin);
		} else {
			lifeInterval.setObservedBegin(currentInterval.getBeginTime());
		}
		lifeInterval = null;
		nextSamplingTime(currentInterval.getEndTime(), "end of current interval");
		startFill();
	}

	/*********************************************************
	 * Called by a child when it ends its active interval.
	 * If the parent's simple time is unresolved because it
	 * depends on its endsync logic, whenever a child ends,
	 * the parent recalculates the implicit dur.
	 *********************************************************/
	public void childHasEnded() {
		writer.println(getID()+" a child has ended");
		nextSamplingTime(getCurrentTime(), "child has ended");
	}
		
	/***********************************************************
	 * Element is starting a new simple duration now.
	 ***********************************************************/
	public void beginSimpleDuration() {
		writer.println(getID()+" beginning a new simple dur");

		lifeInterval.getSimpleDurs().add(new Time(Time.INDEFINITE));
		ELASid elasid = new ELASid(getID(), new Integer(lifeHistory.size()-1),
											new Integer(containerLife.size() - 1), new Integer(lifeInterval.size() - 1));
		lifeInterval.getELASids().add(elasid);

		if (computeSimpleDur().isDefinite()) {
			Time nextRepeatBegin = Time.add(getCurrentTime(), simpleDur);
			if (nextRepeatBegin.isGreater(currentInterval.getEndTime())) {
				nextSamplingTime(currentInterval.getEndTime(), "end of current interval");
			} else if (nextRepeatBegin.isGreater(naturalEnd)) {
				nextSamplingTime(naturalEnd, "natural end");
			} else {
				nextSamplingTime(nextRepeatBegin, "end of a simple dur");
			}
		} else if (naturalEnd.isDefinite()) {
			nextSamplingTime(naturalEnd, "natural end");
		} else if (currentInterval.getEndTime().isDefinite()) {
			nextSamplingTime(currentInterval.getEndTime(), "end the current interval");
		} else {
			try {
				if (currentInterval.getEndTime().isIndefinite()) {
					nextSamplingTime(new Time(Time.INFINITY), "indefinite end");
				}
			} catch (Exception e) {}
		}
		lifeInterval.setLastComputedSimpleDur(getSimpleDur());
		// reset all children
		for (Enumeration e = children(); e.hasMoreElements();) {
			TimeContainer child = (TimeContainer) e.nextElement();
			child.reset(this, elasid);
		}
	}

	/***********************************************************
	 * Element is ending the current simple duration now.
	 ***********************************************************/
	public void endSimpleDuration() {
		writer.println(getID()+" ending cur simple dur");

		Time lifeSimpleDur = Time.subtract(Time.subtract(getCurrentTime(), accIterationDur),
										currentInterval.getBeginTime());
		lifeInterval.getSimpleDurs().removeElementAt(lifeInterval.getSimpleDurs().size() - 1);
		lifeInterval.getSimpleDurs().add(lifeSimpleDur);
		lifeInterval.setLastComputedSimpleDur(getSimpleDur());
//		ELASid elasid = new ELASid(getID(), new Integer(lifeHistory.size()-1),
//											new Integer(containerLife.size() - 1), new Integer(lifeInterval.size() - 1));
//		lifeInterval.getELASids().add(elasid);


		// note: this method is only called during the active state
		// stop all children
		accIterationDur = Time.subtract(getCurrentTime(), currentInterval.getBeginTime());
		accIterationCount++;
		
		for (Enumeration e = children(); e.hasMoreElements();) {
			TimeContainer child = (TimeContainer) e.nextElement();
//			child.stop(elasid);
			child.stop((ELASid) (lifeInterval.getELASids().lastElement()));
		}
	}

	public void repeatSimpleDuration() {
		writer.println(getID()+" repeating");
		// only called when the element is supposed to repeat
		endSimpleDuration();
		// send repeat(accIterationCount) event to dependents here
		beginSimpleDuration();
	}
		
	/***************************************************************
	 * Element begins earlier than the parent simple time.
	 * Pretend it started earlier and fast forward to the current
	 * time, which also happens to be the beginning of the current
	 * life cycle.  Only media elements with an intrinsic duration
	 * with no timed children are processed.  A hack.
	 ***************************************************************/
	public void fastForwardMedia() {
		try {
			long offset = Time.subtract(getCurrentTime(), currentInterval.getBeginTime()).getValue();
			writer.println(getID()+" offset="+offset);
			try {
				accIterationCount = (int) (offset / computeSimpleDur().getValue());
			} catch (Exception e) {
				accIterationCount = 0;
			}
			writer.println("completed iterations="+accIterationCount);
			accIterationDur = Time.multiply(simpleDur, new Time(accIterationCount));
			writer.println("accIterationDur="+accIterationDur);
			Time curIterationBegan = Time.add(currentInterval.getBeginTime(), accIterationDur);
			lifeInterval.setBeginTime(curIterationBegan);
			lifeInterval.getSimpleDurs().add(new Time(Time.INDEFINITE));
			ELASid elasid = new ELASid(getID(), new Integer(lifeHistory.size()-1),
											new Integer(containerLife.size() - 1), new Integer(lifeInterval.size() - 1));
			lifeInterval.getELASids().add(elasid);
			lifeInterval.setLastComputedSimpleDur(getSimpleDur());
			if (simpleDur.isDefinite()) {
				nextSamplingTime(Time.add(curIterationBegan, simpleDur), "end of simple interval");
			} else if (currentInterval.getEndTime().isDefinite()) {
				nextSamplingTime(currentInterval.getEndTime(), "end of active interval");
			} else {
				nextSamplingTime(new Time(Time.INFINITY), "indefinite");
			}
		} catch (Exception e) {}
	}
			
	/****************************************************************
	 * Notify all syncbase and event dependents of a new current interval.
	 ****************************************************************/
	public void sendNewIntervalNotice() {
		writer.println(getID()+" sending new interval notice "+currentInterval);
		for (Enumeration e = syncbaseDependents.elements(); e.hasMoreElements();) {
			TimeContainer dependent = (TimeContainer) e.nextElement();
			dependent.newIntervalNotice(this, currentInterval);
		}
		for (Enumeration e = eventDependents.elements(); e.hasMoreElements();) {
			TimeContainer dependent = (TimeContainer) e.nextElement();
			dependent.newIntervalNotice(this, currentInterval);
		}
	}

	/****************************************************************
	 * Called by a syncbase element.
	 * A syncbase element created a new active interval.  The interval
	 * time is in the global time space without the local syncbase offset added.
	 ****************************************************************/
	public void newIntervalNotice(TimeContainer syncbase, Interval interval) {
		writer.println(getID()+" got new interval notice " + interval);
		if (syncbase == null || interval == null) {
			return;
		} 
		boolean beginAdded = addInstanceTime(syncbase, interval, true);
		boolean endAdded = addInstanceTime(syncbase, interval, false);
		if (beginAdded && (state == TimeContainer.STATE_ACTIVE && restart == TimeContainer.RESTART_ALWAYS)) {
			boolean skipZeroSizeInterval = false;
			if (currentInterval.getDuration().isEqual(new Time(0)) && currentInterval.getEndTime().isEqual(getCurrentTime())) {
				skipZeroSizeInterval = true;
			}
			Interval newInterval = getNextActiveInterval(getCurrentTime(), getCurrentTime(), skipZeroSizeInterval);
			if ((newInterval != null) && newInterval.getBeginTime().isLess(currentInterval.getEndTime())) {
				// end the current interval and set currentInterval = newInterval
				currentInterval.setEndTime(newInterval.getBeginTime());
				sendChangeIntervalNotice();
				if (currentInterval.getEndTime().isLess(curSamplingTime)) {
					nextSamplingTime(currentInterval.getEndTime(), "changed interval end time");
				}
			}
		} else if (beginAdded && (state == TimeContainer.STATE_WAITING)) {
			recomputeNextWaitingInterval();
		} else if (endAdded) {
			// recompute current active end >= current begin, and < current end
			if (state == TimeContainer.STATE_ACTIVE) {
				recomputeActiveCurIntervalEndTime();
			} else if (state == TimeContainer.STATE_WAITING) {
				recomputeNextWaitingInterval();
			}			
		}
	}

	/*****************************************************************
	 * add a new syncbase interval to the begin or end instance time list.
	 *****************************************************************/
	private boolean addInstanceTime(TimeContainer syncbase, Interval interval, boolean begin) {
		if (begin && !interval.getBeginTime().isDefinite()) {
			return false;
		}
		boolean added = false;
		Vector conditions = (begin ? syncConditions[BEGIN] : syncConditions[END]);
		InstanceTimeList curInstanceTimeList = (begin ? instanceTimeList[BEGIN] : instanceTimeList[END]);
		for (Enumeration e = conditions.elements(); e.hasMoreElements();) {
			Condition condition = (Condition) e.nextElement();
			if ((TimeContainer) condition.getSyncbase() == syncbase) {
				if (condition.getType() == Condition.SYNCBASE_BEGIN) {
					Time newInstanceTime = Time.add(
									interval.getBeginTime(), new Time(condition.getOffsetValue()));
					condition.getInstanceTimes().add(newInstanceTime);
					if (!curInstanceTimeList.containsKey(newInstanceTime)) {
						added = true;
					}
					curInstanceTimeList.putTime(newInstanceTime, condition);
					condition.setCurrentInstanceTime(newInstanceTime);
				} else if (condition.getType() == Condition.SYNCBASE_END) {
					Time newInstanceTime = Time.add(
									interval.getEndTime(), new Time(condition.getOffsetValue()));
					condition.getInstanceTimes().add(newInstanceTime);
					if (!curInstanceTimeList.containsKey(newInstanceTime)) {
						added = true;
					}
					curInstanceTimeList.putTime(newInstanceTime, condition);
					condition.setCurrentInstanceTime(newInstanceTime);
				}
			}
		}
		return added;					
	}

	private void recomputeNextWaitingInterval() {	
		writer.println(getID()+" recomputing next waiting interval");
		if (!hasStarted() || (restart != TimeContainer.RESTART_NEVER)) {
			Interval newInterval = getNextActiveInterval(lastActiveEnd, getCurrentTime(), false);
			writer.println(getID()+" recomputed interval="+currentInterval);
			if (newInterval != null) {
				if (currentInterval == null) {
					currentInterval = newInterval;
					sendNewIntervalNotice();
					// next scheduled sampling time is new interval begin time
					nextSamplingTime(currentInterval.getBeginTime(), "new interval begin");
				} else if (newInterval.getBeginTime().isLess(currentInterval.getBeginTime()) ||
								newInterval.getEndTime().isLess(currentInterval.getEndTime())) {
					currentInterval = newInterval;
					sendChangeIntervalNotice();
					// next scheduled sampling time is the new interval time
					nextSamplingTime(currentInterval.getBeginTime(), "changed interval begin");
				}
			}
		}
	}

	private void recomputeActiveCurIntervalEndTime() {
		if (state != TimeContainer.STATE_ACTIVE) {
			return;
		}
		boolean skipZeroSizeInterval = false;
		if (currentInterval.getDuration().isEqual(new Time(0)) && currentInterval.getEndTime().isEqual(getCurrentTime())) {
			skipZeroSizeInterval = true;
		}
		Interval newInterval = getNextActiveInterval(currentInterval.getBeginTime(), getCurrentTime(), skipZeroSizeInterval);
		if ((newInterval != null) && (newInterval.getBeginTime().isEqual(currentInterval.getBeginTime())) &&
				(newInterval.getEndTime().isLess(currentInterval.getEndTime())) &&
				(newInterval.getEndTime().isGreaterEqual(getCurrentTime()))) {
			currentInterval.setEndTime(newInterval.getEndTime());
			sendChangeIntervalNotice();
			if (currentInterval.getEndTime().isLess(curSamplingTime)) {
				nextSamplingTime(currentInterval.getEndTime(), "changed interval end time");
			}
		}			
	}

	/****************************************************************
	 * Notify all syncbase and event dependents of a change in current interval.
	 ****************************************************************/
	public void sendChangeIntervalNotice() {
		writer.println(getID()+" sending change interval notice "+currentInterval);
		for (Enumeration e = syncbaseDependents.elements(); e.hasMoreElements();) {
			TimeContainer dependent = (TimeContainer) e.nextElement();
			dependent.changeIntervalNotice(this, currentInterval);
		}
		for (Enumeration e = eventDependents.elements(); e.hasMoreElements();) {
			TimeContainer dependent = (TimeContainer) e.nextElement();
			dependent.changeIntervalNotice(this, currentInterval);
		}
	}

	/****************************************************************
	 * Called by a syncbase element.
	 * syncbase or event element sent a change interval notice.
	 * update the corresponding the begin and end instance times.
	 ***************************************************************/
	public void changeIntervalNotice(TimeContainer syncbase, Interval interval) {
		writer.println(getID()+" got change interval notice");
		if (syncbase == null || interval == null) {
			return;
		} 
		boolean beginChanged = changeCurInstanceTime(syncbase, interval, true);
		boolean endChanged = changeCurInstanceTime(syncbase, interval, false);
		if (beginChanged && (state == TimeContainer.STATE_WAITING)) {
			writer.println(getID()+" begin changed");
			recomputeNextWaitingInterval();
		} else if (endChanged) {
			writer.println(getID()+" end changed");
			// recompute current active end >= current begin, and < current end
			if (state == TimeContainer.STATE_ACTIVE) {
				recomputeActiveCurIntervalEndTime();
			} else if (state == TimeContainer.STATE_WAITING) {
				recomputeNextWaitingInterval();
			}			
		}

	}

	/*****************************************************************
	 * A syncbase element changed its current interval.
	 * We must change our corresponding instance times.
	 *****************************************************************/
	private boolean changeCurInstanceTime(TimeContainer syncbase, Interval interval, boolean begin) {
		if (begin && !interval.getBeginTime().isDefinite()) {
			return false;
		}
		boolean changed = false;
		Vector conditions = (begin ? syncConditions[BEGIN] : syncConditions[END]);
		InstanceTimeList curInstanceTimeList = (begin ? instanceTimeList[BEGIN] : instanceTimeList[END]);
		for (Enumeration e = conditions.elements(); e.hasMoreElements();) {
			Condition condition = (Condition) e.nextElement();
			if ((TimeContainer) condition.getSyncbase() == syncbase) {
				Time curInstanceTime = condition.getCurrentInstanceTime();
				if (curInstanceTime != null) {
					if (condition.getType() == Condition.SYNCBASE_BEGIN) {
						Time newInstanceTime = Time.add(
										interval.getBeginTime(), new Time(condition.getOffsetValue()));
						if (!newInstanceTime.isEqual(curInstanceTime)) {
							changed = true;
							condition.getInstanceTimes().remove(curInstanceTime);
							condition.getInstanceTimes().add(newInstanceTime);
							curInstanceTimeList.removeTime(curInstanceTime, condition);
							curInstanceTimeList.putTime(newInstanceTime, condition);
							condition.setCurrentInstanceTime(newInstanceTime);
						}
					} else if (condition.getType() == Condition.SYNCBASE_END) {
						Time newInstanceTime = Time.add(
										interval.getEndTime(), new Time(condition.getOffsetValue()));
						if (!newInstanceTime.isEqual(curInstanceTime)) {
							changed = true;
							condition.getInstanceTimes().remove(curInstanceTime);
							condition.getInstanceTimes().add(newInstanceTime);
							curInstanceTimeList.removeTime(curInstanceTime, condition);
							curInstanceTimeList.putTime(newInstanceTime, condition);
							condition.setCurrentInstanceTime(newInstanceTime);
						}
					}
				}
			}
		}
		return changed;					
	}

	/****************************************************************
	 * Notify all syncbase dependents that the current interval is deleted.
	 ****************************************************************/
	public void sendDeleteIntervalNotice() {
		writer.println(getID()+" sending delete interval notice");
		for (Enumeration e = syncbaseDependents.elements(); e.hasMoreElements();) {
			TimeContainer dependent = (TimeContainer) e.nextElement();
			dependent.deleteIntervalNotice(this);
		}
		for (Enumeration e = eventDependents.elements(); e.hasMoreElements();) {
			TimeContainer dependent = (TimeContainer) e.nextElement();
			dependent.deleteIntervalNotice(this);
		}
	}

	/****************************************************************
	 * Called by a syncbase element.
	 * Syncbase or event element sent delete the current interval notice.
	 * Delete the corresponding the begin and end instance times.
	 ***************************************************************/
	public void deleteIntervalNotice(TimeContainer syncbase) {
		writer.println(getID()+" got delete interval notice");
		if (syncbase == null) {
			return;
		} 
		deleteCurInstanceTime(syncbase, true);
		deleteCurInstanceTime(syncbase, false);
	}

	/*****************************************************************
	 * A syncbase element deleted its current interval.
	 * We must delete our corresponding instance times.
	 *****************************************************************/
	private void deleteCurInstanceTime(TimeContainer syncbase, boolean begin) {
		Vector conditions = (begin ? syncConditions[BEGIN] : syncConditions[END]);
		InstanceTimeList curInstanceTimeList = (begin ? instanceTimeList[BEGIN] : instanceTimeList[END]);
		for (Enumeration e = conditions.elements(); e.hasMoreElements();) {
			Condition condition = (Condition) e.nextElement();
			if ((TimeContainer) condition.getSyncbase() == syncbase) {
				Time curInstanceTime = condition.getCurrentInstanceTime();
				if (curInstanceTime != null) {
					condition.getInstanceTimes().remove(curInstanceTime);
					curInstanceTimeList.removeTime(curInstanceTime, condition);
					condition.setCurrentInstanceTime(null);
				}
			}
		}					
	}

	private Interval getNextActiveInterval(Time beginAfter, Time minEndTime, boolean skipZeroSizeInterval) {
		writer.println(getID()+" computing next interval from "+beginAfter + " and "+minEndTime);
		while (true) {
			Time tempBegin = instanceTimeList[BEGIN].getNextResolvedTime(beginAfter);
			if (tempBegin == null || !tempBegin.isDefinite()) return null;
			Time[] endTimes = getNextEndTime(tempBegin);
			Time tempEnd = endTimes[0];
			boolean useNaturalInterval = (endTimes[1] != null);
			if ((tempEnd == null) && !endSpecified()) {
				writer.println(getID()+" no end specified, calculate active end with no end constraints");
				if (useNaturalInterval) {
					tempEnd = calcActiveEnd(tempBegin)[1];
				} else {
					tempEnd = calcActiveEnd(tempBegin)[0];
				}
				if (skipZeroSizeInterval) {
					writer.println(getID()+" in getnextactive interval, skipzerointerval="+skipZeroSizeInterval);
					writer.println(getID()+" tempbegin="+tempBegin+", tempend="+tempEnd);
					if (tempEnd.isEqual(tempBegin)) {  // zero duration interval
						writer.println(getID()+" yes it's zero");
						// may have to send new/changed interval notice to dependents here?
						beginAfter = Time.add(tempEnd, new Time(1));
						continue;
					}
				}
			} else {
				writer.println(getID()+" calculate active end with end contraints");
				// tempEnd = getNextEndTime(tempBegin);
				if (tempEnd == null) {
					return null;
				} else {
					if (useNaturalInterval) {
						tempEnd = calcActiveEnd(tempBegin, tempEnd)[1];
					} else {
						tempEnd = calcActiveEnd(tempBegin, tempEnd)[0];
					}
					if (skipZeroSizeInterval) {
						writer.println(getID()+" in getnextactive interval, skipzerointerval="+skipZeroSizeInterval);
						writer.println(getID()+" tempbegin="+tempBegin+", tempend="+tempEnd);
						if (tempEnd.isEqual(tempBegin)) {  // zero duration interval
							writer.println(getID()+" yes it's zero");
							// may have to send new/changed interval notice to dependents here?
							endTimes = getNextEndTime(Time.add(tempEnd, new Time(1)));
							tempEnd = endTimes[0];
							useNaturalInterval = (endTimes[1] != null);
							writer.println(getID()+" next end after zero interval="+tempEnd);
							if (tempEnd == null) {
									return null;
							}
							if (useNaturalInterval) {
								tempEnd = calcActiveEnd(tempBegin, tempEnd)[1];
							} else {
								tempEnd = calcActiveEnd(tempBegin, tempEnd)[0];
							}
							if (tempEnd.isEqual(tempBegin)) {
								writer.println(getID()+" anther zero interval");
								beginAfter = Time.add(tempEnd, new Time(1));
								continue;
							}
						}
					}
				}
			}
			// We have an end.  Is it after minEndTime?
			if (tempEnd.isGreaterEqual(minEndTime)) {
				if (tempBegin.isLess(minEndTime) && !(this instanceof MediaTimeContainer)) {
					writer.println(getID()+" negative offset and not a media element!");
					beginAfter = Time.add(tempEnd, new Time(1));
				} else {
					writer.println(getID()+" got a good interval");
					return new Interval(tempBegin, tempEnd);
				}
			} else if (restart == TimeContainer.RESTART_NEVER) {
				writer.println(getID()+" got a bad interval and restart=never");
				return null;
			} else {
				// may have to send new/changed interval notice to dependents here?
				beginAfter = Time.add(tempEnd, new Time(1));
			}
		}
	}

	/******************************************************************************
	 * Given a begin time, compute the nearest end time that comes at or after
	 * the begin time.  If restart == always, a time in the begin instance time
	 * list that comes before any times in the end instance time list is returned.
	 * If no such end time exists, but there are unresolved event-based end conditions
	 * or the end instance time list is empty, unresolved time is returned.
	 * The second element of the return data is simply a flag indicating whether
	 * the returned end time is from the begin instance time list or not.
	 * If it's not null, it indicates it comes from the begin list.  This flag
	 * can be used to determine if the active interval that is computed based on
	 * this end time can be extended by the min symantics.  I.e., if the end time
	 * is actually another begin instance time, and restart==always, the interval
	 * cannot be extended by the min.
	 *******************************************************************************/
	private Time[] getNextEndTime(Time begin) {
		Time tempEnd = instanceTimeList[END].getNextTime(begin);
		Time tempBegin = null;
		if (restart == TimeContainer.RESTART_ALWAYS) {
			Time nextBegin = instanceTimeList[BEGIN].getNextResolvedTimeAfter(begin);
			if (tempEnd == null) {
				tempEnd = nextBegin;
				tempBegin = tempEnd;
			} else if (nextBegin != null && nextBegin.isLess(tempEnd)) {
				tempEnd = nextBegin;
				tempBegin = tempEnd;
			}
		}
		if (tempEnd == null) {
			if (endHasEventConditions() || instanceTimeList[END].isEmpty()) {
				tempEnd = new Time();
			}
		}
		Time[] result = new Time[2];
		result[0] = tempEnd;
		result[1] = tempBegin;
		return result;
	}

	private Time[] calcActiveEnd(Time begin) {
		Time[] durs = computeActiveDur(begin);
		durs[0] = Time.add(begin, durs[0]);
		durs[1] = Time.add(begin, durs[1]);
		return durs;
	}

	private Time[] calcActiveEnd(Time begin, Time endConstraint) {
		Time[] durs = computeActiveDur(begin, endConstraint);
		durs[0] = Time.add(begin, durs[0]);
		durs[1] = Time.add(begin, durs[1]);
		return durs;
	}
		
	public Time computeSimpleDur() {
		writer.println(getID()+" computing simple dur");
		if (simpleDur.isResolved()) return simpleDur;
		if (durType == DURTYPE_UNSPECIFIED) {
			if (endSpecified() && !repeatDurSpecified() && !repeatCountSpecified()) {
				simpleDur = new Time(Time.INDEFINITE);
				writer.println(getID()+" simple dur = indefinite");
			} else {
				simpleDur = computeImplicitDur();
				writer.println(getID()+" simple dur = implicit dur =" + simpleDur);
			}
		} else if (durType == DURTYPE_TIMEVALUE) {
			simpleDur = dur;
			writer.println(getID()+" simple dur = time value = " + simpleDur);
		} else if (durType == DURTYPE_MEDIA) {
			simpleDur = computeImplicitDur();
			writer.println(getID()+" simple dur = media (implicit) = " + simpleDur);
		} else {
			writer.println(getID()+" invalid durtype="+durType);
		}
		return simpleDur;
	}

	public Time computeIntrinsicDur() {
		return intrinsicDur;
	}

	/*******************************************************************
	 * Compute the implicit dur of the element.  This is only needed
	 * when the computation of the simple dur requires the implicit dur.
	 * If the implicit dur is already resolved, just return the current
	 * value.  Otherwise, calculate the value based on the endsync rule,
	 * and in the case of media time containers, the intrinsic dur
	 * is also considered.
	 *******************************************************************/
	protected Time computeImplicitDur() {
		writer.println(getID()+" computing implicit dur");
		if (implicitDur.isResolved()) return implicitDur;
		if (endsyncRule == TimeContainer.ENDSYNC_MEDIA) {
			implicitDur = computeIntrinsicDur();
			writer.println(getID()+" implicit dur = intrinsic");
		} else {
			if (timeContainerHasEnded()) {
				if (currentInterval != null) {
					implicitDur = Time.subtract(getCurrentTime(), currentInterval.getBeginTime());
				} else {
					implicitDur = new Time(0);
				}
				writer.println(getID()+" time container has ended");
			} // else it is still unresolved at this time
		}
		writer.println(getID()+" implicitdur = "+implicitDur);
		return implicitDur;
	}

	/*****************************************************************
	 * Decide if the time container simple interval has ended by now
	 * based on the endsync rule.
	 *****************************************************************/
	protected boolean timeContainerHasEnded() {
		boolean assumedResult;
		if ((endsyncRule == TimeContainer.ENDSYNC_FIRST) ||
			(endsyncRule == TimeContainer.ENDSYNC_ID)) {
			assumedResult = false;
	    } else {
			assumedResult = true;
	    }
		for (Enumeration enum = children(); enum.hasMoreElements();) {
			TimeContainer child = (TimeContainer) enum.nextElement();
			switch (endsyncRule) {
				case TimeContainer.ENDSYNC_FIRST:
					// as soon as we find an ended child, return true;
					if (child.hasStarted() && !child.isActiveOrDead()) {
						return true;
					}
					// else keep looking (assumedResult is false)
					break;
				case TimeContainer.ENDSYNC_ID:
					// if we find the matching child, just return result
					if (endsyncID.equals(child.getID())) {
						return (child.hasStarted() & !child.isActiveOrDead());
					}
					break;
				case TimeContainer.ENDSYNC_LAST:
					// we just test for disqualifying children.
					// if the child is active, we're definitely not done.
					// if the child has not yet begun but has a resolved begin,
					// then we're not done.
					if (child.isActiveOrDead() || child.hasResolvedBegin(getCurrentTime())) {
						return false;
					}
					// else, keep checking (assumed result is true)
					break;
				case TimeContainer.ENDSYNC_ALL:
					// we just test for disqualifying children.
					// all means last done after all begin.
					// if the child is active, we're definitely not done.
					// if the child has not yet begun then we're not done.
					// Note that if it has already begun,
					// then we still have to wait for any more resolved begins.
					if (child.isActiveOrDead() || !child.hasStarted() ||
						child.hasResolvedBegin(getCurrentTime())) {
						return false;
					}
					// else, keep checking (the assumed result is true)
					break;
			} // close switch
		} // close for each child loop
		return assumedResult;
	} // timeContainerHasEnded()

	public Time[] computeActiveDur(Time begin) {
		return computeActiveDur(begin, new Time(Time.INDEFINITE));
	}

	/****************************************************************
	 * Compute active duration that begins at a definite time, and
	 * constrained by the end time (if the end time is resolved).
	 ****************************************************************/
	public Time[] computeActiveDur(Time begin, Time end) {
		Time activeDur = new Time();
		Time naturalDur = new Time();
		Time pad = new Time();
		if (endSpecified() && !(durSpecified() | repeatCountSpecified() | repeatDurSpecified())) {
			// The simple duration is indefinite by the simple duration calculation table
			pad = Time.subtract(end, begin);
		} else if (!end.isDefinite()) {
			pad = computeIntermediateActiveDur();
		} else {	// we have a definite end value and at least one of dur, repeatDur, or repeatCount
					// attributes.
			Vector times = new Vector();
			times.add(computeIntermediateActiveDur());
			times.add(Time.subtract(end, begin));
			pad = Time.min(times);
		}
		writer.println(getID()+" pad="+pad);
		if (pad.isResolved()) {
			Vector times2 = new Vector();
			times2.add(max);
			times2.add(pad);
			naturalDur = Time.min(times2);
			if ((restart == TimeContainer.RESTART_ALWAYS) &&
					(instanceTimeList[BEGIN].containsKey(end))) {
				activeDur = naturalDur;	// min cannot override restart
			} else {
				Vector times = new Vector();
				times.add(min);
				times.add(naturalDur);
				activeDur = Time.max(times);
			}
		}
		writer.println(getID()+" activeDur="+activeDur+", naturalDur="+naturalDur);
		Time[] durs = new Time[2];
		durs[0] = activeDur;
		durs[1] = naturalDur;
		return durs;
	}

	private Time computeIntermediateActiveDur() {
		writer.println(getID()+" computing intermediate active dur");
		Time p0 = computeSimpleDur();
		Time p1 = null;
		Time p2 = null;
		if (repeatCountSpecified()) {
			if (isRepeatCountIndefinite()) {
				p1 = new Time(Time.INDEFINITE);
			} else {
				float floatAccIterationCount = new Integer(accIterationCount).floatValue();
				if (repeatCount <= floatAccIterationCount) { // have run all iterations
					p1 = accIterationDur;
				} else {
					p1 = Time.add(accIterationDur, Time.multiply(simpleDur, repeatCount - floatAccIterationCount));
				}
			}
		} else {
			p1 = new Time(Time.INDEFINITE);
		}
		if (repeatDurSpecified()) {
			p2 = repeatDur;
		} else {
			p2 = new Time(Time.INDEFINITE);
		}
		Time iad = null;
		if (p0.isEqual(new Time(0))) {
			iad = new Time(0);
		} else if (!(repeatDurSpecified() || repeatCountSpecified())) {
			iad = p0;
		} else {
			Vector times = new Vector();
			times.add(p1);
			times.add(p2);
			times.add(new Time(Time.INDEFINITE));
			iad = Time.min(times);
		}
		writer.println(getID()+" iad="+iad);
		return iad;
	}

	private boolean endHasEventConditions() {
		for (Enumeration e = syncConditions[END].elements(); e.hasMoreElements();) {
			Condition condition = (Condition) e.nextElement();
			if (condition.isEventType()) {
				return true;
			}
		}
		return false;
	}

	private void startFill() {
		writer.println(getID()+" starting fill at "+getCurrentTime());
		stopFill();
		if ((fill != TimeContainer.FILL_REMOVE) && hasStarted() && !siblingSaysNoFill) {
			fillInterval = new LifeInterval(getCurrentTime(), new Time(Time.INDEFINITE));
			if (fill == TimeContainer.FILL_HOLD) {
				fillInterval.setType(TimeContainer.FILL_FREEZE);
			} else {
				fillInterval.setType(fill);
			}
			containerLife.addInterval(fillInterval);
			fillTime = true;
			for (Enumeration e = children(); e.hasMoreElements();) {
				TimeContainer child = (TimeContainer) e.nextElement();
				child.freeze();
			}
		} else if (siblingSaysNoFill) {	// ok I satisfied the sibling's condition
			siblingSaysNoFill = false;
		}
	}

	private void stopFill() {
		if (fillTime) {
			fillInterval.setEndTime(getCurrentTime());
			if (fillInterval.getDuration().isEqual(new Time(0))) {
				containerLife.removeIntervalAt(containerLife.size() - 1);
			}
			fillInterval = null;
			fillTime = false;
			for (Enumeration e = children(); e.hasMoreElements();) {
				TimeContainer child = (TimeContainer) e.nextElement();
				child.release();
			}
		}
	}
		
	private void freeze() {
		startFill();
	}

	private void release() {
		stopFill();
	}

	private void stopFillBySibling() {
		writer.println(getID()+" stopping fill by next sibling at time="+getCurrentTime());
		if ((fill == TimeContainer.FILL_FREEZE) || (fill == TimeContainer.FILL_TRANSITION)) {
			if (fillTime) {	// element has already started a fill interval.  stop it now
				stopFill();
			}
			siblingSaysNoFill = true;
//			} else {		// element is supposed to end now, but has yet to do the cleanup.
//				siblingSaysNoFill = true;
//			}
		}
	}

	private void stopPrevSiblingFill() {
		if (!hasStarted()) {
			TimeContainer parent = (TimeContainer) getParent();
			if ((parent != null) && (parent instanceof SeqTimeContainer)) {
				TimeContainer prev = (TimeContainer) parent.getPrevChild(this);
				if (prev != null) {
					prev.stopFillBySibling();
				}
			}
		}
	}

	// Time utilities
	/***********************************************************
	 * Get the current global time.
	 ***********************************************************/
	public Time getCurrentTime() {
		return getTimeGraph().getCurrentTime();
	}

	/***********************************************************
	 * Get the current local time in parent simple time.
	 ***********************************************************/
	public Time getCurrentLocalTime() {
		return convertFromGlobalToLocalTime(getCurrentTime());
	}

	/****************************************************************
	 * Convert another element's local time (i.e., parent simple time),
	 * to this element's local time.
	 ****************************************************************/
	public Time convertRemoteToLocalTime(TimeContainer fromContainer,
										Time fromTime) {
		if (fromContainer == null || fromTime == null) {
			return new Time();
		}
		TimeContainer commonAscendant = (TimeContainer) getCommonAscendant(fromContainer);
		if (commonAscendant == null) {
			return new Time();
		}
		Time ascendantSimpleTime = fromContainer.convertLocalToAscendantTime(commonAscendant, fromTime);
		writer.println("ascendant simple time=" + ascendantSimpleTime.toString());
		return convertFromAscendantToLocalTime(commonAscendant, ascendantSimpleTime);
	}

	/*****************************************************************
	 * Convert an ascendant's local time to my local time.
	 *****************************************************************/
	public Time convertFromAscendantToLocalTime(TimeContainer ascendant,
												Time ascendantTime) {
		if (ascendant == null || ascendantTime == null) {
			return new Time();
		}
		Time offsetTime = convertLocalToAscendantTime(ascendant, new Time(0));
		return Time.subtract(ascendantTime, offsetTime);
	}

	/****************************************************************
	 * Convert a global time to a local time and return the local time.
	 * If it's not in the cache already, do the conversion, and store it there.
	 *****************************************************************/
	public Time convertFromGlobalToLocalTime(Time globalTime) {
		if (!globalToLocalTimeCache.containsKey(globalTime)) {
			Time localTime = convertFromAscendantToLocalTime((TimeContainer) getRoot(), globalTime);
			globalToLocalTimeCache.put(globalTime, localTime);
		}
		return (Time) globalToLocalTimeCache.get(globalTime);
	}

	public Time convertFromLocalToGlobalTime(Time localTime) {
		return convertLocalToAscendantTime((TimeContainer) getRoot(), localTime);
	}

	/***********************************************************
	 * Convert an element's local time (in parent's simple time)
	 * to the ascendant's simple time.
	 ***********************************************************/
	public Time convertLocalToAscendantTime(TimeContainer ascendant,
											Time myLocalTime) {
		// writer.println(getID()+" convert local to ascendant");
		if (myLocalTime == null || ascendant == null) {
			return new Time();
		}
		Time tmpSimpleTime = new Time(myLocalTime);
		if (this == ascendant) { // special root node case
			return tmpSimpleTime;
		}
		TimeContainer tmpContainer = (TimeContainer) this.getParent();
		while (tmpContainer != ascendant) {
			if (tmpContainer == null) { // whoa, it's not an ascendant
				return new Time();
			}
			Time tmpSimpleDur = tmpContainer.getSimpleDur();
			Time tmpActiveTime = Time.add(tmpSimpleTime, tmpContainer.getAccIterationDur());
			Interval tmpCurrentInterval = tmpContainer.getCurrentInterval();
			// writer.println(getID()+" tmp active time="+tmpActiveTime);
			// writer.println(getID()+" tmpContainer id="+tmpContainer.getID());
			// writer.println(getID()+" tmp current interval="+tmpCurrentInterval);
			if (tmpCurrentInterval == null) {  // parent is not active
				return new Time();
			}
			tmpSimpleTime = Time.add(tmpActiveTime, tmpCurrentInterval.getBeginTime());
			tmpContainer = (TimeContainer) tmpContainer.getParent();
		}
		return tmpSimpleTime;
	}

	/****************************************************************
	 * Convert an interval in global time to local time.
	 ****************************************************************/
	private Interval convertFromGlobalIntervalToLocal(Interval interval) {
		Time localBeginTime = convertFromGlobalToLocalTime(interval.getBeginTime());
		Time localEndTime = convertFromGlobalToLocalTime(interval.getEndTime());
		return new Interval(localBeginTime, localEndTime);
	}

}
