/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  TimeGraph.java
//
//=============================================================================
package xmto2xmta.timegraph;

import java.util.*;
import java.io.*;
import xmto2xmta.ErrorEvent;

/*********************************************************************** 
 * Time graph class.  Time graph models the timing relationships.
 *
 * <p><b>History:</b><ul>
 * <li>15 Mar 2001 - Created file.
 * </ul><p>
 *********************************************************************/

public class TimeGraph {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	private Container root = null;
	private Hashtable idTab = new Hashtable();
	private Time now = new Time(0);
	private InstanceTimeList eventQ = new InstanceTimeList();
	private PrintWriter writer = null;
	private PrintWriter errWriter = null;

	public TimeGraph(PrintWriter writer, PrintWriter errWriter) {
		this.writer = writer;
		this.errWriter = errWriter;
   }
    
	public void setPrintWriter(PrintWriter writer) {
		this.writer = writer;
	}

	public void setErrPrintWriter(PrintWriter errWriter) {
		this.errWriter = errWriter;
	}

	public void setRoot(Container root) {
		this.root = root;
		idTab.put(root.getID(), root);
	}

	public Container getRoot() {
		return root;
	}

	public void setContainerID(Integer ID, Container container) {
		idTab.put(ID, container);
	}

	public Container getContainerByID(Integer ID) {
		return (Container) idTab.get(ID);
	}

	public Hashtable getIDTab() {
		return idTab;
	}

	/*************************************************************
	 * Return the current global time.
	 *************************************************************/
	public Time getCurrentTime() {
		return now;
	}

	public PrintWriter getPrintWriter() {
		return writer;
	}

	public PrintWriter getErrPrintWriter() {
		return errWriter;
	}

	public void initialize() {
	    
		// Now force in the default syncbase condition into the root container
		// so that it can start at time 0 and run.
		Condition condition = new Condition();
		condition.setType(Condition.SYNCBASE_BEGIN);
		HashSet instanceTimes = new HashSet();
		instanceTimes.add(now);
		condition.setInstanceTimes(instanceTimes);
		TimeContainer rootTimeContainer = (TimeContainer) root;
		Vector conditions = new Vector();
		conditions.add(condition);
		rootTimeContainer.setBeginConditions(conditions);
		rootTimeContainer.getBeginInstanceTimeList().putTime(now, condition);
		rootTimeContainer.setDur(TimeContainer.DURTYPE_UNSPECIFIED, new Time());
		// rootTimeContainer.setDur(TimeContainer.DURTYPE_TIMEVALUE, new Time(Time.INDEFINITE));
	    
		// update the sync dependent list for each container with
		// information from the dependents' conditions
		clearDependents((TimeContainer) root);
		updateDependents((TimeContainer) root);

	}
		
	private void updateDependents(TimeContainer container) {
		if (container == null) return;
		Vector conditions = container.getBeginConditions();
		if (conditions == null) writer.println("null conditions");
		for (Enumeration e = conditions.elements(); e.hasMoreElements(); ) {
			Condition condition = (Condition) e.nextElement();
			int type = condition.getType();
			if (type == Condition.SYNCBASE_BEGIN || type == Condition.SYNCBASE_END) {
				Integer syncbaseID = condition.getSyncbaseID();
				TimeContainer syncbase = (TimeContainer) getContainerByID(syncbaseID);
				if (syncbase != null) {
					condition.setSyncbase(syncbase);
					Vector dependents = syncbase.getSyncbaseDependents();
					if (!dependents.contains(container)) {
						dependents.add(container);
					}
				}
			} else if (type == Condition.BEGIN_EVENT || type == Condition.END_EVENT || type == Condition.ACTIVATE_EVENT) {
				Integer syncbaseID = condition.getSyncbaseID();
				TimeContainer syncbase = (TimeContainer) getContainerByID(syncbaseID);
				if (syncbase != null) {
					condition.setSyncbase(syncbase);
					Vector dependents = syncbase.getEventDependents();
					if (!dependents.contains(container)) {
						dependents.add(container);
					}
				}
			}
		}
		conditions = container.getEndConditions();
		for (Enumeration e = conditions.elements(); e.hasMoreElements(); ) {
			Condition condition = (Condition) e.nextElement();
			int type = condition.getType();
			if (type == Condition.SYNCBASE_BEGIN || type == Condition.SYNCBASE_END) {
				Integer syncbaseID = condition.getSyncbaseID();
				TimeContainer syncbase = (TimeContainer) getContainerByID(syncbaseID);
				if (syncbase != null) {
					condition.setSyncbase(syncbase);
					Vector dependents = syncbase.getSyncbaseDependents();
					if (!dependents.contains(container)) {
						dependents.add(container);
					}
				}
			} else if (type == Condition.BEGIN_EVENT || type == Condition.END_EVENT || type == Condition.ACTIVATE_EVENT) {
				Integer syncbaseID = condition.getSyncbaseID();
				TimeContainer syncbase = (TimeContainer) getContainerByID(syncbaseID);
				if (syncbase != null) {
					condition.setSyncbase(syncbase);
					Vector dependents = syncbase.getEventDependents();
					if (!dependents.contains(container)) {
						dependents.add(container);
					}
				}
			}
		}

		if (container.getFill() == TimeContainer.FILL_AUTO) {
			if (!(container.durSpecified() || container.endSpecified() || container.repeatCountSpecified() ||
						container.repeatDurSpecified())) {
				container.setFill(TimeContainer.FILL_FREEZE);
			} else {
				container.setFill(TimeContainer.FILL_REMOVE);
			}
		}

		if (container.getMinType() == TimeContainer.MINMAX_MEDIA) {
			container.setMin(TimeContainer.MINMAX_TIMEVALUE, container.getIntrinsicDur());
		}
		if (container.getMaxType() == TimeContainer.MINMAX_MEDIA) {
			container.setMax(TimeContainer.MINMAX_TIMEVALUE, container.getIntrinsicDur());
		}
		if (container.getMin().isGreater(container.getMax())) {
			container.setMin(TimeContainer.MINMAX_TIMEVALUE, new Time(0));
			container.setMax(TimeContainer.MINMAX_TIMEVALUE, new Time(Time.INDEFINITE));
		}
		for (Enumeration e = container.children(); e.hasMoreElements(); ) {
			TimeContainer child = (TimeContainer) e.nextElement();
			updateDependents(child);
		}
	}
		
	private void clearDependents(TimeContainer container) {
		if (container == null) return;
		container.setSyncbaseDependents(new Vector());	
		container.setEventDependents(new Vector());	
		for (Enumeration e = container.children(); e.hasMoreElements(); ) {
			TimeContainer child = (TimeContainer) e.nextElement();
			clearDependents(child);
		}
	}

	public void play(Time endTime) {
	    
		now = new Time(0);

		printTimeGraph(true);

		TimeContainer rootTimeContainer = (TimeContainer) root;
		rootTimeContainer.reset(null, null);

		eventQ.putTime(now, rootTimeContainer);

		while (now.isLessEqual(endTime) && !eventQ.isEmpty()) {
			printTimeGraph(false);
			try {
				writer.println("getting next event from q");
				writer.println();
				Time eventTime = (Time) eventQ.firstKey();
				now = eventTime;
				writer.println("got event time="+eventTime);
				HashSet timeContainers = (HashSet) eventQ.get(eventTime);
				eventQ.remove(eventTime);
				if (eventTime.isDefinite()) {
					now = eventTime;
					for (Iterator i = timeContainers.iterator(); i.hasNext();) {
						TimeContainer timeContainer = (TimeContainer) i.next();
						writer.println("event container id="+timeContainer.getID());
						timeContainer.sample();
					}
				} else {
					writer.println("indefinite or unresolved event time received");
				}
			} catch (Exception e) {
				writer.println("problem with the event queue e="+e);
				e.printStackTrace(writer);
			}
		}

		if (now.isLessEqual(endTime)) {
			writer.println("terminating simulation");
			rootTimeContainer.stop();
		} // else it may be an indefinite presentation.  leave the time graph unterminated.
		printTimeGraph(true);
	}

	public void scheduleSamplingTime(Container container, Time eventTime) {
		eventQ.putTime(eventTime, container);
	}

	public void removeSamplingTime(Container container, Time eventTime) {
		eventQ.removeTime(eventTime, container);
	}

	private void printTimeGraph(boolean all) {
		writer.println();
		writer.println("at time " + now.toString());
		printContainer(root, 0, writer, all);
		printContainerHistory(root, 1, writer);
		writer.flush();
	}

	private void printContainer(Container container, int depth, PrintWriter out, boolean all) {
		TimeContainer timeContainer = (TimeContainer) container;
		if (timeContainer != null) {
			printIndent(depth, out);
			out.print("CONTAINER ID="+timeContainer.getID()+" type=");
			if (timeContainer instanceof ParTimeContainer) {
				out.println("par");
			} else if (timeContainer instanceof SeqTimeContainer) {
				out.println("seq");
			} else if (timeContainer instanceof MediaTimeContainer) {
				out.println("media");
			} else {
				out.println("unknown");
			}

			Vector conditions = timeContainer.getBeginConditions();
			for (Enumeration e = conditions.elements(); e.hasMoreElements(); ) {
				Condition condition = (Condition) e.nextElement();
				printIndent(depth+1, out);
				out.println("begin " + condition);
			}
			conditions = timeContainer.getEndConditions();
			for (Enumeration e = conditions.elements(); e.hasMoreElements(); ) {
				Condition condition = (Condition) e.nextElement();
				printIndent(depth+1, out);
				out.println("end " + condition);
			}
			if (all) {
				Vector dependents = timeContainer.getSyncbaseDependents();
				printIndent(depth+1, out);
				out.print("syncbase dependents ");
				for (Enumeration e = dependents.elements(); e.hasMoreElements(); ) {
					TimeContainer dependent = (TimeContainer) e.nextElement();
					out.print(dependent.getID()+", ");
				}
				out.println();

				printIndent(depth+1, out);
				out.print("dur type="+ TimeContainer.DURTYPE_NAME[timeContainer.getDurType()]);
				if (timeContainer.getDurType() == TimeContainer.DURTYPE_TIMEVALUE) {
					out.println(", value="+timeContainer.getDur());
				} else {
					out.println();
				}
				printIndent(depth+1, out);
				out.print("endsync="+ TimeContainer.ENDSYNC_NAME[timeContainer.getEndsyncRule()]);
				if (timeContainer.getEndsyncRule() == TimeContainer.ENDSYNC_ID) {
					out.println(", id="+timeContainer.getEndsyncID());
				} else {
					out.println();
				}
				printIndent(depth+1, out);
				out.print("min type="+ TimeContainer.MINMAX_NAME[timeContainer.getMinType()]);
				if (timeContainer.getMinType() == TimeContainer.MINMAX_TIMEVALUE) {
					out.print(", value="+timeContainer.getMin()+", ");
				} else {
					out.print(", ");
				}
				out.print("max type="+ TimeContainer.MINMAX_NAME[timeContainer.getMaxType()]);
				if (timeContainer.getMaxType() == TimeContainer.MINMAX_TIMEVALUE) {
					out.println(", value="+timeContainer.getMax());
				} else {
					out.println();
				}
				printIndent(depth+1, out);
				out.println("fill="+TimeContainer.FILL_NAME[timeContainer.getFill()]+", default="+TimeContainer.FILL_NAME[timeContainer.getFillDefault()]);
				printIndent(depth+1, out);
				out.println("restart="+TimeContainer.RESTART_NAME[timeContainer.getRestart()]+", default="+TimeContainer.RESTART_NAME[timeContainer.getRestartDefault()]);
				printIndent(depth+1, out);
				out.println("repeatDur="+
					(timeContainer.repeatDurSpecified()?timeContainer.getRepeatDur().toString():"unspecified")+
					", repeatCount="+(timeContainer.repeatCountSpecified()?String.valueOf(timeContainer.getRepeatCount()):"unspecified"));
			}			
			printIndent(depth+1, out);
			out.println("state="+TimeContainer.STATE_NAME[timeContainer.getState()]);
			printIndent(depth+1, out);
			out.print("begin instance times ");
			for (Iterator i = timeContainer.getBeginInstanceTimeList().keySet().iterator(); i.hasNext();) {
				out.print((Time)i.next() + ", ");
			}
			out.println();
			printIndent(depth+1, out);
			out.print("end instance times ");
			for (Iterator i = timeContainer.getEndInstanceTimeList().keySet().iterator(); i.hasNext();) {
				out.print((Time)i.next() + ", ");
			}
			out.println();
			printIndent(depth+1, out);
			out.println("simple dur="+timeContainer.getSimpleDur()+", implicit="+timeContainer.getImplicitDur()
							+", intrinsic="+timeContainer.getIntrinsicDur());
			printIndent(depth+1, out);
			out.println("current interval="+timeContainer.getCurrentInterval());
			printIndent(depth+1, out);
			out.println("accumulated iteration count="+timeContainer.getAccIterationCount()+", time="+timeContainer.getAccIterationDur());
			printIndent(depth+1, out);
			out.println("has started="+timeContainer.hasStarted());
			printInstanceTimeList("begin", timeContainer.getBeginInstanceTimeList(), depth+1, out);
			printInstanceTimeList("end", timeContainer.getEndInstanceTimeList(), depth+1, out);
			for (Enumeration e = container.children(); e.hasMoreElements(); ) {
				printContainer((Container) e.nextElement(), depth+1, out, all);
			}
		}
	}
	
	private void printInstanceTimeList(String label, InstanceTimeList instanceTimeList, int depth, PrintWriter out) {
		printIndent(depth, out);
		out.print(label+" instance times=");
		for (Iterator i = instanceTimeList.keySet().iterator(); i.hasNext();) {
			out.print(i.next()+", ");
		}
		out.println();
	}

	private void printContainerHistory(Container container, int depth, PrintWriter out) {
		TimeContainer timeContainer = (TimeContainer) container;
		if (timeContainer != null) {
			printLifeHistory(timeContainer.getID(), timeContainer.getLifeHistory(), depth, out);
			for (Enumeration e = container.children(); e.hasMoreElements(); ) {
				printContainerHistory((Container) e.nextElement(), depth, out);
			}
		}
	}

	private void printLifeHistory(Integer id, Vector history, int depth, PrintWriter out) {
		printIndent(depth, out);
		out.println("life history for container id="+id);
		for (Enumeration e = history.elements(); e.hasMoreElements();) {
			ContainerLife containerLife = (ContainerLife) e.nextElement();
			printIndent(depth+1, out);
			out.println("life cycle "+containerLife);
			printIndent(depth+1, out);
			out.println("parent ID="+containerLife.getParentELASid());
			for (Enumeration i = containerLife.getLifeIntervals().elements(); i.hasMoreElements();) {
				LifeInterval lifeInterval = (LifeInterval) i.nextElement();
				printIndent(depth+2, out);
				if (lifeInterval.getType() == LifeInterval.ACTIVE) {
					out.print("active interval "+lifeInterval+" observed begin="+lifeInterval.getObservedBegin()+
					" intrinsic dur="+lifeInterval.getIntrinsicDur()+" simple durs=");
					for (Enumeration j = lifeInterval.getSimpleDurs().elements(); j.hasMoreElements();) {
						Time simpleDur = (Time) j.nextElement();
						out.print(simpleDur+", ");
					}
					out.println();
					printIndent(depth+2, out);
					out.print("elasids=");
					for (Enumeration k = lifeInterval.getELASids().elements(); k.hasMoreElements();) {
						ELASid elasid = (ELASid) k.nextElement();
						out.print(elasid+", ");
					}
				} else {
					out.print("fill interval "+lifeInterval+", type="+TimeContainer.FILL_NAME[lifeInterval.getType()]);
				}
				out.println();
			}
		}
	}
		
	private void printIndent(int depth, PrintWriter out) {
		for (int i = 0; i < depth*3; i++) out.print(' ');
	}

}   // class TimeGraph
