/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  AnimationBaseElement.java
//
//=============================================================================
package xmto2xmta.tree;

import java.text.MessageFormat;
import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import xmto2xmta.ErrorEvent;
import xmto2xmta.IdManager;
import xmto2xmta.Xmto2Xmta;
import xmto2xmta.resources.Xmto2XmtaRes;
import xmto2xmta.resources.AnimationBaseElementRes;
import xmto2xmta.timegraph.ELASid;
import xmtutil.DOMXMTO;
import xmtutil.TimingAttribute;
/*********************************************************************** 
 * Class to group together the various type of animation elements.
 *
 * <p><b>History:</b><ul>
 * <li>13 Dec 2000 - Created file.
 * <li>19 Apr 2001 - Add simple lifecycle support.  
 * <li>30 Jun 2001 - Add interpolator support
 * <li>30 Jun 2001 - Fix bug, on validating syntax if problem with parent then
 *                   return directly the problem.
 * </ul>
 *********************************************************************/
public abstract class AnimationBaseElement extends TimingElement {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	/*********************************************************************** 
	 **********************************************************************/
	private String xmtoAttrTo_ = null;
	private String xmtoAttrXlinkHref_ = null;
	private String xmtoAttrTargetElement_ = null;
	private Animatable xmtoAttrTargetAnimatableElement_ = null;

	/*********************************************************************** 
	 **********************************************************************/
	private ResourceBundle xmto2XmtaRes_ = ResourceBundle.getBundle("xmto2xmta.resources.Xmto2XmtaRes", Locale.getDefault());
	private ResourceBundle animationBaseElementRes_ = ResourceBundle.getBundle("xmto2xmta.resources.AnimationBaseElementRes", Locale.getDefault());
/*********************************************************************** 
 * constructor.
 * @param xmtoElement XMT-O element for conversion.
 * @param xmtoIds     hashtable to contain XMT-O Conversion elements 
 *                    with the key of their XMT-O id.
 * @param xmtaIds     hashtable to contain XMT-A elements with the key
 *                    of their XMT-A id.
 * @param xmtaDoc     DOM Document for the new XMT-A elements.
 * @param xmto2Xmta   Converter, used for firing error events.
 *********************************************************************/
public AnimationBaseElement(Element xmtoElement, Hashtable xmtoIds, Hashtable xmtaIds, Hashtable convertIndexTable, Document xmtaDoc, Xmto2Xmta xmto2Xmta) {
	super(xmtoElement, xmtoIds, xmtaIds, convertIndexTable, xmtaDoc, xmto2Xmta);
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrTargetElement() {
	return xmtoAttrTargetElement_;
}
/*********************************************************************** 
 **********************************************************************/
public Animatable getXmtoAttrTargetObject() {
	return xmtoAttrTargetAnimatableElement_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrTo() {
	return xmtoAttrTo_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrXlinkHref() {
	return xmtoAttrXlinkHref_;
}
/*********************************************************************** 
 * Handle the attribute by checking its syntax.  This is to catch all syntax 
 * errors which the schema can not check for, and thus the DOM parser 
 * doesn't find any problems.  Call the parent handler to see if it can
 * handle the attribute or not.
 *
 * @param  attributeName  node name of attribute.
 * @param  attributeValue node value of attribute.
 * @return whether this attribute is handled or not.
 **********************************************************************/
protected boolean handleValidateSyntax(String attributeName, String attributeValue) {
	boolean success = super.handleValidateSyntax(attributeName, attributeValue);

	// Determine the to
	if (DOMXMTO.to.equals(attributeName)) {
		xmtoAttrTo_ = attributeValue;
		return true;
	}

	// Determine the targetElement
	if (DOMXMTO.targetElement.equals(attributeName)) {
		xmtoAttrTargetElement_ = attributeValue;
		return true;
	}

	// Determine the xlink:href
	if (DOMXMTO.xlink_colon_href.equals(attributeName)) {
		xmtoAttrXlinkHref_ = attributeValue;
		return true;
	}

	// 
	return success;
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
/*********************************************************************** 
 * Validate the syntax of the element.  This is to catch all syntax 
 * errors which the schema can not check for, and thus the DOM parser 
 * doesn't find any problems.
 * Mainly check the begin, dur and end values to see they are valid timing
 * attributes.  If this is a problem then remove this node from the tree.
 **********************************************************************/
public boolean validateSyntax() {
	boolean success = super.validateSyntax();
	if (!success)
		return success;

	// Check the xlink:href and target Element attribute. Validate that
	// the object pointed to actually exists.
	// first check if either parameter exists.
	if (null != xmtoAttrXlinkHref_ || null != xmtoAttrTargetElement_) {
		// If they both exist then use the target element attribute and ignore the xlink:href attribute
		if (null != xmtoAttrXlinkHref_ && null != xmtoAttrTargetElement_) {
			MessageFormat msgFormat = new MessageFormat(animationBaseElementRes_.getString(AnimationBaseElementRes.MSG_WARNING_BOTH_HREF_TARGET_SPECIFIED));
			StringBuffer buffer = new StringBuffer();
			buffer.append(getXmtoElement().getNodeName());
			String id = getXmtoAttrId();
			if (null != id && !id.equals(""))
				buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
			StringBuffer xhBuffer = new StringBuffer();
			xhBuffer.append(" ").append(DOMXMTO.xlink_colon_href).append("=\"").append(xmtoAttrXlinkHref_).append("\"");
			StringBuffer teBuffer = new StringBuffer();
			teBuffer.append(" ").append(DOMXMTO.targetElement).append("=\"").append(xmtoAttrTargetElement_).append("\"");
			Object[] msgArgs = {buffer, xhBuffer, teBuffer, xhBuffer};
			String message = msgFormat.format(msgArgs);
			fireErrorEvent(ErrorEvent.ERROR, 0, message, getXmtoElement());
			xmtoAttrXlinkHref_ = null;
		}

		// Now get the id of the target element.  If the xlink:href contains the id then
		// strip off the # to get the id.		
		if (null != xmtoAttrXlinkHref_) {
			// Next check the the xlink:href attribute is in the supported form, i.e. not specifying an external file
			if ('#' != xmtoAttrXlinkHref_.charAt(0)) {
				MessageFormat msgFormat = new MessageFormat(animationBaseElementRes_.getString(AnimationBaseElementRes.MSG_ERROR_HREF_FILE_NOT_SUPPORTED));
				StringBuffer buffer = new StringBuffer();
				buffer.append(getXmtoElement().getNodeName());
				String id = getXmtoAttrId();
				if (null != id && !id.equals(""))
					buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
				StringBuffer attrBuffer = new StringBuffer();
				attrBuffer.append(DOMXMTO.xlink_colon_href).append("=\"").append(xmtoAttrXlinkHref_).append("\"");
				Object[] msgArgs = {buffer.toString(), attrBuffer.toString()};
				String message = msgFormat.format(msgArgs);
				fireErrorEvent(ErrorEvent.ERROR, 0, message, getXmtoElement(), DOMXMTO.xlink_colon_href);
				xmtoAttrXlinkHref_ = null;
				remove();
				return false;
			}
		}
	}

	// Next check that the object to be animated is a valid animatable object.
	// First get the object 

	// if the targetElment or xlink:href are specified get the object via the specified id.
	ConvertElement targetElement = null;
	if (null != xmtoAttrXlinkHref_ || null != xmtoAttrTargetElement_) {
		String targetId = (null == xmtoAttrXlinkHref_) ? xmtoAttrTargetElement_ : xmtoAttrXlinkHref_.substring(1);
		// Next check the the id for the target specifies a valid id found in the file.
		if (null != targetId && !targetId.equals("")) {
			// First is it in the table xmto ids? Also is it a DefsObject class.
			Hashtable xmtoTableIds = getXmtoIds();
			targetElement = (ConvertElement) xmtoTableIds.get(targetId);
		}
		if (null == targetElement) {
			MessageFormat msgFormat = new MessageFormat(animationBaseElementRes_.getString(AnimationBaseElementRes.MSG_ERROR_HREF_ID_NOT_FOUND));
			StringBuffer buffer = new StringBuffer();
			buffer.append(getXmtoElement().getNodeName());
			String id = getXmtoAttrId();
			if (null != id && !id.equals(""))
				buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
			StringBuffer attrBuffer = new StringBuffer();
			String attr = null;
			if (null != xmtoAttrTargetElement_) {
				attrBuffer.append(DOMXMTO.targetElement).append("=\"").append(xmtoAttrTargetElement_).append("\"");
				attr = DOMXMTO.targetElement;
			} else {
				attrBuffer.append(DOMXMTO.xlink_colon_href).append("=\"").append(xmtoAttrXlinkHref_).append("\"");
				attr = DOMXMTO.xlink_colon_href;
			}
			Object[] msgArgs = {buffer.toString(), attrBuffer, targetId};
			String message = msgFormat.format(msgArgs);
			fireErrorEvent(ErrorEvent.ERROR, 0, message, getXmtoElement(), attr);
			xmtoAttrTargetElement_ = null;
			xmtoAttrXlinkHref_ = null;
			remove();
			return false;
		}
	} else {
		targetElement = (ConvertElement) getParent();
	}

	// Verify the targetElement is animatable
	if (targetElement instanceof Animatable) {
		// Found the object
		xmtoAttrTargetAnimatableElement_ = (Animatable) targetElement;
	} else {
		// targetElement isn't a valid animatable object.
		MessageFormat msgFormat = new MessageFormat(animationBaseElementRes_.getString(AnimationBaseElementRes.MSG_ERROR_HREF_ANIMATION_NOT_SUPPORTED));
		StringBuffer buffer = new StringBuffer();
		buffer.append(getXmtoElement().getNodeName());
		String id = getXmtoAttrId();
		if (null != id && !id.equals(""))
			buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
		StringBuffer parentBuffer = new StringBuffer();
		parentBuffer.append(targetElement.getXmtoElement().getNodeName());
		id = targetElement.getXmtoAttrId();
		if (null != id && !id.equals(""))
			parentBuffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
		Object[] msgArgs = {buffer.toString(), parentBuffer};
		String message = msgFormat.format(msgArgs);
		fireErrorEvent(ErrorEvent.ERROR, 0, message, getXmtoElement());
		remove();
		return false;
	}

	// Print out the information on the object.
	if (debug_) {
		StringBuffer buffer = new StringBuffer();
		buffer.append("AnimationBaseElement.validateSyntax() - <");
		buffer.append(getXmtoElement().getNodeName());
		String id = getXmtoAttrId();
		if (null != id && !id.equals(""))
			buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
		buffer.append("> animates <");
		buffer.append(((ConvertElement) xmtoAttrTargetAnimatableElement_).getXmtoElement().getNodeName());
		id = ((ConvertElement) xmtoAttrTargetAnimatableElement_).getXmtoAttrId();
		if (null != id && !id.equals(""))
			buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
		buffer.append(">");
		String attr = (null != xmtoAttrTargetElement_) ? DOMXMTO.targetElement : DOMXMTO.xlink_colon_href;
		fireErrorEvent(ErrorEvent.TRACE, 0, buffer.toString(), getXmtoElement(), attr);
	}

	// dur attribute can't have a MEDIA_VALUE.  Other errors should have already been captured.
	TimingAttribute dur = getXmtoAttrDur();
	if (null != dur && dur.getTimeValueType() == TimingAttribute.MEDIA_VALUE) {
		// targetElement isn't a valid animatable object.
		MessageFormat msgFormat = new MessageFormat(xmto2XmtaRes_.getString(Xmto2XmtaRes.MSG_ERROR_ATTRIBUTE_VALUE_NOT_SUPPORTED));
		StringBuffer buffer = new StringBuffer();
		buffer.append(getXmtoElement().getNodeName());
		String id = getXmtoAttrId();
		if (null != id && !id.equals(""))
			buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
		Object[] msgArgs = {buffer, DOMXMTO.media, DOMXMTO.dur};
		String message = msgFormat.format(msgArgs);
		fireErrorEvent(ErrorEvent.ERROR, 0, message, getXmtoElement());
		remove();
		return false;
	}



	// 
	return success;
}
}
