/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  AnimationElement.java
//
//=============================================================================
package xmto2xmta.tree;

import java.awt.Color;
import java.text.MessageFormat;
import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import xmto2xmta.ErrorEvent;
import xmto2xmta.IdManager;
import xmto2xmta.Xmto2Xmta;
import xmto2xmta.resources.AnimationElementRes;
import xmto2xmta.resources.Xmto2XmtaRes;
import xmto2xmta.timegraph.ELASid;
import xmtutil.AnimationInterpolation;
import xmtutil.ColorAnimationInterpolation;
import xmtutil.Position2DAnimationInterpolation;
import xmtutil.ScalarAnimationInterpolation;
import xmtutil.DOMXMTAParameter;
import xmtutil.DOMXMTO;
import xmtutil.DOMXMTOParameter;
import xmtutil.DOMXMTA;
import xmtutil.TimingAttribute;
/*********************************************************************** 
 * Class to group together the various type of animation elements.
 *
 * <p><b>History:</b><ul>
 * <li>13 Dec 2000 - Created file.
 * <li>19 Apr 2001 - Add simple lifecycle support.  
 * <li>30 Jun 2001 - Add interpolator support
 * <li>30 Jun 2001 - Fix bug, on validating syntax if problem with parent then
 *                   return directly the problem.
 * </ul>                                                                     
 *********************************************************************/
public abstract class AnimationElement extends AnimationBaseElement {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	/*********************************************************************** 
	 **********************************************************************/
	private String xmtoAttrFrom_ = null;
	private String xmtoAttrBy_ = null;
	private String xmtoAttrValues_ = null;
	private String xmtoAttrCalcMode_ = null;
	private String xmtoAttrKeyTimes_ = null;
	private String xmtoAttrKeySplines_ = null;
	private String xmtoAttrAdditive_ = null;
	private String xmtoAttrAccumulate_ = null;

	/***********************************************************************
	 **********************************************************************/
	private AnimationInterpolation animationInterpolation_ = null;
	private String xmtaAttrKey_ = null;
	private String xmtaAttrKeyValue_ = null;
	private TimingAttribute xmtaAttrCycleInterval_ = null;
	private String xmtaAttrLoop_ = null;


	/***********************************************************************
	 **********************************************************************/
	private Hashtable xmtaInterpolatorIds_ = new Hashtable(89);
	private Hashtable xmtaTimeSensorIds_ = new Hashtable(89);

	/***********************************************************************
	 **********************************************************************/
	private ResourceBundle animationElementRes_ = ResourceBundle.getBundle("xmto2xmta.resources.AnimationElementRes", Locale.getDefault());
	private ResourceBundle xmto2XmtaRes_ = ResourceBundle.getBundle("xmto2xmta.resources.Xmto2XmtaRes", Locale.getDefault());
/*********************************************************************** 
 * constructor.
 * @param xmtoElement XMT-O element for conversion.
 * @param xmtoIds     hashtable to contain XMT-O Conversion elements 
 *                    with the key of their XMT-O id.
 * @param xmtaIds     hashtable to contain XMT-A elements with the key
 *                    of their XMT-A id.
 * @param xmtaDoc     DOM Document for the new XMT-A elements.
 * @param xmto2Xmta   Converter, used for firing error events.
 *********************************************************************/
public AnimationElement(Element xmtoElement, Hashtable xmtoIds, Hashtable xmtaIds, Hashtable convertIndexTable, Document xmtaDoc, Xmto2Xmta xmto2Xmta) {
	super(xmtoElement, xmtoIds, xmtaIds, convertIndexTable, xmtaDoc, xmto2Xmta);
}
/*********************************************************************** 
 * Validate the syntax of the interpolator keys and values.
 * This will generate the keys and values and check if there are any errors
 * with these values.  
 * If this is a problem then remove this node from the tree.
 **********************************************************************/
private int getMFTypeField(int sfType) {
	switch (sfType) {
		case Animatable.TYPE_SFColor :
			return Animatable.TYPE_MFColor;
		case Animatable.TYPE_SFFloat :
			return Animatable.TYPE_MFFloat;
		case Animatable.TYPE_SFVec2f :
			return Animatable.TYPE_MFVec2f;
		case Animatable.TYPE_SFBool :
			return Animatable.TYPE_MFBool;
		case Animatable.TYPE_SFInt32 :
			return Animatable.TYPE_MFInt32;
		case Animatable.TYPE_SFString :
			return Animatable.TYPE_MFString;
		default :
			return Animatable.TYPE_UNSUPPORTED;
	}
}
/*********************************************************************** 
 **********************************************************************/
public TimingAttribute getXmtaAttrCycleInterval() {
	return xmtaAttrCycleInterval_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtaAttrKey() {
	return xmtaAttrKey_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtaAttrKeyValue() {
	return xmtaAttrKeyValue_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtaAttrLoop() {
	return xmtaAttrLoop_;
}
/*********************************************************************** 
 * Get the id for the Interpolator element.
 *
 * @return the id (DEF) of the XMT-A time sensor.
 **********************************************************************/
public String getXmtaInterpolatorId(ELASid elasId) {
	String xmtaInterpolatorId = (String) xmtaInterpolatorIds_.get(elasId);

	// If couldn't find one then create it.
	if (null == xmtaInterpolatorId) {
		xmtaInterpolatorId = IdManager.getNextId(IdManager.InterpolatorID);
		// make the group ids based on object id or make them unique.
		if (null != getXmtoAttrId() && !getXmtoAttrId().equals(""))
			xmtaInterpolatorId = xmtaInterpolatorId + ":" + getXmtoAttrId();
		xmtaInterpolatorIds_.put(elasId, xmtaInterpolatorId);
	}

	//
	return xmtaInterpolatorId;
}
/*********************************************************************** 
 * Get the id for the TimeSensor element.
 *
 * @return the id (DEF) of the XMT-A time sensor.
 **********************************************************************/
public String getXmtaTimeSensorId(ELASid elasId) {
	String xmtaTimeSensorId= (String) xmtaTimeSensorIds_.get(elasId);

	// If couldn't find one then create it.
	if (null == xmtaTimeSensorId) {
		xmtaTimeSensorId = IdManager.getNextId(IdManager.TimeSensorID);
		// make the group ids based on object id or make them unique.
		if (null != getXmtoAttrId() && !getXmtoAttrId().equals(""))
			xmtaTimeSensorId = xmtaTimeSensorId + ":" + getXmtoAttrId();
		xmtaTimeSensorIds_.put(elasId, xmtaTimeSensorId);
	}

	//
	return xmtaTimeSensorId;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrAccumulate() {
	return xmtoAttrAccumulate_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrAdditive() {
	return xmtoAttrAdditive_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrBy() {
	return xmtoAttrBy_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrCalcMode() {
	return xmtoAttrCalcMode_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrFrom() {
	return xmtoAttrFrom_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrKeySplines() {
	return xmtoAttrKeySplines_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrKeyTimes() {
	return xmtoAttrKeyTimes_;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrPath() {
	return null;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrValues() {
	return xmtoAttrValues_;
}
/*********************************************************************** 
 * Handle the attribute by checking its syntax.  This is to catch all syntax 
 * errors which the schema can not check for, and thus the DOM parser 
 * doesn't find any problems.  Call the parent handler to see if it can
 * handle the attribute or not.
 *
 * @param  attributeName  node name of attribute.
 * @param  attributeValue node value of attribute.
 * @return whether this attribute is handled or not.
 **********************************************************************/
protected boolean handleValidateSyntax(String attributeName, String attributeValue) {
	boolean success = super.handleValidateSyntax(attributeName, attributeValue);

	// Determine the from field
	if (DOMXMTO.from.equals(attributeName)) {
		xmtoAttrFrom_ = attributeValue;
		return true;
	}
	// Determine the by field    
	if (DOMXMTO.by.equals(attributeName)) {
		xmtoAttrBy_ = attributeValue;
		return true;
	}
	// Determine the values field    
	if (DOMXMTO.values.equals(attributeName)) {
		xmtoAttrValues_ = attributeValue;
		return true;
	}
	// Determine the calcMode field    
	if (DOMXMTO.calcMode.equals(attributeName)) {
		xmtoAttrCalcMode_ = attributeValue;
		return true;
	}
	// Determine the accumulate field    
	if (DOMXMTO.accumulate.equals(attributeName)) {
		xmtoAttrAccumulate_ = attributeValue;
		return true;
	}
	// Determine the additive field    
	if (DOMXMTO.additive.equals(attributeName)) {
		xmtoAttrAdditive_ = attributeValue;
		return true;
	}
	// Determine the keyTimes field    
	if (DOMXMTO.keyTimes.equals(attributeName)) {
		xmtoAttrKeyTimes_ = attributeValue;
		return true;
	}
	// Determine the keySplines field    
	if (DOMXMTO.keySplines.equals(attributeName)) {
		xmtoAttrKeySplines_ = attributeValue;
		return true;
	}

	//
	return success;
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
/*********************************************************************** 
 * Validate the syntax of the interpolator keys and values.
 * This will generate the keys and values and check if there are any errors
 * with these values.  
 * If this is a problem then remove this node from the tree.
 **********************************************************************/
protected boolean validateSyntaxInterpolatorKeysValues(String xmtoAttrAttributeName) {
	Animatable animatable = getXmtoAttrTargetObject();
	int xmtaType = animatable.getAnimXmtaAttrFieldType(xmtoAttrAttributeName);

	// Check the variou types of attributes which we support animation.
	switch (xmtaType) {
		case Animatable.TYPE_SFColor :
			return validateSyntaxInterpolatorKeysValuesSFColor(xmtoAttrAttributeName);
		case Animatable.TYPE_SFVec2f :
			return validateSyntaxInterpolatorKeysValuesSFVec2f(xmtoAttrAttributeName);
		case Animatable.TYPE_SFFloat :
			return validateSyntaxInterpolatorKeysValuesSFFloat(xmtoAttrAttributeName);
	}

	// We don't support the attribute of this type so give an error message and then
	// remove the animation.
	MessageFormat msgFormat = new MessageFormat(xmto2XmtaRes_.getString(Xmto2XmtaRes.MSG_ERROR_ATTRIBUTE_VALUE_NOT_SUPPORTED));
	StringBuffer buffer = new StringBuffer();
	buffer.append(getXmtoElement().getNodeName());
	String id = getXmtoAttrId();
	if (null != id && !id.equals(""))
		buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
	Object[] msgArgs = {buffer.toString(), DOMXMTO.attributeName, xmtoAttrAttributeName};
	String message = msgFormat.format(msgArgs);
	fireErrorEvent(ErrorEvent.ERROR, 0, message, getXmtoElement(), DOMXMTO.attributeName);
	remove();
	return false;
}
/*********************************************************************** 
 * Validate the syntax of the interpolator keys and values.
 * This will generate the keys and values and check if there are any errors
 * with these values.  
 * If this is a problem then remove this node from the tree.
 **********************************************************************/
private boolean validateSyntaxInterpolatorKeysValuesSFColor(String xmtoAttrAttributeName) {
	Animatable animatable = getXmtoAttrTargetObject();

	// get the various attributes and convert them to the Color attributes
	// The problem is that it is possible that not all of the inputs (xmto attributes) are colors. If this is
	// the case then the only one who can convert them is the animable object. So let them.  Since it should also
	// be able to do Colors then might as well make them convert the color attributes as well.
	int xmtoType = animatable.getAnimXmtoAttrFieldType(xmtoAttrAttributeName);
	String sfColorStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrFrom(), xmtoType, Animatable.TYPE_SFColor);
	Color from = DOMXMTOParameter.param2Color(sfColorStr);
	sfColorStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrTo(), xmtoType, Animatable.TYPE_SFColor);
	Color to = DOMXMTOParameter.param2Color(sfColorStr);
	sfColorStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrBy(), xmtoType, Animatable.TYPE_SFColor);
	Color by = DOMXMTOParameter.param2Color(sfColorStr);

	// The values is a MF type field so we also have to figure out what is the correct MF type the input is.
	// Convert the SF type into a MF type field
	int xmtoMFType = getMFTypeField(xmtoType);
	String mfColorStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrValues(), xmtoMFType, Animatable.TYPE_MFColor);
	Color[] values = DOMXMTOParameter.param2Colors(mfColorStr);

	// The initial Value is already in the xmta color format
	Color initialValue = DOMXMTOParameter.param2Color(animatable.getAnimXmtaAttrInitialValue(xmtoAttrAttributeName));

	// These attributes are easy since they are the same for all types of xmto types.
	String calcMode = getXmtoAttrCalcMode();
	String accumulate = getXmtoAttrAccumulate();
	String additive = getXmtoAttrAdditive();
	Float[] keyTimes = DOMXMTOParameter.param2MFFloat(getXmtoAttrKeyTimes());
	Float[] keySplines = DOMXMTOParameter.param2KeySplines(getXmtoAttrKeySplines());
	String path = getXmtoAttrPath();
	TimingAttribute dur = getXmtoAttrDur();
	TimingAttribute repeatDur = getXmtoAttrRepeatDur();
	Float repeatCount = DOMXMTOParameter.param2Float(getXmtoAttrRepeatCount());

	//
	try {
		animationInterpolation_ = new ColorAnimationInterpolation(initialValue, from, to, by, values, calcMode, accumulate, additive, keyTimes, keySplines, path, dur, repeatDur, repeatCount);

		//
		Float[] keys = animationInterpolation_.getInterpolatorKeys();
		Color[] keyValues = (Color[]) animationInterpolation_.getInterpolatorKeyValues();
		xmtaAttrKey_ = DOMXMTAParameter.mfFloat2Param(keys);
		xmtaAttrKeyValue_ = DOMXMTAParameter.colors2Param(keyValues);

		//
		xmtaAttrCycleInterval_ = animationInterpolation_.getCycleInterval();
		xmtaAttrLoop_ = (animationInterpolation_.getLoop()) ? DOMXMTA.string_true : DOMXMTA.string_false;

		//
		if (debug_) {
			StringBuffer msg = new StringBuffer();
			msg.append("AnimationElement.validateSyntaxInterpolatorKeysValuesSFColor() -");
			msg.append(" loop=" + xmtaAttrLoop_);
			msg.append(" cycleInterval=" + xmtaAttrCycleInterval_);
			msg.append(" key=" + xmtaAttrKey_);
			msg.append(" keyValue=" + xmtaAttrKeyValue_);
			fireErrorEvent(ErrorEvent.TRACE, 0, msg.toString());
		}
	} catch (Exception e) {
		MessageFormat msgFormat = new MessageFormat(animationElementRes_.getString(AnimationElementRes.MSG_ERROR_ANIMATION_INVALID_INTERPOLATOR_ATTRIBUTES));
		StringBuffer buffer = new StringBuffer();
		buffer.append(getXmtoElement().getNodeName());
		String id = getXmtoAttrId();
		if (null != id && !id.equals(""))
			buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
		Object[] msgArgs = {buffer, e};
		String message = msgFormat.format(msgArgs);
		fireErrorEvent(ErrorEvent.ERROR, 0, message, getXmtoElement(), DOMXMTO.attributeName);
		remove();

		//
		animationInterpolation_ = null;
		return false;
	}

	// 
	return true;
}
/*********************************************************************** 
 * Validate the syntax of the interpolator keys and values.
 * This will generate the keys and values and check if there are any errors
 * with these values.  
 * If this is a problem then remove this node from the tree.
 **********************************************************************/
private boolean validateSyntaxInterpolatorKeysValuesSFFloat(String xmtoAttrAttributeName) {
	Animatable animatable = getXmtoAttrTargetObject();

	// get the various attributes and convert them to the SFvec2f attributes
	// The problem is that it is possible that not all of the inputs (xmto attributes) are SFvec2f. If this is
	// the case then the only one who can convert them is the animable object. So let them.  Since it should also
	// be able to do SFFloat then might as well make them convert the SFvec2f attributes as well.
	int xmtoType = animatable.getAnimXmtoAttrFieldType(xmtoAttrAttributeName);
	String sfFloatStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrFrom(), xmtoType, Animatable.TYPE_SFFloat);
	Float from = DOMXMTOParameter.param2Float(sfFloatStr);
	sfFloatStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrTo(), xmtoType, Animatable.TYPE_SFFloat);
	Float to = DOMXMTOParameter.param2Float(sfFloatStr);
	sfFloatStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrBy(), xmtoType, Animatable.TYPE_SFFloat);
	Float by = DOMXMTOParameter.param2Float(sfFloatStr);

	// The values is a MF type field so we also have to figure out what is the correct MF type the input is.
	// Convert the SF type into a MF type field
	int xmtoMFType = getMFTypeField(xmtoType);
	String mfFloatStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrValues(), xmtoMFType, Animatable.TYPE_MFFloat);
	Float[] values = DOMXMTOParameter.param2MFFloat(mfFloatStr);

	// The initial Value is already in the xmta SFFloat format
	Float initialValue = DOMXMTOParameter.param2Float(animatable.getAnimXmtaAttrInitialValue(xmtoAttrAttributeName));

	// These attributes are easy since they are the same for all types of xmto types.
	String calcMode = getXmtoAttrCalcMode();
	String accumulate = getXmtoAttrAccumulate();
	String additive = getXmtoAttrAdditive();
	Float[] keyTimes = DOMXMTOParameter.param2MFFloat(getXmtoAttrKeyTimes());
	Float[] keySplines = DOMXMTOParameter.param2KeySplines(getXmtoAttrKeySplines());
	String path = getXmtoAttrPath();
	TimingAttribute dur = getXmtoAttrDur();
	TimingAttribute repeatDur = getXmtoAttrRepeatDur();
	Float repeatCount = DOMXMTOParameter.param2Float(getXmtoAttrRepeatCount());

	//
	try {
		animationInterpolation_ = new ScalarAnimationInterpolation(initialValue, from, to, by, values, calcMode, accumulate, additive, keyTimes, keySplines, path, dur, repeatDur, repeatCount);

		//
		Float[] keys = animationInterpolation_.getInterpolatorKeys();
		Float[] keyValues = (Float[]) animationInterpolation_.getInterpolatorKeyValues();
		xmtaAttrKey_ = DOMXMTAParameter.mfFloat2Param(keys);
		xmtaAttrKeyValue_ = DOMXMTAParameter.mfFloat2Param(keyValues);

		//
		xmtaAttrCycleInterval_ = animationInterpolation_.getCycleInterval();
		xmtaAttrLoop_ = (animationInterpolation_.getLoop()) ? DOMXMTA.string_true : DOMXMTA.string_false;

		//
		if (debug_) {
			StringBuffer msg = new StringBuffer();
			msg.append("AnimationElement.validateSyntaxInterpolatorKeysValuesSFColor() -");
			msg.append(" loop=" + xmtaAttrLoop_);
			msg.append(" cycleInterval=" + xmtaAttrCycleInterval_);
			msg.append(" key=" + xmtaAttrKey_);
			msg.append(" keyValue=" + xmtaAttrKeyValue_);
			fireErrorEvent(ErrorEvent.TRACE, 0, msg.toString());
		}
	} catch (Exception e) {
		MessageFormat msgFormat = new MessageFormat(animationElementRes_.getString(AnimationElementRes.MSG_ERROR_ANIMATION_INVALID_INTERPOLATOR_ATTRIBUTES));
		StringBuffer buffer = new StringBuffer();
		buffer.append(getXmtoElement().getNodeName());
		String id = getXmtoAttrId();
		if (null != id && !id.equals(""))
			buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
		Object[] msgArgs = {buffer, e};
		String message = msgFormat.format(msgArgs);
		fireErrorEvent(ErrorEvent.ERROR, 0, message, getXmtoElement(), DOMXMTO.attributeName);
		remove();

		//
		animationInterpolation_ = null;
		return false;
	}

	// 
	return true;
}
/*********************************************************************** 
 * Validate the syntax of the interpolator keys and values.
 * This will generate the keys and values and check if there are any errors
 * with these values.  
 * If this is a problem then remove this node from the tree.
 **********************************************************************/
private boolean validateSyntaxInterpolatorKeysValuesSFVec2f(String xmtoAttrAttributeName) {
	Animatable animatable = getXmtoAttrTargetObject();

	// get the various attributes and convert them to the SFvec2f attributes
	// The problem is that it is possible that not all of the inputs (xmto attributes) are SFvec2f. If this is
	// the case then the only one who can convert them is the animable object. So let them.  Since it should also
	// be able to do SFvec2f then might as well make them convert the SFvec2f attributes as well.
	int xmtoType = animatable.getAnimXmtoAttrFieldType(xmtoAttrAttributeName);
	String sfVec2fStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrFrom(), xmtoType, Animatable.TYPE_SFVec2f);
	Vector from = DOMXMTOParameter.param2TwoFloats(sfVec2fStr);
	sfVec2fStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrTo(), xmtoType, Animatable.TYPE_SFVec2f);
	Vector to = DOMXMTOParameter.param2TwoFloats(sfVec2fStr);
	sfVec2fStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrBy(), xmtoType, Animatable.TYPE_SFVec2f);
	Vector by = DOMXMTOParameter.param2TwoFloats(sfVec2fStr);

	// The values is a MF type field so we also have to figure out what is the correct MF type the input is.
	// Convert the SF type into a MF type field
	int xmtoMFType = getMFTypeField(xmtoType);
	String mfVec2fStr = animatable.getAnimConvertValueXmto2Xmta(xmtoAttrAttributeName, getXmtoAttrValues(), xmtoMFType, Animatable.TYPE_MFVec2f);
	Vector[] values = DOMXMTOParameter.param2ArrayOfTwoFloats(mfVec2fStr);

	// The initial Value is already in the xmta sfVec2f format
	Vector initialValue = DOMXMTOParameter.param2TwoFloats(animatable.getAnimXmtaAttrInitialValue(xmtoAttrAttributeName));

	// These attributes are easy since they are the same for all types of xmto types.
	String calcMode = getXmtoAttrCalcMode();
	String accumulate = getXmtoAttrAccumulate();
	String additive = getXmtoAttrAdditive();
	Float[] keyTimes = DOMXMTOParameter.param2MFFloat(getXmtoAttrKeyTimes());
	Float[] keySplines = DOMXMTOParameter.param2KeySplines(getXmtoAttrKeySplines());
	String path = getXmtoAttrPath();
	TimingAttribute dur = getXmtoAttrDur();
	TimingAttribute repeatDur = getXmtoAttrRepeatDur();
	Float repeatCount = DOMXMTOParameter.param2Float(getXmtoAttrRepeatCount());

	//
	try {
		animationInterpolation_ = new Position2DAnimationInterpolation(initialValue, from, to, by, values, calcMode, accumulate, additive, keyTimes, keySplines, path, dur, repeatDur, repeatCount);

		//
		Float[] keys = animationInterpolation_.getInterpolatorKeys();
		Vector[] keyValues = (Vector[]) animationInterpolation_.getInterpolatorKeyValues();
		xmtaAttrKey_ = DOMXMTAParameter.mfFloat2Param(keys);
		xmtaAttrKeyValue_ = DOMXMTAParameter.arrayOfTwoFloats2Param(keyValues);

		//
		xmtaAttrCycleInterval_ = animationInterpolation_.getCycleInterval();
		xmtaAttrLoop_ = (animationInterpolation_.getLoop()) ? DOMXMTA.string_true : DOMXMTA.string_false;

		//
		if (debug_) {
			StringBuffer msg = new StringBuffer();
			msg.append("AnimationElement.validateSyntaxInterpolatorKeysValuesSFColor() -");
			msg.append(" loop=" + xmtaAttrLoop_);
			msg.append(" cycleInterval=" + xmtaAttrCycleInterval_);
			msg.append(" key=" + xmtaAttrKey_);
			msg.append(" keyValue=" + xmtaAttrKeyValue_);
			fireErrorEvent(ErrorEvent.TRACE, 0, msg.toString());
		}
	} catch (Exception e) {
		MessageFormat msgFormat = new MessageFormat(animationElementRes_.getString(AnimationElementRes.MSG_ERROR_ANIMATION_INVALID_INTERPOLATOR_ATTRIBUTES));
		StringBuffer buffer = new StringBuffer();
		buffer.append(getXmtoElement().getNodeName());
		String id = getXmtoAttrId();
		if (null != id && !id.equals(""))
			buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
		Object[] msgArgs = {buffer, e};
		String message = msgFormat.format(msgArgs);
		fireErrorEvent(ErrorEvent.ERROR, 0, message, getXmtoElement(), DOMXMTO.attributeName);
		remove();

		//
		animationInterpolation_ = null;
		return false;
	}

	// 
	return true;
}
}
