/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  GroupElement.java
//
//=============================================================================
package xmto2xmta.tree;

import java.lang.reflect.Method;
import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import xmto2xmta.ActiveInterval;
import xmto2xmta.ErrorEvent;
import xmto2xmta.EventModel;
import xmto2xmta.IdManager;
import xmto2xmta.LifeCycle;
import xmto2xmta.TimeScheduler;
import xmto2xmta.Xmto2Xmta;
import xmto2xmta.resources.Xmto2XmtaRes;
import xmto2xmta.resources.GroupElementRes;
import xmto2xmta.timegraph.ELASid;
import xmtutil.DOMXMTO;
import xmtutil.TimingAttribute;
/*********************************************************************** 
 * Class to group together the various type elements who group together other elements.
 *
 * <p><b>History:</b><ul>
 * <li>13 Dec 2000 - Created file.
 * <li>04 Jun 2001 - fix problem with Z-order for group.  Problem with if statements.
 * <li>21 Jun 2001 - Add xmtaMedia child element.
 * </ul>
 *********************************************************************/
public abstract class GroupElement extends TimingElement implements ZorderSupport {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	/*********************************************************************** 
	 **********************************************************************/
	private Hashtable deletePositions_ = null;
	private boolean positionsDetermined_ = false;
	private boolean usePositions_ = false;

	/*********************************************************************** 
	 **********************************************************************/
	static private Class transformationClass_ = null;
	static private Class groupClass_ = null;
	static private Method getXmtoAttrOrderMethodTransformation_ = null;
	static private Method getXmtoAttrOrderMethodGroup_ = null;
	static private Method getXmtoAttrRegionMethodGroup_ = null;
	static private boolean classesLoaded_ = false;

	/*********************************************************************** 
	 **********************************************************************/
	private ResourceBundle groupElementRes_ = ResourceBundle.getBundle("xmto2xmta.resources.GroupElementRes", Locale.getDefault());
	private ResourceBundle xmto2XmtaRes_ = ResourceBundle.getBundle("xmto2xmta.resources.Xmto2XmtaRes", Locale.getDefault());
/**********************************************************************
 * constructor.
 * @param xmtoElement XMT-O element for conversion.
 * @param xmtoIds     hashtable to contain XMT-O Conversion elements 
 *                    with the key of their XMT-O id.
 * @param xmtaIds     hashtable to contain XMT-A elements with the key
 *                    of their XMT-A id.
 * @param xmtaDoc     DOM Document for the new XMT-A elements.
 * @param xmto2Xmta   Converter, used for firing error events.
 *********************************************************************/
public GroupElement(Element xmtoElement, Hashtable xmtoIds, Hashtable xmtaIds,Hashtable convertIndexTable, Document xmtaDoc, Xmto2Xmta xmto2Xmta) {
	super(xmtoElement, xmtoIds, xmtaIds, convertIndexTable, xmtaDoc, xmto2Xmta);
}
/*********************************************************************** 
 * Create a list of the timing element children of this GroupElement node.
 * This includes all media elements and group elements.  If a Time container
 * element is encountered then recurse through the time container element adding
 * its elements.  There can be par elements because the &lt.body&gt. elements is
 * a GroupElement.  The list should be in document order.
 *
 * @return list of the children elements.
 **********************************************************************/
private Vector determineChildrenElements(ConvertElement parent) {
	// empty children list.
	Vector deleteChildrenElements = new Vector();

	// get all the children
	Vector children = parent.getChildren();

	// loop through the children adding the appropraiate children to the list
	for (Enumeration enum = children.elements(); enum.hasMoreElements();) {
		Object child = enum.nextElement();
		if (child instanceof MediaElement) {
			String xmtoAttrRegion = ((MediaElement) child).getXmtoAttrRegion();
			// only add it if it is not slated for a region 
			if (null == xmtoAttrRegion || xmtoAttrRegion.equals(""))
				deleteChildrenElements.add(child);
		} else
			if (child instanceof TimeContainerElement) {
				// If a Time container element is encountered then recurse through the 
				// time container element adding its elements.  There can be par 
				// elements because the &lt.body&gt. elements is a GroupElement.
				Vector newChildren = determineChildrenElements((ConvertElement) child);
				for (Enumeration newEnum = newChildren.elements(); newEnum.hasMoreElements();)
					deleteChildrenElements.add(newEnum.nextElement());
			} else {
				if (TreeClassObjects.getClassObject(TreeClassObjects.XMTO2XMTA_TREE_xmtaMedia).isInstance(child)) {
					deleteChildrenElements.add(child);
				} else {
					if (!classesLoaded_)
						loadClasses();
					if (null != getXmtoAttrRegionMethodGroup_) {
						if (groupClass_.isInstance(child)) {
							try {
								String xmtoAttrRegion = (String) getXmtoAttrRegionMethodGroup_.invoke(child, null);
								// only add it if it is not slated for a region 
								if (null == xmtoAttrRegion || xmtoAttrRegion.equals(""))
									deleteChildrenElements.add(child);
							} catch (Exception e) {
							}
						}
					}
				}
			}
			// else we don't do add the element.
	}

	//
	return deleteChildrenElements;
}
/*********************************************************************** 
 * Add a media element into the ordered group.
 *
 * @param mediaElement XMT-A media object to be added into ordered group.
 **********************************************************************/
private void determineDeleteFieldPositions() {
	positionsDetermined_ = true;

	// First need to create a list of the timing element children of this GroupElement node.
	// This includes all media elements and group elements.
	// the list is in document order.
	Vector children = determineChildrenElements(this);
	if (debug_) {
		StringBuffer elementString = new StringBuffer();
		elementString.append(" Element=<").append(getXmtoElement().getNodeName());
		String id = getXmtoAttrId();
		if (null != id && !id.equals(""))
			elementString.append(" id=").append(id);
		elementString.append(">");
		for (int i = 0; i < children.size(); i++) {
			ConvertElement child = (ConvertElement) children.elementAt(i);
			StringBuffer buffer = new StringBuffer();
			buffer.append("GroupElement.determineDeleteFieldPositions() -").append(elementString).append(" child[").append(i).append("]=<");
			buffer.append(child.getXmtoElement().getNodeName());
			if (null != child.getXmtoAttrId() && !child.getXmtoAttrId().equals(""))
				buffer.append(" id=").append(child.getXmtoAttrId());
			buffer.append(">");
			fireErrorEvent(ErrorEvent.TRACE, 0, buffer.toString(), getXmtoElement());
		}
	}

	// Check whether we use the z-order positions or not.
	usePositions_ = determineUsePosition(this, children);
	if (debug_)
		fireErrorEvent(ErrorEvent.TRACE, 0, ("GroupElement.determineDeleteFieldPositions() - usePositions=" + usePositions_), getXmtoElement());

	// If we're not using positions then there is no need to continue.
	if (!usePositions_)
		return;

	// Next go through the list of children enumerating all their simple durations
	// This should be a list of all the ELASid ids in the order in which they are added to the
	// list.
	// Next go through the list of simple durations determining their position at deletion time.
	deletePositions_ = determineDeletePositions(this, children);

	//
	return;
}
/*********************************************************************** 
 * Create a table of all the (z-order) delete positions for all of the simple 
 * durations for the specified children.
 * <p>
 * First create a list of all the ELASid ids for the children in their add order.
 * The add order is the order in which the simple durations are added
 * to the scene.  The list of the children is in document order.
 * Go through the list of children enumerating all their simple durations
 * This should be a list of all the ELASid ids in the order in which they are added to the
 * list.
 * <p>
 * Next go through the list of simple durations determining their position at deletion time.
 * <p>
 * NOTE: this method is a static method so that it can be shared with
 * &lt;region&gt;
 *
 * @param element  the parent of the list of children parameter.  Used for debug purposes
 *                 to specify which element has problems with determing the 
 *                 delete positions of the children.
 * @param children list of the children is in document order.
 * @return         sorted list of the ELASid ids in delete order.
 **********************************************************************/
static public Hashtable determineDeletePositions(ConvertElement element, Vector children) {
	// empty children list.

	//
	TimeScheduler timeScheduler = element.getXmto2Xmta().getTimeScheduler();

	//
	Hashtable insertTimes = new Hashtable(89);
	Hashtable deleteTimes = new Hashtable(89);
	Vector elasIdsInAddOrder = new Vector();

	// loop through the children adding the appropraiate children to the list
	for (Enumeration enum = children.elements(); enum.hasMoreElements();) {

		// Get the child
		TimingElement child = (TimingElement) enum.nextElement();
		Integer childIndex = child.getIndex();

		// get the lifecycles of the child
		Vector lifeCycles = timeScheduler.getLifeCycles(childIndex);
		for (int lcIndex = 0; lcIndex < lifeCycles.size(); lcIndex++) {
			LifeCycle lifeCycle = (LifeCycle) lifeCycles.elementAt(lcIndex);
			Integer lifeCycleIndex = new Integer(lcIndex);

			// get the Active Intervals
			Vector activeIntervals = lifeCycle.getActiveIntervals();
			for (int aiIndex = 0; aiIndex < activeIntervals.size(); aiIndex++) {
				ActiveInterval activeInterval = (ActiveInterval) activeIntervals.elementAt(aiIndex);
				Integer activeIntervalIndex = new Integer(aiIndex);

				// go through the simple durations
				Vector elasIds = activeInterval.getELASids();
				Vector repeatTimes = activeInterval.getRepeatTimes();
				int sdCount = elasIds.size();
				for (int sdIndex = 0; sdIndex < sdCount; sdIndex++) {
					ELASid elasId = (ELASid) elasIds.elementAt(sdIndex);
					TimingAttribute insertTime;
					TimingAttribute deleteTime;

					// Get the insert time
					if (0 == sdIndex) {
						// Insert time is the observed time not the computed time.  This is the time
						// the node is placed in the Bifs tree for the z-ordering.
						insertTime = activeInterval.getObservedBeginTime();
					} else
						insertTime = (TimingAttribute) repeatTimes.elementAt(sdIndex - 1);


					// get the delete time	
					if (sdCount - 1 == sdIndex)
						deleteTime = activeInterval.getFillEndTime();
					else
						deleteTime = (TimingAttribute) repeatTimes.elementAt(sdIndex);

					// Add the times to the tables
					insertTimes.put(elasId, insertTime);
					deleteTimes.put(elasId, deleteTime);

					// Loop through all the elasIds in the vector to place this one in the time order.
					boolean added = false;
					int addCount = elasIdsInAddOrder.size();
					for (int addIndex = 0; addIndex < addCount; addIndex++) {
						ELASid addElasId = (ELASid) elasIdsInAddOrder.elementAt(addIndex);
						TimingAttribute addInsertTime = (TimingAttribute) insertTimes.get(addElasId);
						if (insertTime.lessThan(addInsertTime)) {
							elasIdsInAddOrder.add(addIndex, elasId);
							added = true;
							break;
						}
					}
					// If is after all the events in the list then add it to the end of the list.
					if (!added)
						elasIdsInAddOrder.add(elasId);
				}

				//
				if (debug_) {
					StringBuffer sb = new StringBuffer();
					sb.append("GroupElement.determineELASidsInAddOrder() -");
					sb.append(" Element=<").append(element.getXmtoElement().getNodeName());
					String id = element.getXmtoAttrId();
					if (null != id && !id.equals(""))
						sb.append(" id=").append(id);
					sb.append(">");
					sb.append(" child=<").append(child.getXmtoElement().getNodeName());
					if (null != child.getXmtoAttrId() && !child.getXmtoAttrId().equals(""))
						sb.append(" id=").append(child.getXmtoAttrId());
					sb.append(">");
					sb.append(" lifeCycle[").append(lifeCycleIndex).append("]");
					sb.append(" activeInterval[").append(activeIntervalIndex).append("]");
					sb.append(" computedBeginTime=").append(activeInterval.getComputedBeginTime());
					sb.append(" observedBeginTime=").append(activeInterval.getObservedBeginTime());
					sb.append(" intrinsicDuration=").append(activeInterval.getIntrinsicDuration());
					for (Enumeration repeatEnum = activeInterval.getRepeatTimes().elements(); repeatEnum.hasMoreElements();) {
						TimingAttribute repeatTime = (TimingAttribute) repeatEnum.nextElement();
						sb.append(" repeatTime=").append(repeatTime);
					}
					sb.append(" activeDuration=").append(activeInterval.getActiveDuration());
					sb.append(" fillEndTime=").append(activeInterval.getFillEndTime());
					element.fireErrorEvent(ErrorEvent.TRACE, 0, sb.toString());
				}
			}
		}
	}

	//
	if (debug_) {
		Hashtable xmtoIds = element.getXmtoIds();
		StringBuffer elementString = new StringBuffer();
		elementString.append(" Element=<").append(element.getXmtoElement().getNodeName());
		String id = element.getXmtoAttrId();
		if (null != id && !id.equals(""))
			elementString.append(" id=").append(id);
		elementString.append(">");
		for (Enumeration enum = elasIdsInAddOrder.elements(); enum.hasMoreElements();) {
			ELASid elasId = (ELASid) enum.nextElement();
			TimingAttribute insertTime = (TimingAttribute) insertTimes.get(elasId);
			TimingAttribute deleteTime = (TimingAttribute) deleteTimes.get(elasId);
			ConvertElement convertElement = (ConvertElement) element.getConvertIndexTable().get(elasId.getElementIndex());
			StringBuffer sb = new StringBuffer();
			sb.append("GroupElement.determineELASidsInAddOrder() -").append(elementString);
			sb.append(" child=<").append(convertElement.getXmtoElement().getNodeName());
			if (null != convertElement.getXmtoAttrId() && !convertElement.getXmtoAttrId().equals(""))
				sb.append(" id=").append(convertElement.getXmtoAttrId());
			sb.append(">");
			sb.append(" insertTime=").append(insertTime);
			sb.append(" deleteTime=").append(deleteTime);
			element.fireErrorEvent(ErrorEvent.TRACE, 0, sb.toString());
		}
	}

	// Next go through the list of simple durations determining their position at deletion time.
	Hashtable deletePositions = determineDeletePositions(element, elasIdsInAddOrder, deleteTimes);
	return deletePositions;
}
/*********************************************************************** 
 * Create list of all the ELASid ids for the children in their add order.
 * The add order is the order in which the simple durations are added
 * to the scene.  The list of the children is in document order.
 *
 * @return sorted list of the ELASid ids.
 **********************************************************************/
static private Hashtable determineDeletePositions(ConvertElement element, Vector elasIdsInAddOrder, Hashtable deleteTimes) {
	// First need to create a copy of the list of simple durations fro the timing element 
	// children of this GroupElement node.
	// This includes all media elements and group elements.
	Vector elasIdsDelete = (Vector) elasIdsInAddOrder.clone();
	Hashtable deletePositions = new Hashtable(89);
	Hashtable convertIndexTable = element.getConvertIndexTable();

	// write delete node commands
	// What we do is go through the list finding the next one to be deleted.
	// Delete it and it's position. Then continue to the next object.
	while (!elasIdsDelete.isEmpty()) {
		// Find the next one to be deleted in the list.  It is the
		// first object with the lowest delete time.
		// Initialize the first time to the first object.
		int nextDeleteListIndex = 0;
		ELASid nextDeleteElasId = (ELASid) elasIdsDelete.elementAt(nextDeleteListIndex);
		TimingAttribute nextDeleteTime = (TimingAttribute) deleteTimes.get(nextDeleteElasId);

		// Go through the list looking to see if there are any with a shorter time
		for (int i = 1, count = elasIdsDelete.size(); i < count; i++) {
			ELASid nextElasId = (ELASid) elasIdsDelete.elementAt(i);
			TimingAttribute nextTime = (TimingAttribute) deleteTimes.get(nextElasId);

			// Check if earlier delete	
			if (nextTime.lessThan(nextDeleteTime)) {
				nextDeleteElasId = nextElasId;
				nextDeleteListIndex = i;
				nextDeleteTime = nextTime;
			}
		}

		// Add the delete position to the has table
		deletePositions.put(nextDeleteElasId, new Integer(nextDeleteListIndex));

		// Now remove the element from the list
		elasIdsDelete.removeElementAt(nextDeleteListIndex);

		// print debug info
		if (debug_) {
			StringBuffer str = new StringBuffer();
			str.append("GroupElement.determineDeletePositions() -");
			str.append(" Element=<").append(element.getXmtoElement().getNodeName());
			String id = element.getXmtoElement().getAttribute(DOMXMTO.id);
			if (null != id && !id.equals(""))
				str.append(" id=").append(id);
			str.append(">");
			TimingElement nextDeleteObject = (TimingElement) convertIndexTable.get(nextDeleteElasId.getElementIndex());
			str.append(" delete Element=<").append(nextDeleteObject.getXmtoElement().getNodeName());
			id = nextDeleteObject.getXmtoElement().getAttribute(DOMXMTO.id);
			if (null != id && !id.equals(""))
				str.append(" id=").append(id);
			str.append(">");
			str.append(" index=").append(nextDeleteListIndex).append(" atTime=").append(nextDeleteTime);
			element.fireErrorEvent(ErrorEvent.TRACE, 0, str.toString());
		}
	}
	return deletePositions;
}
/*********************************************************************** 
 * Create list of all the ELASid ids for the children in their add order.
 * The add order is the order in which the simple durations are added
 * to the scene.  The list of the children is in document order.
 * <p>
 * NOTE: this method is a static method so that it can be shared with
 * &lt;region&gt;
 *
 * @return whether to use the z-order positioning or not.
 **********************************************************************/
static public boolean determineUsePosition(ConvertElement element, Vector children) {
	//
	if (!classesLoaded_)
		loadClasses();

	//
	boolean foundPosition = false;
	boolean foundNoPosition = false;

	// loop through the children adding the appropraiate children to the list
	for (Enumeration enum = children.elements(); enum.hasMoreElements();) {
		ConvertElement child = (ConvertElement) enum.nextElement();
		// Ignore audio elements since they don't change the ordering
		if (child instanceof VisualElement) {
			if (null != getXmtoAttrOrderMethodTransformation_) {
				// Check to see if it has a transformation
				for (Enumeration transEnum = child.getChildren().elements(); transEnum.hasMoreElements();) {
					Object obj = transEnum.nextElement();
					if (transformationClass_.isInstance(obj)) {
						try {
							String order = (String) getXmtoAttrOrderMethodTransformation_.invoke(obj, null);
							if (null == order || order.equals(""))
								foundNoPosition = true;
							else
								foundPosition = true;
						} catch (Exception e) {
						}
					}
				}
			}
		} else {
			if (null != getXmtoAttrOrderMethodGroup_ && groupClass_.isInstance(child)) {
				try {
					String order = (String) getXmtoAttrOrderMethodGroup_.invoke(child, null);
					if (null == order || order.equals(""))
						foundNoPosition = true;
					else
						foundPosition = true;
				} catch (Exception e) {
				}
			} else {
				if (TreeClassObjects.getClassObject(TreeClassObjects.XMTO2XMTA_TREE_xmtaMedia).isInstance(child)) {
					// The xmtaMedia don't have any z-ordering positioning.
					foundNoPosition = true;
				}
			}
		}
		// If we find that we are both using and not using position then break out.
		if (foundPosition && foundNoPosition)
			break;
	}

	// If we didn't find any positions then we aren't using a position
	// It is posible we didn't find either in that case then we still
	// don't use a position.
	if (!foundPosition)
		return false;

	// We found that all thechildren are using z-order position.
	if (!foundNoPosition)
		return true;

	// What we are left with is a mixture of some children using z-order
	// and others not using z-order. Now the questions is is there user 
	// interaction or not that begins or ends the children objects.
	// * In the case where there is no user interaction.
	//   It is possible to figure out the z-order positioning by
	//   determining what is being displayed on top at the time and
	//   assigning the z-order to be the same as it thus the last
	//   item added is displayed.
	// * In the case where there is user interaction.
	//   What is needed is runtime support to determine what is the
	//   the top and add it to it.
	// We don't do either.  What we do is display an error message
	// and ignore the z-order positioning.
	ResourceBundle groupElementRes = ResourceBundle.getBundle("xmto2xmta.resources.GroupElementRes", Locale.getDefault());
	MessageFormat msgFormat = new MessageFormat(groupElementRes.getString(GroupElementRes.MSG_WARNING_MIXED_ORDER));
	StringBuffer buffer = new StringBuffer();
	buffer.append(element.getXmtoElement().getNodeName());
	String id = element.getXmtoAttrId();
	if (null != id && !id.equals(""))
		buffer.append(" ").append(DOMXMTO.id).append("=\"").append(id).append("\"");
	Object[] msgArgs = {buffer};
	String message = msgFormat.format(msgArgs);
	element.fireErrorEvent(ErrorEvent.WARNING, 0, message, element.getXmtoElement());

	//
	return false;
}
/*********************************************************************** 
 * Determine the field position at time of deletion for the media/group element
 * in the group.
 *
 * @param  elasId id to specify which simple duration of a child element.
 * @return the field position at time of deletion for the media/group element.
 **********************************************************************/
public int getDeleteFieldPosition(ELASid elasId) {
	if (!positionsDetermined_)
		determineDeleteFieldPositions();

	//
	Object position = deletePositions_.get(elasId);

	//
	if (null == position)
		return 0;
	else
		return ((Integer) position).intValue();
}
/*********************************************************************** 
 * Get the id for the XMT-A ordered group to which media objects are added.
 *
 * @param  elasId id to specify which simple duration of a child element.
 * @return the id (DEF) of the XMT-A ordered group.
 **********************************************************************/
public abstract String getXmtaOrderedGroupId(ELASid elasId);
/*********************************************************************** 
 **********************************************************************/
static private void loadClasses() {
	//
	if (classesLoaded_)
		return;

	//
	classesLoaded_ = true;

	//
	try {
		transformationClass_ = Class.forName("xmto2xmta.tree.transformation");
		getXmtoAttrOrderMethodTransformation_ = transformationClass_.getDeclaredMethod("getXmtoAttrOrder", null);
	} catch (Exception e) {
	}

	//
	try {
		groupClass_ = Class.forName("xmto2xmta.tree.group");
		getXmtoAttrOrderMethodGroup_ = groupClass_.getDeclaredMethod("getXmtoAttrOrder", null);
		getXmtoAttrRegionMethodGroup_ = groupClass_.getDeclaredMethod("getXmtoAttrRegion", null);
	} catch (Exception e) {
	}
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
/*********************************************************************** 
 * Return whether the z-ordering positioning is used for the children 
 * or not.
 * 
 * @return whether z-ordering is used for this GroupElement or not.
 **********************************************************************/
public boolean usePositions() {
	if (!positionsDetermined_)
		determineDeleteFieldPositions();

	//
	return usePositions_;
}
}
