/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  InternalElement.java
//
//=============================================================================
package xmto2xmta.tree;

import java.util.Hashtable;
import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import xmto2xmta.ErrorEvent;
import xmto2xmta.IdManager;
import xmto2xmta.Xmto2Xmta;
import xmto2xmta.ActiveInterval;
import xmto2xmta.LifeCycle;
import xmto2xmta.TimeScheduler;
import xmto2xmta.timegraph.ELASid;
/*********************************************************************** 
 * Class to group together the various type media elements whose data is internal to the XMT file.
 *
 * <p><b>History:</b><ul>
 * <li>13 Dec 2000 - Created file.
 * <li>06 Jun 2001 - Add id on Appearance element
 * <li>12 Jun 2001 - Add defs support.
 * </ul>
 *********************************************************************/
public abstract class InternalElement extends MediaElement implements DefsObject, VisualElement {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	/*********************************************************************** 
	 **********************************************************************/
	private Hashtable xmtaMaterial2dIds_ = new Hashtable(89);
	private Hashtable xmtaAppearanceIds_ = new Hashtable(89);

	/*********************************************************************** 
	 **********************************************************************/
	private boolean isInDefsTreeInitialized_ = false;
	private boolean isInDefsTree_ = false;
	private boolean isDefsParent_ = false;
/**********************************************************************
 * constructor.
 * @param xmtoElement XMT-O element for conversion.
 * @param xmtoIds     hashtable to contain XMT-O Conversion elements 
 *                    with the key of their XMT-O id.
 * @param xmtaIds     hashtable to contain XMT-A elements with the key
 *                    of their XMT-A id.
 * @param xmtaDoc     DOM Document for the new XMT-A elements.
 * @param xmto2Xmta   Converter, used for firing error events.
 *********************************************************************/
public InternalElement(Element xmtoElement, Hashtable xmtoIds, Hashtable xmtaIds, Hashtable convertIndexTable, Document xmtaDoc, Xmto2Xmta xmto2Xmta) {
	super(xmtoElement, xmtoIds, xmtaIds, convertIndexTable, xmtaDoc, xmto2Xmta);
}
/***********************************************************************
 * get the id for the material element.
 **********************************************************************/
public String getXmtaAppearanceId(ELASid elasId) {
	String xmtaAppearanceId = (String) xmtaAppearanceIds_.get(elasId);

	// If couldn't find one then create it.
	if (null == xmtaAppearanceId) {
		xmtaAppearanceId = IdManager.getNextId(IdManager.AppearanceID);
		// make the group ids based on object id or make them unique.
		if (null != getXmtoAttrId() && !getXmtoAttrId().equals(""))
			xmtaAppearanceId = xmtaAppearanceId + ":" + getXmtoAttrId();
		xmtaAppearanceIds_.put(elasId, xmtaAppearanceId);
	}

	//
	return xmtaAppearanceId;
}
/***********************************************************************
 * Get the defs id of the object.
 * @return the unique id of this object.
 **********************************************************************/
public String getXmtaDefsId() {
	if (!isInDefsTree()) {
		if (debug_)
			fireErrorEvent(ErrorEvent.TRACE, 0, "InternalElement.getXmtaDefsId() - is being called on a none defs object", getXmtoElement());

		// This should never be called but just in case return empty string.
		return "";
	}

	//
	TimeScheduler timeScheduler = getXmto2Xmta().getTimeScheduler();
	if (null == timeScheduler)
		return "";
	Vector lifeCycles = timeScheduler.getLifeCycles(getIndex());
	if (null == lifeCycles)
		return "";
	LifeCycle lifeCycle = (LifeCycle) lifeCycles.firstElement();
	Vector activeIntervals = lifeCycle.getActiveIntervals();
	if (null == activeIntervals)
		return "";
	ActiveInterval activeInterval = (ActiveInterval) activeIntervals.firstElement();
	Vector elasIds = activeInterval.getELASids();
	if (null == elasIds)
		return "";
	ELASid elasId = (ELASid) elasIds.firstElement();
	String xmtaDefsId = getXmtaGroupId(elasId);
	return xmtaDefsId;
}
/***********************************************************************
 * Routine to initialize the defs query flags.  This is added to provide
 * consistency for all the classes that implement the DefsObject interface.
 **********************************************************************/
public void initDefsFlags() {
	if (isInDefsTreeInitialized_)
		return;

	//
	isInDefsTreeInitialized_ = true;
	isInDefsTree_ = false;
	isDefsParent_ = false;

	// First check if parent is defs
	TreeElement parent = getParent();
	if (null != parent && parent instanceof defs) {
		isInDefsTree_ = true;
		isDefsParent_ = true;
		return;
	}


	// Next check if we are in the defs tree	
	while (null != parent) {
		if (parent instanceof defs) {
			isInDefsTree_ = true;
			return;
		}
		parent = parent.getParent();
	}
}
/***********************************************************************
 * Determine whether the object is a child of the defs element or not.  
 * This means that the object is a direct child of the defs element.
 * It isn't a child if it isn't in the defs tree or is a descendent 
 * (starting at grandchild) of the defs element.
 * @return whether the object is a child of the defs element.  True if 
 * it is a child of the defs element, false otherwise.
 **********************************************************************/
public boolean isDefsParent() {
	if (!isInDefsTreeInitialized_)
		initDefsFlags();

	//
	return isDefsParent_;
}
/***********************************************************************
 * Determine whether the object is in the defs tree or not.  This means
 * that the object is defined somewhere in the Defs tree.  It may be a
 * child or descendent of the defs element.
 * @return whether the object is in the defs tree.  True if a descendent
 * of the defs element, false otherwise.
 **********************************************************************/
public boolean isInDefsTree() {
	if (!isInDefsTreeInitialized_)
		initDefsFlags();

	//
	return isInDefsTree_;
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
}
