/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  XmediaChildrenExternalElement.java
//
//=============================================================================
package xmto2xmta.tree;

import java.text.MessageFormat;
import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import xmto2xmta.ActiveInterval;
import xmto2xmta.ErrorEvent;
import xmto2xmta.IdManager;
import xmto2xmta.LifeCycle;
import xmto2xmta.TimeScheduler;
import xmto2xmta.Xmto2Xmta;
import xmto2xmta.resources.ExternalElementRes;
import xmto2xmta.timegraph.ELASid;
import xmtutil.DOMXMTO;
/*********************************************************************** 
 * Class to group together the various type media elements whose data is internal to the XMT file.
 *
 * <p><b>History:</b><ul>
 * <li>13 Dec 2000 - Created file.
 * <li>06 Jun 2001 - Add id on Appearance element
 * <li>06 Jun 2001 - Add defs support.
 * <li>12 Jun 2001 - Add support for defs id.
 * <li>30 Jun 2001 - Fix bug, on validating syntax if problem with parent then
 *                   return directly the problem.
 * </ul><p>
 *********************************************************************/
public abstract class XmediaChildrenExternalElement extends XmediaChildrenElement implements DefsObject {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	/*********************************************************************** 
	 **********************************************************************/
	private String xmtoAttrSrc_ = null;

	/*********************************************************************** 
	 **********************************************************************/
	private Hashtable xmtaObjectDescriptorIds_ = new Hashtable(89);
	private Hashtable xmtaEsDescriptorIds_ = new Hashtable(89);
	private String xmtaObjectDescriptorId_ = null;
	private String xmtaEsDescriptorId_ = null;

	/*********************************************************************** 
	 **********************************************************************/
	private ResourceBundle externalElementRes_ = ResourceBundle.getBundle("xmto2xmta.resources.ExternalElementRes", Locale.getDefault());
/**********************************************************************
 * constructor.
 * @param xmtoElement XMT-O element for conversion.
 * @param xmtoIds     hashtable to contain XMT-O Conversion elements 
 *                    with the key of their XMT-O id.
 * @param xmtaIds     hashtable to contain XMT-A elements with the key
 *                    of their XMT-A id.
 * @param xmtaDoc     DOM Document for the new XMT-A elements.
 * @param xmto2Xmta   Converter, used for firing error events.
 *********************************************************************/
public XmediaChildrenExternalElement(Element xmtoElement, Hashtable xmtoIds, Hashtable xmtaIds, Hashtable convertIndexTable, Document xmtaDoc, Xmto2Xmta xmto2Xmta) {
	super(xmtoElement, xmtoIds, xmtaIds, convertIndexTable, xmtaDoc, xmto2Xmta);
}
/***********************************************************************
 * Get the defs id of the object.
 * @return the unique id of this object.
 **********************************************************************/
public String getXmtaEsDescriptorId() {
	if (null != xmtaEsDescriptorId_)
		return xmtaEsDescriptorId_;

	// Not a valid defs object
	if (!isInDefsTree()) {
		if (debug_)
			fireErrorEvent(ErrorEvent.TRACE, 0, "XmediaChildrenExternalElement.getXmtaEsDescriptorId() - is being called on a none defs object", getXmtoElement());

		// This should never be called but just in case return empty string.
		return "";
	}

	// Check whether the defs is a child of the
	// <defs> element or a descendent of a TimingElement.  If it is a child of the
	// defs then it doesn't have a elasid.  If it is a descendent of a timingelement then
	// use the elasId of the timingElement.
	// First check if it is a child of the <defs> element.
	if (isDefsParent()) {
		// If couldn't find one then create it.
		if (null == xmtaEsDescriptorId_) {
			xmtaEsDescriptorId_ = IdManager.getNextId(IdManager.ESID);
			// make the group ids based on object id or make them unique.
			String id = getXmtoAttrId();
			if (null != id && !id.equals(""))
				xmtaEsDescriptorId_ = xmtaEsDescriptorId_ + ":" + id;
		}
		//
		return xmtaEsDescriptorId_;
	}

	// See if the element is a descendent of defs but not through a timing element.
	TreeElement parent = getParent();
	while (null != parent && !(parent instanceof TimingElement))
		parent = parent.getParent();

	// This is a defs element under another defs element.  Thus look for its
	// If we couldn't find the parent then this must be a descendent of the defs
	// through other non timing elements, e.g. outline.
	// Treat it the same as a child of the defs.
	if (null == parent) {
		// If couldn't find one then create it.
		if (null == xmtaEsDescriptorId_) {
			xmtaEsDescriptorId_ = IdManager.getNextId(IdManager.ESID);
			// make the group ids based on object id or make them unique.
			String id = getXmtoAttrId();
			if (null != id && !id.equals(""))
				xmtaEsDescriptorId_ = xmtaEsDescriptorId_ + ":" + id;
		}
		//
		return xmtaEsDescriptorId_;
	}

	// Now we must be a descendent in the defs tree and a descendent of a timing
	// element.  Thus we need the elasid to distinguish this id.
	TimeScheduler timeScheduler = getXmto2Xmta().getTimeScheduler();
	if (null == timeScheduler)
		return "";
	Vector lifeCycles = timeScheduler.getLifeCycles(((ConvertElement) parent).getIndex());
	if (null == lifeCycles)
		return "";
	LifeCycle lifeCycle = (LifeCycle) lifeCycles.firstElement();
	Vector activeIntervals = lifeCycle.getActiveIntervals();
	if (null == activeIntervals)
		return "";
	ActiveInterval activeInterval = (ActiveInterval) activeIntervals.firstElement();
	Vector elasIds = activeInterval.getELASids();
	if (null == elasIds)
		return "";
	ELASid elasId = (ELASid) elasIds.firstElement();
	xmtaEsDescriptorId_ = getXmtaEsDescriptorId(elasId);
	return xmtaEsDescriptorId_;
}
/*********************************************************************** 
 * Get the id for the Es Descriptor element.
 **********************************************************************/
public String getXmtaEsDescriptorId(ELASid elasId) {
	String xmtaEsDescriptorId = (String) xmtaEsDescriptorIds_.get(elasId);

	// If couldn't find one then create it.
	if (null == xmtaEsDescriptorId) {
		xmtaEsDescriptorId = IdManager.getNextId(IdManager.ESID);
		// make the group ids based on object id or make them unique.
		if (null != getXmtoAttrId() && !getXmtoAttrId().equals(""))
			xmtaEsDescriptorId = xmtaEsDescriptorId + ":" + getXmtoAttrId();
		xmtaEsDescriptorIds_.put(elasId, xmtaEsDescriptorId);
	}

	//
	return xmtaEsDescriptorId;
}
/***********************************************************************
 * Get the defs id of the object.
 * @return the unique id of this object.
 **********************************************************************/
public String getXmtaObjectDescriptorId() {
	if (null != xmtaObjectDescriptorId_)
		return xmtaObjectDescriptorId_;

	// Not a valid defs object
	if (!isInDefsTree()) {
		if (debug_)
			fireErrorEvent(ErrorEvent.TRACE, 0, "XmediaChildrenExternalElement.getXmtaObjectDescriptorId() - is being called on a none defs object", getXmtoElement());

		// This should never be called but just in case return empty string.
		return "";
	}

	// Check whether the defs is a child of the
	// <defs> element or a descendent of a TimingElement.  If it is a child of the
	// defs then it doesn't have a elasid.  If it is a descendent of a timingelement then
	// use the elasId of the timingElement.
	// First check if it is a child of the <defs> element.
	if (isDefsParent()) {
		// If couldn't find one then create it.
		if (null == xmtaObjectDescriptorId_) {
			xmtaObjectDescriptorId_ = IdManager.getNextId(IdManager.ODID);
			// make the group ids based on object id or make them unique.
			String id = getXmtoAttrId();
			if (null != id && !id.equals(""))
				xmtaObjectDescriptorId_ = xmtaObjectDescriptorId_ + ":" + id;
		}
		//
		return xmtaObjectDescriptorId_;
	}

	// See if the element is a descendent of defs but not through a timing element.
	TreeElement parent = getParent();
	while (null != parent && !(parent instanceof TimingElement))
		parent = parent.getParent();

	// This is a defs element under another defs element.  Thus look for its
	// If we couldn't find the parent then this must be a descendent of the defs
	// through other non timing elements, e.g. outline.
	// Treat it the same as a child of the defs.
	if (null == parent) {
		// If couldn't find one then create it.
		if (null == xmtaObjectDescriptorId_) {
			xmtaObjectDescriptorId_ = IdManager.getNextId(IdManager.ODID);
			// make the group ids based on object id or make them unique.
			String id = getXmtoAttrId();
			if (null != id && !id.equals(""))
				xmtaObjectDescriptorId_ = xmtaObjectDescriptorId_ + ":" + id;
		}
		//
		return xmtaObjectDescriptorId_;
	}

	// Now we must be a descendent in the defs tree and a descendent of a timing
	// element.  Thus we need the elasid to distinguish this id.
	TimeScheduler timeScheduler = getXmto2Xmta().getTimeScheduler();
	if (null == timeScheduler)
		return "";
	Vector lifeCycles = timeScheduler.getLifeCycles(((ConvertElement) parent).getIndex());
	if (null == lifeCycles)
		return "";
	LifeCycle lifeCycle = (LifeCycle) lifeCycles.firstElement();
	Vector activeIntervals = lifeCycle.getActiveIntervals();
	if (null == activeIntervals)
		return "";
	ActiveInterval activeInterval = (ActiveInterval) activeIntervals.firstElement();
	Vector elasIds = activeInterval.getELASids();
	if (null == elasIds)
		return "";
	ELASid elasId = (ELASid) elasIds.firstElement();
	xmtaObjectDescriptorId_ = getXmtaObjectDescriptorId(elasId);
	return xmtaObjectDescriptorId_;
}
/*********************************************************************** 
 * Get the id for the Object Descriptor element.
 **********************************************************************/
public String getXmtaObjectDescriptorId(ELASid elasId) {
	String xmtaObjectDescriptorId = (String) xmtaObjectDescriptorIds_.get(elasId);

	// If couldn't find one then create it.
	if (null == xmtaObjectDescriptorId) {
		xmtaObjectDescriptorId = IdManager.getNextId(IdManager.ODID);
		// make the group ids based on object id or make them unique.
		if (null != getXmtoAttrId() && !getXmtoAttrId().equals(""))
			xmtaObjectDescriptorId = xmtaObjectDescriptorId + ":" + getXmtoAttrId();
		xmtaObjectDescriptorIds_.put(elasId, xmtaObjectDescriptorId);
	}

	//
	return xmtaObjectDescriptorId;
}
/*********************************************************************** 
 **********************************************************************/
public String getXmtoAttrSrc() {
	return xmtoAttrSrc_;
}
/*********************************************************************** 
 * Return the canonical path of the file that the src attribute refers to.
 * The file is relative to the document path.
 **********************************************************************/
public String getXmtoAttrSrcCanonical() {
	try {
		return new java.io.File(xmtutil.DOMUtil.getDirectory(getXmto2Xmta().getXmtODocPath()), getXmtoAttrSrc()).getCanonicalPath();
	} catch (Exception e) {
		return getXmtoAttrSrc();
	}
}

/*********************************************************************** 
 * Handle the attribute by checking its syntax.  This is to catch all syntax 
 * errors which the schema can not check for, and thus the DOM parser 
 * doesn't find any problems.  Call the parent handler to see if it can
 * handle the attribute or not.
 *
 * @param  attributeName  node name of attribute.
 * @param  attributeValue node value of attribute.
 * @return whether this attribute is handled or not.
 **********************************************************************/
protected boolean handleValidateSyntax(String attributeName, String attributeValue) {
	boolean success = super.handleValidateSyntax(attributeName, attributeValue);

	// Determine the src
	if (DOMXMTO.src.equals(attributeName)) {
		xmtoAttrSrc_ = attributeValue;
		return true;
	}

	//
	return success;
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
/*********************************************************************** 
 * Validate the syntax of the element.  This is to catch all syntax 
 * errors which the schema can not check for, and thus the DOM parser 
 * doesn't find any problems.
 * Mainly check the begin, dur and end values to see they are valid timing
 * attributes.  If this is a problem then remove this node from the tree.
 **********************************************************************/
public boolean validateSyntax() {
	boolean success = super.validateSyntax();
	if (!success)
		return success;

	// Validate that we have a src attribute
	if (null == xmtoAttrSrc_ || xmtoAttrSrc_.equals("")) {
		MessageFormat msgFormat = new MessageFormat(externalElementRes_.getString(ExternalElementRes.MSG_ERROR_SRC_ATTRIBUTE_MISSING));
		StringBuffer buffer = new StringBuffer();
		buffer.append(getXmtoElement().getNodeName());
		if (null != getXmtoAttrId() && !getXmtoAttrId().equals(""))
			buffer.append(" ").append(DOMXMTO.id).append("=\"").append(getXmtoAttrId()).append("\"");
		Object[] msgArgs = {buffer};
		String message = msgFormat.format(msgArgs);
		fireErrorEvent(ErrorEvent.ERROR, 0, message, getXmtoElement());

		// Since we don't have a src for this element remove it from the tree
		remove();
		return false;
	}

	// 
	return success;
}
}
