/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  AnimationInterpolation.java
//
//  Overview  :
//    The abstract base class maps XMT-O animation into a form that can be
//    plugged into an MPEG-4 Interpolator
//=============================================================================
package xmtutil;

public abstract class AnimationInterpolation
{
  protected static final int kCalcOpUnknown            = 0;
  protected static final int kCalcOpFromTo             = 1;
  protected static final int kCalcOpFromBy             = 2;
  protected static final int kCalcOpTo                 = 3;
  protected static final int kCalcOpBy                 = 4;
  protected static final int kCalcOpValues             = 5;
  protected static final int kCalcOpValuesKeyed        = 6;
  protected static final int kCalcOpValuesKeyedSplined = 7;

  private Object   initialValue;
  private Object   from;
  private Object   to;
  private Object   by;
  private Object[] values;

  private TimingAttribute dur;
  private TimingAttribute repeatDur;
  private Float repeatCount;

  protected AnimationInterpolation(Object initialValue, Object from, Object to, Object by,
                                   Object[] values,
                                   String  calcMode, String accumulate, String additive,
                                   Float[] keyTimes, Float[] keySplines, String path,
                                   TimingAttribute dur, TimingAttribute repeatDur, Float repeatCount)
  {
    if (values == null && from == null && to == null && by == null && path == null)
    {
      throw new java.lang.IllegalArgumentException("All animation attributes null");
    }

    if (values == null && from != null && to == null && by == null && path == null)
    {
      throw new java.lang.IllegalArgumentException("From but no by or to");
    }

    if (values == null && from == null && (to != null || by != null) && path == null && initialValue == null)
    {
      throw new java.lang.IllegalArgumentException("By or To but no initial value");
    }

    this.initialValue = initialValue;
    this.from         = from;
    this.to           = to;
    this.by           = by;
    this.values       = values;
    //
    setCalcMode(calcMode);
    setAccumulate(accumulate);
    setAdditive(additive);
    //
    this.keyTimes   = keyTimes;
    this.keySplines = keySplines;
    this.path       = path;

    if (dur == null)
    {
      throw new java.lang.IllegalArgumentException("Dur must be specified for animation");
    }
    if (dur.getTimeValueType() != TimingAttribute.CLOCK_VALUE)
    {
      throw new java.lang.IllegalArgumentException("Dur must be a clock value for animation");
    }
    if (dur.getClockValue() <= 0)
    {
      throw new java.lang.IllegalArgumentException("Dur must be > 0 for animation");
    }
    this.dur       = dur;

    if (repeatDur != null)
    {
      if (   repeatDur.getTimeValueType() != TimingAttribute.CLOCK_VALUE
          && repeatDur.getTimeValueType() != TimingAttribute.INDEFINITE_VALUE
         )
      {
        throw new java.lang.IllegalArgumentException("repeatDur must be a clock value or indefinite for animation");
      }
      if (   repeatDur.getTimeValueType() == TimingAttribute.CLOCK_VALUE
          && repeatDur.getClockValue()    <= 0
         )
      {
        repeatDur = null; // Ignore it
      }
    }
    this.repeatDur   = repeatDur;

    if (   repeatCount != null
        && repeatCount.floatValue() <= 0.0F
       )
    {
      repeatCount = null; // Ignore it
    }
    this.repeatCount = repeatCount;

    // If we cannot track the value we fall back to Replace rather than additive
    //
    if (getAdditive() == kAdditiveSum && initialValue == null)
    {
      setAdditive("Replace");
    }

    // If we have values we are not the simple From/To/By so ignore if set
    //
    if (values != null)
    {
      this.from = null;
      this.to   = null;
      this.by   = null;
    }

    // To overrides by so forget the by value
    //
    if (to != null)
    {
      this.by = null;
    }

    // Spline mode specified but no splines
    //
    if (getCalcMode() == kCalcModeSpline && keySplines == null)
    {
      setCalcMode("Discrete");
    }

    // This may be From/To, From/By, To or By
    //
    if (this.values == null)
    {
      if (this.from != null)
      {
        if      (this.to != null) calcOp = kCalcOpFromTo;
        else if (this.by != null) calcOp = kCalcOpFromBy;
      }
      else
      {
        if      (this.to != null) calcOp = kCalcOpTo;
        else if (this.by != null) calcOp = kCalcOpBy;
      }
    }
    else
    {
      // TO DO
    }

    calculateCycleIntervalLoop();

    calculateInterpolation();
  }

  private   Float[] keyTimes;
  protected Float[] getKeyTimes() { return keyTimes; }

  private   Float[] keySplines;
  protected Float[] getKeySplines() { return keySplines; }

  private   String path;
  protected String getPath() { return path; }

  private   int  calcOp;
  private   void setCalcOp(int calcOp)  { this.calcOp = calcOp; }
  protected int  getCalcOp()            { return calcOp;
  }

  protected static final int kCalcModeDiscrete = 0;
  protected static final int kCalcModeLinear   = 1;
  protected static final int kCalcModePaced    = 2;
  protected static final int kCalcModeSpline   = 3;

  private String calcMode;
  private int    calcModeK;

  private void setCalcMode(String calcMode)
  {
    if      (calcMode == null)
    {
      calcModeK     = kCalcModeDiscrete;
      this.calcMode = "discrete";
    }
    else if (calcMode.equalsIgnoreCase("discrete")) calcModeK = kCalcModeDiscrete;
    else if (calcMode.equalsIgnoreCase("linear"))   calcModeK = kCalcModeLinear;
    else if (calcMode.equalsIgnoreCase("paced"))    calcModeK = kCalcModePaced;
    else if (calcMode.equalsIgnoreCase("spline"))   calcModeK = kCalcModeSpline;
    else
    {
      throw new java.lang.IllegalArgumentException("Invalid animation calcMode: " + calcMode);
    }
  }

  protected int getCalcMode() { return calcModeK; }


  protected static final int kAccumulateSum  = 0;
  protected static final int kAccumulateNone = 1;

  private String accumulate;
  private int    accumulateK;

  private void setAccumulate(String accumulate)
  {
    if      (accumulate == null)
    {
      accumulateK     = kAccumulateNone;
      this.accumulate = "none";
    }
    else if (accumulate.equalsIgnoreCase("sum"))  accumulateK = kAccumulateSum;
    else if (accumulate.equalsIgnoreCase("none")) accumulateK = kAccumulateNone;
    else
    {
      throw new java.lang.IllegalArgumentException("Invalid accumulate value: " + accumulate);
    }
  }

  protected int getAccumulate() { return accumulateK; }


  protected static final int kAdditiveSum     = 0;
  protected static final int kAdditiveReplace = 1;

  private String additive;
  private int    additiveK;

  private void setAdditive(String additive)
  {
    if      (additive == null)
    {
      additiveK     = kAdditiveReplace;
      this.additive = "replace";
    }
    else if (additive.equalsIgnoreCase("sum"))     additiveK = kAdditiveSum;
    else if (additive.equalsIgnoreCase("replace")) additiveK = kAdditiveReplace;
    else
    {
      throw new java.lang.IllegalArgumentException("Invalid additive value: " + additive);
    }
  }

  protected int getAdditive() { return additiveK; }

  private int loopCount = 1;

  private void setCount(int loopCount)
  {
    if   (loopCount > 0)  this.loopCount = loopCount;
    else
    {
      throw new java.lang.IllegalArgumentException("Invalid loop count : " + loopCount);
    }
  }

  protected int getLoopCount() { return loopCount; }


  protected void calculateCycleIntervalLoop()
  {
    long dur = this.dur.getClockValue();

    if (repeatDur != null)
    {
      if (repeatDur.getTimeValueType() == TimingAttribute.INDEFINITE_VALUE)
      {
        loop = true;
      }
      else if (repeatDur.getClockValue() > dur)
      {
        loop = true;
      }
    }

    if (repeatCount != null && repeatCount.floatValue() > 1.0)
    {
      loop = true;
    }

    cycleInterval = new TimingAttribute(TimingAttribute.CLOCK_VALUE, dur);
  }


  protected void calculateInterpolation()
  {
    calculateAsFromByTo();

    // TODO
  }


  private static final Float zeroF = new Float(0.0f);
  private static final Float oneF  = new Float(1.0f);

  protected boolean calculateAsFromByTo()
  {
    Object  startValue = null;
    Object  endValue   = null;
    boolean done       = false;

    switch (getCalcOp())
    {
      case kCalcOpFromTo:
        if (getCalcMode() == kCalcModeDiscrete)
        {
          startValue = to;
          endValue   = startValue;
        }
        else
        {
          startValue = from;
          endValue   = to;
        }
        done = true;
        break;

      case kCalcOpFromBy:
        if (getCalcMode() == kCalcModeDiscrete)
        {
          startValue = sum(from, by);
          endValue   = startValue;
        }
        else
        {
          startValue = from;
          endValue   = sum(from, by);;
        }
        done = true;
        break;

      case kCalcOpBy:
        startValue = initialValue;
        endValue   = sum(initialValue, by);
        done = true;
        break;

      case kCalcOpTo:
        startValue = initialValue;
        endValue   = to;
        done = true;
        break;
    }

    if (getAdditive() == kAdditiveSum)
    {
      switch (getCalcOp())
      {
        case kCalcOpFromTo:
        case kCalcOpFromBy:
          startValue = sum(initialValue, startValue);
          endValue   = sum(initialValue, endValue);
          break;
      }
    }

    if (done)
    {
      keys      = new Float[2];
      keyValues = alloc(2);

      keys[0] = zeroF;
      keys[1] = oneF;

      keyValues[0] = startValue;
      keyValues[1] = endValue;
    }

    return done;
  }

  private Float[]         keys;
  private Object[]        keyValues;
  private boolean         loop;
  private TimingAttribute cycleInterval;

  public Float[]         getInterpolatorKeys()      { return keys; }
  public Object[]        getInterpolatorKeyValues() { return keyValues; }
  public boolean         getLoop()                  { return loop; }
  public TimingAttribute getCycleInterval()         { return cycleInterval; }

  protected abstract Object[] alloc(int size);

  protected abstract Object sum(Object a, Object b);
}
