/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  DOMXMTAParameter.java
//
//=============================================================================
package xmtutil;

import java.awt.Color;
import java.awt.Font;
import java.awt.Point;

import java.util.Vector;
/*********************************************************************** 
 * DOMParameter converts the parameter strings to java objects and back.
 *
 * <p><b>History:</b><ul>
 * <li>16 Jun 2000 - Creation.
 * <li>19 Jun 2000 - Add some minimal time parameter support.
 * <li>20 Jun 2000 - Fix converting from Color to parameter.
 * <li>22 Jun 2000 - Fix converting Point, TwoFloats and Integer.
 * <li>23 Jun 2000 - Finish converting TwoFloats.
 * <li>15 Aug 2000 - Add Font and Alignment support
 * <li>03 Oct 2000 - Use SFTime (double in seconds) in XMT-A instead of SMIL timing (13ms)
 * <li>30 Jan 2002 - Change SFString and MFString handling to conform to latest spec.
 *                 - XMT-O - SFString doesn't have quotes around strings  i.e. style="BOLD" not style="&quot;BOLD&quot;".
 *                         - MFString have quotes around SFString elements and uses semi-colon delimiter between SFStrings.
 *                         - MFString exception for justify.  No quotes around elements but has the semi-colon delimiter.
 *                         - i.e. justify="BEGIN; BEGIN" not justify="&quot;BEGIN&quot;; &quot;BEGIN&quot;".
 *                 - XMT-A - SFString has have quotes around strings  i.e. style="&quot;BOLD&quot;" not style="BOLD".
 *                         - MFString have quotes around SFString elements and does not use a delimiter (other than whitespace) between SFStrings. 
 * </ul>
 *********************************************************************/
public class DOMXMTAParameter {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;
	static final String separator = "";

/***********************************************************************
 * Converts the parameter into an Integer of Java font name.  If something goes
 * wrong then return serif font.
 **********************************************************************/
static public String alignment2Param(String horizontalAlignment, String verticalAlignment) {
	return "\"" + horizontalAlignment + "\" \"" + verticalAlignment + "\"";
}
/***********************************************************************
 * Converts an array of Float pairs to an XMTA parameter string.
 * Each pair is a Vector of two Float objects.
 * The elements in the output are separated by the separator char.
 ***********************************************************************/
static public String arrayOfTwoFloats2Param(Vector[] twoFloats) {
	return DOMXMTOParameter.arrayOfTwoFloats2Param(twoFloats, separator);
}
/*********************************************************************** 
 * Converts the parameter into a Color.  If something goes
 * wrong then return null.
 **********************************************************************/
static public String color2Param(Color color) {
	return DOMXMTOParameter.color2Param(color);
}
/*********************************************************************** 
 * Converts an array of colors to an XMTA parameter string.
 * The colors in the output are separated by the separator char.
 **********************************************************************/
static public String colors2Param(Color[] color) {
	return DOMXMTOParameter.colors2Param(color, separator);
	
}
/***********************************************************************
 * Converts the parameter into an Integer.  If something goes
 * wrong then return null.
 **********************************************************************/
static public String float2Param(float value) {
	return DOMXMTOParameter.float2Param(value);
}
/***********************************************************************
 * Converts the parameter into an Integer of Java font style.  If something goes
 * wrong then return plain style.
 **********************************************************************/
static public String fontStyle2Param(int fontStyle) {
	StringBuffer sb = new StringBuffer();
	sb.append("\"");
	switch (fontStyle) {
		case Font.BOLD :
			sb.append(DOMXMTA.BOLD);
			break;
		case Font.ITALIC :
			sb.append(DOMXMTA.ITALIC);
			break;
		case (Font.BOLD | Font.ITALIC) :
			sb.append(DOMXMTA.BOLDITALIC);
			break;
		default :
			sb.append(DOMXMTA.PLAIN);
			break;
	}
	sb.append("\"");
	return sb.toString();
}
/*********************************************************************** 
 * Converts an array of Integers to an XMTA parameter string.
 **********************************************************************/
static public String integers2Param(Integer[] integer) {
	return DOMXMTOParameter.integers2Param(integer, separator);
}
/***********************************************************************
 * Converts an array of Float to an XMTO parameter string.
 * The elements in the output are separated by the separator char.
 ***********************************************************************/
static public String mfBool2Param(Boolean[] bools) {
	return DOMXMTOParameter.mfBool2Param(bools, "");
}
/***********************************************************************
 * Converts an array of Float to an XMTO parameter string.
 * The elements in the output are separated by the separator char.
 ***********************************************************************/
static public String mfFloat2Param(Float[] floats) {
	return DOMXMTOParameter.mfFloat2Param(floats, "");
}
/*********************************************************************** 
 * Converts the parameter into a Color.  If something goes
 * wrong then return null.
 **********************************************************************/
static public Color param2Color(String param) {
	return DOMXMTOParameter.param2Color(param);
}
/***********************************************************************
 * Convert a parameter string to an array of Floats (MFFloat).
 * If something goes wrong then return null.
 **********************************************************************/
static public Float[] param2MFFloat(String param) {
	return DOMXMTOParameter.param2MFFloat(param);
}
/***********************************************************************
 **********************************************************************/
static public String point2Param(Point point) {
	return DOMXMTOParameter.point2Param(point);
}
/*********************************************************************** 
 * Converts an array of points to an XMTA parameter string.
 * The points in the output are separated by the separator char.
 **********************************************************************/
static public String points2Param(Point[] point) {
	return DOMXMTOParameter.points2Param(point, separator);
	
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
/***********************************************************************
 * Converts the array of lines of text into a parameter for the textLines
 * attribute of the string element.  The parameter is a MFString.
 **********************************************************************/
static public String textLines2Param(String[] textLines) {
	if (null == textLines || 0 == textLines.length)
		return "";

	// First we have to change all the \ or " characters with \\ or \"
	// In order to do this create an array of string buffers since it is
	// easier to insert escape (\) characters into string buffers than
	// strings.
	StringBuffer[] stringBuffers = new StringBuffer[textLines.length];
	for (int i = 0; i < textLines.length; i++)
		stringBuffers[i] = new StringBuffer(textLines[i]);

	// Now loop through each string buffer and then through each
	// through the characters in the string buffer replacing
	// the \ or " characters with \\ or \"
	// This is to escape out all the MPEG-4 escapes.
	for (int i = 0; i < stringBuffers.length; i++) {
		StringBuffer sb = stringBuffers[i];
		for (int j = 0, count = sb.length(); j < count; j++) {
			char character = sb.charAt(j);
			if ('\\' == character || '\"' == character) {
				sb.insert(j, '\\');
				j++;
				count++;
			}
		}
	}

	// Now create the parameter.  This is a concatenation
	// of all the text lines surrounded by quote (") characters.
	StringBuffer param = new StringBuffer();
	for (int i = 0; i < stringBuffers.length - 1; i++) {
		param.append("\"").append(stringBuffers[i]).append("\" ");
	}
	// append the last line.  Done so there is no "; " at the end.
	param.append("\"").append(stringBuffers[stringBuffers.length - 1]).append("\"");

	//
	return param.toString();
}
/*********************************************************************** 
 **********************************************************************/
static public String time2Param(long time) {
	double t = time;
	t /= 1000; // convert from milliseconds to seconds
	StringBuffer param = new StringBuffer();
	param.append(t);
	return param.toString();
}
/***********************************************************************
 * Converts the parameter into an Integer.  If something goes
 * wrong then return null.
 **********************************************************************/
static public String twoFloats2Param(float first, float second) {
	return DOMXMTOParameter.twoFloats2Param(first, second);
}
}
