/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  DOMXMTOParameter.java
//
//=============================================================================
package xmtutil;

import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.awt.Toolkit;
import java.text.DecimalFormat;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.StringTokenizer;
/*********************************************************************** 
 * DOMParameter converts the parameter strings to java objects and back.
 *
 * <p><b>History:</b><ul>
 * <li>16 Jun 2000 - Creation.
 * <li>19 Jun 2000 - Add some minimal time parameter support.
 * <li>20 Jun 2000 - Fix converting from Color to parameter.
 * <li>22 Jun 2000 - Fix converting Point, TwoFloats and Integer.
 * <li>23 Jun 2000 - Finish converting TwoFloats.
 * <li>15 Aug 2000 - Add Font and Alignment support
 * <li>05 Sep 2000 - Fix decimal formatting to eliminate 1000's separator (grouping)
 *                 - Fix sanserif font conversion problem
 * <li>25 Sep 2000 - Fix font name conversion
 * <li>23 Oct 2000 - Add support for FIRST for text alignment
 *                 - Fix font size problem of difference in sizes between MP4 and Java
 *                 - Add better support for MFString delimiter in Alignment.
 * <li>30 Oct 2000 - Change XMT-O to have text placed in the attribute textLines instead of
 *                   text nodes.
 * <li>23 Jan 2001 - Fix default time to be seconds instead of milliseconds (if no time indicator).
 * <li>02 Feb 2001 - Fix font name and style so they are correct for SFStrings (enclosed in quotes).
 * <li>05 Jun 2001 - Add color value in format "0xFFFFFF".
 * <li>08 Jun 2001 - Changed color string support to support the standard (16) color string as defined
 *                   in HTML 4.0.  See http://www.w3.org/TR/html4/types.html#type-color
 *                 - Also make color string case insensitive.
 * <li>27 Jun 2001 - Add support for MFFloat (array of floats).  Start calling the fields by their
 *                   correct name.
 * </ul>
 *********************************************************************/
public class DOMXMTOParameter {
	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;
	static private DecimalFormat decimalFormat_ = null;
	static final private String JAVA_SERIF = "serif";
	static final private String JAVA_SANSSERIF = "sansserif";
	static final private String JAVA_MONOSPACED = "monospaced";
	static private boolean initPixelsPerPt_ = false;
	static private float pixelsPerPt_ = 1;
	static private boolean initColorStrings_ = false;
	static private Hashtable colorStrings_ = null;
	static private final String separator = ";";
	static final private String outlineStyles_[] = {
	/* 0 */ "solid",
	/* 1 */ "dash",
	/* 2 */ "dot",
	/* 3 */ "dash-dot",
	/* 4 */ "dash-dash-dot",
	/* 5 */ "dash-dot-dot",
	/*   */};
/***********************************************************************
 * Converts the parameter into an Integer of Java font name.  If something goes
 * wrong then return serif font.
 **********************************************************************/
static public String alignment2Param(String horizontalAlignment, String verticalAlignment) {
	return horizontalAlignment + "; " + verticalAlignment;
}
/***********************************************************************
 * Converts an array of Float pairs to an XMTO parameter string.
 * Each pair is a Vector of two Float objects.
 * The elements in the output are separated by the separator char.
 * The parameter string is of the form "x y z, x y z, ..."
 ***********************************************************************/
static public String arrayOfTwoFloats2Param(Vector[] twoFloats) {
	return arrayOfTwoFloats2Param(twoFloats, separator);
}
/***********************************************************************
 ***********************************************************************/
static public String arrayOfTwoFloats2Param(Vector[] twoFloats, String separator) {
	if (twoFloats == null) {
		return "";
	}
	StringBuffer xyzBuffer = new StringBuffer();
	for (int i = 0; i < twoFloats.length; i++) {
		float first = ((Float) twoFloats[i].elementAt(0)).floatValue();
		float second = ((Float) twoFloats[i].elementAt(1)).floatValue();
		String xyz = twoFloats2Param(first, second);
		if (xyz.length() > 0) {
			if (i > 0) {
				xyzBuffer.append(separator);
				xyzBuffer.append(" ");
			}
			xyzBuffer.append(xyz);
		}
	}
	return xyzBuffer.toString();
}
/*********************************************************************** 
 **********************************************************************/
static public String color2Param(Color color) {
	if (null == color)
		return "";
	float[] colors = color.getRGBColorComponents(null);
	if (colors.length < 3)
		return "";

	//
	StringBuffer rgb = new StringBuffer();
	rgb.append(colors[0]);
	rgb.append(" ");
	rgb.append(colors[1]);
	rgb.append(" ");
	rgb.append(colors[2]);
	//	rgb.append(float2Param(colors[0]));
	//	rgb.append(" ");
	//	rgb.append(float2Param(colors[1]));
	//	rgb.append(" ");
	//	rgb.append(float2Param(colors[2]));

	//
	return rgb.toString();
}
/***********************************************************************
 * Converts an array of Colors to an XMTO parameter string.
 * The elements in the output are separated by the separator char.
 * The parameter string is of the form "r g b, r g b, ..."
 ***********************************************************************/
static public String colors2Param(Color[] color) {
	return colors2Param(color, separator);
}
/***********************************************************************
 ***********************************************************************/
static public String colors2Param(Color[] color, String separator) {
	if (color == null) {
		return "";
	}
	StringBuffer rgbBuffer = new StringBuffer();
	for (int i = 0; i < color.length; i++) {
		String rgb = color2Param(color[i]);
		if (rgb.length() > 0) {
			if (i > 0) {
				rgbBuffer.append(separator);
				rgbBuffer.append(" ");
			}
			rgbBuffer.append(rgb);
		}
	}
	return rgbBuffer.toString();
}
/***********************************************************************
 * Converts the parameter into an Integer.  If something goes
 * wrong then return null.
 **********************************************************************/
static public String double2Param(double value) {
	if (null == decimalFormat_)
		initDecimalFormat();
	return decimalFormat_.format(value);
}
/***********************************************************************
 * Converts the parameter into an Integer.  If something goes
 * wrong then return null.
 **********************************************************************/
static public String float2Param(float value) {
	if (null == decimalFormat_)
		initDecimalFormat();
	//	return decimalFormat_.format(value);
	return Float.toString(value);
}
/***********************************************************************
 * Converts the parameter into an Integer of Java font name.  If something goes
 * wrong then return serif font.
 **********************************************************************/
static public String font2ParamFontName(Font font) {
	String fontName = font.getName();
	if (fontName.equals(JAVA_MONOSPACED))
		return "\"" + DOMXMTO.TYPEWRITER + "\"";
	if (fontName.equals(JAVA_SANSSERIF))
		return "\"" + DOMXMTO.SANS + "\"";
	return "\"" + DOMXMTO.SERIF + "\"";
}
/***********************************************************************
 * Converts the parameter into an Integer of Java font style.  If something goes
 * wrong then return plain style.
 **********************************************************************/
static public String font2ParamFontSize(Font font) {
	if (!initPixelsPerPt_)
		initPixelsPerPt();
	float javaFontSize = font.getSize();

	// size below is size of font as a float
	float mp4Size = javaFontSize * pixelsPerPt_;

	// Maintain a minimum pt size - 0 just does not cut it!
	if (0 >= mp4Size)
		mp4Size = 1;
	return float2Param(mp4Size);
}
/***********************************************************************
 * Converts the parameter into an Integer of Java font style.  If something goes
 * wrong then return plain style.
 **********************************************************************/
static public String font2ParamFontStyle(Font font) {
	int fontStyle = font.getStyle();
	if (Font.BOLD == fontStyle)
		return DOMXMTO.BOLD;
	if (Font.ITALIC == fontStyle)
		return DOMXMTO.ITALIC;
	if ((Font.BOLD | Font.ITALIC) == fontStyle)
		return DOMXMTO.BOLDITALIC;
	return DOMXMTO.PLAIN;
}
/***********************************************************************
 * Initialize the stings for color parameters.
 **********************************************************************/
static private void initColorStrings() {
	if (!initColorStrings_) {
		initColorStrings_ = true;
		colorStrings_ = new Hashtable(1439); // A prime number

		// From XConsortium: rgb.txt,v 10.41 94/02/20
		//
		colorStrings_.put("aliceblue", new Color(240, 248, 255));
		colorStrings_.put("aliceblue", new Color(240, 248, 255));
		colorStrings_.put("antiquewhite", new Color(250, 235, 215));
		colorStrings_.put("antiquewhite", new Color(250, 235, 215));
		colorStrings_.put("antiquewhite1", new Color(255, 239, 219));
		colorStrings_.put("antiquewhite2", new Color(238, 223, 204));
		colorStrings_.put("antiquewhite3", new Color(205, 192, 176));
		colorStrings_.put("antiquewhite4", new Color(139, 131, 120));
		colorStrings_.put("aquamarine", new Color(127, 255, 212));
		colorStrings_.put("aquamarine1", new Color(127, 255, 212));
		colorStrings_.put("aquamarine2", new Color(118, 238, 198));
		colorStrings_.put("aquamarine3", new Color(102, 205, 170));
		colorStrings_.put("aquamarine4", new Color(69, 139, 116));
		colorStrings_.put("azure", new Color(240, 255, 255));
		colorStrings_.put("azure1", new Color(240, 255, 255));
		colorStrings_.put("azure2", new Color(224, 238, 238));
		colorStrings_.put("azure3", new Color(193, 205, 205));
		colorStrings_.put("azure4", new Color(131, 139, 139));
		colorStrings_.put("beige", new Color(245, 245, 220));
		colorStrings_.put("bisque", new Color(255, 228, 196));
		colorStrings_.put("bisque1", new Color(255, 228, 196));
		colorStrings_.put("bisque2", new Color(238, 213, 183));
		colorStrings_.put("bisque3", new Color(205, 183, 158));
		colorStrings_.put("bisque4", new Color(139, 125, 107));
		colorStrings_.put("black", new Color(0, 0, 0));
		colorStrings_.put("blanchedalmond", new Color(255, 235, 205));
		colorStrings_.put("blanchedalmond", new Color(255, 235, 205));
		colorStrings_.put("blue", new Color(0, 0, 255));
		colorStrings_.put("blue1", new Color(0, 0, 255));
		colorStrings_.put("blue2", new Color(0, 0, 238));
		colorStrings_.put("blue3", new Color(0, 0, 205));
		colorStrings_.put("blue4", new Color(0, 0, 139));
		colorStrings_.put("blueviolet", new Color(138, 43, 226));
		colorStrings_.put("blueviolet", new Color(138, 43, 226));
		colorStrings_.put("brown", new Color(165, 42, 42));
		colorStrings_.put("brown1", new Color(255, 64, 64));
		colorStrings_.put("brown2", new Color(238, 59, 59));
		colorStrings_.put("brown3", new Color(205, 51, 51));
		colorStrings_.put("brown4", new Color(139, 35, 35));
		colorStrings_.put("burlywood", new Color(222, 184, 135));
		colorStrings_.put("burlywood1", new Color(255, 211, 155));
		colorStrings_.put("burlywood2", new Color(238, 197, 145));
		colorStrings_.put("burlywood3", new Color(205, 170, 125));
		colorStrings_.put("burlywood4", new Color(139, 115, 85));
		colorStrings_.put("cadetblue", new Color(95, 158, 160));
		colorStrings_.put("cadetblue", new Color(95, 158, 160));
		colorStrings_.put("cadetblue1", new Color(152, 245, 255));
		colorStrings_.put("cadetblue2", new Color(142, 229, 238));
		colorStrings_.put("cadetblue3", new Color(122, 197, 205));
		colorStrings_.put("cadetblue4", new Color(83, 134, 139));
		colorStrings_.put("chartreuse", new Color(127, 255, 0));
		colorStrings_.put("chartreuse1", new Color(127, 255, 0));
		colorStrings_.put("chartreuse2", new Color(118, 238, 0));
		colorStrings_.put("chartreuse3", new Color(102, 205, 0));
		colorStrings_.put("chartreuse4", new Color(69, 139, 0));
		colorStrings_.put("chocolate", new Color(210, 105, 30));
		colorStrings_.put("chocolate1", new Color(255, 127, 36));
		colorStrings_.put("chocolate2", new Color(238, 118, 33));
		colorStrings_.put("chocolate3", new Color(205, 102, 29));
		colorStrings_.put("chocolate4", new Color(139, 69, 19));
		colorStrings_.put("coral", new Color(255, 127, 80));
		colorStrings_.put("coral1", new Color(255, 114, 86));
		colorStrings_.put("coral2", new Color(238, 106, 80));
		colorStrings_.put("coral3", new Color(205, 91, 69));
		colorStrings_.put("coral4", new Color(139, 62, 47));
		colorStrings_.put("cornflowerblue", new Color(100, 149, 237));
		colorStrings_.put("cornflowerblue", new Color(100, 149, 237));
		colorStrings_.put("cornsilk", new Color(255, 248, 220));
		colorStrings_.put("cornsilk1", new Color(255, 248, 220));
		colorStrings_.put("cornsilk2", new Color(238, 232, 205));
		colorStrings_.put("cornsilk3", new Color(205, 200, 177));
		colorStrings_.put("cornsilk4", new Color(139, 136, 120));
		colorStrings_.put("cyan", new Color(0, 255, 255));
		colorStrings_.put("cyan1", new Color(0, 255, 255));
		colorStrings_.put("cyan2", new Color(0, 238, 238));
		colorStrings_.put("cyan3", new Color(0, 205, 205));
		colorStrings_.put("cyan4", new Color(0, 139, 139));
		colorStrings_.put("darkblue", new Color(0, 0, 139));
		colorStrings_.put("darkblue", new Color(0, 0, 139));
		colorStrings_.put("darkcyan", new Color(0, 139, 139));
		colorStrings_.put("darkcyan", new Color(0, 139, 139));
		colorStrings_.put("darkgoldenrod", new Color(184, 134, 11));
		colorStrings_.put("darkgoldenrod", new Color(184, 134, 11));
		colorStrings_.put("darkgoldenrod1", new Color(255, 185, 15));
		colorStrings_.put("darkgoldenrod2", new Color(238, 173, 14));
		colorStrings_.put("darkgoldenrod3", new Color(205, 149, 12));
		colorStrings_.put("darkgoldenrod4", new Color(139, 101, 8));
		colorStrings_.put("darkgray", new Color(169, 169, 169));
		colorStrings_.put("darkgray", new Color(169, 169, 169));
		colorStrings_.put("darkgreen", new Color(0, 100, 0));
		colorStrings_.put("darkgreen", new Color(0, 100, 0));
		colorStrings_.put("darkgrey", new Color(169, 169, 169));
		colorStrings_.put("darkgrey", new Color(169, 169, 169));
		colorStrings_.put("darkkhaki", new Color(189, 183, 107));
		colorStrings_.put("darkkhaki", new Color(189, 183, 107));
		colorStrings_.put("darkmagenta", new Color(139, 0, 139));
		colorStrings_.put("darkmagenta", new Color(139, 0, 139));
		colorStrings_.put("darkolivegreen", new Color(85, 107, 47));
		colorStrings_.put("darkolivegreen", new Color(85, 107, 47));
		colorStrings_.put("darkolivegreen1", new Color(202, 255, 112));
		colorStrings_.put("darkolivegreen2", new Color(188, 238, 104));
		colorStrings_.put("darkolivegreen3", new Color(162, 205, 90));
		colorStrings_.put("darkolivegreen4", new Color(110, 139, 61));
		colorStrings_.put("darkorange", new Color(255, 140, 0));
		colorStrings_.put("darkorange", new Color(255, 140, 0));
		colorStrings_.put("darkorange1", new Color(255, 127, 0));
		colorStrings_.put("darkorange2", new Color(238, 118, 0));
		colorStrings_.put("darkorange3", new Color(205, 102, 0));
		colorStrings_.put("darkorange4", new Color(139, 69, 0));
		colorStrings_.put("darkorchid", new Color(153, 50, 204));
		colorStrings_.put("darkorchid", new Color(153, 50, 204));
		colorStrings_.put("darkorchid1", new Color(191, 62, 255));
		colorStrings_.put("darkorchid2", new Color(178, 58, 238));
		colorStrings_.put("darkorchid3", new Color(154, 50, 205));
		colorStrings_.put("darkorchid4", new Color(104, 34, 139));
		colorStrings_.put("darkred", new Color(139, 0, 0));
		colorStrings_.put("darkred", new Color(139, 0, 0));
		colorStrings_.put("darksalmon", new Color(233, 150, 122));
		colorStrings_.put("darksalmon", new Color(233, 150, 122));
		colorStrings_.put("darkseagreen", new Color(143, 188, 143));
		colorStrings_.put("darkseagreen", new Color(143, 188, 143));
		colorStrings_.put("darkseagreen1", new Color(193, 255, 193));
		colorStrings_.put("darkseagreen2", new Color(180, 238, 180));
		colorStrings_.put("darkseagreen3", new Color(155, 205, 155));
		colorStrings_.put("darkseagreen4", new Color(105, 139, 105));
		colorStrings_.put("darkslateblue", new Color(72, 61, 139));
		colorStrings_.put("darkslateblue", new Color(72, 61, 139));
		colorStrings_.put("darkslategray", new Color(47, 79, 79));
		colorStrings_.put("darkslategray", new Color(47, 79, 79));
		colorStrings_.put("darkslategray1", new Color(151, 255, 255));
		colorStrings_.put("darkslategray2", new Color(141, 238, 238));
		colorStrings_.put("darkslategray3", new Color(121, 205, 205));
		colorStrings_.put("darkslategray4", new Color(82, 139, 139));
		colorStrings_.put("darkslategrey", new Color(47, 79, 79));
		colorStrings_.put("darkslategrey", new Color(47, 79, 79));
		colorStrings_.put("darkturquoise", new Color(0, 206, 209));
		colorStrings_.put("darkturquoise", new Color(0, 206, 209));
		colorStrings_.put("darkviolet", new Color(148, 0, 211));
		colorStrings_.put("darkviolet", new Color(148, 0, 211));
		colorStrings_.put("deeppink", new Color(255, 20, 147));
		colorStrings_.put("deeppink", new Color(255, 20, 147));
		colorStrings_.put("deeppink1", new Color(255, 20, 147));
		colorStrings_.put("deeppink2", new Color(238, 18, 137));
		colorStrings_.put("deeppink3", new Color(205, 16, 118));
		colorStrings_.put("deeppink4", new Color(139, 10, 80));
		colorStrings_.put("deepskyblue", new Color(0, 191, 255));
		colorStrings_.put("deepskyblue", new Color(0, 191, 255));
		colorStrings_.put("deepskyblue1", new Color(0, 191, 255));
		colorStrings_.put("deepskyblue2", new Color(0, 178, 238));
		colorStrings_.put("deepskyblue3", new Color(0, 154, 205));
		colorStrings_.put("deepskyblue4", new Color(0, 104, 139));
		colorStrings_.put("dimgray", new Color(105, 105, 105));
		colorStrings_.put("dimgray", new Color(105, 105, 105));
		colorStrings_.put("dimgrey", new Color(105, 105, 105));
		colorStrings_.put("dimgrey", new Color(105, 105, 105));
		colorStrings_.put("dodgerblue", new Color(30, 144, 255));
		colorStrings_.put("dodgerblue", new Color(30, 144, 255));
		colorStrings_.put("dodgerblue1", new Color(30, 144, 255));
		colorStrings_.put("dodgerblue2", new Color(28, 134, 238));
		colorStrings_.put("dodgerblue3", new Color(24, 116, 205));
		colorStrings_.put("dodgerblue4", new Color(16, 78, 139));
		colorStrings_.put("firebrick", new Color(178, 34, 34));
		colorStrings_.put("firebrick1", new Color(255, 48, 48));
		colorStrings_.put("firebrick2", new Color(238, 44, 44));
		colorStrings_.put("firebrick3", new Color(205, 38, 38));
		colorStrings_.put("firebrick4", new Color(139, 26, 26));
		colorStrings_.put("floralwhite", new Color(255, 250, 240));
		colorStrings_.put("floralwhite", new Color(255, 250, 240));
		colorStrings_.put("forestgreen", new Color(34, 139, 34));
		colorStrings_.put("forestgreen", new Color(34, 139, 34));
		colorStrings_.put("gainsboro", new Color(220, 220, 220));
		colorStrings_.put("ghostwhite", new Color(248, 248, 255));
		colorStrings_.put("ghostwhite", new Color(248, 248, 255));
		colorStrings_.put("gold", new Color(255, 215, 0));
		colorStrings_.put("gold1", new Color(255, 215, 0));
		colorStrings_.put("gold2", new Color(238, 201, 0));
		colorStrings_.put("gold3", new Color(205, 173, 0));
		colorStrings_.put("gold4", new Color(139, 117, 0));
		colorStrings_.put("goldenrod", new Color(218, 165, 32));
		colorStrings_.put("goldenrod1", new Color(255, 193, 37));
		colorStrings_.put("goldenrod2", new Color(238, 180, 34));
		colorStrings_.put("goldenrod3", new Color(205, 155, 29));
		colorStrings_.put("goldenrod4", new Color(139, 105, 20));
		colorStrings_.put("gray", new Color(190, 190, 190));
		colorStrings_.put("gray0", new Color(0, 0, 0));
		colorStrings_.put("gray1", new Color(3, 3, 3));
		colorStrings_.put("gray10", new Color(26, 26, 26));
		colorStrings_.put("gray100", new Color(255, 255, 255));
		colorStrings_.put("gray11", new Color(28, 28, 28));
		colorStrings_.put("gray12", new Color(31, 31, 31));
		colorStrings_.put("gray13", new Color(33, 33, 33));
		colorStrings_.put("gray14", new Color(36, 36, 36));
		colorStrings_.put("gray15", new Color(38, 38, 38));
		colorStrings_.put("gray16", new Color(41, 41, 41));
		colorStrings_.put("gray17", new Color(43, 43, 43));
		colorStrings_.put("gray18", new Color(46, 46, 46));
		colorStrings_.put("gray19", new Color(48, 48, 48));
		colorStrings_.put("gray2", new Color(5, 5, 5));
		colorStrings_.put("gray20", new Color(51, 51, 51));
		colorStrings_.put("gray21", new Color(54, 54, 54));
		colorStrings_.put("gray22", new Color(56, 56, 56));
		colorStrings_.put("gray23", new Color(59, 59, 59));
		colorStrings_.put("gray24", new Color(61, 61, 61));
		colorStrings_.put("gray25", new Color(64, 64, 64));
		colorStrings_.put("gray26", new Color(66, 66, 66));
		colorStrings_.put("gray27", new Color(69, 69, 69));
		colorStrings_.put("gray28", new Color(71, 71, 71));
		colorStrings_.put("gray29", new Color(74, 74, 74));
		colorStrings_.put("gray3", new Color(8, 8, 8));
		colorStrings_.put("gray30", new Color(77, 77, 77));
		colorStrings_.put("gray31", new Color(79, 79, 79));
		colorStrings_.put("gray32", new Color(82, 82, 82));
		colorStrings_.put("gray33", new Color(84, 84, 84));
		colorStrings_.put("gray34", new Color(87, 87, 87));
		colorStrings_.put("gray35", new Color(89, 89, 89));
		colorStrings_.put("gray36", new Color(92, 92, 92));
		colorStrings_.put("gray37", new Color(94, 94, 94));
		colorStrings_.put("gray38", new Color(97, 97, 97));
		colorStrings_.put("gray39", new Color(99, 99, 99));
		colorStrings_.put("gray4", new Color(10, 10, 10));
		colorStrings_.put("gray40", new Color(102, 102, 102));
		colorStrings_.put("gray41", new Color(105, 105, 105));
		colorStrings_.put("gray42", new Color(107, 107, 107));
		colorStrings_.put("gray43", new Color(110, 110, 110));
		colorStrings_.put("gray44", new Color(112, 112, 112));
		colorStrings_.put("gray45", new Color(115, 115, 115));
		colorStrings_.put("gray46", new Color(117, 117, 117));
		colorStrings_.put("gray47", new Color(120, 120, 120));
		colorStrings_.put("gray48", new Color(122, 122, 122));
		colorStrings_.put("gray49", new Color(125, 125, 125));
		colorStrings_.put("gray5", new Color(13, 13, 13));
		colorStrings_.put("gray50", new Color(127, 127, 127));
		colorStrings_.put("gray51", new Color(130, 130, 130));
		colorStrings_.put("gray52", new Color(133, 133, 133));
		colorStrings_.put("gray53", new Color(135, 135, 135));
		colorStrings_.put("gray54", new Color(138, 138, 138));
		colorStrings_.put("gray55", new Color(140, 140, 140));
		colorStrings_.put("gray56", new Color(143, 143, 143));
		colorStrings_.put("gray57", new Color(145, 145, 145));
		colorStrings_.put("gray58", new Color(148, 148, 148));
		colorStrings_.put("gray59", new Color(150, 150, 150));
		colorStrings_.put("gray6", new Color(15, 15, 15));
		colorStrings_.put("gray60", new Color(153, 153, 153));
		colorStrings_.put("gray61", new Color(156, 156, 156));
		colorStrings_.put("gray62", new Color(158, 158, 158));
		colorStrings_.put("gray63", new Color(161, 161, 161));
		colorStrings_.put("gray64", new Color(163, 163, 163));
		colorStrings_.put("gray65", new Color(166, 166, 166));
		colorStrings_.put("gray66", new Color(168, 168, 168));
		colorStrings_.put("gray67", new Color(171, 171, 171));
		colorStrings_.put("gray68", new Color(173, 173, 173));
		colorStrings_.put("gray69", new Color(176, 176, 176));
		colorStrings_.put("gray7", new Color(18, 18, 18));
		colorStrings_.put("gray70", new Color(179, 179, 179));
		colorStrings_.put("gray71", new Color(181, 181, 181));
		colorStrings_.put("gray72", new Color(184, 184, 184));
		colorStrings_.put("gray73", new Color(186, 186, 186));
		colorStrings_.put("gray74", new Color(189, 189, 189));
		colorStrings_.put("gray75", new Color(191, 191, 191));
		colorStrings_.put("gray76", new Color(194, 194, 194));
		colorStrings_.put("gray77", new Color(196, 196, 196));
		colorStrings_.put("gray78", new Color(199, 199, 199));
		colorStrings_.put("gray79", new Color(201, 201, 201));
		colorStrings_.put("gray8", new Color(20, 20, 20));
		colorStrings_.put("gray80", new Color(204, 204, 204));
		colorStrings_.put("gray81", new Color(207, 207, 207));
		colorStrings_.put("gray82", new Color(209, 209, 209));
		colorStrings_.put("gray83", new Color(212, 212, 212));
		colorStrings_.put("gray84", new Color(214, 214, 214));
		colorStrings_.put("gray85", new Color(217, 217, 217));
		colorStrings_.put("gray86", new Color(219, 219, 219));
		colorStrings_.put("gray87", new Color(222, 222, 222));
		colorStrings_.put("gray88", new Color(224, 224, 224));
		colorStrings_.put("gray89", new Color(227, 227, 227));
		colorStrings_.put("gray9", new Color(23, 23, 23));
		colorStrings_.put("gray90", new Color(229, 229, 229));
		colorStrings_.put("gray91", new Color(232, 232, 232));
		colorStrings_.put("gray92", new Color(235, 235, 235));
		colorStrings_.put("gray93", new Color(237, 237, 237));
		colorStrings_.put("gray94", new Color(240, 240, 240));
		colorStrings_.put("gray95", new Color(242, 242, 242));
		colorStrings_.put("gray96", new Color(245, 245, 245));
		colorStrings_.put("gray97", new Color(247, 247, 247));
		colorStrings_.put("gray98", new Color(250, 250, 250));
		colorStrings_.put("gray99", new Color(252, 252, 252));
		colorStrings_.put("green", new Color(0, 255, 0));
		colorStrings_.put("green1", new Color(0, 255, 0));
		colorStrings_.put("green2", new Color(0, 238, 0));
		colorStrings_.put("green3", new Color(0, 205, 0));
		colorStrings_.put("green4", new Color(0, 139, 0));
		colorStrings_.put("greenyellow", new Color(173, 255, 47));
		colorStrings_.put("greenyellow", new Color(173, 255, 47));
		colorStrings_.put("grey", new Color(190, 190, 190));
		colorStrings_.put("grey0", new Color(0, 0, 0));
		colorStrings_.put("grey1", new Color(3, 3, 3));
		colorStrings_.put("grey10", new Color(26, 26, 26));
		colorStrings_.put("grey100", new Color(255, 255, 255));
		colorStrings_.put("grey11", new Color(28, 28, 28));
		colorStrings_.put("grey12", new Color(31, 31, 31));
		colorStrings_.put("grey13", new Color(33, 33, 33));
		colorStrings_.put("grey14", new Color(36, 36, 36));
		colorStrings_.put("grey15", new Color(38, 38, 38));
		colorStrings_.put("grey16", new Color(41, 41, 41));
		colorStrings_.put("grey17", new Color(43, 43, 43));
		colorStrings_.put("grey18", new Color(46, 46, 46));
		colorStrings_.put("grey19", new Color(48, 48, 48));
		colorStrings_.put("grey2", new Color(5, 5, 5));
		colorStrings_.put("grey20", new Color(51, 51, 51));
		colorStrings_.put("grey21", new Color(54, 54, 54));
		colorStrings_.put("grey22", new Color(56, 56, 56));
		colorStrings_.put("grey23", new Color(59, 59, 59));
		colorStrings_.put("grey24", new Color(61, 61, 61));
		colorStrings_.put("grey25", new Color(64, 64, 64));
		colorStrings_.put("grey26", new Color(66, 66, 66));
		colorStrings_.put("grey27", new Color(69, 69, 69));
		colorStrings_.put("grey28", new Color(71, 71, 71));
		colorStrings_.put("grey29", new Color(74, 74, 74));
		colorStrings_.put("grey3", new Color(8, 8, 8));
		colorStrings_.put("grey30", new Color(77, 77, 77));
		colorStrings_.put("grey31", new Color(79, 79, 79));
		colorStrings_.put("grey32", new Color(82, 82, 82));
		colorStrings_.put("grey33", new Color(84, 84, 84));
		colorStrings_.put("grey34", new Color(87, 87, 87));
		colorStrings_.put("grey35", new Color(89, 89, 89));
		colorStrings_.put("grey36", new Color(92, 92, 92));
		colorStrings_.put("grey37", new Color(94, 94, 94));
		colorStrings_.put("grey38", new Color(97, 97, 97));
		colorStrings_.put("grey39", new Color(99, 99, 99));
		colorStrings_.put("grey4", new Color(10, 10, 10));
		colorStrings_.put("grey40", new Color(102, 102, 102));
		colorStrings_.put("grey41", new Color(105, 105, 105));
		colorStrings_.put("grey42", new Color(107, 107, 107));
		colorStrings_.put("grey43", new Color(110, 110, 110));
		colorStrings_.put("grey44", new Color(112, 112, 112));
		colorStrings_.put("grey45", new Color(115, 115, 115));
		colorStrings_.put("grey46", new Color(117, 117, 117));
		colorStrings_.put("grey47", new Color(120, 120, 120));
		colorStrings_.put("grey48", new Color(122, 122, 122));
		colorStrings_.put("grey49", new Color(125, 125, 125));
		colorStrings_.put("grey5", new Color(13, 13, 13));
		colorStrings_.put("grey50", new Color(127, 127, 127));
		colorStrings_.put("grey51", new Color(130, 130, 130));
		colorStrings_.put("grey52", new Color(133, 133, 133));
		colorStrings_.put("grey53", new Color(135, 135, 135));
		colorStrings_.put("grey54", new Color(138, 138, 138));
		colorStrings_.put("grey55", new Color(140, 140, 140));
		colorStrings_.put("grey56", new Color(143, 143, 143));
		colorStrings_.put("grey57", new Color(145, 145, 145));
		colorStrings_.put("grey58", new Color(148, 148, 148));
		colorStrings_.put("grey59", new Color(150, 150, 150));
		colorStrings_.put("grey6", new Color(15, 15, 15));
		colorStrings_.put("grey60", new Color(153, 153, 153));
		colorStrings_.put("grey61", new Color(156, 156, 156));
		colorStrings_.put("grey62", new Color(158, 158, 158));
		colorStrings_.put("grey63", new Color(161, 161, 161));
		colorStrings_.put("grey64", new Color(163, 163, 163));
		colorStrings_.put("grey65", new Color(166, 166, 166));
		colorStrings_.put("grey66", new Color(168, 168, 168));
		colorStrings_.put("grey67", new Color(171, 171, 171));
		colorStrings_.put("grey68", new Color(173, 173, 173));
		colorStrings_.put("grey69", new Color(176, 176, 176));
		colorStrings_.put("grey7", new Color(18, 18, 18));
		colorStrings_.put("grey70", new Color(179, 179, 179));
		colorStrings_.put("grey71", new Color(181, 181, 181));
		colorStrings_.put("grey72", new Color(184, 184, 184));
		colorStrings_.put("grey73", new Color(186, 186, 186));
		colorStrings_.put("grey74", new Color(189, 189, 189));
		colorStrings_.put("grey75", new Color(191, 191, 191));
		colorStrings_.put("grey76", new Color(194, 194, 194));
		colorStrings_.put("grey77", new Color(196, 196, 196));
		colorStrings_.put("grey78", new Color(199, 199, 199));
		colorStrings_.put("grey79", new Color(201, 201, 201));
		colorStrings_.put("grey8", new Color(20, 20, 20));
		colorStrings_.put("grey80", new Color(204, 204, 204));
		colorStrings_.put("grey81", new Color(207, 207, 207));
		colorStrings_.put("grey82", new Color(209, 209, 209));
		colorStrings_.put("grey83", new Color(212, 212, 212));
		colorStrings_.put("grey84", new Color(214, 214, 214));
		colorStrings_.put("grey85", new Color(217, 217, 217));
		colorStrings_.put("grey86", new Color(219, 219, 219));
		colorStrings_.put("grey87", new Color(222, 222, 222));
		colorStrings_.put("grey88", new Color(224, 224, 224));
		colorStrings_.put("grey89", new Color(227, 227, 227));
		colorStrings_.put("grey9", new Color(23, 23, 23));
		colorStrings_.put("grey90", new Color(229, 229, 229));
		colorStrings_.put("grey91", new Color(232, 232, 232));
		colorStrings_.put("grey92", new Color(235, 235, 235));
		colorStrings_.put("grey93", new Color(237, 237, 237));
		colorStrings_.put("grey94", new Color(240, 240, 240));
		colorStrings_.put("grey95", new Color(242, 242, 242));
		colorStrings_.put("grey96", new Color(245, 245, 245));
		colorStrings_.put("grey97", new Color(247, 247, 247));
		colorStrings_.put("grey98", new Color(250, 250, 250));
		colorStrings_.put("grey99", new Color(252, 252, 252));
		colorStrings_.put("honeydew", new Color(240, 255, 240));
		colorStrings_.put("honeydew1", new Color(240, 255, 240));
		colorStrings_.put("honeydew2", new Color(224, 238, 224));
		colorStrings_.put("honeydew3", new Color(193, 205, 193));
		colorStrings_.put("honeydew4", new Color(131, 139, 131));
		colorStrings_.put("hotpink", new Color(255, 105, 180));
		colorStrings_.put("hotpink", new Color(255, 105, 180));
		colorStrings_.put("hotpink1", new Color(255, 110, 180));
		colorStrings_.put("hotpink2", new Color(238, 106, 167));
		colorStrings_.put("hotpink3", new Color(205, 96, 144));
		colorStrings_.put("hotpink4", new Color(139, 58, 98));
		colorStrings_.put("indianred", new Color(205, 92, 92));
		colorStrings_.put("indianred", new Color(205, 92, 92));
		colorStrings_.put("indianred1", new Color(255, 106, 106));
		colorStrings_.put("indianred2", new Color(238, 99, 99));
		colorStrings_.put("indianred3", new Color(205, 85, 85));
		colorStrings_.put("indianred4", new Color(139, 58, 58));
		colorStrings_.put("ivory", new Color(255, 255, 240));
		colorStrings_.put("ivory1", new Color(255, 255, 240));
		colorStrings_.put("ivory2", new Color(238, 238, 224));
		colorStrings_.put("ivory3", new Color(205, 205, 193));
		colorStrings_.put("ivory4", new Color(139, 139, 131));
		colorStrings_.put("khaki", new Color(240, 230, 140));
		colorStrings_.put("khaki1", new Color(255, 246, 143));
		colorStrings_.put("khaki2", new Color(238, 230, 133));
		colorStrings_.put("khaki3", new Color(205, 198, 115));
		colorStrings_.put("khaki4", new Color(139, 134, 78));
		colorStrings_.put("lavender", new Color(230, 230, 250));
		colorStrings_.put("lavenderblush", new Color(255, 240, 245));
		colorStrings_.put("lavenderblush", new Color(255, 240, 245));
		colorStrings_.put("lavenderblush1", new Color(255, 240, 245));
		colorStrings_.put("lavenderblush2", new Color(238, 224, 229));
		colorStrings_.put("lavenderblush3", new Color(205, 193, 197));
		colorStrings_.put("lavenderblush4", new Color(139, 131, 134));
		colorStrings_.put("lawngreen", new Color(124, 252, 0));
		colorStrings_.put("lawngreen", new Color(124, 252, 0));
		colorStrings_.put("lemonchiffon", new Color(255, 250, 205));
		colorStrings_.put("lemonchiffon", new Color(255, 250, 205));
		colorStrings_.put("lemonchiffon1", new Color(255, 250, 205));
		colorStrings_.put("lemonchiffon2", new Color(238, 233, 191));
		colorStrings_.put("lemonchiffon3", new Color(205, 201, 165));
		colorStrings_.put("lemonchiffon4", new Color(139, 137, 112));
		colorStrings_.put("lightblue", new Color(173, 216, 230));
		colorStrings_.put("lightblue", new Color(173, 216, 230));
		colorStrings_.put("lightblue1", new Color(191, 239, 255));
		colorStrings_.put("lightblue2", new Color(178, 223, 238));
		colorStrings_.put("lightblue3", new Color(154, 192, 205));
		colorStrings_.put("lightblue4", new Color(104, 131, 139));
		colorStrings_.put("lightcoral", new Color(240, 128, 128));
		colorStrings_.put("lightcoral", new Color(240, 128, 128));
		colorStrings_.put("lightcyan", new Color(224, 255, 255));
		colorStrings_.put("lightcyan", new Color(224, 255, 255));
		colorStrings_.put("lightcyan1", new Color(224, 255, 255));
		colorStrings_.put("lightcyan2", new Color(209, 238, 238));
		colorStrings_.put("lightcyan3", new Color(180, 205, 205));
		colorStrings_.put("lightcyan4", new Color(122, 139, 139));
		colorStrings_.put("lightgoldenrod", new Color(238, 221, 130));
		colorStrings_.put("lightgoldenrod", new Color(238, 221, 130));
		colorStrings_.put("lightgoldenrod1", new Color(255, 236, 139));
		colorStrings_.put("lightgoldenrod2", new Color(238, 220, 130));
		colorStrings_.put("lightgoldenrod3", new Color(205, 190, 112));
		colorStrings_.put("lightgoldenrod4", new Color(139, 129, 76));
		colorStrings_.put("lightgoldenrodyellow", new Color(250, 250, 210));
		colorStrings_.put("lightgoldenrodyellow", new Color(250, 250, 210));
		colorStrings_.put("lightgray", new Color(211, 211, 211));
		colorStrings_.put("lightgray", new Color(211, 211, 211));
		colorStrings_.put("lightgreen", new Color(144, 238, 144));
		colorStrings_.put("lightgreen", new Color(144, 238, 144));
		colorStrings_.put("lightgrey", new Color(211, 211, 211));
		colorStrings_.put("lightgrey", new Color(211, 211, 211));
		colorStrings_.put("lightpink", new Color(255, 182, 193));
		colorStrings_.put("lightpink", new Color(255, 182, 193));
		colorStrings_.put("lightpink1", new Color(255, 174, 185));
		colorStrings_.put("lightpink2", new Color(238, 162, 173));
		colorStrings_.put("lightpink3", new Color(205, 140, 149));
		colorStrings_.put("lightpink4", new Color(139, 95, 101));
		colorStrings_.put("lightsalmon", new Color(255, 160, 122));
		colorStrings_.put("lightsalmon", new Color(255, 160, 122));
		colorStrings_.put("lightsalmon1", new Color(255, 160, 122));
		colorStrings_.put("lightsalmon2", new Color(238, 149, 114));
		colorStrings_.put("lightsalmon3", new Color(205, 129, 98));
		colorStrings_.put("lightsalmon4", new Color(139, 87, 66));
		colorStrings_.put("lightseagreen", new Color(32, 178, 170));
		colorStrings_.put("lightseagreen", new Color(32, 178, 170));
		colorStrings_.put("lightskyblue", new Color(135, 206, 250));
		colorStrings_.put("lightskyblue", new Color(135, 206, 250));
		colorStrings_.put("lightskyblue1", new Color(176, 226, 255));
		colorStrings_.put("lightskyblue2", new Color(164, 211, 238));
		colorStrings_.put("lightskyblue3", new Color(141, 182, 205));
		colorStrings_.put("lightskyblue4", new Color(96, 123, 139));
		colorStrings_.put("lightslateblue", new Color(132, 112, 255));
		colorStrings_.put("lightslateblue", new Color(132, 112, 255));
		colorStrings_.put("lightslategray", new Color(119, 136, 153));
		colorStrings_.put("lightslategray", new Color(119, 136, 153));
		colorStrings_.put("lightslategrey", new Color(119, 136, 153));
		colorStrings_.put("lightslategrey", new Color(119, 136, 153));
		colorStrings_.put("lightsteelblue", new Color(176, 196, 222));
		colorStrings_.put("lightsteelblue", new Color(176, 196, 222));
		colorStrings_.put("lightsteelblue1", new Color(202, 225, 255));
		colorStrings_.put("lightsteelblue2", new Color(188, 210, 238));
		colorStrings_.put("lightsteelblue3", new Color(162, 181, 205));
		colorStrings_.put("lightsteelblue4", new Color(110, 123, 139));
		colorStrings_.put("lightyellow", new Color(255, 255, 224));
		colorStrings_.put("lightyellow", new Color(255, 255, 224));
		colorStrings_.put("lightyellow1", new Color(255, 255, 224));
		colorStrings_.put("lightyellow2", new Color(238, 238, 209));
		colorStrings_.put("lightyellow3", new Color(205, 205, 180));
		colorStrings_.put("lightyellow4", new Color(139, 139, 122));
		colorStrings_.put("limegreen", new Color(50, 205, 50));
		colorStrings_.put("limegreen", new Color(50, 205, 50));
		colorStrings_.put("linen", new Color(250, 240, 230));
		colorStrings_.put("magenta", new Color(255, 0, 255));
		colorStrings_.put("magenta1", new Color(255, 0, 255));
		colorStrings_.put("magenta2", new Color(238, 0, 238));
		colorStrings_.put("magenta3", new Color(205, 0, 205));
		colorStrings_.put("magenta4", new Color(139, 0, 139));
		colorStrings_.put("maroon", new Color(176, 48, 96));
		colorStrings_.put("maroon1", new Color(255, 52, 179));
		colorStrings_.put("maroon2", new Color(238, 48, 167));
		colorStrings_.put("maroon3", new Color(205, 41, 144));
		colorStrings_.put("maroon4", new Color(139, 28, 98));
		colorStrings_.put("mediumaquamarine", new Color(102, 205, 170));
		colorStrings_.put("mediumaquamarine", new Color(102, 205, 170));
		colorStrings_.put("mediumblue", new Color(0, 0, 205));
		colorStrings_.put("mediumblue", new Color(0, 0, 205));
		colorStrings_.put("mediumorchid", new Color(186, 85, 211));
		colorStrings_.put("mediumorchid", new Color(186, 85, 211));
		colorStrings_.put("mediumorchid1", new Color(224, 102, 255));
		colorStrings_.put("mediumorchid2", new Color(209, 95, 238));
		colorStrings_.put("mediumorchid3", new Color(180, 82, 205));
		colorStrings_.put("mediumorchid4", new Color(122, 55, 139));
		colorStrings_.put("mediumpurple", new Color(147, 112, 219));
		colorStrings_.put("mediumpurple", new Color(147, 112, 219));
		colorStrings_.put("mediumpurple1", new Color(171, 130, 255));
		colorStrings_.put("mediumpurple2", new Color(159, 121, 238));
		colorStrings_.put("mediumpurple3", new Color(137, 104, 205));
		colorStrings_.put("mediumpurple4", new Color(93, 71, 139));
		colorStrings_.put("mediumseagreen", new Color(60, 179, 113));
		colorStrings_.put("mediumseagreen", new Color(60, 179, 113));
		colorStrings_.put("mediumslateblue", new Color(123, 104, 238));
		colorStrings_.put("mediumslateblue", new Color(123, 104, 238));
		colorStrings_.put("mediumspringgreen", new Color(0, 250, 154));
		colorStrings_.put("mediumspringgreen", new Color(0, 250, 154));
		colorStrings_.put("mediumturquoise", new Color(72, 209, 204));
		colorStrings_.put("mediumturquoise", new Color(72, 209, 204));
		colorStrings_.put("mediumvioletred", new Color(199, 21, 133));
		colorStrings_.put("mediumvioletred", new Color(199, 21, 133));
		colorStrings_.put("midnightblue", new Color(25, 25, 112));
		colorStrings_.put("midnightblue", new Color(25, 25, 112));
		colorStrings_.put("mintcream", new Color(245, 255, 250));
		colorStrings_.put("mintcream", new Color(245, 255, 250));
		colorStrings_.put("mistyrose", new Color(255, 228, 225));
		colorStrings_.put("mistyrose", new Color(255, 228, 225));
		colorStrings_.put("mistyrose1", new Color(255, 228, 225));
		colorStrings_.put("mistyrose2", new Color(238, 213, 210));
		colorStrings_.put("mistyrose3", new Color(205, 183, 181));
		colorStrings_.put("mistyrose4", new Color(139, 125, 123));
		colorStrings_.put("moccasin", new Color(255, 228, 181));
		colorStrings_.put("navajowhite", new Color(255, 222, 173));
		colorStrings_.put("navajowhite", new Color(255, 222, 173));
		colorStrings_.put("navajowhite1", new Color(255, 222, 173));
		colorStrings_.put("navajowhite2", new Color(238, 207, 161));
		colorStrings_.put("navajowhite3", new Color(205, 179, 139));
		colorStrings_.put("navajowhite4", new Color(139, 121, 94));
		colorStrings_.put("navy", new Color(0, 0, 128));
		colorStrings_.put("navyblue", new Color(0, 0, 128));
		colorStrings_.put("navyblue", new Color(0, 0, 128));
		colorStrings_.put("oldlace", new Color(253, 245, 230));
		colorStrings_.put("oldlace", new Color(253, 245, 230));
		colorStrings_.put("olivedrab", new Color(107, 142, 35));
		colorStrings_.put("olivedrab", new Color(107, 142, 35));
		colorStrings_.put("olivedrab1", new Color(192, 255, 62));
		colorStrings_.put("olivedrab2", new Color(179, 238, 58));
		colorStrings_.put("olivedrab3", new Color(154, 205, 50));
		colorStrings_.put("olivedrab4", new Color(105, 139, 34));
		colorStrings_.put("orange", new Color(255, 165, 0));
		colorStrings_.put("orange1", new Color(255, 165, 0));
		colorStrings_.put("orange2", new Color(238, 154, 0));
		colorStrings_.put("orange3", new Color(205, 133, 0));
		colorStrings_.put("orange4", new Color(139, 90, 0));
		colorStrings_.put("orangered", new Color(255, 69, 0));
		colorStrings_.put("orangered", new Color(255, 69, 0));
		colorStrings_.put("orangered1", new Color(255, 69, 0));
		colorStrings_.put("orangered2", new Color(238, 64, 0));
		colorStrings_.put("orangered3", new Color(205, 55, 0));
		colorStrings_.put("orangered4", new Color(139, 37, 0));
		colorStrings_.put("orchid", new Color(218, 112, 214));
		colorStrings_.put("orchid1", new Color(255, 131, 250));
		colorStrings_.put("orchid2", new Color(238, 122, 233));
		colorStrings_.put("orchid3", new Color(205, 105, 201));
		colorStrings_.put("orchid4", new Color(139, 71, 137));
		colorStrings_.put("palegoldenrod", new Color(238, 232, 170));
		colorStrings_.put("palegoldenrod", new Color(238, 232, 170));
		colorStrings_.put("palegreen", new Color(152, 251, 152));
		colorStrings_.put("palegreen", new Color(152, 251, 152));
		colorStrings_.put("palegreen1", new Color(154, 255, 154));
		colorStrings_.put("palegreen2", new Color(144, 238, 144));
		colorStrings_.put("palegreen3", new Color(124, 205, 124));
		colorStrings_.put("palegreen4", new Color(84, 139, 84));
		colorStrings_.put("paleturquoise", new Color(175, 238, 238));
		colorStrings_.put("paleturquoise", new Color(175, 238, 238));
		colorStrings_.put("paleturquoise1", new Color(187, 255, 255));
		colorStrings_.put("paleturquoise2", new Color(174, 238, 238));
		colorStrings_.put("paleturquoise3", new Color(150, 205, 205));
		colorStrings_.put("paleturquoise4", new Color(102, 139, 139));
		colorStrings_.put("palevioletred", new Color(219, 112, 147));
		colorStrings_.put("palevioletred", new Color(219, 112, 147));
		colorStrings_.put("palevioletred1", new Color(255, 130, 171));
		colorStrings_.put("palevioletred2", new Color(238, 121, 159));
		colorStrings_.put("palevioletred3", new Color(205, 104, 137));
		colorStrings_.put("palevioletred4", new Color(139, 71, 93));
		colorStrings_.put("papayawhip", new Color(255, 239, 213));
		colorStrings_.put("papayawhip", new Color(255, 239, 213));
		colorStrings_.put("peachpuff", new Color(255, 218, 185));
		colorStrings_.put("peachpuff", new Color(255, 218, 185));
		colorStrings_.put("peachpuff1", new Color(255, 218, 185));
		colorStrings_.put("peachpuff2", new Color(238, 203, 173));
		colorStrings_.put("peachpuff3", new Color(205, 175, 149));
		colorStrings_.put("peachpuff4", new Color(139, 119, 101));
		colorStrings_.put("peru", new Color(205, 133, 63));
		colorStrings_.put("pink", new Color(255, 192, 203));
		colorStrings_.put("pink1", new Color(255, 181, 197));
		colorStrings_.put("pink2", new Color(238, 169, 184));
		colorStrings_.put("pink3", new Color(205, 145, 158));
		colorStrings_.put("pink4", new Color(139, 99, 108));
		colorStrings_.put("plum", new Color(221, 160, 221));
		colorStrings_.put("plum1", new Color(255, 187, 255));
		colorStrings_.put("plum2", new Color(238, 174, 238));
		colorStrings_.put("plum3", new Color(205, 150, 205));
		colorStrings_.put("plum4", new Color(139, 102, 139));
		colorStrings_.put("powderblue", new Color(176, 224, 230));
		colorStrings_.put("powderblue", new Color(176, 224, 230));
		colorStrings_.put("purple", new Color(160, 32, 240));
		colorStrings_.put("purple1", new Color(155, 48, 255));
		colorStrings_.put("purple2", new Color(145, 44, 238));
		colorStrings_.put("purple3", new Color(125, 38, 205));
		colorStrings_.put("purple4", new Color(85, 26, 139));
		colorStrings_.put("red", new Color(255, 0, 0));
		colorStrings_.put("red1", new Color(255, 0, 0));
		colorStrings_.put("red2", new Color(238, 0, 0));
		colorStrings_.put("red3", new Color(205, 0, 0));
		colorStrings_.put("red4", new Color(139, 0, 0));
		colorStrings_.put("rosybrown", new Color(188, 143, 143));
		colorStrings_.put("rosybrown", new Color(188, 143, 143));
		colorStrings_.put("rosybrown1", new Color(255, 193, 193));
		colorStrings_.put("rosybrown2", new Color(238, 180, 180));
		colorStrings_.put("rosybrown3", new Color(205, 155, 155));
		colorStrings_.put("rosybrown4", new Color(139, 105, 105));
		colorStrings_.put("royalblue", new Color(65, 105, 225));
		colorStrings_.put("royalblue", new Color(65, 105, 225));
		colorStrings_.put("royalblue1", new Color(72, 118, 255));
		colorStrings_.put("royalblue2", new Color(67, 110, 238));
		colorStrings_.put("royalblue3", new Color(58, 95, 205));
		colorStrings_.put("royalblue4", new Color(39, 64, 139));
		colorStrings_.put("saddlebrown", new Color(139, 69, 19));
		colorStrings_.put("saddlebrown", new Color(139, 69, 19));
		colorStrings_.put("salmon", new Color(250, 128, 114));
		colorStrings_.put("salmon1", new Color(255, 140, 105));
		colorStrings_.put("salmon2", new Color(238, 130, 98));
		colorStrings_.put("salmon3", new Color(205, 112, 84));
		colorStrings_.put("salmon4", new Color(139, 76, 57));
		colorStrings_.put("sandybrown", new Color(244, 164, 96));
		colorStrings_.put("sandybrown", new Color(244, 164, 96));
		colorStrings_.put("seagreen", new Color(46, 139, 87));
		colorStrings_.put("seagreen", new Color(46, 139, 87));
		colorStrings_.put("seagreen1", new Color(84, 255, 159));
		colorStrings_.put("seagreen2", new Color(78, 238, 148));
		colorStrings_.put("seagreen3", new Color(67, 205, 128));
		colorStrings_.put("seagreen4", new Color(46, 139, 87));
		colorStrings_.put("seashell", new Color(255, 245, 238));
		colorStrings_.put("seashell1", new Color(255, 245, 238));
		colorStrings_.put("seashell2", new Color(238, 229, 222));
		colorStrings_.put("seashell3", new Color(205, 197, 191));
		colorStrings_.put("seashell4", new Color(139, 134, 130));
		colorStrings_.put("sienna", new Color(160, 82, 45));
		colorStrings_.put("sienna1", new Color(255, 130, 71));
		colorStrings_.put("sienna2", new Color(238, 121, 66));
		colorStrings_.put("sienna3", new Color(205, 104, 57));
		colorStrings_.put("sienna4", new Color(139, 71, 38));
		colorStrings_.put("skyblue", new Color(135, 206, 235));
		colorStrings_.put("skyblue", new Color(135, 206, 235));
		colorStrings_.put("skyblue1", new Color(135, 206, 255));
		colorStrings_.put("skyblue2", new Color(126, 192, 238));
		colorStrings_.put("skyblue3", new Color(108, 166, 205));
		colorStrings_.put("skyblue4", new Color(74, 112, 139));
		colorStrings_.put("slateblue", new Color(106, 90, 205));
		colorStrings_.put("slateblue", new Color(106, 90, 205));
		colorStrings_.put("slateblue1", new Color(131, 111, 255));
		colorStrings_.put("slateblue2", new Color(122, 103, 238));
		colorStrings_.put("slateblue3", new Color(105, 89, 205));
		colorStrings_.put("slateblue4", new Color(71, 60, 139));
		colorStrings_.put("slategray", new Color(112, 128, 144));
		colorStrings_.put("slategray", new Color(112, 128, 144));
		colorStrings_.put("slategray1", new Color(198, 226, 255));
		colorStrings_.put("slategray2", new Color(185, 211, 238));
		colorStrings_.put("slategray3", new Color(159, 182, 205));
		colorStrings_.put("slategray4", new Color(108, 123, 139));
		colorStrings_.put("slategrey", new Color(112, 128, 144));
		colorStrings_.put("slategrey", new Color(112, 128, 144));
		colorStrings_.put("snow", new Color(255, 250, 250));
		colorStrings_.put("snow1", new Color(255, 250, 250));
		colorStrings_.put("snow2", new Color(238, 233, 233));
		colorStrings_.put("snow3", new Color(205, 201, 201));
		colorStrings_.put("snow4", new Color(139, 137, 137));
		colorStrings_.put("springgreen", new Color(0, 255, 127));
		colorStrings_.put("springgreen", new Color(0, 255, 127));
		colorStrings_.put("springgreen1", new Color(0, 255, 127));
		colorStrings_.put("springgreen2", new Color(0, 238, 118));
		colorStrings_.put("springgreen3", new Color(0, 205, 102));
		colorStrings_.put("springgreen4", new Color(0, 139, 69));
		colorStrings_.put("steelblue", new Color(70, 130, 180));
		colorStrings_.put("steelblue", new Color(70, 130, 180));
		colorStrings_.put("steelblue1", new Color(99, 184, 255));
		colorStrings_.put("steelblue2", new Color(92, 172, 238));
		colorStrings_.put("steelblue3", new Color(79, 148, 205));
		colorStrings_.put("steelblue4", new Color(54, 100, 139));
		colorStrings_.put("tan", new Color(210, 180, 140));
		colorStrings_.put("tan1", new Color(255, 165, 79));
		colorStrings_.put("tan2", new Color(238, 154, 73));
		colorStrings_.put("tan3", new Color(205, 133, 63));
		colorStrings_.put("tan4", new Color(139, 90, 43));
		colorStrings_.put("thistle", new Color(216, 191, 216));
		colorStrings_.put("thistle1", new Color(255, 225, 255));
		colorStrings_.put("thistle2", new Color(238, 210, 238));
		colorStrings_.put("thistle3", new Color(205, 181, 205));
		colorStrings_.put("thistle4", new Color(139, 123, 139));
		colorStrings_.put("tomato", new Color(255, 99, 71));
		colorStrings_.put("tomato1", new Color(255, 99, 71));
		colorStrings_.put("tomato2", new Color(238, 92, 66));
		colorStrings_.put("tomato3", new Color(205, 79, 57));
		colorStrings_.put("tomato4", new Color(139, 54, 38));
		colorStrings_.put("turquoise", new Color(64, 224, 208));
		colorStrings_.put("turquoise1", new Color(0, 245, 255));
		colorStrings_.put("turquoise2", new Color(0, 229, 238));
		colorStrings_.put("turquoise3", new Color(0, 197, 205));
		colorStrings_.put("turquoise4", new Color(0, 134, 139));
		colorStrings_.put("violet", new Color(238, 130, 238));
		colorStrings_.put("violetred", new Color(208, 32, 144));
		colorStrings_.put("violetred", new Color(208, 32, 144));
		colorStrings_.put("violetred1", new Color(255, 62, 150));
		colorStrings_.put("violetred2", new Color(238, 58, 140));
		colorStrings_.put("violetred3", new Color(205, 50, 120));
		colorStrings_.put("violetred4", new Color(139, 34, 82));
		colorStrings_.put("wheat", new Color(245, 222, 179));
		colorStrings_.put("wheat1", new Color(255, 231, 186));
		colorStrings_.put("wheat2", new Color(238, 216, 174));
		colorStrings_.put("wheat3", new Color(205, 186, 150));
		colorStrings_.put("wheat4", new Color(139, 126, 102));
		colorStrings_.put("white", new Color(255, 255, 255));
		colorStrings_.put("whitesmoke", new Color(245, 245, 245));
		colorStrings_.put("whitesmoke", new Color(245, 245, 245));
		colorStrings_.put("yellow", new Color(255, 255, 0));
		colorStrings_.put("yellow1", new Color(255, 255, 0));
		colorStrings_.put("yellow2", new Color(238, 238, 0));
		colorStrings_.put("yellow3", new Color(205, 205, 0));
		colorStrings_.put("yellow4", new Color(139, 139, 0));
		colorStrings_.put("yellowgreen", new Color(154, 205, 50));
		colorStrings_.put("yellowgreen", new Color(154, 205, 50));

		// Standard HTML colors
		//
		colorStrings_.put("black", new Color(Integer.parseInt("000000", 16)));
		colorStrings_.put("green", new Color(Integer.parseInt("008000", 16)));
		colorStrings_.put("silver", new Color(Integer.parseInt("C0C0C0", 16)));
		colorStrings_.put("lime", new Color(Integer.parseInt("00FF00", 16)));
		colorStrings_.put("gray", new Color(Integer.parseInt("808080", 16)));
		colorStrings_.put("olive", new Color(Integer.parseInt("808000", 16)));
		colorStrings_.put("white", new Color(Integer.parseInt("FFFFFF", 16)));
		colorStrings_.put("yellow", new Color(Integer.parseInt("FFFF00", 16)));
		colorStrings_.put("maroon", new Color(Integer.parseInt("800000", 16)));
		colorStrings_.put("navy", new Color(Integer.parseInt("000080", 16)));
		colorStrings_.put("red", new Color(Integer.parseInt("FF0000", 16)));
		colorStrings_.put("blue", new Color(Integer.parseInt("0000FF", 16)));
		colorStrings_.put("purple", new Color(Integer.parseInt("800080", 16)));
		colorStrings_.put("teal", new Color(Integer.parseInt("008080", 16)));
		colorStrings_.put("fuchsia", new Color(Integer.parseInt("FF00FF", 16)));
		colorStrings_.put("aqua", new Color(Integer.parseInt("00FFFF", 16)));
	}
}
/***********************************************************************
 **********************************************************************/
static private void initDecimalFormat() {
	if (null == decimalFormat_) {
		decimalFormat_ = new DecimalFormat();
		decimalFormat_.setGroupingUsed(false);
		decimalFormat_.setMinimumFractionDigits(0);
		//		decimalFormat_.setMaximumFractionDigits(10);
	}
}
/***********************************************************************
 * Initialize the pixels per point size according to the screen resolution.
 **********************************************************************/
static private void initPixelsPerPt() {
	if (!initPixelsPerPt_) {
		Toolkit toolkit = Toolkit.getDefaultToolkit();
		float screenRes = toolkit.getScreenResolution();
		pixelsPerPt_ = screenRes / 72.0f;
	}
}
/***********************************************************************
 ***********************************************************************/
static public String integers2Param(Integer[] integer) {
	return integers2Param(integer, separator);
}
/***********************************************************************
 ***********************************************************************/
static public String integers2Param(Integer[] integer, String separator) {
	if (integer == null) {
		return "";
	}
	StringBuffer buffer = new StringBuffer();
	for (int i = 0; i < integer.length; i++) {
		String intString = integer[i].toString();
		if (i > 0) {
			buffer.append(separator);
			buffer.append(" ");
		}
		buffer.append(intString);
	}
	return buffer.toString();
}
/***********************************************************************
 * Converts an array of Float to an XMTO parameter string.
 * The elements in the output are separated by the separator char.
 ***********************************************************************/
static public String mfBool2Param(Boolean[] bools) {
	return mfBool2Param(bools, separator);
}
/***********************************************************************
 * Converts an array of Float to an XMTO parameter string.
 * The elements in the output are separated by the separator char.
 ***********************************************************************/
static public String mfBool2Param(Boolean[] bools, String separator) {
	if (null == bools) {
		return "";
	}
	StringBuffer buffer = new StringBuffer();
	for (int i = 0; i < bools.length; i++) {
		if (i > 0)
			buffer.append(separator).append(" ");
		buffer.append(bools[i].toString());
	}
	return buffer.toString();
}
/***********************************************************************
 * Converts an array of Float to an XMTO parameter string.
 * The elements in the output are separated by the separator char.
 ***********************************************************************/
static public String mfFloat2Param(Float[] floats) {
	return mfFloat2Param(floats, separator);
}
/***********************************************************************
 * Converts an array of Float to an XMTO parameter string.
 * The elements in the output are separated by the separator char.
 ***********************************************************************/
static public String mfFloat2Param(Float[] floats, String separator) {
	if (null == floats) {
		return "";
	}
	StringBuffer buffer = new StringBuffer();
	for (int i = 0; i < floats.length; i++) {
		if (i > 0)
			buffer.append(separator).append(" ");
		buffer.append(floats[i].toString());
	}
	return buffer.toString();
}
/***********************************************************************
 * Converts the style into the appropriate style string, otherwise 
 * convert it into a Integer string.
 **********************************************************************/
static public String outlineStyle2param(int style) {
	if (style < 0 || outlineStyles_.length <= style)
		return "" + style;
	else
		return outlineStyles_[style];
}
/***********************************************************************
 ***********************************************************************/
static public Integer[] param2ArrayOfOutlineStyles(String param) {
	return param2ArrayOfOutlineStyles(param, separator);
}
/***********************************************************************
 * Convert a parameter string to an array of pairs of Floats.
 * Each such pair is represented as a Vector of two Float objects.
 ***********************************************************************/
static public Integer[] param2ArrayOfOutlineStyles(String param, String separator) {
	if (param == null) {
		return null;
	}
	Vector outlineStyles = new Vector();
	StringTokenizer tokens = new StringTokenizer(param, separator);
	while (tokens.hasMoreTokens()) {
		String token = tokens.nextToken().trim();
		Integer outlineStyle = param2OutlineStyle(token);
		if (outlineStyle != null) {
			outlineStyles.add(outlineStyle);
		}
	}
	Integer[] outlineStylesArray = null;
	if (outlineStyles.size() > 0) {
		outlineStylesArray = new Integer[outlineStyles.size()];
		for (int i = 0; i < outlineStyles.size(); outlineStylesArray[i] = (Integer) outlineStyles.elementAt(i), i++);
	}
	return outlineStylesArray;
}
/***********************************************************************
 ***********************************************************************/
static public Vector[] param2ArrayOfTwoFloats(String param) {
	return param2ArrayOfTwoFloats(param, separator);
}
/***********************************************************************
 * Convert a parameter string to an array of pairs of Floats.
 * Each such pair is represented as a Vector of two Float objects.
 ***********************************************************************/
static public Vector[] param2ArrayOfTwoFloats(String param, String separator) {
	if (param == null) {
		return null;
	}
	Vector twoFloats = new Vector();
	StringTokenizer tokens = new StringTokenizer(param, separator);
	while (tokens.hasMoreTokens()) {
		String token = tokens.nextToken().trim();
		Vector twoFloat = param2TwoFloats(token);
		if (twoFloat != null) {
			twoFloats.add(twoFloat);
		}
	}
	Vector[] twoFloatsArray = null;
	if (twoFloats.size() > 0) {
		twoFloatsArray = new Vector[twoFloats.size()];
		for (int i = 0; i < twoFloats.size(); twoFloatsArray[i] = (Vector) twoFloats.elementAt(i), i++); 
	}
	return twoFloatsArray;
}
/*********************************************************************** 
 * Converts the parameter into a Color.  If something goes
 * wrong then return null.
 **********************************************************************/
static public Color param2Color(String param) {
	if (null == param)
		return null;
	try {
		if (!initColorStrings_)
			initColorStrings();
		Object colorObj = colorStrings_.get(param.toLowerCase());
		if (null != colorObj)
			return (Color) colorObj;
		if (param.charAt(0) == '#') {
			int intColor = Integer.parseInt(param.substring(1), 16);
			return new Color(intColor);
		} else {
			if (param.startsWith("0x")) {
				int intColor = Integer.parseInt(param.substring(2), 16);
				return new Color(intColor);
			} else {
				int firstSpace = param.indexOf(' ');
				int secondSpace = param.indexOf(' ', firstSpace + 1);
				float red = Float.parseFloat(param.substring(0, firstSpace));
				float green = Float.parseFloat(param.substring(firstSpace + 1, secondSpace));
				float blue = Float.parseFloat(param.substring(secondSpace + 1));
				return new Color(red, green, blue);
			}
		}
	} catch (NumberFormatException e) {
		if (debug_) {
			System.out.println("DOMParameter.param2Color() - param=\"" + param + "\" NumberFormatException=" + e);
			e.printStackTrace();
		}
	} catch (Exception e) {
		if (debug_) {
			System.out.println("DOMParameter.param2Color() - param=\"" + param + "\" Exception=" + e);
			e.printStackTrace();
		}
	}
	return null;
}
/***********************************************************************
 * Convert a parameter to an array of color objects.
 * The elements in the parameter are separated by the separator char.
 * The parameter string is of the form "r g b, r g b, ..."
 ***********************************************************************/
static public Color[] param2Colors(String param) {
	return param2Colors(param, separator);
}
/***********************************************************************
 ***********************************************************************/
static public Color[] param2Colors(String param, String separator) {
	if (param == null) {
		return null;
	}
	Vector colors = new Vector();
	StringTokenizer tokens = new StringTokenizer(param, separator);
	while (tokens.hasMoreTokens()) {
		String token = tokens.nextToken().trim();
		Color color = param2Color(token);
		if (color != null) {
			colors.add(color);
		}
	}
	Color[] colorArray = null;
	if (colors.size() > 0) {
		colorArray = new Color[colors.size()];
		for (int i = 0; i < colors.size(); colorArray[i] = (Color) colors.elementAt(i), i++); 
	}
	return colorArray;
}
/***********************************************************************
 * Converts the parameter into an Float.  If something goes
 * wrong then return null.
 **********************************************************************/
static public Double param2Double(String param) {
	if (null == param)
		return null;
	try {
		return Double.valueOf(param);
	} catch (NumberFormatException e) {
		if (debug_) {
			System.out.println("DOMParameter.param2Double() - param=\"" + param + "\" NumberFormatException=" + e);
			e.printStackTrace();
		}
	}
	return null;
}
/***********************************************************************
 * Converts the parameter into an Float.  If something goes
 * wrong then return null.
 **********************************************************************/
static public Float param2Float(String param) {
	if (null == param)
		return null;
	try {
		return Float.valueOf(param);
	} catch (NumberFormatException e) {
		if (debug_) {
			System.out.println("DOMParameter.param2Float() - param=\"" + param + "\" NumberFormatException=" + e);
			e.printStackTrace();
		}
	}
	return null;
}
/***********************************************************************
 * Converts the parameter into an Integer of Java font name.  If something goes
 * wrong then return serif font.
 **********************************************************************/
static public Font param2Font(String paramFontName, String paramFontStyle, String paramSize) {
	String fontName = param2FontName(paramFontName);
	Integer fontStyle = param2FontStyle(paramFontStyle);
	Float size = param2Float(paramSize);
	int fontSize = 12;
	if (null != size) {
		if (!initPixelsPerPt_)
			initPixelsPerPt();
		fontSize = size.intValue();
		// size below is size of font as a float
		fontSize = (int) (size.floatValue() / pixelsPerPt_);
		// Maintain a minimum pt size - 0 just does not cut it!
		if (0 >= fontSize)
			fontSize = 1;
	}
	return (new Font(fontName, fontStyle.intValue(), fontSize));
}
/***********************************************************************
 * Converts the parameter into an Integer of Java font name.  If something goes
 * wrong then return serif font.
 **********************************************************************/
static public String param2FontName(String param) {
	if (null == param)
		return JAVA_SERIF;

	//
	if (param.startsWith("\""))
		param = param.substring(1).trim();
	if (param.endsWith("\"")) {
		int index = param.lastIndexOf("\"");
		param = param.substring(0, index).trim();
	}

	//
	if (param.equals(DOMXMTO.SANS))
		return JAVA_SANSSERIF;
	if (param.equals(DOMXMTO.TYPEWRITER))
		return JAVA_MONOSPACED;
	//	if (param.equals(DOMXMTO.SERIF))
	return JAVA_SERIF;
}
/***********************************************************************
 * Converts the parameter into an Integer of Java font style.  If something goes
 * wrong then return plain style.
 **********************************************************************/
static public Integer param2FontStyle(String param) {
	if (null == param)
		return new Integer(Font.PLAIN);

	//
	param = param.trim();

	//
	if (param.equals(DOMXMTO.BOLD))
		return new Integer(Font.BOLD);
	if (param.equals(DOMXMTO.ITALIC))
		return new Integer(Font.ITALIC);
	if (param.equals(DOMXMTO.BOLDITALIC))
		return new Integer(Font.BOLD | Font.ITALIC);
	//	if (param.equals(DOMXMTO.PLAIN))
	return new Integer(Font.PLAIN);
}
/***********************************************************************
 * Converts the parameter into an Integer of Java font name.  If something goes
 * wrong then return serif font.
 * NOTE: this should be a MFfield, but for now try our best at determining
 * alignment
 **********************************************************************/
static public String param2HorizontalAlignment(String param) {
	if (null == param)
		return DOMXMTO.BEGIN;

	//
	int commaPos = param.indexOf(";");

	// If can't find comma or space then consider it one string.
	String horizontalAlignment;
	if (0 > commaPos)
		horizontalAlignment = param.trim();
	else
		horizontalAlignment = param.substring(0, commaPos).trim();

	//
	if (null == horizontalAlignment)
		return DOMXMTO.BEGIN;
	if (horizontalAlignment.equals(DOMXMTO.BEGIN) || horizontalAlignment.equals(DOMXMTO.MIDDLE) || horizontalAlignment.equals(DOMXMTO.END) || horizontalAlignment.equals(DOMXMTO.FIRST))
		return horizontalAlignment;
	else
		return DOMXMTO.BEGIN;
}
/***********************************************************************
 * Converts the parameter into an Integer.  If something goes
 * wrong then return null.
 **********************************************************************/
static public Integer param2Integer(String param) {
	if (null == param)
		return null;
	try {
		return Integer.valueOf(param);
	} catch (NumberFormatException e) {
		if (debug_) {
			System.out.println("DOMParameter.param2Integer() - param=\"" + param + "\" NumberFormatException=" + e);
			e.printStackTrace();
		}
	} catch (Exception e) {
		if (debug_) {
			System.out.println("DOMParameter.param2Integer() - param=\"" + param + "\" Exception=" + e);
			e.printStackTrace();
		}
	}
	return null;
}
/***********************************************************************
 ***********************************************************************/
static public Integer[] param2Integers(String param) {
	return param2Integers(param, separator);
}
/***********************************************************************
 ***********************************************************************/
static public Integer[] param2Integers(String param, String separator) {
	if (param == null) {
		return null;
	}
	Vector integers = new Vector();
	StringTokenizer tokens = new StringTokenizer(param, separator);
	while (tokens.hasMoreTokens()) {
		String token = tokens.nextToken().trim();
		Integer integer = param2Integer(token);
		if (integer != null) {
			integers.add(integer);
		}
	}
	Integer[] integerArray = null;
	if (integers.size() > 0) {
		integerArray = new Integer[integers.size()];
		for (int i = 0; i < integers.size(); integerArray[i] = (Integer) integers.elementAt(i), i++); 
	}
	return integerArray;
}
/***********************************************************************
 * Convert a parameter string to an array of Floats (MFFloat).
 * If something goes wrong then return null.
 **********************************************************************/
static public Float[] param2KeySplines(String param) {
	return param2MFFloat(param, " ;");
}
/***********************************************************************
 * Convert a parameter string to an array of Booleans (MFBool).
 * If something goes wrong then return null.
 **********************************************************************/
static public Boolean[] param2MFBool(String param) {
	return param2MFBool(param, separator);
}
/***********************************************************************
 * Convert a parameter string to an array of Floats (MFFloat).
 * If something goes wrong then return null.
 ***********************************************************************/
static public Boolean[] param2MFBool(String param, String separator) {
	if (null == param)
		return null;
	Vector bools = new Vector();
	StringTokenizer tokens = new StringTokenizer(param, separator);
	while (tokens.hasMoreTokens()) {
		String token = tokens.nextToken().trim();
		bools.addElement(new Boolean(token));
	}
	Boolean[] boolsArray = null;
	if (bools.size() > 0) {
		boolsArray = new Boolean[bools.size()];
		for (int i = 0; i < bools.size(); i++)
			boolsArray[i] = (Boolean) bools.elementAt(i);
	}
	return boolsArray;
}
/***********************************************************************
 * Convert a parameter string to an array of Floats (MFFloat).
 * If something goes wrong then return null.
 **********************************************************************/
static public Float[] param2MFFloat(String param) {
	return param2MFFloat(param, separator);
}
/***********************************************************************
 * Convert a parameter string to an array of Floats (MFFloat).
 * If something goes wrong then return null.
 ***********************************************************************/
static public Float[] param2MFFloat(String param, String separator) {
	if (null == param)
		return null;
	try {
		Vector floats = new Vector();
		StringTokenizer tokens = new StringTokenizer(param, separator);
		while (tokens.hasMoreTokens()) {
			String token = tokens.nextToken().trim();
			floats.addElement(new Float(token));
		}
		Float[] floatsArray = null;
		if (floats.size() > 0) {
			floatsArray = new Float[floats.size()];
			for (int i = 0; i < floats.size(); i++)
				floatsArray[i] = (Float) floats.elementAt(i);
		}
		return floatsArray;
	} catch (NumberFormatException e) {
		if (debug_) {
			System.out.println("DOMParameter.param2Float() - param=\"" + param + "\" NumberFormatException=" + e);
			e.printStackTrace();
		}
	}
	return null;
}
/***********************************************************************
 * Converts the parameter into an Integer.  The parameter can be one of the
 * valid line style strings. If something goes
 * wrong then return null.
 **********************************************************************/
static public Integer param2OutlineStyle(String param) {
	if (null == param)
		return null;

	// Check if it is one of the strings
	String lowerCaseParam = param.toLowerCase();
	for (int i = 0; i < outlineStyles_.length; i++) {
		if (outlineStyles_[i].equals(lowerCaseParam)) {
			return new Integer(i);
		}
	}

	// If not then try and convert it from as integer
	return param2Integer(param);
}
/***********************************************************************
 * Converts the parameter into an Integer.  If something goes
 * wrong then return null.
 **********************************************************************/
static public Point param2Point(String param) {
	if (null == param)
		return null;
	try {
		int firstSpace = param.indexOf(' ');
		int secondSpace = param.indexOf(' ', firstSpace + 1);
		double x = Double.parseDouble(param.substring(0, firstSpace));
		double y;
		// if there are only 2 parameters get use the substring to the end 
		// otherwise use the substring to the 3rd parameter
		if (-1 == secondSpace)
			y = Double.parseDouble(param.substring(firstSpace + 1));
		else
			y = Double.parseDouble(param.substring(firstSpace + 1, secondSpace));
		Point point = new Point();
		point.setLocation(x, y);
		return point;
	} catch (NumberFormatException e) {
		if (debug_) {
			System.out.println("DOMParameter.param2Point() - param=\"" + param + "\" NumberFormatException=" + e);
			e.printStackTrace();
		}
	} catch (Exception e) {
		if (debug_) {
			System.out.println("DOMParameter.param2Point() - param=\"" + param + "\" Exception=" + e);
			e.printStackTrace();
		}
	}
	return null;
}
/***********************************************************************
 * Convert a parameter to an array of Points.
 * The elements in the parameter are separated by the separator char.
 * The parameter string is of the form "x y z, x y z, ..."
 * For 2D points, "z" is not specified.
 ***********************************************************************/
static public Point[] param2Points(String param) {
	return param2Points(param, separator);
}
/***********************************************************************
 ***********************************************************************/
static public Point[] param2Points(String param, String separator) {
	if (param == null) {
		return null;
	}
	Vector points = new Vector();
	StringTokenizer tokens = new StringTokenizer(param, separator);
	while (tokens.hasMoreTokens()) {
		String token = tokens.nextToken().trim();
		Point point = param2Point(token);
		if (point != null) {
			points.add(point);
		}
	}
	Point[] pointArray = null;
	if (points.size() > 0) {
		pointArray = new Point[points.size()];
		for (int i = 0; i < points.size(); pointArray[i] = (Point) points.elementAt(i), i++);
	}
	return pointArray;
}
/***********************************************************************
 * Converts the parameter for the textLines attribute of the string 
 * element into an array of lines of text.  The parameter is a MFString.
 **********************************************************************/
static public String[] param2TextLines(String param) {
	if (null == param)
		return new String[0];


	// First separate the string into lines
	// extract the Strings, but ignoring \n's in the first and last position
	Vector textLinesVector = new Vector();
	int beginIndex = param.indexOf('"');
	int nextBackslashIndex = param.indexOf('\\', beginIndex + 1);
	int nextQuoteIndex = param.indexOf('"', beginIndex + 1);
	// look for the end quote. Skip escaped quote characters.
	while (-1 != nextBackslashIndex && nextQuoteIndex > nextBackslashIndex) {
		// If this is a escaped quote go to the next quote character
		// Otherwise ignore the escaped backslash skip the two backslashes
		if (nextBackslashIndex + 1 == nextQuoteIndex) {
			nextBackslashIndex = param.indexOf('\\', nextQuoteIndex + 1);
			nextQuoteIndex = param.indexOf('"', nextQuoteIndex + 1);
		} else {
			nextQuoteIndex = param.indexOf('"', nextBackslashIndex + 2);
			nextBackslashIndex = param.indexOf('\\', nextBackslashIndex + 2);
		}
	}
	int endIndex = nextQuoteIndex;

	// extract the lines
	while (-1 < beginIndex && -1 < endIndex) {
		// Use string buffer so that we can later get rid of escape characters
		StringBuffer line = new StringBuffer(param.substring(beginIndex + 1, endIndex));
		textLinesVector.addElement(line);

		//
		beginIndex = param.indexOf('"', nextQuoteIndex + 1);
		nextBackslashIndex = param.indexOf('\\', beginIndex + 1);
		nextQuoteIndex = param.indexOf('"', beginIndex + 1);
		// look for the end quote. Skip escaped quote characters.
		while (-1 < nextBackslashIndex && nextQuoteIndex > nextBackslashIndex) {
			// If this is a escaped quote go to the next quote character
			// Otherwise ignore the escaped backslash skip the two backslashes
			if (nextBackslashIndex + 1 == nextQuoteIndex) {
				nextBackslashIndex = param.indexOf('\\', nextQuoteIndex + 1);
				nextQuoteIndex = param.indexOf('"', nextQuoteIndex + 1);
			} else {
				nextQuoteIndex = param.indexOf('"', nextBackslashIndex + 2);
				nextBackslashIndex = param.indexOf('\\', nextBackslashIndex + 2);
			}
		}
		endIndex = nextQuoteIndex;
	}


	// We have the list of lines now go through them to strip out all the
	// MPEG-4 escape characters (backslash) for "\"" and "\\"
	// Also add the new string to the array of lines we want to pass back
	String[] textLines = new String[textLinesVector.size()];
	for (int i = 0; i < textLinesVector.size(); i++) {
		StringBuffer sb = (StringBuffer) textLinesVector.elementAt(i);
		for (int j = 0; j < sb.length(); j++) {
			char character = sb.charAt(j);
			if ('\\' == character) {
				sb.deleteCharAt(j);
			}
		}
		textLines[i] = sb.toString();
	}

	//
	return textLines;
}
/***********************************************************************
 * Converts the parameter for the textLines attribute of the string 
 * element into an single text string with new line character separating the
 * lines of text.  The parameter is a MFString.
 **********************************************************************/
static public String param2TextString(String param) {
	if (null == param)
		return "";

	// Get the lines
	String[] lines = param2TextLines(param);
	StringBuffer sb = new StringBuffer("");

	// Concatenate the lines onto the String buffer.  
	// After each line except the last line add a new line character.
	if (0 < lines.length)
		sb.append(lines[0]);
	for (int i = 1, count = lines.length; i < count; i++)
		sb.append("\n").append(lines[i]);

	// return the string
	return sb.toString();
}
/***********************************************************************
 * Converts the parameter into a ms time in a Long.  If something goes
 * wrong then return null.
 **********************************************************************/
static public Long param2Time(String param) {
	if (null == param)
		return null;
	int multiplier = 1000;
	int truncateCharacters = 0;
	if (param.endsWith("ms")) {
		truncateCharacters = 2;
		multiplier = 1;
	} else
		if (param.endsWith("s")) {
			truncateCharacters = 1;
			multiplier = 1000;
		} else
			if (param.endsWith("min")) {
				truncateCharacters = 3;
				multiplier = 1000 * 60;
			} else
				if (param.endsWith("h")) {
					truncateCharacters = 1;
					multiplier = 1000 * 60 * 60;
				}
	try {
		float floatTime = Float.parseFloat(param.substring(0, param.length() - truncateCharacters));
		Float value = new Float(floatTime * multiplier);
		return new Long(value.longValue());
	} catch (NumberFormatException e) {
		if (debug_) {
			System.out.println("DOMParameter.param2Time() - param=\"" + param + "\" NumberFormatException=" + e);
			e.printStackTrace();
		}
	}
	return null;
}
/***********************************************************************
 * Converts the parameter into an Integer.  If something goes
 * wrong then return null.
 **********************************************************************/
static public Vector param2TwoDoubles(String param) {
	if (null == param)
		return null;
	try {
		int firstSpace = param.indexOf(' ');
		int secondSpace = param.indexOf(' ', firstSpace + 1);
		Double first = new Double(param.substring(0, firstSpace));
		Double second;
		// if there are only 2 parameters get use the substring to the end 
		// otherwise use the substring to the 3rd parameter
		if (-1 == secondSpace) {
			String str = param.substring(firstSpace + 1);
			second = new Double(str);
		} else
			second = new Double(param.substring(firstSpace + 1, secondSpace));
		Vector vector = new Vector();
		vector.addElement(first);
		vector.addElement(second);
		return vector;
	} catch (NumberFormatException e) {
		if (debug_) {
			System.out.println("DOMParameter.param2TwoDoubles() - param=\"" + param + "\" NumberFormatException=" + e);
			e.printStackTrace();
		}
	} catch (Exception e) {
		if (debug_) {
			System.out.println("DOMParameter.param2TwoDoubles() - param=\"" + param + "\" Exception=" + e);
			e.printStackTrace();
		}
	}
	return null;
}
/***********************************************************************
 * Converts the parameter into an Integer.  If something goes
 * wrong then return null.
 **********************************************************************/
static public Vector param2TwoFloats(String param) {
	if (null == param)
		return null;
	try {
		StringTokenizer tokens = new StringTokenizer(param, " ");
		Vector vector = new Vector();
		// Add first element
		if (!tokens.hasMoreTokens())
			return null;
		String token = tokens.nextToken().trim();
		vector.addElement(new Float(token));
		// add Second element
		if (!tokens.hasMoreTokens())
			return null;
		token = tokens.nextToken().trim();
		vector.addElement(new Float(token));
		return vector;
	} catch (NumberFormatException e) {
		if (debug_) {
			System.out.println("DOMParameter.param2TwoFloats() - param=\"" + param + "\" NumberFormatException=" + e);
			e.printStackTrace();
		}
	} catch (Exception e) {
		if (debug_) {
			System.out.println("DOMParameter.param2TwoFloats() - param=\"" + param + "\" Exception=" + e);
			e.printStackTrace();
		}
	}
	return null;
}
/***********************************************************************
 * Converts the parameter into an Integer of Java font name.  If something goes
 * wrong then return serif font.
 * NOTE: this should be a MFfield, but for now try our best at determining
 * alignment
 **********************************************************************/
static public String param2VerticalAlignment(String param) {
	if (null == param)
		return DOMXMTO.BEGIN;

	//
	int commaPos = param.indexOf(";");

	// If can't find comma or space then consider it one string.
	String verticalAlignment;
	if (0 > commaPos)
		verticalAlignment = null;
	else
		verticalAlignment = param.substring(commaPos + 1).trim();

	//
	if (null == verticalAlignment)
		return DOMXMTO.BEGIN;
	if (verticalAlignment.equals(DOMXMTO.BEGIN) || verticalAlignment.equals(DOMXMTO.MIDDLE) || verticalAlignment.equals(DOMXMTO.END) || verticalAlignment.equals(DOMXMTO.FIRST))
		return verticalAlignment;
	else
		return DOMXMTO.BEGIN;
}
/***********************************************************************
 **********************************************************************/
static public String point2Param(Point point) {
	return ("" + point.x + " " + point.y);
}
/***********************************************************************
 * Converts an array of Points to an XMTO parameter string.
 * The elements in the output are separated by the separator char.
 * The parameter string is of the form "x y z, x y z, ..."
 ***********************************************************************/
static public String points2Param(Point[] point) {
	return points2Param(point, separator);
}
/***********************************************************************
 ***********************************************************************/
static public String points2Param(Point[] point, String separator) {
	if (point == null) {
		return "";
	}
	StringBuffer xyzBuffer = new StringBuffer();
	for (int i = 0; i < point.length; i++) {
		String xyz = point2Param(point[i]);
		if (xyz.length() > 0) {
			if (i > 0) {
				xyzBuffer.append(separator);
				xyzBuffer.append(" ");
			}
			xyzBuffer.append(xyz);
		}
	}
	return xyzBuffer.toString();
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
/***********************************************************************
 * Converts the array of lines of text into a parameter for the textLines 
 * attribute of the string element.  The parameter is a MFString.
 **********************************************************************/
static public String textLines2Param(String[] textLines) {
	if (null == textLines || 0 == textLines.length)
		return "";

	// First we have to change all the \ or " characters with \\ or \"
	// In order to do this create an array of string buffers since it is
	// easier to insert escape (\) characters into string buffers than
	// strings.
	StringBuffer[] stringBuffers = new StringBuffer[textLines.length];
	for (int i = 0; i < textLines.length; i++)
		stringBuffers[i] = new StringBuffer(textLines[i]);

	// Now loop through each string buffer and then through each
	// through the characters in the string buffer replacing 
	// the \ or " characters with \\ or \"
	// This is to escape out all the MPEG-4 escapes.
	for (int i = 0; i < stringBuffers.length; i++) {
		StringBuffer sb = stringBuffers[i];
		for (int j = 0, count = sb.length(); j < count; j++) {
			char character = sb.charAt(j);
			if ('\\' == character || '\"' == character) {
				sb.insert(j, '\\');
				j++;
				count++;
			}
		}
	}

	// Now create the parameter.  This is a concatenation
	// of all the text lines surrounded by quote (") characters.
	StringBuffer param = new StringBuffer();
	for (int i = 0; i < stringBuffers.length - 1; i++) {
		param.append("\"").append(stringBuffers[i]).append("\"; ");
	}
	// append the last line.  Done so there is no "; " at the end.
	param.append("\"").append(stringBuffers[stringBuffers.length - 1]).append("\"");

	//
	return param.toString();
}
/***********************************************************************
 * Converts the Text string into a parameter for the textLines attribute 
 * of the string element.  The input text string is a single string containing
 * multiple lines of text.  The multiple lines are delimted by new line ('\n') 
 * characters.
 **********************************************************************/
static public String textString2Param(String textString) {
	if (null == textString)
		return "";

	// First separate the string into lines
	// extract the Strings, but ignoring \n's in the first and last position
	Vector textLinesVector= new Vector();
	int beginIndex = 0;
	int endIndex = -1;
	while ((endIndex = textString.indexOf('\n', beginIndex)) != -1) {
		String line = textString.substring(beginIndex, endIndex);
		textLinesVector.addElement(line);
		beginIndex = endIndex + 1;
	}

	//
	endIndex = textString.length();
	// if source string was NOT terminated with a \n, include the last line
	if (beginIndex != textString.length()) {
		String line = textString.substring(beginIndex, endIndex);
		textLinesVector.addElement(line);
	}


	// allocate the result array, fill and return it
	String[] textLines = new String[textLinesVector.size()];
	textLinesVector.copyInto(textLines);

	// Now convert the text lines into the parameter
	return textLines2Param(textLines);
}
/*********************************************************************** 
 **********************************************************************/
static public String time2Param(long time) {
	StringBuffer param = new StringBuffer();
	param.append(time).append("ms");
	return param.toString();
}
/***********************************************************************
 * Converts the parameter into an Integer.  If something goes
 * wrong then return null.
 **********************************************************************/
static public String twoDoubles2Param(double first, double second) {
	if (null == decimalFormat_)
		initDecimalFormat();
	StringBuffer param = new StringBuffer();
	param.append(decimalFormat_.format(first)).append(" ").append(decimalFormat_.format(second));
	return param.toString();
}
/***********************************************************************
 * Converts the parameter into an Integer.  If something goes
 * wrong then return null.
 **********************************************************************/
static public String twoFloats2Param(float first, float second) {
	if (null == decimalFormat_)
		initDecimalFormat();
	StringBuffer param = new StringBuffer();
	//	param.append(decimalFormat_.format(first)).append(" ").append(decimalFormat_.format(second));
	param.append(first);
	param.append(" ");
	param.append(second);
	return param.toString();
}
}
