/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  MediaObjectInformation.java
//
//  Overview  : Class to determine the information about a media object source file.
//    
//=============================================================================
package xmtutil;

import java.awt.Dimension;
import java.io.File;
import javax.swing.ImageIcon;

public class MediaObjectInformation {

  ////////////////////////////////////////////////////////////////////
  // objectTypeIndication values (table 8).
  ////////////////////////////////////////////////////////////////////

  public static final int Visual_Mpeg4                      = 0x20;
  public static final int Audio_Mpeg4                       = 0x40;
  public static final int Visual_Mpeg2_Simple_profile       = 0x60;
  public static final int Visual_Mpeg2_Main_Profile         = 0x61;
  public static final int Visual_Mpeg2_SNR_Profile          = 0x62;
  public static final int Visual_Mpeg2_Spatial_Profile      = 0x63;
  public static final int Visual_Mpeg2_High_Profile         = 0x64;
  public static final int Visual_Mpeg2_422_Profile          = 0x65;
  public static final int Audio_Mpeg2_Main_Profile          = 0x66;
  public static final int Audio_Mpeg2_LowComplexity_Profile = 0x67;
  public static final int Audio_Mpeg2_SSR_Profile           = 0x68;
  public static final int Audio_Mpeg2                       = 0x69;
  public static final int Visual_Mpeg1                      = 0x6A;
  public static final int Audio_Mpeg1                       = 0x6B;
  public static final int Visual_Jpeg                       = 0x6C;
  public static final int no_profile_specified              = 0xFF;

  ////////////////////////////////////////////////////////////////////
  // objectTypeIndication values defined by me or by agreement with others.
  ////////////////////////////////////////////////////////////////////

  public static final int Visual_GIF                        = 0xC0;
  public static final int Audio_G723                        = 0xC1;
  public static final int Visual_H263                       = 0xC2;
  public static final int Audio_G711                        = 0xC3;

  ////////////////////////////////////////////////////////////////////
  // streamType values (table 9).
  ////////////////////////////////////////////////////////////////////

  public static final int ObjectDescriptorStream  = 0x01;
  public static final int ClockReferenceStream    = 0x02;
  public static final int SceneDescriptionStream  = 0x03;
  public static final int VisualStream            = 0x04;
  public static final int AudioStream             = 0x05;
  public static final int MPEG7Stream             = 0x06;
  public static final int IPMPStream              = 0x07;
  public static final int ObjectContentInfoStream = 0x08;

  ////////////////////////////////////////////////////////////////////
  // streamType values defined by me or by agreement with others.
  ////////////////////////////////////////////////////////////////////

  public static final int MPEGJStream             = 0x09;

	/*********************************************************************** 
	 * Whether debug information is to be printed to the console. 
	 * default is not have debug information printed (false).
	 **********************************************************************/
	static private boolean debug_ = false;

	/***********************************************************************
	 * media file name 
	 **********************************************************************/
	private String fileName_ = null;
	/***********************************************************************
	 * media object type 
	 **********************************************************************/
	private int mediaObjectType = 0;
	/***********************************************************************
	 * media stream type 
	 **********************************************************************/
	private int mediaStreamType = 0;

	/***********************************************************************
	 * Generic type of file.
	 **********************************************************************/
	static final public int UNKNOWN_TYPE = 0;
	static final public int VIDEO = 1;
	static final public int AUDIO = 2;
	static final public int IMAGE = 3;

/*******************************************************************
 *******************************************************************/
public MediaObjectInformation(String fileName) {
	fileName_ = fileName;
	if (debug_)
		System.out.println("MediaObjectInformation constructor - Entered fileName=" + fileName);

	// If the file doesn't exist then don't even try and figure out what it is.
	if (null == fileName_) {
		if (debug_)
			System.out.println("MediaObjectInformation constructor - Exiting fileName=null");
		return;
	}
	File file = new File(fileName_);
	if (!file.exists()) {
		if (debug_)
			System.out.println("MediaObjectInformation constructor - Exiting file doesn't exist");
		return;
	}


	// First try to use the file extension to pretest the type of the file.
	String upcaseFileName = fileName_.toUpperCase();
	if (upcaseFileName.endsWith("263")) {
			mediaObjectType = Visual_Mpeg4;
			mediaStreamType = VisualStream;
	} else {
		if (upcaseFileName.endsWith("723")) {
			mediaObjectType = Audio_G723;
			mediaStreamType = AudioStream;
		} else
			if (upcaseFileName.endsWith("MP3")) {
				mediaObjectType = Audio_Mpeg1;
				mediaStreamType = AudioStream;
			} else
				if (upcaseFileName.endsWith("CMP")) {
					mediaObjectType = Visual_Mpeg4;
					mediaStreamType = VisualStream;
				} else
					if (upcaseFileName.endsWith("AAC")) {
						mediaObjectType = Audio_Mpeg4;
						mediaStreamType = AudioStream;
					} else
						if (upcaseFileName.endsWith("JPG")) {
							mediaObjectType = Visual_Jpeg;
							mediaStreamType = VisualStream;
						}
	}

}
/*******************************************************************
 * Get the dimension of the object.
 * NOTE: for audio objects this returns a dimension of (1,1).
 * @return the dimension for the media object.
 *******************************************************************/
public Dimension getDimension() {
	return new Dimension(1, 1);
}
/*******************************************************************
 * Get the duration of the media object.
 * @return the duration for the media object.
 *******************************************************************/
public long getDuration() {
	return 0;
}
/*******************************************************************
 * Get the name of the file.
 * @return the file name for the media object.
 *******************************************************************/
public String getFileName() {
	return fileName_;
}
/*******************************************************************
 * Get the image at the specified time.
 * Currently only support image from the begining of the video.
 * Actually we don't even get that for now.
 *******************************************************************/
public ImageIcon getImage(long time) {
	return null;
}
/*******************************************************************
 * Get the media object type description of the object.
 * Object Type Indication Values from MPEG-4 Systems specification, Table 8.
 * @return object type of the media object.
 *******************************************************************/
public int getMediaObjectType() {
	return mediaObjectType;
}
/*******************************************************************
 * Get the media stream type of the object.
 * Stream Type Values from MPEG-4 Systems specification, Table 9.
 * @return stream type of the media object.
 *******************************************************************/
public int getMediaStreamType() {
	return mediaStreamType;
}
/*******************************************************************
 * Get a generic description of the media type for the authoring tool.
 * @return the generic media type (VIDEO, AUDIO or UNKNOWN_TYPE.
 *******************************************************************/
public int getMediaType() {
	switch (getMediaObjectType()) {
		case Visual_H263 :
		case Visual_Mpeg4 :
		case Visual_Mpeg2_Simple_profile :
		case Visual_Mpeg2_Main_Profile :
		case Visual_Mpeg2_SNR_Profile :
		case Visual_Mpeg2_Spatial_Profile :
		case Visual_Mpeg2_High_Profile :
		case Visual_Mpeg2_422_Profile :
		case Visual_Mpeg1 :
			return VIDEO;
		case Audio_G723 :
		case Audio_Mpeg1 :
		case Audio_G711 :
		case Audio_Mpeg4 :
		case Audio_Mpeg2_Main_Profile :
		case Audio_Mpeg2_LowComplexity_Profile :
		case Audio_Mpeg2_SSR_Profile :
		case Audio_Mpeg2 :
			return AUDIO;
		case Visual_Jpeg :
			return IMAGE;
		default :
			return UNKNOWN_TYPE;
	}
}
/***********************************************************************
 * add a Media Object
 **********************************************************************/
static public String[] getSupportedMediaFileTypes(int mediaType) {
	switch (mediaType) {
		case MediaObjectInformation.IMAGE :
			return new String[] {
			/* */ "jpg",
			/* */};
		case MediaObjectInformation.VIDEO :
			return new String[] {
			/* */ "263",
			/* */ "m4v",
			/* */ "cmp",
			/* */};
		case MediaObjectInformation.AUDIO :
			return new String[] {
			/* */ "723",
			/* */ "mp3",
			/* */ "aac",
			/* */};
		default :
			return new String[0];
	}
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	debug_ = debug;
}
}
