/******************************************************************************\
** This software module was originally developed by IBM Research (T.J.Watson) **
** in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)       **
** standard.  This software module is an implementation of a part of one or   **
** more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4     **
** Systems (ISO/IEC 14496-1) standard.  ISO/IEC gives users of the MPEG-4     **
** Systems (ISO/IEC 14496-1) free license to this software module or          **
** modifications thereof for use in hardware or software products claiming    **
** conformance to the MPEG-4 Systems (ISO/IEC 14496-1).  Those intending to   **
** use this software module in hardware or software products are advised that **
** its use may infringe existing patents.  The original developer of this     **
** software module and his/her company, the subsequent editors and their      **
** companies, and ISO/IEC have no liability for use of this software module   **
** or modifications thereof in an implementation.  Copyright is not released  **
** for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  IBM retain  **
** full right to use the code for their own purpose, assign or donate the     **
** code to a third party and to inhibit third parties from using the code for **
** non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.  This copyright  **
** notice must be included in all copies or derivative works.                 **
** Copyright  2000-2001.                                                     **
** -------------------------------------------------------------------------- **
**  IBM  TJ Watson Research, USA                                              **
**  Contact :                                                                 **
**    Dr. Michelle Kim ( mykim@us.ibm.com )                                   **
\******************************************************************************/

//=============================================================================
//  TimingAttribute.java
//
//=============================================================================
package xmtutil;
                                                                        
import java.text.MessageFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Vector;
import org.w3c.dom.Document;
import xmtutil.resources.TimingAttributeRes;
/*********************************************************************** 
 * Class to interpret the various timing that can occur in the timing attribute.
 * NOTE: the grammar for the dur is not explicitly so it has been generated here
 * to be based on the spec.
 *
 * <p>
 * NOTE: the ";", "+", "(", ")" and "\" are not valid characters in XML ids.  This is important when
 * looking for delimiter variables in ids.  Other delimiter characters "-", ":" and "." are valid
 * characters in XML ids.  In order to use them in ids located in TimingAttributes they need to be
 * escaped with the "\" character.  Thus they would appear in the id part of the attribute string as 
 * "\-", "\:" and "\.".
 *
 * <pre>
 * begin-value-list     ::= begin-value (S ";" S begin-value-list )?
 * begin-value          ::= (offset-value | syncbase-value | syncToPrev-value | event-value | repeat-value
 *                           | accessKey-value | media-marker-value | wallclock-sync-value | "indefinite" )
 * end-value-list       ::= end-value (S ";" S end-value-list )?
 * end-value            ::= (offset-value | syncbase-value | syncToPrev-value | event-value | repeat-value 
 *                           | accessKey-value | media-marker-value | wallclock-sync-value | "indefinite" )
 * dur-value            ::= ( S ) ( Clock-value | "media" | "indefinite" ) ( S )
 *                     
 * S                    ::= (#x20 | #x9 | #xD | #xA)*
 * 
 * offset-value         ::= (( S "+" | "-" S )? ( Clock-value )
 *                     
 * Clock-value          ::= ( Full-clock-value | Partial-clock-value | Timecount-value )
 * Full-clock-value     ::= Hours ":" Minutes ":" Seconds ("." Fraction)?
 * Partial-clock-value  ::= Minutes ":" Seconds ("." Fraction)?
 * Timecount-value      ::= Timecount ("." Fraction)? (Metric)?
 * Metric               ::= "h" | "min" | "s" | "ms"
 * Hours                ::= DIGIT+; any positive number
 * Minutes              ::= 2DIGIT; range from 00 to 59
 * Seconds              ::= 2DIGIT; range from 00 to 59
 * Fraction             ::= DIGIT+
 * Timecount            ::= DIGIT+
 * 2DIGIT               ::= DIGIT DIGIT
 * DIGIT                ::= [0-9]
 *                     
 * Syncbase-value       ::= ( Syncbase-element "." Time-symbol ) ( S ("+"|"-") S Clock-value )? 
 * Syncbase-element     ::= Id-value
 * Time-symbol          ::= "begin" | "end" 
 * Id-value             ::= IDREF
 *                     
 * SyncToPrev-value     ::= ( "prev." Time-symbol ) ( S ("+"|"-") S Clock-value )?  
 *                     
 * Event-value          ::= ( Eventbase-element "." )? Event-symbol ( S ("+"|"-") S Clock-value )? 
 * Eventbase-element    ::= ID 
 *                     
 * Repeat-value         ::= ( Eventbase-element "." )? "repeat(" iteration ")" ( S ("+"|"-") S Clock-value )? 
 * iteration            ::= DIGIT+  
 *                     
 * AccessKey-value      ::= "accessKey(" character ")" ( S ("+"|"-") S Clock-value )? 
 * 
 * Media-Marker-value   ::= Id-value ".marker(" S marker-name S ")" ) 
 * 
 * wallclock-sync-value ::= "wallclock(" S (DateTime | WallTime)  S ")"
 * DateTime             ::= Date "T" WallTime
 * Date                 ::= Years "-" Months "-" Days
 * WallTime             ::= (HHMM-Time | HHMMSS-Time)(TZD)?
 * HHMM-Time            ::= Hours24 ":" Minutes
 * HHMMSS-Time          ::= Hours24 ":" Minutes ":" Seconds ("." Fraction)?
 * Years                ::= 4DIGIT;
 * Months               ::= 2DIGIT; range from 01 to 12
 * Days                 ::= 2DIGIT; range from 01 to 31
 * Hours24              ::= 2DIGIT; range from 00 to 23
 * 4DIGIT               ::= DIGIT DIGIT DIGIT DIGIT
 * TZD                  ::= "Z" | (("+" | "-") Hours24 ":" Minutes )
 * </pre>
 *
 * <p><b>History:</b><ul>
 * <li>13 Dec 2000 - Created file.
 * <li>08 Feb 2001 - Add escaping of id values for characters which are also key characters in timing attributes
 * <li>27 Feb 2001 - Add more event types SMIL 2.0 ones.
 *                 - Fix syncbase so that event types that begin with ".begin" or ".end" work ok.  (e.g. ".beginEvent")
 * <li>23 Mar 2001 - Add support for repeat time values.
 * </ul>
 *********************************************************************/
public class TimingAttribute {
	/*********************************************************************** 
	* Whether debug information is to be printed to the console. 
	* default is not have debug information printed (false).
	**********************************************************************/
	static private boolean debug_ = false;

	/*********************************************************************** 
	 * Various Strings used in timing values
	 **********************************************************************/
	static private final String CLICK = "click";
	static private final String MOUSEUP = "mouseup";
	static private final String MOUSEDOWN = "mousedown";
	static private final String MOUSEOUT = "mouseout";
	static private final String MOUSEOVER = "mouseover";
	static private final String MOUSEDRAG = "mousedrag";
	static private final String VIEWABLE = "viewable";
	static private final String NEAR = "near";
	static private final String COLLIDE = "collide";
	static private final String FOCUS_IN_EVENT = "focusInEvent";
	static private final String FOCUS_OUT_EVENT = "focusOutEvent";
	static private final String ACTIVATE_EVENT = "activateEvent";
	static private final String BEGIN_EVENT = "beginEvent";
	static private final String END_EVENT = "endEvent";
	static private final String IN_BOUNDS_EVENT = "inBoundsEvent";
	static private final String OUT_OF_BOUNDS_EVENT = "outOfBoundsEvent";
	static private final String REPEAT_EVENT = "repeatEvent";
	static private final String VIEWPORT_CLOSED_EVENT = "viewportClosedEvent";
	static private final String VIEWPORT_OPENED_EVENT = "viewportOpenedEvent";
	static private final String WALLCLOCK = "wallclock";
	static private final String INDEFINITE = "indefinite";
	static private final String MEDIA = "media";
	static private final String BEGIN = "begin";
	static private final String END = "end";
	static private final String PREV = "prev";
	static private final String ACCESSKEY = "accessKey";
	static private final String REPEAT = "repeat";
	static private final String MARKER = "marker";
	static private final String ACCESSKEY_UNIQUE_KEY = ACCESSKEY + "(";
	static private final String WALLCLOCK_UNIQUE_KEY = WALLCLOCK + "(";
	static private final String PREV_BEGIN_UNIQUE_KEY = PREV + "." + BEGIN;
	static private final String PREV_END_UNIQUE_KEY = PREV + "." + END;
	static private final String SYNCBASE_BEGIN_UNIQUE_KEY = "." + BEGIN;
	static private final String SYNCBASE_END_UNIQUE_KEY = "." + END;
	static private final String REPEAT_UNIQUE_KEY = REPEAT + "(";
	static private final String MARKER_UNIQUE_KEY = "." + MARKER + "(";

	/*********************************************************************** 
	 * Various Types of timing values
	 **********************************************************************/
	static public final int INVALID = 0;
	static public final int CLOCK_VALUE = 1;
	static public final int SYNCBASE_VALUE = 2;
	static public final int SYNC_TO_PREV_VALUE = 3;
	static public final int EVENT_VALUE = 4;
	static public final int REPEAT_VALUE = 5;
	static public final int ACCESS_KEY_VALUE = 6;
	static public final int MEDIA_MARKER_VALUE = 7;
	static public final int WALLCLOCK_VALUE = 8;
	static public final int INDEFINITE_VALUE = 9;
	static public final int MEDIA_VALUE = 10;
	private int timingValueType_ = INVALID;

	/*********************************************************************** 
	 **********************************************************************/
	private long clockValue_ = 0;

	/*********************************************************************** 
	 **********************************************************************/
	private String idValue_ = null;

	/*********************************************************************** 
	 * Various types of Syncbase values
	 **********************************************************************/
	static public final int TIME_SYMBOL_BEGIN = 1;
	static public final int TIME_SYMBOL_END = 2;
	private int timeSymbol_ = 0;

	/*********************************************************************** 
	 * Various Types of known event symbols
	 * There is not a complete list of these symbols since they can be host defined
	 * Node the values for the ids need to line up with the string array.
	 **********************************************************************/
	static public final int SYMBOL_EVENT_UNKNOWN = -1;
	static public final int SYMBOL_EVENT_CLICK = 0;
	static public final int SYMBOL_EVENT_MOUSEUP = 1;
	static public final int SYMBOL_EVENT_MOUSEDOWN = 2;
	static public final int SYMBOL_EVENT_MOUSEOUT = 3;
	static public final int SYMBOL_EVENT_MOUSEOVER = 4;
	static public final int SYMBOL_EVENT_MOUSEDRAG = 5;
	static public final int SYMBOL_EVENT_VIEWABLE = 6;
	static public final int SYMBOL_EVENT_NEAR = 7;
	static public final int SYMBOL_EVENT_COLLIDE = 8;
	static public final int SYMBOL_EVENT_FOCUS_IN_EVENT = 9;
	static public final int SYMBOL_EVENT_FOCUS_OUT_EVENT = 10;
	static public final int SYMBOL_EVENT_ACTIVATE_EVENT = 11;
	static public final int SYMBOL_EVENT_BEGIN_EVENT = 12;
	static public final int SYMBOL_EVENT_END_EVENT = 13;
	static public final int SYMBOL_EVENT_REPEAT_EVENT = 14;
	static public final int SYMBOL_EVENT_IN_BOUNDS_EVENT = 15;
	static public final int SYMBOL_EVENT_OUT_OF_BOUNDS_EVENT = 16;
	static public final int SYMBOL_EVENT_VIEWPORT_CLOSED_EVENT = 17;
	static public final int SYMBOL_EVENT_VIEWPORT_OPENED_EVENT = 18;
	private int eventSymbol_ = SYMBOL_EVENT_UNKNOWN;

	/*********************************************************************** 
	 **********************************************************************/
	static private final String eventSymbolStrings_[] = {
	/* */ CLICK,
	/* */ MOUSEUP,
	/* */ MOUSEDOWN,
	/* */ MOUSEOUT,
	/* */ MOUSEOVER,
	/* */ MOUSEDRAG,
	/* */ VIEWABLE,
	/* */ NEAR,
	/* */ COLLIDE,
	/* */ FOCUS_IN_EVENT,
	/* */ FOCUS_OUT_EVENT,
	/* */ ACTIVATE_EVENT,
	/* */ BEGIN_EVENT,
	/* */ END_EVENT,
	/* */ REPEAT_EVENT,
	/* */ IN_BOUNDS_EVENT,
	/* */ OUT_OF_BOUNDS_EVENT,
	/* */ VIEWPORT_CLOSED_EVENT,
	/* */ VIEWPORT_OPENED_EVENT,
	/* */};
	private String eventSymbolString_ = null;

	/*********************************************************************** 
	 **********************************************************************/
	private long repeatIteration_ = 0;

	/*********************************************************************** 
	 **********************************************************************/
	private char accessKeyCharacter_ = ' ';

	/*********************************************************************** 
	 **********************************************************************/
	private String mediaMarkerName_ = null;

	/*********************************************************************** 
	 **********************************************************************/
	private Date wallClock_ = null;

	/*********************************************************************** 
	 **********************************************************************/
	private Hashtable idTable_ = null;
	private Document idDocument_ = null;

	/*********************************************************************** 
	 **********************************************************************/
	private Vector warningMessages_ = null;
	private Vector errorMessages_ = null;

	/*********************************************************************** 
	 **********************************************************************/
	private ResourceBundle timingAttributeRes_ = ResourceBundle.getBundle("xmtutil.resources.TimingAttributeRes", Locale.getDefault());
	static public final int MULTI_VALUE = 11;
	/*********************************************************************** 
	 **********************************************************************/
	private Vector multiTimingAttributes_ = new Vector();
/*********************************************************************** 
 **********************************************************************/
public TimingAttribute(int timingValueType) {
	this(null, null, timingValueType, 0, null, 0, 0, 0, ' ', null, null);
}
/*********************************************************************** 
 **********************************************************************/
public TimingAttribute(int timingValueType, long clockValue) {
	this(null, null, timingValueType, clockValue, null, 0, 0, 0, ' ', null, null);
}
/*********************************************************************** 
 * Class to interpret the various timing that can occur in the timing attribute.
 *
 * <pre>
 * begin-value-list     ::= begin-value (S ";" S begin-value-list )?
 * begin-value          ::= (offset-value | syncbase-value | syncToPrev-value | event-value | repeat-value
 *                           | accessKey-value | media-marker-value | wallclock-sync-value | "indefinite" )                                          
 * 
 * S                    ::= (#x20 | #x9 | #xD | #xA)*
 * 
 * offset-value         ::= (( S "+" | "-" S )? ( Clock-value )
 *                     
 * Clock-value          ::= ( Full-clock-value | Partial-clock-value | Timecount-value )
 * Full-clock-value     ::= Hours ":" Minutes ":" Seconds ("." Fraction)?
 * Partial-clock-value  ::= Minutes ":" Seconds ("." Fraction)?
 * Timecount-value      ::= Timecount ("." Fraction)? (Metric)?
 * Metric               ::= "h" | "min" | "s" | "ms"
 * Hours                ::= DIGIT+; any positive number
 * Minutes              ::= 2DIGIT; range from 00 to 59
 * Seconds              ::= 2DIGIT; range from 00 to 59
 * Fraction             ::= DIGIT+
 * Timecount            ::= DIGIT+
 * 2DIGIT               ::= DIGIT DIGIT
 * DIGIT                ::= [0-9]
 *                     
 * Syncbase-value       ::= ( Syncbase-element "." Time-symbol ) ( S ("+"|"-") S Clock-value )? 
 * Syncbase-element     ::= Id-value
 * Time-symbol          ::= "begin" | "end" 
 * Id-value             ::= IDREF
 *                     
 * SyncToPrev-value     ::= ( "prev." Time-symbol ) ( S ("+"|"-") S Clock-value )?  
 *                     
 * Event-value          ::= ( Eventbase-element "." )? Event-symbol ( S ("+"|"-") S Clock-value )? 
 * Eventbase-element    ::= ID 
 *                     
 * Repeat-value         ::= ( Eventbase-element "." )? "repeat(" iteration ")" ( S ("+"|"-") S Clock-value )? 
 * iteration            ::= DIGIT+  
 *                     
 * AccessKey-value      ::= "accessKey(" character ")" ( S ("+"|"-") S Clock-value )? 
 * 
 * Media-Marker-value   ::= Id-value ".marker(" S marker-name S ")" ) 
 * 
 * wallclock-sync-value ::= "wallclock(" S (DateTime | WallTime)  S ")"
 * DateTime             ::= Date "T" WallTime
 * Date                 ::= Years "-" Months "-" Days
 * WallTime             ::= (HHMM-Time | HHMMSS-Time)(TZD)?
 * HHMM-Time            ::= Hours24 ":" Minutes
 * HHMMSS-Time          ::= Hours24 ":" Minutes ":" Seconds ("." Fraction)?
 * Years                ::= 4DIGIT;
 * Months               ::= 2DIGIT; range from 01 to 12
 * Days                 ::= 2DIGIT; range from 01 to 31
 * Hours24              ::= 2DIGIT; range from 00 to 23
 * 4DIGIT               ::= DIGIT DIGIT DIGIT DIGIT
 * TZD                  ::= "Z" | (("+" | "-") Hours24 ":" Minutes )
 * </pre>
 *********************************************************************/
public TimingAttribute(Hashtable tableIds, String param) {
	idTable_ = tableIds;

	//
	parseAttribute(param);
}
/*********************************************************************** 
 * Create a timing attribute from various parameters.
 *********************************************************************/
public TimingAttribute(Hashtable tableIds, Document docIds, int timingValueType, long clockValue, String idValue, int symbol) {
	idTable_ = tableIds;
	idDocument_ = docIds;
	boolean hasIdValue = false;
	switch (timingValueType) {
		case INDEFINITE_VALUE :
		case MEDIA_VALUE :
			timingValueType_ = timingValueType;
			break;
		case CLOCK_VALUE :
			clockValue_ = clockValue;
			timingValueType_ = timingValueType;
			break;
		case SYNCBASE_VALUE :
			switch (symbol) {
				case TIME_SYMBOL_BEGIN :
				case TIME_SYMBOL_END :
					break;
				default :
					addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_SYNCBASE_TIMING));
					return;
			}
			hasIdValue = true;
			idValue_ = idValue;
			timeSymbol_ = symbol;
			clockValue_ = clockValue;
			timingValueType_ = timingValueType;
			break;
		case EVENT_VALUE :
			if (0 > symbol || symbol >= eventSymbolStrings_.length) {
				MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_EVENT_SYMBOL));
				Object[] msgArgs = {eventSymbolString_};
				addErrorMessage(msgFormat.format(msgArgs));
				return;
			}
			hasIdValue = true;
			eventSymbol_ = symbol;
			eventSymbolString_ = eventSymbolStrings_[eventSymbol_];
			idValue_ = idValue;
			clockValue_ = clockValue;
			timingValueType_ = timingValueType;
			break;
		case SYNC_TO_PREV_VALUE :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_SYNC_TO_PREV_TIMING_NOT_SUPPORTED));
			timingValueType_ = INVALID;
			return;
		case REPEAT_VALUE :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_REPEAT_TIMING_NOT_SUPPORTED));
			timingValueType_ = INVALID;
			return;
		case ACCESS_KEY_VALUE :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_ACCESS_KEY_TIMING_NOT_SUPPORTED));
			timingValueType_ = INVALID;
			return;
		case MEDIA_MARKER_VALUE :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_MARKER_TIMING_NOT_SUPPORTED));
			timingValueType_ = INVALID;
			return;
		case WALLCLOCK_VALUE :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_WALL_CLOCK_TIMING_NOT_SUPPORTED));
			timingValueType_ = INVALID;
			return;
		default :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
			timingValueType_ = INVALID;
			return;
	}

	// Check if the id value is valid
	if (hasIdValue) {
		boolean error = false;
		if (null == idValue_)
			error = true;
		else {
			if (null != idTable_)
				if (!idTable_.containsKey(idValue_))
					error = true;
			if (null != idDocument_)
				if (null == idDocument_.getElementById(idValue_))
					error = true;
		}
		if (error) {
			MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_ID_VALUE));
			Object[] msgArgs = {idValue_};
			addErrorMessage(msgFormat.format(msgArgs));
			timingValueType_ = INVALID;
			return;
		}
	}
}
/*********************************************************************** 
 * Create a timing attribute from various parameters.
 *********************************************************************/
public TimingAttribute(
	Hashtable tableIds,
	Document docIds,
	int timingValueType,
	long clockValue,
	String idValue,
	int timeSymbol,
	int eventSymbol,
	long repeatInteration,
	char accessKeyCharacter,
	String mediaMarkerName,
	Date wallClock) {
	idTable_ = tableIds;
	idDocument_ = docIds;
	boolean hasIdValue = false;
	switch (timingValueType) {
		case INDEFINITE_VALUE :
		case MEDIA_VALUE :
			timingValueType_ = timingValueType;
			break;
		case CLOCK_VALUE :
			clockValue_ = clockValue;
			timingValueType_ = timingValueType;
			break;
		case SYNCBASE_VALUE :
			switch (timeSymbol) {
				case TIME_SYMBOL_BEGIN :
				case TIME_SYMBOL_END :
					break;
				default :
					addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_SYNCBASE_TIMING));
					return;
			}
			hasIdValue = true;
			idValue_ = idValue;
			timeSymbol_ = timeSymbol;
			clockValue_ = clockValue;
			timingValueType_ = timingValueType;
			break;
		case EVENT_VALUE :
			if (0 > eventSymbol || eventSymbol >= eventSymbolStrings_.length) {
				MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_EVENT_SYMBOL));
				Object[] msgArgs = { eventSymbolString_ };
				addErrorMessage(msgFormat.format(msgArgs));
				return;
			}
			hasIdValue = true;
			eventSymbol_ = eventSymbol;
			eventSymbolString_ = eventSymbolStrings_[eventSymbol_];
			idValue_ = idValue;
			clockValue_ = clockValue;
			timingValueType_ = timingValueType;
			break;
		case SYNC_TO_PREV_VALUE :
			switch (timeSymbol) {
				case TIME_SYMBOL_BEGIN :
				case TIME_SYMBOL_END :
					break;
				default :
					addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_SYNCBASE_TIMING));
					return;
			}
			timeSymbol_ = timeSymbol;
			clockValue_ = clockValue;
			timingValueType_ = timingValueType;
			return;
		case REPEAT_VALUE :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_REPEAT_TIMING_NOT_SUPPORTED));
			timingValueType_ = INVALID;
			return;
		case ACCESS_KEY_VALUE :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_ACCESS_KEY_TIMING_NOT_SUPPORTED));
			timingValueType_ = INVALID;
			return;
		case MEDIA_MARKER_VALUE :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_MARKER_TIMING_NOT_SUPPORTED));
			timingValueType_ = INVALID;
			return;
		case WALLCLOCK_VALUE :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_WALL_CLOCK_TIMING_NOT_SUPPORTED));
			timingValueType_ = INVALID;
			return;
		default :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
			timingValueType_ = INVALID;
			return;
	}

	// Check if the id value is valid
	if (hasIdValue) {
		boolean error = false;
		if (null == idValue_)
			error = true;
		else {
			if (null != idTable_)
				if (!idTable_.containsKey(idValue_))
					error = true;
			if (null != idDocument_)
				if (null == idDocument_.getElementById(idValue_))
					error = true;
		}
		if (error) {
			MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_ID_VALUE));
			Object[] msgArgs = { idValue_ };
			addErrorMessage(msgFormat.format(msgArgs));
			timingValueType_ = INVALID;
			return;
		}
	}
}
/*********************************************************************** 
 * Create a timing attribute from various parameters.
 *********************************************************************/
public TimingAttribute(Vector timingAttributes) {
	//
	addTimingAttributesFromVector(timingAttributes);

	// Now make sure it worked ok
	switch (multiTimingAttributes_.size()) {
		case 0 :
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
			break;
		case 1 :
			TimingAttribute time = (TimingAttribute) multiTimingAttributes_.firstElement();
			multiTimingAttributes_.clear();
			clockValue_ = time.getClockValue();
			idValue_ = time.getIdValue();
			timeSymbol_ = time.getTimeSymbol();
			eventSymbol_ = time.getEventSymbol();
			eventSymbolString_ = time.getEventSymbolString();
			repeatIteration_ = time.getRepeatIteration();
			accessKeyCharacter_ = time.getAccessKeyCharacter();
			mediaMarkerName_ = time.getMarkerName();
			wallClock_ = time.getWallClock();
			timingValueType_ = time.getTimeValueType();
			break;
		default :
			timingValueType_ = MULTI_VALUE;
			break;
	}
}
/*********************************************************************** 
 * Class to interpret the various timing that can occur in the timing attribute.
 *********************************************************************/
public TimingAttribute(Document docIds,  String param) {
	idDocument_ = docIds;

	//
	parseAttribute(param);
}
/*********************************************************************** 
 * Determine whether this is a normal time value attribute.
 *
 * @return whether this is a normal time value attribute.
 **********************************************************************/
public TimingAttribute add(TimingAttribute time) {
	if (INDEFINITE_VALUE == timingValueType_)
		return this;
	if (INDEFINITE_VALUE == time.getTimeValueType())
		return time;

	// Do some minimal checking to see if we support the timing model
	if (!(CLOCK_VALUE == timingValueType_ || CLOCK_VALUE == time.getTimeValueType()))
		return null;

	//
	TimingAttribute baseTime;
	long additionalClockValue;
	if (CLOCK_VALUE != timingValueType_) {
		baseTime = time;
		additionalClockValue = clockValue_;
	} else {
		baseTime = this;
		additionalClockValue = time.getClockValue();
	}

	//
	if (0 == additionalClockValue)
		return baseTime;

	//
	switch (baseTime.getTimeValueType()) {
		case CLOCK_VALUE :
			return new TimingAttribute(baseTime.getTimeValueType(), baseTime.getClockValue() + additionalClockValue);
		case SYNCBASE_VALUE :
			return new TimingAttribute(null, null, baseTime.getTimeValueType(), baseTime.getClockValue() + additionalClockValue, baseTime.getIdValue(), baseTime.getTimeSymbol());
		case EVENT_VALUE :
			return new TimingAttribute(null, null, baseTime.getTimeValueType(), baseTime.getClockValue() + additionalClockValue, baseTime.getIdValue(), baseTime.getEventSymbol());
	}


	//
	return null;
}
/*********************************************************************** 
 **********************************************************************/
private void addErrorMessage(String message) {
	if (null == message || message.equals(""))
		return;

	//
	if (null == errorMessages_)
		errorMessages_ = new Vector();

	//
	errorMessages_.add(message);
}
/*********************************************************************** 
 * Create a timing attribute from various parameters.
 *********************************************************************/
private void addTimingAttributesFromVector(Vector timingAttributes) {
	//
	StringBuffer buf = new StringBuffer();
	for (Enumeration enum = timingAttributes.elements(); enum.hasMoreElements();) {
		Object obj = enum.nextElement();
		if (!(obj instanceof TimingAttribute)) {
			MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_MULTI_TIMING_ATTRIBUTE_ELEMENT));
			Object[] msgArgs = {obj};
			addErrorMessage(msgFormat.format(msgArgs));
		} else {
			TimingAttribute time = (TimingAttribute) obj;
			switch (time.getTimeValueType()) {
				case MULTI_VALUE :
					addTimingAttributesFromVector(time.getMultiTimingAttributes());
					break;
				case INVALID :
					MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_MULTI_TIMING_ATTRIBUTE_ELEMENT));
					Object[] msgArgs = {time};
					addErrorMessage(msgFormat.format(msgArgs));
					// append messages
					Vector messages = time.getErrorMessages();
					if (null != messages)
						for (Enumeration msgEnum = messages.elements(); msgEnum.hasMoreElements();)
							addErrorMessage((String) msgEnum.nextElement());
					messages = time.getWarningMessages();
					if (null != messages)
						for (Enumeration msgEnum = messages.elements(); msgEnum.hasMoreElements();)
							addWarningMessage((String) msgEnum.nextElement());
					break;
				default :
					multiTimingAttributes_.add(time);
					break;
			}
		}
	}
}
/*********************************************************************** 
 **********************************************************************/
private void addWarningMessage(String message) {
	if (null == message || message.equals(""))
		return;

	//
	if (null == warningMessages_)
		warningMessages_ = new Vector();

	//
	warningMessages_.add(message);
}
/*********************************************************************** 
 * Determine whether this is a normal time value attribute.
 *
 * @return whether this is a normal time value attribute.
 **********************************************************************/
public boolean equals(long time) {
	//	
	switch (timingValueType_) {
		case CLOCK_VALUE :
			return (time == clockValue_);
	}

	//
	return false;
}
/*********************************************************************** 
 * Determine whether this is a normal time value attribute.
 *
 * @return whether this is a normal time value attribute.
 **********************************************************************/
public boolean equals(TimingAttribute time) {
	//	
	if (timingValueType_ == time.getTimeValueType()) {
		switch (timingValueType_) {
			case INDEFINITE_VALUE :
			case MEDIA_VALUE :
				return true;
			case CLOCK_VALUE :
				return (time.getClockValue() == clockValue_);
			case SYNCBASE_VALUE :
				if (time.getTimeSymbol() == timeSymbol_ && null != idValue_ && null != time.getIdValue() && idValue_.equals(time.getIdValue()))
					return (time.getClockValue() == clockValue_);
				break;
			case EVENT_VALUE :
				if (time.getEventSymbol() == eventSymbol_ && null != idValue_ && null != time.getIdValue() && idValue_.equals(time.getIdValue()))
					return (time.getClockValue() == clockValue_);
				break;
		}
	}

	//
	return false;
}
/***********************************************************************
 * Add the escape characters for ids.
 * <p>
 * NOTE: the ";", "+", "(", ")" and "\" are not valid characters in XML ids.  This is important when
 * looking for delimiter variables in ids.  Other delimiter characters "-", ":" and "." are valid
 * characters in XML ids.  In order to use them in ids located in TimingAttributes they need to be
 * escaped with the "\" character.  Thus they would appear in the id part of the attribute string as 
 * "\-", "\:" and "\.".
 *********************************************************************/
static private String escapeCharactersAdd(String id) {
	char[] escapedCharacters = {'.', ':', '-',};
	String str = id;

	//
	for (int i = 0; i < escapedCharacters.length; i++) {
		StringBuffer sb = new StringBuffer();
		int prev = 0;
		int next = str.indexOf(escapedCharacters[i], prev);
		while (-1 < next) {
			sb.append(str.substring(prev, next)).append('\\');
			prev = next;
			next = str.indexOf(escapedCharacters[i], prev + 1);
		}
		sb.append(str.substring(prev));
		str = sb.toString();
	}
	return str;
}
/***********************************************************************
 * Remove the escape characters for ids.
 * <p>
 * NOTE: the ";", "+", "(", ")" and "\" are not valid characters in XML ids.  This is important when
 * looking for delimiter variables in ids.  Other delimiter characters "-", ":" and "." are valid
 * characters in XML ids.  In order to use them in ids located in TimingAttributes they need to be
 * escaped with the "\" character.  Thus they would appear in the id part of the attribute string as 
 * "\-", "\:" and "\.".
 *********************************************************************/
static private String escapeCharactersStrip(String id) {
	StringBuffer sb = new StringBuffer();
	int prev = 0;
	int next = id.indexOf('\\', prev);
	while (-1 < next) {
		sb.append(id.substring(prev, next));
		prev = next + 1;
		next = id.indexOf('\\', prev);
	}
	sb.append(id.substring(prev));
	return sb.toString();
}
/*********************************************************************** 
 **********************************************************************/
private char getAccessKeyCharacter() {
	return accessKeyCharacter_;
}
/*********************************************************************** 
 * Determine whether this is a normal time value attribute.
 *
 * @return whether this is a normal time value attribute.
 **********************************************************************/
public long getClockValue() {
	return clockValue_;
}
/*********************************************************************** 
 **********************************************************************/
public Vector getErrorMessages() {
	//
	if (null == errorMessages_)
		return null;

	//
	return (Vector) errorMessages_.clone();
}
/*********************************************************************** 
 **********************************************************************/
public int getEventSymbol() {
	return eventSymbol_;
}
/*********************************************************************** 
 **********************************************************************/
public String getEventSymbolString() {
	return eventSymbolString_;
}
/*********************************************************************** 
 **********************************************************************/
public String getIdValue() {
	return idValue_;
}
/*********************************************************************** 
 **********************************************************************/
private String getMarkerName() {
	return mediaMarkerName_;
}
/*********************************************************************** 
 * Determine whether this is a normal time value attribute.
 *
 * @return whether this is a normal time value attribute.
 **********************************************************************/
public Vector getMultiTimingAttributes() {
	return (Vector) multiTimingAttributes_.clone();
}
/*********************************************************************** 
 **********************************************************************/
public long getRepeatIteration() {
	return repeatIteration_;
}
/*********************************************************************** 
 **********************************************************************/
public int getTimeSymbol() {
	return timeSymbol_;
}
/*********************************************************************** 
 **********************************************************************/
public int getTimeValueType() {
	return timingValueType_;
}
/*********************************************************************** 
 **********************************************************************/
private Date getWallClock() {
	return wallClock_;
}
/*********************************************************************** 
 **********************************************************************/
public Vector getWarningMessages() {
	//
	if (null == warningMessages_)
		return null;

	//
	return (Vector) warningMessages_.clone();
}
/*********************************************************************** 
 **********************************************************************/
public boolean greaterThan(long time) {
	//	
	switch (timingValueType_) {
		case CLOCK_VALUE :
			return (clockValue_ > time);
	}

	//
	return false;
}
/*********************************************************************** 
 **********************************************************************/
public boolean greaterThan(TimingAttribute time) {
	if (INDEFINITE_VALUE == timingValueType_ && INDEFINITE_VALUE != time.getTimeValueType())
		return true;

	//	
	if (timingValueType_ == time.getTimeValueType()) {
		switch (timingValueType_) {
			case CLOCK_VALUE :
				return (clockValue_ > time.getClockValue());
			case SYNCBASE_VALUE :
				if (time.getTimeSymbol() == timeSymbol_ && null != idValue_ && null != time.getIdValue() && idValue_.equals(time.getIdValue()))
					return (clockValue_ > time.getClockValue());
				break;
			case EVENT_VALUE :
				if (time.getEventSymbol() == eventSymbol_ && null != idValue_ && null != time.getIdValue() && idValue_.equals(time.getIdValue()))
					return (clockValue_ > time.getClockValue());
				break;
		}
	}

	//
	return false;
}
/*********************************************************************** 
 **********************************************************************/
public boolean lessThan(long time) {
	//	
	switch (timingValueType_) {
		case CLOCK_VALUE :
			return (clockValue_ < time);
	}

	//
	return false;
}
/*********************************************************************** 
 **********************************************************************/
public boolean lessThan(TimingAttribute time) {
	if (INDEFINITE_VALUE != timingValueType_ && INDEFINITE_VALUE == time.getTimeValueType())
		return true;

	//	
	if (timingValueType_ == time.getTimeValueType()) {
		switch (timingValueType_) {
			case CLOCK_VALUE :
				return (clockValue_ < time.getClockValue());
			case SYNCBASE_VALUE :
				if (time.getTimeSymbol() == timeSymbol_ && null != idValue_ && null != time.getIdValue() && idValue_.equals(time.getIdValue()))
					return (clockValue_ < time.getClockValue());
				break;
			case EVENT_VALUE :
				if (time.getEventSymbol() == eventSymbol_ && null != idValue_ && null != time.getIdValue() && idValue_.equals(time.getIdValue()))
					return (clockValue_ < time.getClockValue());
				break;
		}
	}

	//
	return false;
}
/*********************************************************************** 
 * Class to interpret the various timing that can occur in the timing attribute.
 * NOTE: the grammar for the dur is not explicitly so it has been generated here
 * to be based on the spec.
 *
 * <pre>
 * begin-value-list     ::= begin-value (S ";" S begin-value-list )?
 * begin-value          ::= (offset-value | syncbase-value | syncToPrev-value | event-value | repeat-value
 *                           | accessKey-value | media-marker-value | wallclock-sync-value | "indefinite" )
 * dur-value            ::= ( S ) ( Clock-value | "media" | "indefinite" ) ( S )
 * 
 * S                    ::= (#x20 | #x9 | #xD | #xA)*
 * 
 * offset-value         ::= (( S "+" | "-" S )? ( Clock-value )
 *                     
 * Syncbase-value       ::= ( Syncbase-element "." Time-symbol ) ( S ("+"|"-") S Clock-value )? 
 * Syncbase-element     ::= Id-value
 * Time-symbol          ::= "begin" | "end" 
 * Id-value             ::= IDREF
 *                     
 * SyncToPrev-value     ::= ( "prev." Time-symbol ) ( S ("+"|"-") S Clock-value )?  
 *                     
 * Event-value          ::= ( Eventbase-element "." )? Event-symbol ( S ("+"|"-") S Clock-value )? 
 * Eventbase-element    ::= ID 
 *                     
 * Repeat-value         ::= ( Eventbase-element "." )? "repeat(" iteration ")" ( S ("+"|"-") S Clock-value )? 
 * iteration            ::= DIGIT+  
 *                     
 * AccessKey-value      ::= "accessKey(" character ")" ( S ("+"|"-") S Clock-value )? 
 * 
 * Media-Marker-value   ::= Id-value ".marker(" S marker-name S ")" ) 
 * 
 * wallclock-sync-value ::= "wallclock(" S (DateTime | WallTime)  S ")"
 * </pre>
 *********************************************************************/
private boolean parseAttribute(String param) {
	timingValueType_ = INVALID;
	if (null == param) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	// Do some minimal checking to see if we support the timing model
	String timeValue = param;

	// First check if this is multiple timing values, if so only use the
	// first one.
	int semicolonPosition = param.indexOf(";");
	if (-1 != semicolonPosition) {
		boolean success = parseMultiValues(timeValue);
		if (success) {
			//
			if (1 == multiTimingAttributes_.size()) {
				TimingAttribute time = (TimingAttribute) multiTimingAttributes_.firstElement();
				multiTimingAttributes_.clear();
				clockValue_ = time.getClockValue();
				idValue_ = time.getIdValue();
				timeSymbol_ = time.getTimeSymbol();
				eventSymbol_ = time.getEventSymbol();
				eventSymbolString_ = time.getEventSymbolString();
				repeatIteration_ = time.getRepeatIteration();
				accessKeyCharacter_ = time.getAccessKeyCharacter();
				mediaMarkerName_ = time.getMarkerName();
				wallClock_ = time.getWallClock();
				timingValueType_ = time.getTimeValueType();
			} else
				timingValueType_ = MULTI_VALUE;
		}
		return success;
	}

	//
	timeValue = timeValue.trim();

	//
	if (timeValue.equals("")) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	// do the easy checks first.
	// Excact value check
	if (timeValue.equals(INDEFINITE)) {
		timingValueType_ = INDEFINITE_VALUE;
		return true;
	}
	if (timeValue.equals(MEDIA)) {
		timingValueType_ = MEDIA_VALUE;
		return true;
	}
	// values checking at the begining of the string are faster to check
	if (timeValue.startsWith(PREV_BEGIN_UNIQUE_KEY) || timeValue.startsWith(PREV_END_UNIQUE_KEY)) {
		boolean success = parseSyncToPrevValue(timeValue);
		if (success)
			timingValueType_ = SYNC_TO_PREV_VALUE;
		return success;
	}
	if (timeValue.startsWith(ACCESSKEY_UNIQUE_KEY)) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_ACCESS_KEY_TIMING_NOT_SUPPORTED));
		return false;
	}
	if (timeValue.startsWith(WALLCLOCK_UNIQUE_KEY)) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_WALL_CLOCK_TIMING_NOT_SUPPORTED));
		return false;
	}

	// The next checks have to be searched throughout the sting
	if (-1 != timeValue.indexOf(REPEAT_UNIQUE_KEY)) {
		boolean success = parseRepeatValue(timeValue);
		if (success)
			timingValueType_ = REPEAT_VALUE;
		return success;
	}
	if (-1 != timeValue.indexOf(MARKER_UNIQUE_KEY)) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_MARKER_TIMING_NOT_SUPPORTED));
		return false;
	}

	// Check for Syncbase value.  First see if we can find the unique identifier keys.  Also check if this isn't an
	// escaped key.  We have to check to make sure that the period at the begining of the string isn't an escaped 
	// period (preceded by a backslash.
	int index = timeValue.lastIndexOf(SYNCBASE_BEGIN_UNIQUE_KEY);
	// check if an escaped period "\\.begin"
	if (0 < index && '\\' == timeValue.charAt(index - 1))
		index = -1;
	if (0 < index) {
		// check if an event that begins with ".begin" (e.g. ".beginEvent"). 
		// check by making sure the next character is is not a numeric, plus, minus or space character
		int nextCharIndex = index + SYNCBASE_BEGIN_UNIQUE_KEY.length();
		if (timeValue.length() > nextCharIndex) {
			switch (timeValue.charAt(nextCharIndex)) {
				case '1' :
				case '2' :
				case '3' :
				case '4' :
				case '5' :
				case '6' :
				case '7' :
				case '8' :
				case '9' :
				case '0' :
				case '+' :
				case '-' :
				case ' ' : // #x20
				case '\t' : // #x9
				case '\r' : // #xD
				case '\n' : // #xA
					break;
				default :
					index = -1;
					break;
			}
		}
	}
	if (-1 == index) {
		index = timeValue.lastIndexOf(SYNCBASE_END_UNIQUE_KEY);
		// check if an escaped period "\\.end"
		if (0 < index && '\\' == timeValue.charAt(index - 1))
			index = -1;
		if (0 < index) {
			// check if an event that begins with ".end" (e.g. ".endEvent"). 
			// check by making sure the next character is is not a numeric, plus, minus or space character
			int nextCharIndex = index + SYNCBASE_END_UNIQUE_KEY.length();
			if (timeValue.length() > nextCharIndex) {
				switch (timeValue.charAt(nextCharIndex)) {
					case '1' :
					case '2' :
					case '3' :
					case '4' :
					case '5' :
					case '6' :
					case '7' :
					case '8' :
					case '9' :
					case '0' :
					case '+' :
					case '-' :
					case ' ' : // #x20
					case '\t' : // #x9
					case '\r' : // #xD
					case '\n' : // #xA
						break;
					default :
						index = -1;
						break;
				}
			}
		}
	}
	if (-1 != index) {
		boolean success = parseSyncbaseValue(timeValue);
		if (success)
			timingValueType_ = SYNCBASE_VALUE;
		return success;
	}

	// Need to check if event based timing (first character not a +,- or digit)
	switch (timeValue.charAt(0)) {
		case '+' :
		case '-' :
		case '0' :
		case '1' :
		case '2' :
		case '3' :
		case '4' :
		case '5' :
		case '6' :
		case '7' :
		case '8' :
		case '9' :
			{
				boolean success = parseClockValue(timeValue);
				if (success)
					timingValueType_ = CLOCK_VALUE;
				return success;
			}
		default :
			boolean success = parseEventValue(timeValue);
			if (success)
				timingValueType_ = EVENT_VALUE;
			return success;
	}
}
/*********************************************************************** 
 * Class to interpret the various timing that can occur in the timing attribute.
 *
 * <pre>
 * offset-value         ::= (( S "+" | "-" S )? ( Clock-value )
 *                     
 * Clock-value          ::= ( Full-clock-value | Partial-clock-value | Timecount-value )
 * Full-clock-value     ::= Hours ":" Minutes ":" Seconds ("." Fraction)?
 * Partial-clock-value  ::= Minutes ":" Seconds ("." Fraction)?
 * Timecount-value      ::= Timecount ("." Fraction)? (Metric)?
 * Metric               ::= "h" | "min" | "s" | "ms"
 * Hours                ::= DIGIT+; any positive number
 * Minutes              ::= 2DIGIT; range from 00 to 59
 * Seconds              ::= 2DIGIT; range from 00 to 59
 * Fraction             ::= DIGIT+
 * Timecount            ::= DIGIT+
 * 2DIGIT               ::= DIGIT DIGIT
 * DIGIT                ::= [0-9]
 * </pre>
 *********************************************************************/
private boolean parseClockValue(String param) {
	if (null == param) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	//
	String timeValue = param.trim();
	if (timeValue.equals("")) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	//
	boolean isNegative = false;
	if ('-' == timeValue.charAt(0) || '+' == timeValue.charAt(0)) {
		isNegative = ('-' == timeValue.charAt(0));
		// Remove the sign and get rid of the extra spaces.
		// This needs to be for both '-' and '+'
		timeValue = timeValue.substring(1);
		timeValue = timeValue.trim();
		if (timeValue.equals("")) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
			return false;
		}
	}

	//
	try {
		int firstColonPosition = timeValue.indexOf(':');
		if (-1 != firstColonPosition) {
			// See if there is a second colon
			int secondColonPosition = timeValue.indexOf(':', firstColonPosition + 1);
			int hrMinColonPosition = (-1 != secondColonPosition) ? firstColonPosition : -1;
			int minSecColonPosition = (-1 != secondColonPosition) ? secondColonPosition : firstColonPosition;

			// determine the hours 
			float hours = 0;
			if (-1 != hrMinColonPosition) {
				String hrs = timeValue.substring(0, hrMinColonPosition);
				hours = Long.parseLong(hrs);
			}

			// determine the minutes 
			String mins = timeValue.substring(hrMinColonPosition + 1, minSecColonPosition);
			float minutes = Long.parseLong(mins);

			// determine the seconds
			String secs = timeValue.substring(minSecColonPosition + 1);
			float seconds = Float.parseFloat(secs);

			// Set the time value
			Float value = new Float((hours * (60 * 60 * 1000)) + (minutes * (60 * 1000)) + (seconds * (1000)));
			clockValue_ = value.longValue();
		} else {
			// this is a timecount clock value
			int multiplier = 1000;
			int truncateCharacters = 0;
			if (timeValue.endsWith("ms")) {
				truncateCharacters = 2;
				multiplier = 1;
			} else
				if (timeValue.endsWith("s")) {
					truncateCharacters = 1;
					multiplier = 1000;
				} else
					if (timeValue.endsWith("min")) {
						truncateCharacters = 3;
						multiplier = 1000 * 60;
					} else
						if (timeValue.endsWith("h")) {
							truncateCharacters = 1;
							multiplier = 1000 * 60 * 60;
						}

						//
			float floatTime = Float.parseFloat(timeValue.substring(0, timeValue.length() - truncateCharacters));
			Float value = new Float(floatTime * multiplier);
			clockValue_ = value.longValue();
		}

		// if is negative then need to multiply by -1
		if (isNegative)
			clockValue_ *= -1;

		//
		return true;
	} catch (NumberFormatException e) {
		if (debug_) {
			String msg = "TimingAttribute constructor() - timeValue=\"" + timeValue + "\" NumberFormatException=" + e;
			System.out.println(msg);
			e.printStackTrace();
		}
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}
}
/*********************************************************************** 
 * Class to interpret the various timing that can occur in the timing attribute.
 *
 * <pre>
 * Event-value          ::= ( Eventbase-element "." )? Event-symbol ( S ("+"|"-") S Clock-value )? 
 * Eventbase-element    ::= ID 
 * </pre>
 *********************************************************************/
private boolean parseEventValue(String param) {
	if (null == param) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	//
	String timeValue = param.trim();
	if (timeValue.equals("")) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	// First check if we have a clock value at the end (we would then have a + or - sign
	int signPosition = timeValue.lastIndexOf('+');
	if (-1 == signPosition) {
		signPosition = timeValue.lastIndexOf('-');
		// check if an escaped minus "\\-"
		if (0 < signPosition && '\\' == timeValue.charAt(signPosition - 1))
			signPosition = -1;
	}

	// If we have a sign then we definitely have an additional clock value. 
	// Thus determine the clock value and continue on
	if (-1 != signPosition) {
		boolean success = parseClockValue(timeValue.substring(signPosition));
		if (!success)
			return false;

		// get rid of the clock value and trim off the excess from the string
		timeValue = timeValue.substring(0, signPosition);
		timeValue = timeValue.trim();
	}

	// First check if there is an Eventbase-element
	// This indicated by having a period '.' in the remaining string
	// This is the last period in the string, since id's can have periods in them.
	int periodPosition = timeValue.lastIndexOf('.');

	// if we have the period then we have an eventbase-element
	if (-1 != periodPosition) {
		String idValue = timeValue.substring(0, periodPosition);
		idValue_ = escapeCharactersStrip(idValue);
		// If we have a hash table or a document check that the id exists in it.
		// If it doesn't then give a warning message
		if (null != idTable_) {
			if (!idTable_.containsKey(idValue_)) {
				MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_ID_VALUE));
				Object[] msgArgs = {idValue_};
				addErrorMessage(msgFormat.format(msgArgs));
				return false;
			}
		}
		if (null != idDocument_) {
			if (null == idDocument_.getElementById(idValue_)) {
				MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_ID_VALUE));
				Object[] msgArgs = {idValue_};
				addErrorMessage(msgFormat.format(msgArgs));
				return false;
			}
		}

		// get rid of the eventbase element from the string
		timeValue = timeValue.substring(periodPosition + 1);
	}

	// What should be left is the event type
	eventSymbolString_ = timeValue;

	// Now go through the known event symbols and see if this string is one of them.
	for (int i = 0; i < eventSymbolStrings_.length; i++) {
		if (eventSymbolStrings_[i].equals(eventSymbolString_)) {
			eventSymbol_ = i;
			return true;
		}
	}

	// We didn't find the event so give a message
	eventSymbol_ = SYMBOL_EVENT_UNKNOWN;
	MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_EVENT_SYMBOL));
	Object[] msgArgs = {eventSymbolString_};
	addErrorMessage(msgFormat.format(msgArgs));
	return false;
}
/*********************************************************************** 
 * Class to interpret the various timing that can occur in the timing attribute.
 *
 * <pre>
 * begin-value-list     ::= begin-value (S ";" S begin-value-list )?
 * begin-value          ::= (offset-value | syncbase-value | syncToPrev-value | event-value | repeat-value
 *                           | accessKey-value | media-marker-value | wallclock-sync-value | "indefinite" )
 * 
 * S                    ::= (#x20 | #x9 | #xD | #xA)*
 * </pre>
 *********************************************************************/
private boolean parseMultiValues(String param) {
	if (null == param) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	//
	String timeValue = param.trim();
	if (timeValue.equals("")) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	//
	Vector messages = null;
	TimingAttribute time = null;
	int semicolonPosition = param.indexOf(";");
	while (-1 != semicolonPosition) {
		String timeString = timeValue.substring(0, semicolonPosition);
		if (null != idTable_)
			time = new TimingAttribute(idTable_, timeString);
		else
			time = new TimingAttribute(idDocument_, timeString);

		// Only add it if it was good
		if (INVALID != time.getTimeValueType())
			multiTimingAttributes_.add(time);

		// append messages
		messages = time.getErrorMessages();
		if (null != messages)
			for (Enumeration enum = messages.elements(); enum.hasMoreElements();)
				addErrorMessage((String) enum.nextElement());
		messages = time.getWarningMessages();
		if (null != messages)
			for (Enumeration enum = messages.elements(); enum.hasMoreElements();)
				addWarningMessage((String) enum.nextElement());

		// look for next one
		timeValue = timeValue.substring(semicolonPosition + 1);
		timeValue = timeValue.trim();
		semicolonPosition = timeValue.indexOf(";");
	}

	//
	if (null != timeValue && !timeValue.equals("")) {
		if (null != idTable_)
			time = new TimingAttribute(idTable_, timeValue);
		else
			time = new TimingAttribute(idDocument_, timeValue);

		// Only add it if it was good
		if (INVALID != time.getTimeValueType())
			multiTimingAttributes_.add(time);

		// append messages
		messages = time.getErrorMessages();
		if (null != messages)
			for (Enumeration enum = messages.elements(); enum.hasMoreElements();)
				addErrorMessage((String) enum.nextElement());
		messages = time.getWarningMessages();
		if (null != messages)
			for (Enumeration enum = messages.elements(); enum.hasMoreElements();)
				addWarningMessage((String) enum.nextElement());
	}

	//
	if (0 == multiTimingAttributes_.size())
		return false;

	//
	return true;
}
/*********************************************************************** 
 * Class to interpret the various timing that can occur in the timing attribute.
 *
 * <pre>
 * Repeat-value         ::= ( Eventbase-element "." )? "repeat(" iteration ")" ( S ("+"|"-") S Clock-value )? 
 * Eventbase-element    ::= ID 
 * iteration            ::= DIGIT+  
 * </pre>
 *********************************************************************/
private boolean parseRepeatValue(String param) {
	if (null == param) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	//
	String timeValue = param.trim();
	if (timeValue.equals("")) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	// First check if we have a clock value at the end (we would then have a + or - sign
	int signPosition = timeValue.lastIndexOf('+');
	if (-1 == signPosition) {
		signPosition = timeValue.lastIndexOf('-');
		// check if an escaped minus "\\-"
		if (0 < signPosition && '\\' == timeValue.charAt(signPosition - 1))
			signPosition = -1;
	}

	// If we have a sign then we definitely have an additional clock value. 
	// Thus determine the clock value and continue on
	if (-1 != signPosition) {
		boolean success = parseClockValue(timeValue.substring(signPosition));
		if (!success)
			return false;

		// get rid of the clock value and trim off the excess from the string
		timeValue = timeValue.substring(0, signPosition);
		timeValue = timeValue.trim();
	}

	// First check if there is an Eventbase-element
	// This indicated by having a period '.' in the remaining string
	// This is the last period in the string, since id's can have periods in them.
	int periodPosition = timeValue.lastIndexOf('.');

	// if we have the period then we have an eventbase-element
	if (-1 != periodPosition) {
		String idValue = timeValue.substring(0, periodPosition);
		idValue_ = escapeCharactersStrip(idValue);
		// If we have a hash table or a document check that the id exists in it.
		// If it doesn't then give a warning message
		if (null != idTable_) {
			if (!idTable_.containsKey(idValue_)) {
				MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_ID_VALUE));
				Object[] msgArgs = {idValue_};
				addErrorMessage(msgFormat.format(msgArgs));
				return false;
			}
		}
		if (null != idDocument_) {
			if (null == idDocument_.getElementById(idValue_)) {
				MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_ID_VALUE));
				Object[] msgArgs = {idValue_};
				addErrorMessage(msgFormat.format(msgArgs));
				return false;
			}
		}

		// get rid of the eventbase element from the string
		timeValue = timeValue.substring(periodPosition + 1);
	}

	// What should be left is the repeat iteration
	int leftParenPosition = timeValue.lastIndexOf('(');
	int rightParenPosition = timeValue.lastIndexOf(')');
	if (rightParenPosition < rightParenPosition) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	// get the iteration string
	timeValue = timeValue.substring(leftParenPosition+1, rightParenPosition);
	timeValue = timeValue.trim();
	if (timeValue.equals("")) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	// get the repeat value value
	try {
		repeatIteration_ = Long.parseLong(timeValue);
		return true;
	} catch (NumberFormatException e) {
		if (debug_) {
			String msg = "TimingAttribute constructor() - timeValue=\"" + timeValue + "\" NumberFormatException=" + e;
			System.out.println(msg);
			e.printStackTrace();
		}
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}
}
/*********************************************************************** 
 * Class to interpret the various timing that can occur in the timing attribute.
 *
 * <pre>
 * Syncbase-value       ::= ( Syncbase-element "." Time-symbol ) ( S ("+"|"-") S Clock-value )? 
 * Syncbase-element     ::= Id-value
 * Time-symbol          ::= "begin" | "end" 
 * Id-value             ::= IDREF
 * </pre>
 *********************************************************************/
private boolean parseSyncbaseValue(String param) {
	if (null == param) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	//
	String timeValue = param.trim();
	if (timeValue.equals("")) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	// First check if we have a clock value at the end (we would then have a + or - sign
	int signPosition = timeValue.lastIndexOf('+');
	if (-1 == signPosition) {
		signPosition = timeValue.lastIndexOf('-');
		// check if an escaped minus "\\-"
		if (0 < signPosition && '\\' == timeValue.charAt(signPosition - 1))
			signPosition = -1;
	}

	// If we have a sign then we definitely have an additional clock value. 
	// Thus determine the clock value and continue on
	if (-1 != signPosition) {
		boolean success = parseClockValue(timeValue.substring(signPosition));
		if (!success)
			return false;

		// get rid of the clock value and trim off the excess from the string
		timeValue = timeValue.substring(0, signPosition);
		timeValue = timeValue.trim();
	}

	// First check if there is an Eventbase-element
	// This indicated by having a period '.' in the remaining string
	// This is the last period in the string, since id's can have periods in them.
	int periodPosition = timeValue.lastIndexOf('.');

	// if we don't have a period then we have an invalid syncbase element
	if (-1 == periodPosition) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_SYNCBASE_TIMING));
		return false;
	}

	//
	String idValue = timeValue.substring(0, periodPosition);
	idValue_ = escapeCharactersStrip(idValue);
	// If we have a hash table or a document check that the id exists in it.
	// If it doesn't then give a warning message
	if (null != idTable_) {
		if (!idTable_.containsKey(idValue_)) {
			MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_ID_VALUE));
			Object[] msgArgs = {idValue_};
			addErrorMessage(msgFormat.format(msgArgs));
			return false;
		}
	}
	if (null != idDocument_) {
		if (null == idDocument_.getElementById(idValue_)) {
			MessageFormat msgFormat = new MessageFormat(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_ID_VALUE));
			Object[] msgArgs = {idValue_};
			addErrorMessage(msgFormat.format(msgArgs));
			return false;
		}
	}

	// get rid of the eventbase element from the string
	timeValue = timeValue.substring(periodPosition + 1);

	// What should be left is the event type
	if (timeValue.equals(BEGIN))
		timeSymbol_ = TIME_SYMBOL_BEGIN;
	else
		if (timeValue.equals(END))
			timeSymbol_ = TIME_SYMBOL_END;
		else {
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_SYNCBASE_TIMING));
			return false;
		}
	return true;
}
/*********************************************************************** 
 * Class to interpret the various timing that can occur in the timing attribute.
 *
 * <pre>
 * SyncToPrev-value     ::= ( "prev." Time-symbol ) ( S ("+"|"-") S Clock-value )?  
 * Time-symbol          ::= "begin" | "end" 
 * </pre>
 *********************************************************************/
private boolean parseSyncToPrevValue(String param) {
	if (null == param) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	//
	String timeValue = param.trim();
	if (timeValue.equals("")) {
		addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
		return false;
	}

	// values checking at the begining of the string are faster to check
	if (timeValue.startsWith(PREV_BEGIN_UNIQUE_KEY)) {
		timeSymbol_ = TIME_SYMBOL_BEGIN;
		timeValue = timeValue.substring(PREV_BEGIN_UNIQUE_KEY.length());
		timeValue = timeValue.trim();
	} else
		if (timeValue.startsWith(PREV_END_UNIQUE_KEY)) {
			timeSymbol_ = TIME_SYMBOL_END;
			timeValue = timeValue.substring(PREV_END_UNIQUE_KEY.length());
			timeValue = timeValue.trim();
		} else {
			addErrorMessage(timingAttributeRes_.getString(TimingAttributeRes.MSG_ERROR_INVALID_TIMING_ATTRIBUTE));
			return false;
		}


		// We have gotten rid of the prev.begin and prev.end from the string.
		// if there is anything left then it must be a clock value
	if (null == timeValue || timeValue.equals("")) {
		clockValue_ = 0;
		return true;
	}

	// Thus determine the clock value 
	boolean success = parseClockValue(timeValue);
	return success;
}
/*********************************************************************** 
 * Enable debugging information to be printed to the console. 
 *
 * @param debug     sets debug information is printed
 **********************************************************************/
static public void setDebug(boolean debug) {
	   debug_ = debug;
}
/*********************************************************************** 
 * Determine whether this is a normal time value attribute.
 *
 * @return whether this is a normal time value attribute.
 **********************************************************************/
public TimingAttribute subtract(TimingAttribute time) {
	//
	if (INDEFINITE_VALUE == timingValueType_)
		return this;
	if (INDEFINITE_VALUE == time.getTimeValueType())
		return new TimingAttribute(CLOCK_VALUE, 0);

	// Do some minimal checking to see if we support the timing model
	if (CLOCK_VALUE != time.getTimeValueType())
		return null;

	//
	switch (timingValueType_) {
		case CLOCK_VALUE :
			return new TimingAttribute(timingValueType_, clockValue_ - time.getClockValue());
		case SYNCBASE_VALUE :
			return new TimingAttribute(null, null, timingValueType_, clockValue_ - time.getClockValue(), idValue_, timeSymbol_);
		case EVENT_VALUE :
			return new TimingAttribute(null, null, timingValueType_, clockValue_ - time.getClockValue(), idValue_, eventSymbol_);
	}


	//
	return null;
}
/***********************************************************************
 **********************************************************************/
public String toString() {
	switch (timingValueType_) {
		case CLOCK_VALUE :
			return clockValue_ + "ms";
		case EVENT_VALUE :
			{
				StringBuffer sb = new StringBuffer();
				if (null != idValue_)
					sb.append(escapeCharactersAdd(idValue_)).append('.');
				sb.append(eventSymbolString_);
				if (0 != clockValue_) {
					if (0 < clockValue_)
						sb.append('+');
					sb.append(clockValue_).append("ms");
				}
				return sb.toString();
			}
		case SYNCBASE_VALUE :
			{
				StringBuffer sb = new StringBuffer();
				sb.append(escapeCharactersAdd(idValue_)).append('.');
				if (TIME_SYMBOL_BEGIN == timeSymbol_)
					sb.append(BEGIN);
				else
					sb.append(END);
				if (0 != clockValue_) {
					if (0 < clockValue_)
						sb.append('+');
					sb.append(clockValue_).append("ms");
				}
				return sb.toString();
			}
		case SYNC_TO_PREV_VALUE :
			{
				StringBuffer sb = new StringBuffer();
				sb.append(PREV).append('.');
				if (TIME_SYMBOL_BEGIN == timeSymbol_)
					sb.append(BEGIN);
				else
					sb.append(END);
				if (0 != clockValue_) {
					if (0 < clockValue_)
						sb.append('+');
					sb.append(clockValue_).append("ms");
				}
				return sb.toString();
			}
		case INDEFINITE_VALUE :
			return INDEFINITE;
		case MEDIA_VALUE :
			return MEDIA;
		case MULTI_VALUE :
			{
				StringBuffer buf = new StringBuffer();
				Enumeration enum = multiTimingAttributes_.elements();
				if (enum.hasMoreElements())
					buf.append(enum.nextElement());
				for (; enum.hasMoreElements();) {
					buf.append(";").append(enum.nextElement());
				}
				return buf.toString();
			}
		case REPEAT_VALUE :
			{
				StringBuffer sb = new StringBuffer();
				if (null != idValue_)
					sb.append(escapeCharactersAdd(idValue_)).append('.');
				sb.append(REPEAT_UNIQUE_KEY);
				sb.append(repeatIteration_);
				sb.append(")");
				if (0 != clockValue_) {
					if (0 < clockValue_)
						sb.append('+');
					sb.append(clockValue_).append("ms");
				}
				return sb.toString();
			}
		case ACCESS_KEY_VALUE :
		case MEDIA_MARKER_VALUE :
		case WALLCLOCK_VALUE :
		default :
			return "";
	}
}
}
